/***********************************************************************\
*
* $Revision: 2176 $
* $Date: 2017-11-02 08:20:25 +0100 (Thu, 02 Nov 2017) $
* $Author: torsten $
* Contents: command find files
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;

import java.util.Arrays;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.regex.Pattern;
import java.util.Timer;
import java.util.TimerTask;

// graphics
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseWheelListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;


/****************************** Classes ********************************/

/** find files command
 */
public class CommandFindFiles
{
  enum Types
  {
    NAME,
    CONTENT,
    UNKNOWN
  };

  enum MatchModes
  {
    ANY,
    ALL
  };

  /** find data
   */
  class FindData
  {
    RepositoryTab repositoryTab;
    File          file;
    int           lineNumber;
    String        line;

    FindData(RepositoryTab repositoryTab, File file, int lineNumber, String line)
    {
      this.repositoryTab = repositoryTab;
      this.file          = file;
      this.lineNumber    = lineNumber;
      this.line          = line;
    }
  }

  /** find data comparator
   */
  class FindDataComparator implements Comparator<FindData>
  {
    // Note: enum in inner classes are not possible in Java, thus use the old way...
    private final static int SORTMODE_NAME        = 0;
    private final static int SORTMODE_LINE_NUMBER = 1;
    private final static int SORTMODE_PATH        = 2;
    private final static int SORTMODE_DATE        = 3;
    private final static int SORTMODE_SIZE        = 4;
    private final static int SORTMODE_LINE        = 5;

    private int sortMode;

    /** create file data comparator
     * @param table table
     * @param sortColumn column to sort
     */
    FindDataComparator(Table table, TableColumn sortColumn)
    {
      if      (table.getColumn(0) == sortColumn) sortMode = SORTMODE_NAME;
      else if (table.getColumn(1) == sortColumn) sortMode = SORTMODE_LINE_NUMBER;
      else if (table.getColumn(2) == sortColumn) sortMode = SORTMODE_PATH;
      else if (table.getColumn(3) == sortColumn) sortMode = SORTMODE_DATE;
      else if (table.getColumn(4) == sortColumn) sortMode = SORTMODE_SIZE;
      else if (table.getColumn(5) == sortColumn) sortMode = SORTMODE_LINE;
      else                                       sortMode = SORTMODE_NAME;
    }

    /** create file data comparator
     * @param table table
     */
    FindDataComparator(Table table)
    {
      this(table,table.getSortColumn());
    }

    /** compare file tree data without taking care about type
     * @param findData1, findData2 file tree data to compare
     * @return -1 iff findData1 < findData2,
                0 iff findData1 = findData2,
                1 iff findData1 > findData2
     */
    public int compare(FindData findData1, FindData findData2)
    {
      final int[][] SORT_ORDERING = new int[][]{new int[]{SORTMODE_NAME,SORTMODE_LINE_NUMBER,SORTMODE_PATH,SORTMODE_DATE,SORTMODE_SIZE,SORTMODE_LINE},
                                                new int[]{SORTMODE_LINE_NUMBER,SORTMODE_PATH,SORTMODE_NAME,SORTMODE_DATE,SORTMODE_SIZE,SORTMODE_LINE},
                                                new int[]{SORTMODE_PATH,SORTMODE_LINE_NUMBER,SORTMODE_NAME,SORTMODE_DATE,SORTMODE_SIZE,SORTMODE_LINE},
                                                new int[]{SORTMODE_DATE,SORTMODE_LINE_NUMBER,SORTMODE_NAME,SORTMODE_PATH,SORTMODE_SIZE,SORTMODE_LINE},
                                                new int[]{SORTMODE_SIZE,SORTMODE_LINE_NUMBER,SORTMODE_NAME,SORTMODE_PATH,SORTMODE_DATE,SORTMODE_LINE},
                                                new int[]{SORTMODE_LINE,SORTMODE_NAME,SORTMODE_LINE_NUMBER,SORTMODE_PATH,SORTMODE_DATE,SORTMODE_SIZE}
                                               };

      int result = 0;

      int z = 0;
      do
      {
        switch (SORT_ORDERING[sortMode][z])
        {
          case SORTMODE_NAME:
            String name1 = findData1.file.getName();
            String name2 = findData2.file.getName();
            result = name1.compareTo(name2);
            break;
          case SORTMODE_LINE_NUMBER:
            if      (findData1.lineNumber < findData2.lineNumber) result = -1;
            else if (findData1.lineNumber > findData2.lineNumber) result =  1;
            else                                                  result =  0;
            break;
          case SORTMODE_PATH:
            String path1 = findData1.file.getParent();
            String path2 = findData2.file.getParent();
            result = path1.compareTo(path2);
            break;
          case SORTMODE_DATE:
            long lastModified1 = findData1.file.lastModified();
            long lastModified2 = findData2.file.lastModified();
            if      (lastModified1 < lastModified2) result = -1;
            else if (lastModified1 > lastModified2) result =  1;
            else                                    result =  0;
            break;
          case SORTMODE_SIZE:
            long length1 = findData1.file.length();
            long length2 = findData2.file.length();
            if      (length1 < length2) result = -1;
            else if (length1 > length2) result =  1;
            else                        result =  0;
            break;
          case SORTMODE_LINE:
            String line1 = findData1.line;
            String line2 = findData2.line;
            result = line1.compareTo(line2);
            break;
          default:
            break;
        }
        z++;
      }
      while ((z < SORT_ORDERING[sortMode].length) && (result == 0));

      return result;
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return "FindDataComparator {"+sortMode+"}";
    }
  }

  /** dialog data
   */
  class Data
  {
    String     fileNamePatterns;
    String     textNamePatterns;
    MatchModes matchMode;
    boolean    findByNameFlag;
    boolean    findByContentFlag;
    boolean    caseSensitiveFlag;
    boolean    showAllHitsFlag;
    boolean    showAllRepositoriesFlag;
    boolean    showHiddenFilesFlag;
    boolean    forceFlag;
    boolean    quitFlag;

    Data()
    {
      this.fileNamePatterns        = CommandFindFiles.fileNamePatterns;
      this.textNamePatterns        = CommandFindFiles.textNamePatterns;
      this.matchMode               = CommandFindFiles.matchMode;
      this.findByNameFlag          = false;
      this.findByContentFlag       = false;
      this.caseSensitiveFlag       = false;
      this.showAllHitsFlag         = CommandFindFiles.showAllHitsFlag;
      this.showAllRepositoriesFlag = CommandFindFiles.showAllRepositoriesFlag;
      this.showHiddenFilesFlag     = CommandFindFiles.showHiddenFilesFlag;
      this.forceFlag               = false;
      this.quitFlag                = false;
    }
  };

  // --------------------------- constants --------------------------------
  // unique history database ids
  private final String HISTORY_NAME_FILENAME_PATTERNS = "fileNamePatterns";
  private final String HISTORY_NAME_TEXTNAME_PATTERNS = "textNamePatterns";

  // --------------------------- variables --------------------------------

  // stored settings
  private static String                  fileNamePatterns        = "";
  private static String                  textNamePatterns        = "";
  private static MatchModes              matchMode               = MatchModes.ALL;
  private static boolean                 caseSensitiveFlag       = false;
  private static boolean                 showAllHitsFlag         = false;
  private static boolean                 showAllRepositoriesFlag = false;
  private static boolean                 showHiddenFilesFlag     = false;

  // global variable references
  private final Types                    type;
  private final RepositoryTab            repositoryTab;
  private final Display                  display;
  private final Clipboard                clipboard;

  private final RepositoryTab[]          thisRepositoryTabs;
  private final RepositoryTab[]          allRepositoryTabs;

  private Timer                          restartFindFilesTimer;

  // dialog
  private final Data                     data = new Data();
  private final Shell                    dialog;
  private final FindDataComparator       findDataComparator;

  // widgets
  private final Table                    widgetFiles;
  private final Button                   widgetFileNamePatternAny;
  private final Button                   widgetFileNamePatternShort[] = new Button[5];
  private final Combo                    widgetFileNamePatterns;
  private final Combo                    widgetTextNamePatterns;
  private final Listener                 mouseWheelListener;
  private final Button                   widgetMatchModeAll;
  private final Button                   widgetMatchModeAny;
  private final Button                   widgetFindByName;
  private final Button                   widgetFindByContent;
  private final Button                   widgetCaseSensitive;
  private final Button                   widgetShowAllHits;
  private final Button                   widgetShowAllRepositories;
  private final Button                   widgetShowHiddenFiles;

  private final Button                   widgetButtonOpenDirectory;
  private final Button                   widgetButtonOpen;
  private final Button                   widgetButtonOpenWith;
//  private final Button         widgetButtonUpdate;
//  private final Button         widgetButtonCommit;
//  private final Button         widgetButtonCreatePatch;
//  private final Button         widgetButtonAdd;
//  private final Button         widgetButtonRemove;
//  private final Button         widgetButtonRevert;
//  private final Button        widgetButtonDiff;
  private final Button                   widgetButtonRevisions;
//  private final Button        widgetButtonSolve;

  private final Button                   widgetClose;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** find file command
   * @param shell shell
   * @param repositoryTab repository tab
   */
  CommandFindFiles(final Shell shell, final RepositoryTab repositoryTab, Types type)
  {
    Composite   composite,subComposite,subSubComposite;
    MenuItem    menuItem;
    Label       label;
    Button      button;
    TableColumn tableColumn;
    Listener    listener;

    final Menu  widgetFilesHeaderMenu;
    final Menu  widgetFilesBodyMenu;

    // initialize variables
    this.type               = type;
    this.repositoryTab      = repositoryTab;
    this.thisRepositoryTabs = new RepositoryTab[]{repositoryTab};
    this.allRepositoryTabs  = repositoryTab.onzen.getRepositoryTabs();

    // get display, clipboard
    display   = shell.getDisplay();
    clipboard = new Clipboard(display);

    // init restart timer
    restartFindFilesTimer = new Timer(true);

    // add files dialog
    dialog = Dialogs.open(shell,
                          showAllRepositoriesFlag
                            ? "Find files"
                            : "Find files in '"+repositoryTab.repository.rootPath+"'",
                          new double[]{1.0,0.0},
                          1.0
                         );

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0,4));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      final String FILES_COLUMN_NAMES[] = new String[]{"Name","#","Path","Date","Size","Line"};
      widgetFiles = Widgets.newTable(composite);
      widgetFiles.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetFiles,0,0,TableLayoutData.NSWE);
      SelectionListener selectionListener = new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          TableColumn        tableColumn    = (TableColumn)selectionEvent.widget;
          FindDataComparator findDataComparator = new FindDataComparator(widgetFiles,tableColumn);

          synchronized(widgetFiles)
          {
            Widgets.sortTableColumn(widgetFiles,tableColumn,findDataComparator);
          }
        }
      };
      tableColumn = Widgets.addTableColumn(widgetFiles,0,"Name",        SWT.LEFT, true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetFiles,1,"#",           SWT.RIGHT,true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetFiles,2,"Path",        SWT.LEFT, true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetFiles,3,"Date",        SWT.LEFT, true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetFiles,4,"Size [bytes]",SWT.RIGHT,true);
      tableColumn.addSelectionListener(selectionListener);
      tableColumn = Widgets.addTableColumn(widgetFiles,5,"Line",        SWT.LEFT, true);
      tableColumn.addSelectionListener(selectionListener);
      Widgets.setTableColumnWidth(widgetFiles,Settings.geometryFindFilesColumns.width);
      widgetFiles.setColumnOrder(Settings.geometryFindFilesColumnOrder.getMap(FILES_COLUMN_NAMES));
      Listener filesColumnResizeListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Settings.geometryFindFilesColumns = new Settings.ColumnSizes(Widgets.getTableColumnWidth(widgetFiles));
        }
      };
      Listener filesColumnMoveListener = new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          int[]    columnOrder = widgetFiles.getColumnOrder();
          String[] names       = new String[columnOrder.length];
          for (int i = 0; i < columnOrder.length; i++)
          {
            names[i] = FILES_COLUMN_NAMES[columnOrder[i]];
          }
          Settings.geometryFindFilesColumnOrder = new SettingUtils.SimpleStringArray(names);
        }
      };
      for (TableColumn orderTableColumn : widgetFiles.getColumns())
      {
        orderTableColumn.setMoveable(true);
        orderTableColumn.addListener(SWT.Move,filesColumnMoveListener);
        orderTableColumn.addListener(SWT.Resize,filesColumnResizeListener);
      }

      widgetFilesHeaderMenu = Widgets.newPopupMenu(dialog);
      {
        menuItem = Widgets.addMenuItemCheckbox(widgetFilesHeaderMenu,"Name");
        menuItem.setSelection(Settings.geometryFindFilesColumns.get(0) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetFiles,0,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetFilesHeaderMenu,"#");
        menuItem.setSelection(Settings.geometryFindFilesColumns.get(1) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetFiles,1,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetFilesHeaderMenu,"Path");
        menuItem.setSelection(Settings.geometryFindFilesColumns.get(2) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetFiles,2,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetFilesHeaderMenu,"Date");
        menuItem.setSelection(Settings.geometryFindFilesColumns.get(3) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetFiles,3,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetFilesHeaderMenu,"Size");
        menuItem.setSelection(Settings.geometryFindFilesColumns.get(4) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetFiles,4,widget.getSelection());
          }
        });
        menuItem = Widgets.addMenuItemCheckbox(widgetFilesHeaderMenu,"Line");
        menuItem.setSelection(Settings.geometryFindFilesColumns.get(5) > 0);
        menuItem.addSelectionListener(new SelectionListener()
        {
          @Override
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          @Override
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            MenuItem widget = (MenuItem)selectionEvent.widget;
            Widgets.showTableColumn(widgetFiles,5,widget.getSelection());
          }
        });
      }

      widgetFilesBodyMenu = Widgets.newPopupMenu(dialog);
      {
        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Open directory");
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetButtonOpenDirectory);
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetFilesBodyMenu);

        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Revisions"+"\u2026");
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetButtonRevisions);
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetFilesBodyMenu);

        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Open file"+"\u2026");
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetButtonOpen);
          }
        });

        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Open files with"+"\u2026",Settings.keyOpenFilesWith);
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Widgets.invoke(widgetButtonOpenWith);
          }
        });

        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Rename local file/directory"+"\u2026",Settings.keyRenameLocal);
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            FindData findData = getSelectedFindData();
            if (findData != null)
            {
              findData.repositoryTab.renameLocalFile(findData.repositoryTab.repository.getFileDataInstance(findData.file));
            }
          }
        });

        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Delete local files/directories"+"\u2026",Settings.keyDeleteLocal);
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            FindData findData = getSelectedFindData();
            if (findData != null)
            {
              findData.repositoryTab.deleteLocalFile(findData.repositoryTab.repository.getFileDataInstance(findData.file));
            }
          }
        });

        menuItem = Widgets.addMenuItemSeparator(widgetFilesBodyMenu);

        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Copy file name to clipboard");
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            FindData findData = getSelectedFindData();
            if (findData != null)
            {
              Widgets.setClipboard(clipboard,FileUtils.getCanonicalName(findData.file));
            }
          }
        });

        menuItem = Widgets.addMenuItem(widgetFilesBodyMenu,"Convert whitespaces"+"\u2026",Settings.keyDeleteLocal);
        menuItem.setEnabled(false);
        Widgets.addModifyListener(new WidgetModifyListener(menuItem,data)
        {
          public void modified(MenuItem menuItem)
          {
//            menuItem.setEnabled((widgetFiles.getSelectionCount() > 0));
          }
        });
        menuItem.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
Dprintf.dprintf("");
//  public boolean convertWhitespaces(String fileName, String[] fileNames, String message)
//            convertWhitespaces();
          }
        });
      }
      widgetFiles.setMenu(widgetFilesBodyMenu);

      widgetFiles.addListener(SWT.MenuDetect, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          Table widget = (Table)event.widget;

          Point     p          = display.map(null, widget, new Point(event.x, event.y));
          Rectangle clientArea = widget.getClientArea();
          boolean   isInHeader =    (clientArea.y <= p.y)
                                 && (p.y < (clientArea.y + widget.getHeaderHeight()));
          widget.setMenu(isInHeader ? widgetFilesHeaderMenu : widgetFilesBodyMenu);
        }
      });
      widgetFiles.addListener(SWT.Dispose, new Listener()
      {
        @Override
        public void handleEvent(Event event)
        {
          widgetFilesHeaderMenu.dispose();
          widgetFilesBodyMenu.dispose();
        }
      });

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,new double[]{0.0,1.0}));
      Widgets.layout(subComposite,1,0,TableLayoutData.WE);
      {
        java.util.List<String> patternList = HistoryDatabase.getHistory(HISTORY_NAME_FILENAME_PATTERNS,20,HistoryDatabase.Directions.SORTED);
        String patterns[] = new String[5];
        int    i = 0;
        for (String pattern : patternList)
        {
          if (!pattern.equals("*"))
          {
            patterns[i] = pattern;
            i++;
          }
          if (i >= 5) break;
        }

        label = Widgets.newLabel(subComposite,"File patterns:",SWT.NONE,Settings.keyFileNamePatterns);
        Widgets.layout(label,0,0,TableLayoutData.W);

        subSubComposite = Widgets.newComposite(subComposite);
        subSubComposite.setLayout(new TableLayout(null,new double[]{0.0,0.1,0.1,0.1,0.1,0.1,0.5}));
        Widgets.layout(subSubComposite,0,1,TableLayoutData.WE,0,2);
        {
          widgetFileNamePatternAny = Widgets.newRadio(subSubComposite,"*");
          widgetFileNamePatternAny.setToolTipText("Any file.");
          widgetFileNamePatternAny.setSelection(CommandFindFiles.fileNamePatterns.equals("*"));
          Widgets.layout(widgetFileNamePatternAny,0,0,TableLayoutData.W);

          widgetFileNamePatternShort[0] = Widgets.newRadio(subSubComposite,(patterns[0] != null) ? patterns[0] : "");
          widgetFileNamePatternShort[0].setToolTipText("File name filter pattern.");
          widgetFileNamePatternShort[0].setSelection(CommandFindFiles.fileNamePatterns.equals(patterns[0]));
          Widgets.layout(widgetFileNamePatternShort[0],0,1,TableLayoutData.W);

          widgetFileNamePatternShort[1] = Widgets.newRadio(subSubComposite,(patterns[1] != null) ? patterns[1] : "");
          widgetFileNamePatternShort[1].setToolTipText("File name filter pattern.");
          widgetFileNamePatternShort[1].setSelection(CommandFindFiles.fileNamePatterns.equals(patterns[1]));
          Widgets.layout(widgetFileNamePatternShort[1],0,2,TableLayoutData.W);

          widgetFileNamePatternShort[2] = Widgets.newRadio(subSubComposite,(patterns[2] != null) ? patterns[2] : "");
          widgetFileNamePatternShort[2].setToolTipText("File name filter pattern.");
          widgetFileNamePatternShort[2].setSelection(CommandFindFiles.fileNamePatterns.equals(patterns[2]));
          Widgets.layout(widgetFileNamePatternShort[2],0,3,TableLayoutData.W);

          widgetFileNamePatternShort[3] = Widgets.newRadio(subSubComposite,(patterns[3] != null) ? patterns[3] : "");
          widgetFileNamePatternShort[3].setToolTipText("File name filter pattern.");
          widgetFileNamePatternShort[3].setSelection(CommandFindFiles.fileNamePatterns.equals(patterns[3]));
          Widgets.layout(widgetFileNamePatternShort[3],0,4,TableLayoutData.W);

          widgetFileNamePatternShort[4] = Widgets.newRadio(subSubComposite,(patterns[4] != null) ? patterns[4] : "");
          widgetFileNamePatternShort[4].setToolTipText("File name filter pattern.");
          widgetFileNamePatternShort[4].setSelection(CommandFindFiles.fileNamePatterns.equals(patterns[4]));
          Widgets.layout(widgetFileNamePatternShort[4],0,5,TableLayoutData.W);

          widgetFileNamePatterns = Widgets.newCombo(subSubComposite);
          widgetFileNamePatterns.setToolTipText("File name filter patterns. Use * and ? as wildcards. Use space as separator.");
          Widgets.layout(widgetFileNamePatterns,0,6,TableLayoutData.WE);
          widgetFileNamePatterns.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
              Combo  widget  = (Combo)selectionEvent.widget;
              String pattern = widget.getText().trim();

              for (int i = 1; i < 5; i++)
              {
                widgetFileNamePatternShort[i].setText(widgetFileNamePatternShort[i-1].getText());
              }
              widgetFileNamePatternShort[0].setText(pattern);
              widgetFileNamePatternShort[0].getParent().layout();

              CommandFindFiles.fileNamePatterns = pattern;
              restartFindFiles();
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              Combo  widget  = (Combo)selectionEvent.widget;
              String pattern = widget.getText().trim();

              for (int i = 1; i < 5; i++)
              {
                widgetFileNamePatternShort[i].setText(widgetFileNamePatternShort[i-1].getText());
              }
              widgetFileNamePatternShort[0].setText(pattern);
              widgetFileNamePatternShort[0].getParent().layout();

              CommandFindFiles.fileNamePatterns = pattern;
              restartFindFiles();
            }
          });
          for (String pattern : patternList)
          {
            widgetFileNamePatterns.add(pattern);
          }
          widgetFileNamePatterns.setText(CommandFindFiles.fileNamePatterns);

          widgetFileNamePatternAny.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
              Button widget = (Button)selectionEvent.widget;

              widgetFileNamePatterns.setText("*");

              CommandFindFiles.fileNamePatterns = widgetFileNamePatterns.getText().trim();
              restartFindFiles();
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              widgetFileNamePatterns.setText("*");

              CommandFindFiles.fileNamePatterns = widgetFileNamePatterns.getText().trim();
              restartFindFiles();
            }
          });
          for (i = 0; i < 5; i++)
          {
            widgetFileNamePatternShort[i].addSelectionListener(new SelectionListener()
            {
              public void widgetDefaultSelected(SelectionEvent selectionEvent)
              {
                Button widget = (Button)selectionEvent.widget;

                widgetFileNamePatterns.setText(widget.getText());

                CommandFindFiles.fileNamePatterns = widgetFileNamePatterns.getText().trim();
                restartFindFiles();
              }
              public void widgetSelected(SelectionEvent selectionEvent)
              {
                Button widget = (Button)selectionEvent.widget;

                widgetFileNamePatterns.setText(widget.getText());

                CommandFindFiles.fileNamePatterns = widgetFileNamePatterns.getText().trim();
                restartFindFiles();
              }
            });
          }
        }

        label = Widgets.newLabel(subComposite,"Text/Name:",SWT.NONE,Settings.keyTextNamePatterns);
        Widgets.layout(label,1,0,TableLayoutData.W);
        subSubComposite = Widgets.newComposite(subComposite);
        subSubComposite.setLayout(new TableLayout(null,new double[]{1.0,0.0,0.0}));
        Widgets.layout(subSubComposite,1,1,TableLayoutData.WE,0,2);
        {
          widgetTextNamePatterns = Widgets.newCombo(subSubComposite);
          widgetTextNamePatterns.setToolTipText("File name/text patterns. Use * and ? as wildcards. Use space as separator.");
          Widgets.layout(widgetTextNamePatterns,0,0,TableLayoutData.WE);
          widgetTextNamePatterns.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
              restartFindFiles();
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              restartFindFiles();
            }
          });
          String[] history;
          for (String pattern : HistoryDatabase.getHistory(HISTORY_NAME_TEXTNAME_PATTERNS,20,HistoryDatabase.Directions.SORTED))
          {
            widgetTextNamePatterns.add(pattern);
          }
          widgetTextNamePatterns.setText(CommandFindFiles.textNamePatterns);

          widgetMatchModeAll = Widgets.newRadio(subSubComposite,"all");
          widgetMatchModeAll.setSelection(CommandFindFiles.matchMode == MatchModes.ALL);
          Widgets.layout(widgetMatchModeAll,0,1,TableLayoutData.W);
          widgetMatchModeAll.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              restartFindFiles();
            }
          });

          widgetMatchModeAny = Widgets.newRadio(subSubComposite,"any");
          widgetMatchModeAny.setSelection(CommandFindFiles.matchMode == MatchModes.ANY);
          Widgets.layout(widgetMatchModeAny,0,2,TableLayoutData.W);
          widgetMatchModeAny.addSelectionListener(new SelectionListener()
          {
            public void widgetDefaultSelected(SelectionEvent selectionEvent)
            {
            }
            public void widgetSelected(SelectionEvent selectionEvent)
            {
              restartFindFiles();
            }
          });
        }
      }

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(null,0.0));
      Widgets.layout(subComposite,3,0,TableLayoutData.WE);
      {
        widgetFindByName = Widgets.newCheckbox(subComposite,"find by name",SWT.F5);
        widgetFindByName.setSelection(type == Types.NAME);
        Widgets.layout(widgetFindByName,0,0,TableLayoutData.W);
        widgetFindByName.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            restartFindFiles();
          }
        });

        widgetFindByContent = Widgets.newCheckbox(subComposite,"find by content",SWT.F6);
        widgetFindByContent.setSelection(type == Types.CONTENT);
        Widgets.layout(widgetFindByContent,0,1,TableLayoutData.W);
        widgetFindByContent.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            restartFindFiles();
          }
        });

        widgetCaseSensitive = Widgets.newCheckbox(subComposite,"case sensitive",SWT.F7);
        widgetCaseSensitive.setSelection(data.caseSensitiveFlag);
        Widgets.layout(widgetCaseSensitive,0,2,TableLayoutData.W);
        widgetCaseSensitive.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            restartFindFiles();
          }
        });

        widgetShowAllHits = Widgets.newCheckbox(subComposite,"show all hits",SWT.F8);
        widgetShowAllHits.setSelection(data.showAllRepositoriesFlag);
        Widgets.layout(widgetShowAllHits,0,3,TableLayoutData.W);
        widgetShowAllHits.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            restartFindFiles();
          }
        });

        widgetShowAllRepositories = Widgets.newCheckbox(subComposite,"show all repositories",SWT.F9);
        widgetShowAllRepositories.setSelection(data.showAllRepositoriesFlag);
        Widgets.layout(widgetShowAllRepositories,0,4,TableLayoutData.W);
        widgetShowAllRepositories.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            restartFindFiles();
          }
        });

        widgetShowHiddenFiles = Widgets.newCheckbox(subComposite,"show hidden files",SWT.F10);
        widgetShowHiddenFiles.setSelection(data.showHiddenFilesFlag);
        Widgets.layout(widgetShowHiddenFiles,0,5,TableLayoutData.W);
        widgetShowHiddenFiles.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            restartFindFiles();
          }
        });
      }
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,0.0,0.0,0.0,1.0}));
    Widgets.layout(composite,1,0,TableLayoutData.WE,0,0,4);
    {
      widgetButtonOpenDirectory = Widgets.newButton(composite,"Open directory");
      widgetButtonOpenDirectory.setEnabled(false);
      Widgets.layout(widgetButtonOpenDirectory,0,0,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetButtonOpenDirectory,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(widgetFiles.getSelectionCount() > 0));
        }
      });
      widgetButtonOpenDirectory.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          FindData findData = getSelectedFindData();
          if (findData != null)
          {
            findData.repositoryTab.onzen.selectRepositoryTab(findData.repositoryTab);
            findData.repositoryTab.openDirectory(findData.file);
          }
        }
      });

      widgetButtonOpen = Widgets.newButton(composite,"Open file");
      widgetButtonOpen.setEnabled(false);
      Widgets.layout(widgetButtonOpen,0,1,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetButtonOpen,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(widgetFiles.getSelectionCount() > 0));
        }
      });
      widgetButtonOpen.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          FindData findData = getSelectedFindData();
          if (findData != null)
          {
            findData.repositoryTab.openFile(findData.file,
                                            Onzen.getMimeType(findData.file),
                                            findData.lineNumber
                                           );
          }
        }
      });

      widgetButtonOpenWith = Widgets.newButton(composite,"Open files with...",Settings.keyOpenFilesWith);
      widgetButtonOpenWith.setEnabled(false);
      Widgets.layout(widgetButtonOpenWith,0,2,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetButtonOpenWith,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(widgetFiles.getSelectionCount() > 0));
        }
      });
      widgetButtonOpenWith.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          FindData findData = getSelectedFindData();
          if (findData != null)
          {
            findData.repositoryTab.openFileWith(findData.file,findData.lineNumber);
          }
        }
      });

      widgetButtonRevisions = Widgets.newButton(composite,"Revisions",Settings.keyRevisions);
      widgetButtonRevisions.setEnabled(false);
      Widgets.layout(widgetButtonRevisions,0,3,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetButtonRevisions,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(widgetFiles.getSelectionCount() > 0));
        }
      });
      widgetButtonRevisions.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          FindData findData = getSelectedFindData();
          if (findData != null)
          {
            CommandRevisions commandRevisions = new CommandRevisions(shell,
                                                                     findData.repositoryTab,
                                                                     findData.repositoryTab.repository.getFileDataInstance(findData.file)
                                                                    );
            commandRevisions.run();
          }
        }
      });

      button = Widgets.newButton(composite,"Reread");
      Widgets.layout(button,0,4,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          restartFindFiles(true);
        }
      });

      widgetClose = Widgets.newButton(composite,"Close");
      Widgets.layout(widgetClose,0,5,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetClose.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Dialogs.close(dialog);
        }
      });
    }

    // listeners
    widgetFiles.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        Table widget = (Table)selectionEvent.widget;

        int index = widget.getSelectionIndex();
        if (index >= 0)
        {
          TableItem tableItem = widget.getItem(index);
          FindData  findData  = (FindData)tableItem.getData();

          repositoryTab.openFile(findData.file,
                                 Onzen.getMimeType(findData.file),
                                 findData.lineNumber
                                );
        }
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Widgets.modified(data);
      }
    });

    KeyListener keyListener = new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (Widgets.isAccelerator(keyEvent,Settings.keyFileNamePatterns))
        {
          Widgets.setFocus(widgetFileNamePatterns);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyTextNamePatterns))
        {
          Widgets.setFocus(widgetTextNamePatterns);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyOpenFilesWith))
        {
          Widgets.invoke(widgetButtonOpenWith);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyRevisions))
        {
          Widgets.invoke(widgetButtonRevisions);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.F5))
        {
          Widgets.invoke(widgetFindByName);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.F6))
        {
          Widgets.invoke(widgetFindByContent);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.F7))
        {
          Widgets.invoke(widgetCaseSensitive);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.F8))
        {
          Widgets.invoke(widgetShowAllHits);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.F9))
        {
          Widgets.invoke(widgetShowAllRepositories);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.F10))
        {
          Widgets.invoke(widgetShowHiddenFiles);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'a'))
        {
          widgetFiles.selectAll();
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'c'))
        {
          HashSet<FindData> findDataSet = getSelectedFindDataSet();
          if (findDataSet != null)
          {
            StringBuilder buffer = new StringBuilder();
            for (FindData findData : findDataSet)
            {
              buffer.append(findData.file.getAbsolutePath()); buffer.append('\n');
            }
            Widgets.setClipboard(clipboard,buffer.toString());
          }
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    };
    widgetFiles.addKeyListener(keyListener);
    widgetFileNamePatterns.addKeyListener(keyListener);
    widgetTextNamePatterns.addKeyListener(keyListener);
    widgetFindByName.addKeyListener(keyListener);
    widgetFindByContent.addKeyListener(keyListener);
    widgetCaseSensitive.addKeyListener(keyListener);
    widgetShowAllRepositories.addKeyListener(keyListener);
    widgetShowHiddenFiles.addKeyListener(keyListener);

    widgetFileNamePatterns.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (keyEvent.keyCode == SWT.ARROW_UP)
        {
          int index = widgetFiles.getSelectionIndex();
          if (index > 0) index--;
//          widgetFiles.setSelection(index,index);
        }
        else if (keyEvent.keyCode == SWT.ARROW_DOWN)
        {
          int index = widgetFiles.getSelectionIndex();
          if (index < widgetFiles.getItemCount()-1) index++;
//          widgetFiles.setSelection(index,index);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
        {
          widgetFileNamePatterns.setText("");
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
        CommandFindFiles.fileNamePatterns = widgetFileNamePatterns.getText().trim();
        restartFindFiles();
      }
    });
    widgetTextNamePatterns.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (keyEvent.keyCode == SWT.ARROW_UP)
        {
          int index = widgetFiles.getSelectionIndex();
          if (index > 0) index--;
          widgetFiles.setSelection(index,index);
        }
        else if (keyEvent.keyCode == SWT.ARROW_DOWN)
        {
          int index = widgetFiles.getSelectionIndex();
          if (index < widgetFiles.getItemCount()-1) index++;
          widgetFiles.setSelection(index,index);
        }
        else if (Widgets.isAccelerator(keyEvent,SWT.CTRL+'x'))
        {
          widgetTextNamePatterns.setText("");
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
        restartFindFiles();
      }
    });

    // filter mouse-wheel event to avoid scrolling in file name/content-combo
    mouseWheelListener = new Listener()
    {
      public void handleEvent(Event event)
      {
        if (   (event.widget == widgetFileNamePatterns)
            || (event.widget == widgetTextNamePatterns)
           )
        {
          event.doit = false;

/*
??? scroll widgetFiles
          widgetFiles

          Event redirectedEvent = new Event();
          redirectedEvent.type   = SWT.MouseWheel;
          redirectedEvent.widget = widgetFiles;
          redirectedEvent.time   = event.time;
          redirectedEvent.x      = event.x;
          redirectedEvent.y      = event.y;
          redirectedEvent.detail = event.detail;
          redirectedEvent.count  = event.count;
          widgetFiles.notifyListeners(SWT.MouseWheel,redirectedEvent);
*/
        }
      }
    };
    display.addFilter(SWT.MouseWheel,mouseWheelListener);

    findDataComparator = new FindDataComparator(widgetFiles);

    // show dialog
    Dialogs.show(dialog,Settings.geometryFindFiles,Settings.setWindowLocation);

    // find files
    Background.run(new BackgroundRunnable()
    {
      String                  fileNamePatterns          = "";
      String                  textNamePatterns          = "";
      java.util.List<Pattern> fileNamePatternList       = null;
      java.util.List<Pattern> textNamePatternList       = null;
      boolean                 findByNameFlag            = data.findByNameFlag;
      boolean                 findByContentFlag         = data.findByContentFlag;
      MatchModes              matchMode                 = data.matchMode;
      boolean                 caseSensitiveFlag         = data.caseSensitiveFlag;
      boolean                 showAllHitsFlag           = data.showAllHitsFlag;
      boolean                 showAllRepositoriesFlag   = data.showAllRepositoriesFlag;
      boolean                 showHiddenFilesFlag       = data.showHiddenFilesFlag;

      /** check if data is modified
       * @return true iff data is modified
       */
      boolean isDataModified()
      {
        return    data.quitFlag
               || ((data.fileNamePatterns != null) && !data.fileNamePatterns.equals(fileNamePatterns))
               || ((data.textNamePatterns != null) && !data.textNamePatterns.equals(textNamePatterns))
               || (data.matchMode != matchMode)
               || (data.findByNameFlag != findByNameFlag)
               || (data.findByContentFlag != findByContentFlag)
               || (data.caseSensitiveFlag != caseSensitiveFlag)
               || (data.showAllHitsFlag != showAllHitsFlag)
               || (data.showAllRepositoriesFlag != showAllRepositoriesFlag)
               || (data.showHiddenFilesFlag != showHiddenFilesFlag)
               || data.forceFlag;
      }

      /** run method
       */
      public void run()
      {
        HashSet<String> findDataKeySet = new HashSet<String>();

        while (!data.quitFlag)
        {
          // clearfiles
          if (!display.isDisposed())
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                if (!widgetFiles.isDisposed())
                {
                  Widgets.removeAllTableItems(widgetFiles);
                }
              }
            });
          }
          findDataKeySet.clear();

          // find files
          if (!textNamePatterns.isEmpty())
          {
//Dprintf.dprintf("findNamePatterns=%s findContentPatterns=%s",findNamePatterns,findContentPatterns);
            // start find files
            if (!display.isDisposed())
            {
              display.syncExec(new Runnable()
              {
                public void run()
                {
                  if (!dialog.isDisposed())
                  {
                    Widgets.setCursor(dialog,Onzen.CURSOR_WAIT);
                  }
                  if (!widgetFiles.isDisposed())
                  {
                    widgetFiles.setForeground(Onzen.COLOR_DARK_GRAY);
                  }
                }
              });
            }

            // find files
            for (final RepositoryTab repositoryTab : (showAllRepositoriesFlag) ? allRepositoryTabs : thisRepositoryTabs)
            {
              File directory = new File(repositoryTab.repository.rootPath);

              // check for modified data
              if (isDataModified()) break;

              LinkedList<File> directoryList = new LinkedList<File>();
              do
              {
//Dprintf.dprintf("directory=%s",directory);
                File[] files = directory.listFiles();
                if (files != null)
                {
                  for (final File file : files)
                  {
                    // check for modified data
                    if (isDataModified()) break;

                    // get name
                    String fileName = file.getName();

//Dprintf.dprintf("fileName=%s %s",fileName,repositoryTab.repository.isHiddenFile(fileName));
                    if (showHiddenFilesFlag || !repositoryTab.repository.isHiddenFile(fileName))
                    {
                      if     (file.isFile())
                      {
                        if (fileNamePatternList.isEmpty() || matchPatterns(fileName,fileNamePatternList))
                        {
                          // initialise text search
                          TextSearch textSearch = new TextSearch(file,textNamePatterns,textNamePatternList,caseSensitiveFlag,matchMode);

                          // find by name or content
                          boolean nameMatchFlag    = false;
                          boolean contentMatchFlag = false;
                          if (findByNameFlag)
                          {
                            nameMatchFlag = textNamePatternList.isEmpty() || matchPatterns(fileName,textNamePatternList,matchMode);
                          }
                          if (findByContentFlag)
                          {
                            if (!nameMatchFlag)
                            {
                              contentMatchFlag = textSearch.findNext();
                            }
                          }

                          if (nameMatchFlag || contentMatchFlag)
                          {
//Dprintf.dprintf("fileName=%s nameMatchFlag=%s contentMatchFlag=%s",fileName,nameMatchFlag,contentMatchFlag);
                            do
                            {
                              String findDataKey = file.getAbsolutePath()+Integer.toString(textSearch.lineNumber);
                              if (!findDataKeySet.contains(findDataKey))
                              {
//Dprintf.dprintf("file=%s n=%d x=%s",file.getAbsolutePath(),lineNumber,findDataKeySet.contains(findDataKey));
                                // add find data to list
                                final FindData findData = new FindData(repositoryTab,file,textSearch.lineNumber,textSearch.line);
                                display.syncExec(new Runnable()
                                {
                                  public void run()
                                  {
                                    if (!widgetFiles.isDisposed())
                                    {
                                      Widgets.insertTableItem(widgetFiles,
                                                              findDataComparator,
                                                              findData,
                                                              file.getName(),
                                                              Integer.toString(findData.lineNumber),
                                                              file.getParent(),
                                                              Onzen.DATETIME_FORMAT.format(file.lastModified()),
                                                              Long.toString(file.length()),
                                                              findData.line
                                                             );
                                    }
                                  }
                                });

                                findDataKeySet.add(findDataKey);
                              }

                              // check for modified data
                              if (isDataModified()) break;

                              // find next text line
                              if (data.showAllHitsFlag && contentMatchFlag)
                              {
                                contentMatchFlag = textSearch.findNext();
                              }
                            }
                            while (data.showAllHitsFlag && contentMatchFlag);
                          }

                          // free resources
                          textSearch.close();
                        }
                      }
                      else if (file.isDirectory())
                      {
//Dprintf.dprintf("add dir=%s %s",fileName,file.getPath());
                        directoryList.add(file);
                      }
                    }
                  }
                }
              }
              while ((directory = directoryList.pollFirst()) != null);
            }

            // stop find files
            if (!display.isDisposed())
            {
              display.syncExec(new Runnable()
              {
                public void run()
                {
                  if (!widgetFiles.isDisposed())
                  {
                    widgetFiles.setForeground(Onzen.COLOR_BLACK);
                  }
                  if (!dialog.isDisposed())
                  {
                    Widgets.resetCursor(dialog);
                  }
                }
              });
            }
          }

          // wait for new filters/quit
          synchronized(data)
          {
            // wait for new find data
            while (!isDataModified())
            {
              try
              {
                data.wait();
              }
              catch (InterruptedException exception)
              {
                // ignored
              }
            }

            // get new find data, clear data
            if (data.fileNamePatterns != null)
            {
              // get new find file name patterns
              fileNamePatterns    = data.fileNamePatterns;
              fileNamePatternList = compilePatterns(data.fileNamePatterns,data.caseSensitiveFlag);
              data.fileNamePatterns = null;
            }
            if (data.textNamePatterns != null)
            {
              // get new find text/name patterns
              textNamePatterns    = data.textNamePatterns;
              textNamePatternList = compilePatterns(data.textNamePatterns,data.caseSensitiveFlag);
              data.textNamePatterns = null;
            }
            matchMode               = data.matchMode;
            findByNameFlag          = data.findByNameFlag;
            findByContentFlag       = data.findByContentFlag;
            caseSensitiveFlag       = data.caseSensitiveFlag;
            showAllHitsFlag         = data.showAllHitsFlag;
            showAllRepositoriesFlag = data.showAllRepositoriesFlag;
            showHiddenFilesFlag     = data.showHiddenFilesFlag;

            data.forceFlag = false;
          }
        }
      }
    });
  }

  /** run dialog
   */
  public void run()
  {
    if (!dialog.isDisposed())
    {
      Widgets.setFocus(widgetTextNamePatterns);
      Dialogs.run(dialog,new DialogRunnable()
      {
        public void done(Object result)
        {
          // store values
          HistoryDatabase.setHistory(HISTORY_NAME_FILENAME_PATTERNS,
                                     widgetFileNamePatterns.getItems(),
                                     20,
                                     HistoryDatabase.Directions.SORTED,
                                     widgetFileNamePatterns.getText().trim()
                                    );
          HistoryDatabase.setHistory(HISTORY_NAME_TEXTNAME_PATTERNS,
                                     widgetTextNamePatterns.getItems(),
                                     20,
                                     HistoryDatabase.Directions.SORTED,
                                     widgetTextNamePatterns.getText().trim()
                                    );
          CommandFindFiles.fileNamePatterns        = widgetFileNamePatterns.getText().trim();
          CommandFindFiles.textNamePatterns        = widgetTextNamePatterns.getText().trim();
          CommandFindFiles.caseSensitiveFlag       = widgetCaseSensitive.getSelection();
          CommandFindFiles.showAllHitsFlag         = widgetShowAllHits.getSelection();
          CommandFindFiles.showAllRepositoriesFlag = widgetShowAllRepositories.getSelection();
          CommandFindFiles.showHiddenFilesFlag     = widgetShowHiddenFiles.getSelection();

          Settings.geometryFindFiles        = dialog.getSize();
          Settings.geometryFindFilesColumns = new Settings.ColumnSizes(Widgets.getTableColumnWidth(widgetFiles));

          // signal quit to find thread
          synchronized(data)
          {
            data.quitFlag = true;
            data.notifyAll();
          }

          // remove mouse-wheel filter
          display.removeFilter(SWT.MouseWheel,mouseWheelListener);
        }
      });
    }
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "CommandFindFiles {}";
  }

  //-----------------------------------------------------------------------

  /** text search class
   */
  class TextSearch
  {
    private BufferedReader          input;
    private String                  content;
    private java.util.List<Pattern> contentPatternList;
    private MatchModes              matchMode;
    private boolean                 caseSensitiveFlag;
    private int                     lineNumber;
    private String                  line;

    /** create text search
     * @param
     */
    TextSearch(File file, String content, java.util.List<Pattern> contentPatternList, boolean caseSensitiveFlag, MatchModes matchMode)
    {
      this.input              = null;
      this.content            = (caseSensitiveFlag) ? content : content.toLowerCase();
      this.contentPatternList = contentPatternList;
      this.matchMode          = matchMode;
      this.caseSensitiveFlag  = caseSensitiveFlag;
      this.lineNumber         = 0;
      this.line               = null;

      // open file
      try
      {
        input = new BufferedReader(new FileReader(file));
      }
      catch (IOException exception)
      {
        input = null;
      }
    }

    /** create text search
     * @param
     */
    public void close()
    {
      try
      {
        if (input != null) input.close();
      }
      catch (IOException exception)
      {
        // ignored
      }
    }

    /** find next line with search text
     * @return true iff found
     */
    boolean findNext()
    {
      if (input != null)
      {
        try
        {
          // read file and find text
          while ((line = input.readLine()) != null)
          {
            lineNumber++;
            if (   (caseSensitiveFlag && line.contains(content))
                || (!caseSensitiveFlag && line.toLowerCase().contains(content))
                || matchPatterns(line,contentPatternList,matchMode)
               )
            {
              return true;
            }
          }
        }
        catch (IOException exception)
        {
          // not found
        }
      }

      return false;
    }
  }

  /** restart find files matching filters
   * @param forceFlag true to force restart find files
   */
  private void restartFindFiles(final boolean forceFlag)
  {
    // set dialog title
    dialog.setText(widgetShowAllRepositories.getSelection()
                     ? "Find files"
                     : "Find files in '"+repositoryTab.repository.rootPath+"'"
                  );

    // get data
    if (!dialog.isDisposed())
    {
      synchronized(data)
      {
        data.fileNamePatterns        = widgetFileNamePatterns.getText().trim();
        data.textNamePatterns        = widgetTextNamePatterns.getText().trim();
        if      (widgetMatchModeAll.getSelection()) data.matchMode = MatchModes.ALL;
        else if (widgetMatchModeAny.getSelection()) data.matchMode = MatchModes.ANY;
        else                                        data.matchMode = MatchModes.ALL;
        data.findByNameFlag          = widgetFindByName.getSelection();
        data.findByContentFlag       = widgetFindByContent.getSelection();
        data.caseSensitiveFlag       = widgetCaseSensitive.getSelection();
        data.showAllHitsFlag         = widgetShowAllHits.getSelection();
        data.showAllRepositoriesFlag = widgetShowAllRepositories.getSelection();
        data.showHiddenFilesFlag     = widgetShowHiddenFiles.getSelection();
        data.forceFlag               = forceFlag;

        if (!data.textNamePatterns.isEmpty()) Widgets.modified(data);
      }
    }

    // restart find files timer
    restartFindFilesTimer.cancel();
    restartFindFilesTimer = new Timer(true);
    restartFindFilesTimer.schedule(new TimerTask()
                                   {
                                     public void run()
                                     {
                                       synchronized(data)
                                       {
                                         data.notifyAll();
                                       }
                                     }
                                   },
                                   500  // ms
                                  );
  }

  /** restart find files matching filters
   */
  private void restartFindFiles()
  {
    restartFindFiles(false);
  }

  /** get selected file data
   * @return file data or null
   */
  private FindData getSelectedFindData()
  {
    FindData findData;

    int index = widgetFiles.getSelectionIndex();
    if (index >= 0)
    {
      findData = (FindData)widgetFiles.getItem(index).getData();
    }
    else
    {
      findData = null;
    }

    return findData;
  }

  /** get selected find data set
   * @return find data set
   */
  private HashSet<FindData> getSelectedFindDataSet()
  {
    HashSet<FindData> findDataSet = new HashSet<FindData>();

    for (TableItem tableItem : widgetFiles.getSelection())
    {
      findDataSet.add((FindData)tableItem.getData());
    }

    return (findDataSet.size() > 0) ? findDataSet : null;
  }

  /** compile patterns
   * @param textNamePattern text/name patterns string
   * @param caseSensitiveFlag true for case-sensitive matching
   * @return file content pattern list
   */
  private java.util.List<Pattern> compilePatterns(String textNamePattern, boolean caseSensitiveFlag)
  {
    ArrayList<Pattern> patternList = new ArrayList<Pattern>();

    for (String pattern : StringUtils.splitArray(textNamePattern,StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS))
    {
      patternList.add(Pattern.compile(".*"+StringUtils.globToRegex(pattern)+".*",caseSensitiveFlag ? 0 : Pattern.CASE_INSENSITIVE));
    }

    return patternList;
  }

  /** match patterns
   * @param string string to match
   * @param patternList patterns to match
   * @param matchMode match mode
   * @return true iff all/any pattern match with string
   */
  private boolean matchPatterns(String string, java.util.List<Pattern> patternList, MatchModes matchMode)
  {
    if (!patternList.isEmpty())
    {
      for (Pattern pattern : patternList)
      {
        switch (matchMode)
        {
          case ANY:
            if (pattern.matcher(string).matches())
            {
              return true;
            }
            break;
          case ALL:
            if (!pattern.matcher(string).matches())
            {
              return false;
            }
            break;
        }
      }

      switch (matchMode)
      {
        case ANY: return false;
        case ALL: return true;
        default:  return false;
      }
    }
    else
    {
      return false;
    }
  }

  /** match patterns
   * @param string string to match
   * @param patternList patterns to match
   * @return true iff all/any pattern match with string
   */
  private boolean matchPatterns(String string, java.util.List<Pattern> patternList)
  {
    return matchPatterns(string,patternList,MatchModes.ANY);
  }

  /** check if file contains text
   * @param file file to check
   * @param content text to find
   * @param contentPatterns content patterns
   * @param caseSensitiveFlag true for case-sensitive matching
   * @return line number or -1 if not found
   */
/*
  private int fileContains(File file, String content, Pattern[] contentPatterns, boolean caseSensitiveFlag)
  {
    int lineNumber = -1;

    BufferedReader input = null;
    try
    {
      // open file
      input = new BufferedReader(new FileReader(file));

      // read file and find text
      if (!caseSensitiveFlag) content = content.toLowerCase();
      String line;
      int    n = 0;
      while ((line = input.readLine()) != null)
      {
        n++;
        if (   (caseSensitiveFlag && line.contains(content))
            || (!caseSensitiveFlag && line.toLowerCase().contains(content))
            || matchPatterns(line,contentPatterns)
           )
        {
          lineNumber = n;
          break;
        }
      }

      // close file
      input.close(); input = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      try
      {
        if (input != null) input.close();
      }
      catch (IOException exception)
      {
        // ignored
      }
    }

    return lineNumber;
  }
*/
}

/* end of file */
