/***********************************************************************\
*
* $Revision: 2171 $
* $Date: 2017-09-29 15:48:53 +0200 (Fri, 29 Sep 2017) $
* $Author: torsten $
* Contents: command show file revisions tree
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.DataOutputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

// graphics
import org.eclipse.swt.dnd.Clipboard;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.MouseMoveListener;
import org.eclipse.swt.events.MouseTrackListener;
import org.eclipse.swt.events.PaintEvent;
import org.eclipse.swt.events.PaintListener;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Cursor;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Canvas;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.List;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.swt.widgets.MenuItem;
import org.eclipse.swt.widgets.ScrollBar;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.swt.widgets.Text;
import org.eclipse.swt.widgets.Widget;

/****************************** Classes ********************************/

/** view file revisions tree command
 */
class CommandRevisions
{
  /** drawing states
   */
  static enum DrawStates
  {
    ENTRY_POSITIONS,
    ENTRIES,
    LINES
  };

  /** drawing info
   */
  class DrawInfo
  {
    RevisionData revisionData;
    Rectangle    container;
    Rectangle    handle;

    /** create draw info
     * @param revisionData revision data
     * @param container container size
     * @param handle handle size
     */
    DrawInfo(RevisionData revisionData, Rectangle container, Rectangle handle)
    {
      this.revisionData = revisionData;
      this.container    = container;
      this.handle       = handle;
    }

    /** create draw info
     * @param revisionData revision data
     * @param container container size
     */
    DrawInfo(RevisionData revisionData, Rectangle container)
    {
      this(revisionData,container,(Rectangle)null);
    }
  }

  /** dialog data
   */
  class Data
  {
    RevisionDataTree        revisionDataTree;                           // revision data tree
    RevisionData            lastRevisionData;                           // revision data of last commit
    ArrayList<RevisionData> revisionDataList;                           // revision data list (used for searching)
    Rectangle               view;
    Point                   size;
    ArrayList<DrawInfo>     drawInfoList;
    boolean                 containerResizeFlag;
    Point                   containerResizeStart;
    Point                   containerResizeDelta;
    Rectangle               containerResizeRectangle;
    String                  author;
    RevisionData            selectedRevisionData0,selectedRevisionData1;

    Data()
    {
      this.revisionDataTree         = null;
      this.lastRevisionData         = null;
      this.revisionDataList         = null;
      this.view                     = new Rectangle(0,0,0,0);
      this.size                     = new Point(0,0);
      this.drawInfoList             = new ArrayList<DrawInfo>(128);
      this.containerResizeFlag      = false;
      this.containerResizeStart     = new Point(0,0);
      this.containerResizeDelta     = new Point(0,0);
      this.containerResizeRectangle = new Rectangle(0,0,0,0);
      this.author                   = "";
      this.selectedRevisionData0    = null;
      this.selectedRevisionData1    = null;
    }
  };

  // --------------------------- constants --------------------------------
  private final int MARGIN                       = 10;   // margin left/top for containers in canvas
  private final int PADDING_X                    = 10;   // padding between containers
  private final int PADDING_Y                    = 30;   // padding between containers
  private final int CONTAINER_MARGIN             =  4;   // margin left/top for text in container
  private final int CONTAINER_MIN_WIDTH          = 20;   // min. width of container
  private final int CONTAINER_MIN_HEIGHT         = 20;   // min. height of container
  private final int HANDLE_SIZE                  =  8;   // size of resize-handle rectangle
  private final int DOT_SIZE                     = 20;
  private final int MAX_TEXT_LENGTH              = 30;   // max. text length to show [characters]

  private final Color COLOR_BACKGROUND           = Onzen.COLOR_WHITE;
  private final Color COLOR_CONTAINER            = Onzen.COLOR_GRAY;
  private final Color COLOR_HANDLE               = Onzen.COLOR_RED;
  private final Color COLOR_CONTAINER_NORMAL     = Onzen.COLOR_BLACK;
  private final Color COLOR_CONTAINER_CURRENT    = Onzen.COLOR_GRAY;
  private final Color COLOR_CONTAINER_SELECTED0;
  private final Color COLOR_CONTAINER_SELECTED1;
  private final Color COLOR_TEXT                 = Onzen.COLOR_BLACK;
  private final Color COLOR_LINES                = Onzen.COLOR_RED;
  private final Color COLOR_BRANCH_LINES         = Onzen.COLOR_BLUE;
  private final Color COLOR_MERGE_LINES          = Onzen.COLOR_GREEN;
  private final Color COLOR_INFO_BACKGROUND;
  private final Color COLOR_INFO_FOREGROUND;

  private final int   LINES_WIDTH                = 3;
//  private final int[] LINES_DASH_PATTERN1        = new int[]{1,2,1,2};
//  private final int[] LINES_DASH_PATTERN2        = new int[]{2,2,2};
  private final int[] LINES_DASH_PATTERN1        = new int[]{1};
  private final int[] LINES_DASH_PATTERN2        = new int[]{1};

  private final RevisionData LOCAL_REVISION      = new RevisionData("local");

  // --------------------------- variables --------------------------------

  // global variable references
  private final RepositoryTab repositoryTab;
  private final Display       display;
  private final Clipboard     clipboard;
  private final FileData      fileData;
  private Point               maxTreeSize;
  private int                 maxRevisionIdTextWidth;
  private int                 maxAuthorTextWidth;
  private Point               maxRevisionInfoTextSize;
  private final int           fontHeight;

  // dialog
  private final Data          data = new Data();
  private final Shell         dialog;

  // widgets
  private final Text          widgetSelectedRevision0;
  private final Text          widgetSelectedRevision1;
  private final Canvas        widgetRevisions;
  private final ScrollBar     widgetHorizontalScrollBar;
  private final ScrollBar     widgetVerticalScrollBar;
  private final Text          widgetFind;
  private final Button        widgetFindPrev;
  private final Button        widgetFindNext;
  private final Combo         widgetAuthor;
  private final Button        widgetDiff;
  private final Button        widgetPatch;
  private final Text          widgetSelectedRevision;
  private final Button        widgetView;
  private final Button        widgetSave;
  private final Button        widgetRevert;
  private final Button        widgetClose;
  private Shell               widgetRevisionInfoToolTip = null;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create revision view
   * @param shell shell
   * @param repositoryTab repository tab
   * @param fileData file data
   * @param revision revision to show or null
   */
  CommandRevisions(final Shell shell, final RepositoryTab repositoryTab, final FileData fileData, String revision)
  {
    Composite composite,subComposite;
    Label     label;
    Button    button;

    // initialize variables
    this.repositoryTab = repositoryTab;
    this.fileData      = fileData;

    // get display, clipboard
    display   = shell.getDisplay();
    clipboard = new Clipboard(display);

    // init colors
    COLOR_CONTAINER_SELECTED0 = new Color(display,Settings.colorSelect0.background);
    COLOR_CONTAINER_SELECTED1 = new Color(display,Settings.colorSelect1.background);
    COLOR_INFO_FOREGROUND     = display.getSystemColor(SWT.COLOR_INFO_FOREGROUND);
    COLOR_INFO_BACKGROUND     = display.getSystemColor(SWT.COLOR_INFO_BACKGROUND);

    // create dialog
    dialog = Dialogs.open(shell,"Revisions"+((fileData != null) ? ": "+fileData.getName(repositoryTab.repository) : ""),new double[]{1.0,0.0},1.0);

    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(new double[]{1.0,0.0},1.0));
    Widgets.layout(composite,0,0,TableLayoutData.NSWE,0,0,4);
    {
      // revisions graph
      widgetRevisions = Widgets.newCanvas(composite,SWT.H_SCROLL|SWT.V_SCROLL);
      widgetRevisions.setBackground(Onzen.COLOR_WHITE);
      Widgets.layout(widgetRevisions,0,0,TableLayoutData.NSWE,0,0,4);
      widgetHorizontalScrollBar = widgetRevisions.getHorizontalBar();
      widgetVerticalScrollBar   = widgetRevisions.getVerticalBar();

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(1.0,new double[]{0.0,0.6,0.0,0.0,0.0,0.4,0.0,0.0,0.0,0.0}));
      Widgets.layout(subComposite,2,0,TableLayoutData.WE);
      {
        // find
        label = Widgets.newLabel(subComposite,"Find:",SWT.NONE,Settings.keyFind);
        Widgets.layout(label,0,0,TableLayoutData.W);

        widgetFind = Widgets.newText(subComposite,SWT.SEARCH|SWT.ICON_CANCEL);
        widgetFind.setMessage("Enter text to find");
        Widgets.layout(widgetFind,0,1,TableLayoutData.WE);

        widgetFindPrev = Widgets.newButton(subComposite,Onzen.IMAGE_ARROW_UP);
        widgetFindPrev.setToolTipText("Find previous occurrence of text ["+Widgets.acceleratorToText(Settings.keyFindPrev)+"].");
        widgetFindPrev.setEnabled(false);
        Widgets.layout(widgetFindPrev,0,2,TableLayoutData.NSW);
        Widgets.addModifyListener(new WidgetModifyListener(widgetFindPrev,data)
        {
          public void modified(Control control)
          {
            Widgets.setEnabled(control,(data.revisionDataTree != null) && (data.revisionDataList != null));
          }
        });

        widgetFindNext = Widgets.newButton(subComposite,Onzen.IMAGE_ARROW_DOWN);
        widgetFindNext.setToolTipText("Find next occurrence of text  ["+Widgets.acceleratorToText(Settings.keyFindNext)+"].");
        widgetFindNext.setEnabled(false);
        Widgets.layout(widgetFindNext,0,3,TableLayoutData.NSW);
        Widgets.addModifyListener(new WidgetModifyListener(widgetFindNext,data)
        {
          public void modified(Control control)
          {
            Widgets.setEnabled(control,(data.revisionDataTree != null) && (data.revisionDataList != null));
          }
        });

        // author filter
        label = Widgets.newLabel(subComposite,"Author:");
        Widgets.layout(label,0,4,TableLayoutData.W);

        widgetAuthor = Widgets.newCombo(subComposite,SWT.READ_ONLY);
        Widgets.layout(widgetAuthor,0,5,TableLayoutData.WE);

        // legend
        label = Widgets.newLabel(subComposite,"");
        label.setBackground(COLOR_BRANCH_LINES);
        Widgets.layout(label,0,6,TableLayoutData.W,0,0,0,0,20,20);
        label = Widgets.newLabel(subComposite,"Branch");
        Widgets.layout(label,0,7,TableLayoutData.W);

        label = Widgets.newLabel(subComposite,"");
        label.setBackground(COLOR_MERGE_LINES);
        Widgets.layout(label,0,8,TableLayoutData.W,0,0,0,0,20,20);
        label = Widgets.newLabel(subComposite,"Merge");
        Widgets.layout(label,0,9,TableLayoutData.W);
      }
    }

    // settings
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0,0.0}));
    Widgets.layout(composite,1,0,TableLayoutData.WE);
    {
      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
      Widgets.layout(subComposite,0,0,TableLayoutData.W);
      {
        button = Widgets.newCheckbox(subComposite,"revision");
        button.setToolTipText("Set graph style.");
        button.setSelection(Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.REVISION));
        button.setEnabled(   (Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS)
                          || (Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST)
                         );
        Widgets.layout(button,0,0,TableLayoutData.W);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {

            Button button = (Button)selectionEvent.widget;

            if (button.getSelection())
            {
              Settings.revisionGraphTitle.add(Settings.RevisionGraphTitles.REVISION);
            }
            else
            {
              Settings.revisionGraphTitle.remove(Settings.RevisionGraphTitles.REVISION);
            }

            // re-set canvas size
            setSize();

            // redraw
            redraw();
          }
        });
        Widgets.addModifyListener(new WidgetModifyListener(button,Settings.class)
        {
          public void modified(Control control)
          {
            Widgets.setEnabled(control,
                                  (Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS)
                               || (Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST)
                              );
          }
        });

        button = Widgets.newCheckbox(subComposite,"author");
        button.setToolTipText("Set graph style.");
        button.setSelection(Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.AUTHOR));
        button.setEnabled(   (Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS)
                          || (Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST)
                         );
        Widgets.layout(button,0,2,TableLayoutData.W);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button button = (Button)selectionEvent.widget;

            if (button.getSelection())
            {
              Settings.revisionGraphTitle.add(Settings.RevisionGraphTitles.AUTHOR);
            }
            else
            {
              Settings.revisionGraphTitle.remove(Settings.RevisionGraphTitles.AUTHOR);
            }

            // re-set canvas size
            setSize();

            // redraw
            redraw();
          }
        });
        Widgets.addModifyListener(new WidgetModifyListener(button,Settings.class)
        {
          public void modified(Control control)
          {
            Widgets.setEnabled(control,
                                  (Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS)
                               || (Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST)
                              );
          }
        });

        button = Widgets.newCheckbox(subComposite,"message");
        button.setToolTipText("Set graph style.");
        button.setSelection(Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.MESSAGE));
        button.setEnabled(   (Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS)
                          || (Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST)
                         );
        Widgets.layout(button,0,1,TableLayoutData.W);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button button = (Button)selectionEvent.widget;

            if (button.getSelection())
            {
              Settings.revisionGraphTitle.add(Settings.RevisionGraphTitles.MESSAGE);
            }
            else
            {
              Settings.revisionGraphTitle.remove(Settings.RevisionGraphTitles.MESSAGE);
            }

            // re-set canvas size
            setSize();

            // redraw
            redraw();
          }
        });
        Widgets.addModifyListener(new WidgetModifyListener(button,Settings.class)
        {
          public void modified(Control control)
          {
            Widgets.setEnabled(control,
                                  (Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS)
                               || (Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST)
                              );
          }
        });
      }

      subComposite = Widgets.newComposite(composite);
      subComposite.setLayout(new TableLayout(1.0,new double[]{0.0,1.0}));
      Widgets.layout(subComposite,0,1,TableLayoutData.E);
      {
        button = Widgets.newRadio(subComposite,"boxes");
        button.setToolTipText("Set boxes style.");
        button.setSelection(Settings.revisionGraphType == Settings.RevisionGraphTypes.BOXES);
        Widgets.layout(button,0,0,TableLayoutData.E);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button button = (Button)selectionEvent.widget;

            Settings.revisionGraphType = Settings.RevisionGraphTypes.BOXES;

            // re-set canvas size
            setSize();

            // redraw
            redraw();
          }
        });

        button = Widgets.newRadio(subComposite,"dots");
        button.setToolTipText("Set dots style.");
        button.setSelection(Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS);
        Widgets.layout(button,0,1,TableLayoutData.E);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button button = (Button)selectionEvent.widget;

            Settings.revisionGraphType = Settings.RevisionGraphTypes.DOTS;

            // re-set canvas size
            setSize();

            // redraw
            redraw();
          }
        });

        button = Widgets.newRadio(subComposite,"list");
        button.setToolTipText("Set list style.");
        button.setSelection(Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST);
        Widgets.layout(button,0,2,TableLayoutData.E);
        button.addSelectionListener(new SelectionListener()
        {
          public void widgetDefaultSelected(SelectionEvent selectionEvent)
          {
          }
          public void widgetSelected(SelectionEvent selectionEvent)
          {
            Button button = (Button)selectionEvent.widget;

            Settings.revisionGraphType = Settings.RevisionGraphTypes.LIST;

            // re-set canvas size
            setSize();

            // redraw
            redraw();
          }
        });
      }

      button = Widgets.newCheckbox(composite,"branches/merges only");
      button.setToolTipText("Show only branches.");
      button.setSelection(Settings.revisionGraphBranchesMergesOnly);
      Widgets.layout(button,0,3,TableLayoutData.E);
      button.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button button = (Button)selectionEvent.widget;

          Settings.revisionGraphBranchesMergesOnly = button.getSelection();

          // re-set canvas size
          setSize();

          // redraw
          redraw();
        }
      });
    }

    // buttons
    composite = Widgets.newComposite(dialog);
    composite.setLayout(new TableLayout(0.0,new double[]{0.0,1.0,0.0,1.0,0.0,0.0,0.0,0.0,1.0,0.0,0.0,0.0,0.0,0.0}));
    Widgets.layout(composite,2,0,TableLayoutData.WE,0,0,4);
    {
      label = Widgets.newLabel(composite,"For");
      Widgets.layout(label,0,0,TableLayoutData.W);

      widgetSelectedRevision0 = Widgets.newStringView(composite);
      widgetSelectedRevision0.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetSelectedRevision0,0,1,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSelectedRevision0,data)
      {
        public void modified(Control control)
        {
          if (!widgetSelectedRevision0.isDisposed()) widgetSelectedRevision0.setText((data.selectedRevisionData0 != null) ? data.selectedRevisionData0.revisionId : "");
        }
      });

      label = Widgets.newLabel(composite,"->");
      Widgets.layout(label,0,2,TableLayoutData.W);

      widgetSelectedRevision1 = Widgets.newStringView(composite);
      widgetSelectedRevision1.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetSelectedRevision1,0,3,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSelectedRevision1,data)
      {
        public void modified(Control control)
        {
          if (!widgetSelectedRevision1.isDisposed()) widgetSelectedRevision1.setText((data.selectedRevisionData1 != null) ? data.selectedRevisionData1.revisionId : "");
        }
      });

      label = Widgets.newLabel(composite,"do:");
      Widgets.layout(label,0,4,TableLayoutData.W);

      widgetDiff = Widgets.newButton(composite,"Diff",Settings.keyDiff);
      widgetDiff.setEnabled(false);
      Widgets.layout(widgetDiff,0,5,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetDiff,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(data.selectedRevisionData0 != null) && (data.selectedRevisionData1 != null));
        }
      });
      widgetDiff.addSelectionListener(new SelectionListener()
      {
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          CommandDiff commandDiff = new CommandDiff(dialog,
                                                    repositoryTab,
                                                    fileData,
                                                    ((data.selectedRevisionData0 != null) && (data.selectedRevisionData0 != LOCAL_REVISION)) ? data.selectedRevisionData0.revisionId : null,
                                                    ((data.selectedRevisionData1 != null) && (data.selectedRevisionData1 != LOCAL_REVISION)) ? data.selectedRevisionData1.revisionId : null
                                                   );
          commandDiff.run();
        }
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
      });

      widgetPatch = Widgets.newButton(composite,"Patch");
      widgetPatch.setEnabled(false);
      Widgets.layout(widgetPatch,0,6,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetPatch,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(data.selectedRevisionData0 != null) && (data.selectedRevisionData1 != null));
        }
      });
      widgetPatch.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          if (data.selectedRevisionData1 != null)
          {
            CommandCreatePatch commandCreatePatch;
            if (data.selectedRevisionData0 != null)
            {
              commandCreatePatch = new CommandCreatePatch(dialog,
                                                          repositoryTab,
                                                          fileData.toSet(),
                                                          data.selectedRevisionData0.revisionId,
                                                          data.selectedRevisionData1.revisionId,
                                                          false
                                                         );
            }
            else
            {
              commandCreatePatch = new CommandCreatePatch(dialog,
                                                          repositoryTab,
                                                          fileData.toSet(),
                                                          data.selectedRevisionData1.revisionId,
                                                          false
                                                         );
            }
            commandCreatePatch.run();
          }
        }
      });

      label = Widgets.newLabel(composite,"For");
      Widgets.layout(label,0,7,TableLayoutData.W);

      widgetSelectedRevision = Widgets.newStringView(composite);
      widgetSelectedRevision.setBackground(Onzen.COLOR_GRAY);
      Widgets.layout(widgetSelectedRevision,0,8,TableLayoutData.WE);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSelectedRevision,data)
      {
        public void modified(Control control)
        {
          if (!widgetSelectedRevision.isDisposed()) widgetSelectedRevision.setText((data.selectedRevisionData1 != null) ? data.selectedRevisionData1.revisionId : "");
        }
      });

      label = Widgets.newLabel(composite,"do:");
      Widgets.layout(label,0,9,TableLayoutData.W);

      widgetView = Widgets.newButton(composite,"View");
      widgetView.setEnabled(false);
      Widgets.layout(widgetView,0,10,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetView,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,data.selectedRevisionData1 != null);
        }
      });
      widgetView.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          if (data.selectedRevisionData1 != null)
          {
            CommandView commandView = new CommandView(dialog,repositoryTab,fileData,data.selectedRevisionData1.revisionId);
            commandView.run();
          }
        }
      });

      widgetSave = Widgets.newButton(composite,"Save...");
      widgetSave.setEnabled(false);
      Widgets.layout(widgetSave,0,11,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetSave,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(data.selectedRevisionData1 != null));
        }
      });
      widgetSave.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          if (data.selectedRevisionData1 != null)
          {
            try
            {
              // get file
              byte[] fileDataBytes =  repositoryTab.repository.getFileBytes(fileData,data.selectedRevisionData1.revisionId);

              // save to file
              String fileName = Dialogs.fileSave(dialog,"Save file");
              if (fileName != null)
              {
                try
                {
                  DataOutputStream output = new DataOutputStream(new FileOutputStream(fileName));
                  output.write(fileDataBytes);
                  output.close();
                }
                catch (IOException exception)
                {
                  Dialogs.error(dialog,"Cannot save file '%s' (error: %s)",fileData.getName(repositoryTab.repository),exception.getMessage());
                }
              }
            }
            catch (RepositoryException exception)
            {
              Dialogs.error(dialog,"Cannot save file '%s' (error: %s)",fileData.getName(repositoryTab.repository),exception.getMessage());
            }
          }
        }
      });

      widgetRevert = Widgets.newButton(composite,"Revert...",Settings.keyRevert);
      widgetRevert.setEnabled(false);
      Widgets.layout(widgetRevert,0,12,TableLayoutData.W,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      Widgets.addModifyListener(new WidgetModifyListener(widgetRevert,data)
      {
        public void modified(Control control)
        {
          Widgets.setEnabled(control,(data.selectedRevisionData1 != null));
        }
      });
      widgetRevert.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          if (data.selectedRevisionData1 != null)
          {
            if (Dialogs.confirm(dialog,String.format("Revert file '%s' to revision %s?",fileData.getName(repositoryTab.repository),data.selectedRevisionData1.revisionId)))
            {
              try
              {
                // revert file
                repositoryTab.repository.revert(fileData,data.selectedRevisionData1.revisionId);

                // update state
                repositoryTab.updateFileState(fileData);
              }
              catch (RepositoryException exception)
              {
                Dialogs.error(dialog,"Cannot revert file '%s' (error: %s)",fileData.getName(repositoryTab.repository),exception.getMessage());
              }
            }
          }
        }
      });

      widgetClose = Widgets.newButton(composite,"Close");
      Widgets.layout(widgetClose,0,13,TableLayoutData.E,0,0,0,0,SWT.DEFAULT,SWT.DEFAULT,70,SWT.DEFAULT);
      widgetClose.addSelectionListener(new SelectionListener()
      {
        public void widgetDefaultSelected(SelectionEvent selectionEvent)
        {
        }
        public void widgetSelected(SelectionEvent selectionEvent)
        {
          Button widget = (Button)selectionEvent.widget;

          Settings.geometryRevisions = dialog.getSize();

          Dialogs.close(dialog);
        }
      });
    }

    // listeners
    widgetRevisions.addPaintListener(new PaintListener()
    {
      public void paintControl(PaintEvent paintEvent)
      {
        redraw();
      }
    });
    widgetRevisions.addMouseTrackListener(new MouseTrackListener()
    {
      @Override
      public void mouseEnter(MouseEvent mouseEvent)
      {
      }

      @Override
      public void mouseExit(MouseEvent mouseEvent)
      {
      }

      @Override
      public void mouseHover(MouseEvent mouseEvent)
      {
        Canvas widget = (Canvas)mouseEvent.widget;

        if (widgetRevisionInfoToolTip != null)
        {
          widgetRevisionInfoToolTip.dispose();
          widgetRevisionInfoToolTip = null;
        }

        if (   (Settings.revisionGraphType == Settings.RevisionGraphTypes.DOTS)
            || (Settings.revisionGraphType == Settings.RevisionGraphTypes.LIST)
           )
        {
          RevisionData revisionData = getRevisionAt(mouseEvent.x,mouseEvent.y);
          if (revisionData != null)
          {
            Point point = display.getCursorLocation();
            if (point.x > 16) point.x -= 16;
            if (point.y > 16) point.y -= 16;

            showRevisionInfoToolTip(revisionData,point.x,point.y);
          }
        }
      }
    });
    widgetHorizontalScrollBar.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        ScrollBar widget = (ScrollBar)selectionEvent.widget;

        // scroll
        int n = widget.getSelection();
        int dx = -n-data.view.x;
        widgetRevisions.scroll(dx,0,0,0,data.size.x,data.size.y,false);
        widgetRevisions.redraw();

        // save origin
        data.view.x = -n;
      }
    });
    widgetVerticalScrollBar.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        ScrollBar widget = (ScrollBar)selectionEvent.widget;

        // scroll
        int n = widget.getSelection();
        int dy = -n-data.view.y;
        widgetRevisions.scroll(0,dy,0,0,data.size.x,data.size.y,false);
        widgetRevisions.redraw();

        // save origin
        data.view.y = -n;
      }
    });
    widgetRevisions.addListener(SWT.Resize,new Listener()
    {
      public void handleEvent(Event event)
      {
        Rectangle clientArea = widgetRevisions.getClientArea ();

        data.view.width  = clientArea.width;
        data.view.height = clientArea.height;

        widgetHorizontalScrollBar.setThumb(Math.min(data.view.width,data.size.x));
        widgetVerticalScrollBar.setThumb(Math.min(data.view.height,data.size.y));
        widgetRevisions.redraw();
      }
    });
    widgetRevisions.addMouseListener(new MouseListener()
    {
      public void mouseDoubleClick(MouseEvent mouseEvent)
      {
        for (DrawInfo drawInfo : data.drawInfoList)
        {
          if (   drawInfo.container.contains(mouseEvent.x,mouseEvent.y)
              && (drawInfo.revisionData != null)
             )
          {
            CommandRevisionInfo commandRevisionInfo = new CommandRevisionInfo(dialog,
                                                                              repositoryTab,
                                                                              fileData,
                                                                              (drawInfo.revisionData != LOCAL_REVISION) ? drawInfo.revisionData.revisionId : null
                                                                             );
            commandRevisionInfo.run();
            break;
          }
        }
      }
      public void mouseDown(MouseEvent mouseEvent)
      {
        if (!data.containerResizeFlag)
        {
          for (DrawInfo drawInfo : data.drawInfoList)
          {
            if      (drawInfo.container.contains(mouseEvent.x,mouseEvent.y))
            {
              // selected revision
              data.selectedRevisionData0 = data.selectedRevisionData1;
              data.selectedRevisionData1 = drawInfo.revisionData;

              // notify modification
              Widgets.modified(data);

              widgetRevisions.redraw();
              break;
            }
            else if ((drawInfo.handle != null) && drawInfo.handle.contains(mouseEvent.x,mouseEvent.y))
            {
              // start resize container
              data.containerResizeStart.x          = mouseEvent.x;
              data.containerResizeStart.y          = mouseEvent.y;
              data.containerResizeDelta.x          = 0;
              data.containerResizeDelta.y          = 0;
              data.containerResizeRectangle.x      = drawInfo.container.x;
              data.containerResizeRectangle.y      = drawInfo.container.y;
              data.containerResizeRectangle.width  = drawInfo.container.width;
              data.containerResizeRectangle.height = drawInfo.container.height;
              data.containerResizeFlag             = true;
              break;
            }
          }
        }
      }
      public void mouseUp(MouseEvent mouseEvent)
      {
        if (data.containerResizeFlag)
        {
          // stop resize container
          data.containerResizeDelta.x = mouseEvent.x-data.containerResizeStart.x;
          data.containerResizeDelta.y = mouseEvent.y-data.containerResizeStart.y;
          data.containerResizeFlag    = false;

          // store new box size
          Settings.geometryRevisionBox.x = Math.max(Settings.geometryRevisionBox.x+data.containerResizeDelta.x,CONTAINER_MIN_WIDTH );
          Settings.geometryRevisionBox.y = Math.max(Settings.geometryRevisionBox.y+data.containerResizeDelta.y,CONTAINER_MIN_HEIGHT);

          // set canvas size
          setSize();

          // redraw
          widgetRevisions.redraw();
        }
      }
    });
    widgetRevisions.addMouseMoveListener(new MouseMoveListener()
    {
      public void mouseMove(MouseEvent mouseEvent)
      {
        if (data.containerResizeFlag)
        {
          data.containerResizeDelta.x = mouseEvent.x-data.containerResizeStart.x;
          data.containerResizeDelta.y = mouseEvent.y-data.containerResizeStart.y;
          redraw(data.containerResizeDelta.x,data.containerResizeDelta.y);
        }
      }
    });
    widgetRevisions.addKeyListener(new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (Widgets.isAccelerator(keyEvent,Settings.keyFind))
        {
          widgetFind.forceFocus();
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyFindPrev))
        {
          Widgets.invoke(widgetFindPrev);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyFindNext))
        {
          Widgets.invoke(widgetFindNext);
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    });
    widgetFind.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        RevisionData revisionData = findNext(widgetFind);
        if (revisionData != null)
        {
          // select revision and show
          data.selectedRevisionData0 = null;
          data.selectedRevisionData1 = revisionData;
          scrollTo(revisionData);

          // notify modification
          Widgets.modified(data);
        }
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
      }
    });
    widgetFindPrev.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        RevisionData revisionData = findPrev(widgetFind);
        if (revisionData != null)
        {
          // select revision and show
          data.selectedRevisionData0 = null;
          data.selectedRevisionData1 = revisionData;
          scrollTo(revisionData);

          // notify modification
          Widgets.modified(data);
        }
      }
    });
    widgetFindNext.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        RevisionData revisionData = findNext(widgetFind);
        if (revisionData != null)
        {
          // select revision and show
          data.selectedRevisionData0 = null;
          data.selectedRevisionData1 = revisionData;
          scrollTo(revisionData);

          // notify modification
          Widgets.modified(data);
        }
      }
    });
    widgetAuthor.addSelectionListener(new SelectionListener()
    {
      public void widgetDefaultSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        data.author = widget.getText().trim();

        // notify modification
        Widgets.modified(data);
      }
      public void widgetSelected(SelectionEvent selectionEvent)
      {
        Combo widget = (Combo)selectionEvent.widget;

        data.author = widget.getText().trim();

        // redraw
        redraw();
      }
    });

    KeyListener keyListener = new KeyListener()
    {
      public void keyPressed(KeyEvent keyEvent)
      {
        if      (((keyEvent.stateMask & SWT.MODIFIER_MASK) == SWT.NONE) && (keyEvent.keyCode == SWT.PAGE_UP))
        {
          // goto one page up
          int pageIncrement = widgetVerticalScrollBar.getPageIncrement();
          int n = widgetVerticalScrollBar.getSelection();
          n = (n-pageIncrement > 0) ? n-pageIncrement : 0;
          widgetVerticalScrollBar.setSelection(n);

          // save origin
          data.view.y = -n;

          // redraw
          widgetRevisions.redraw();
        }
        else if (((keyEvent.stateMask & SWT.MODIFIER_MASK) == SWT.NONE) && (keyEvent.keyCode == SWT.PAGE_DOWN))
        {
          // goto one page down
          int pageIncrement = widgetVerticalScrollBar.getPageIncrement();
          int maximum = widgetVerticalScrollBar.getMaximum();
          int n = widgetVerticalScrollBar.getSelection();
          n = (n+pageIncrement < maximum-pageIncrement) ? n+pageIncrement : maximum-pageIncrement;
          widgetVerticalScrollBar.setSelection(n);

          // save origin
          data.view.y = -n;

          // redraw
          widgetRevisions.redraw();
        }
        else if (((keyEvent.stateMask & SWT.MODIFIER_MASK) == SWT.NONE) && (keyEvent.keyCode == SWT.HOME))
        {
          // goto top
          widgetVerticalScrollBar.setSelection(0);

          // save origin
          data.view.y = 0;

          // redraw
          widgetRevisions.redraw();
        }
        else if (((keyEvent.stateMask & SWT.MODIFIER_MASK) == SWT.NONE) && (keyEvent.keyCode == SWT.END))
        {
          // goto bottom
          int pageIncrement = widgetVerticalScrollBar.getPageIncrement();
          int n = widgetVerticalScrollBar.getMaximum()-pageIncrement;
          widgetVerticalScrollBar.setSelection(n);

          // save origin
          data.view.y = -n;

          // redraw
          widgetRevisions.redraw();
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyFind))
        {
          widgetFind.forceFocus();
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyFindPrev))
        {
          Widgets.invoke(widgetFindPrev);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyFindNext))
        {
          Widgets.invoke(widgetFindNext);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyDiff))
        {
          Widgets.invoke(widgetDiff);
        }
        else if (Widgets.isAccelerator(keyEvent,Settings.keyRevert))
        {
          Widgets.invoke(widgetRevert);
        }
      }
      public void keyReleased(KeyEvent keyEvent)
      {
      }
    };
    widgetRevisions.addKeyListener(keyListener);
    widgetFind.addKeyListener(keyListener);

    // get font height
    fontHeight = Widgets.getTextHeight(widgetRevisions);

    // show dialog
    Dialogs.show(dialog,Settings.geometryRevisions,Settings.setWindowLocation);

    // show
    show(revision);
  }

  /** create revision view
   * @param shell shell
   * @param repositoryTab repository tab
   * @param fileData file data
   */
  CommandRevisions(Shell shell, RepositoryTab repositoryTab, FileData fileData)
  {
    this(shell,repositoryTab,fileData,(fileData != null) ? fileData.workingRevision : (String)null);
  }

  /** set scroll value
   * @param scrollX,scrollY scroll values
   */
  public void setScroll(int scrollX, int scrollY)
  {
    widgetRevisions.getHorizontalBar().setSelection(scrollX);
    widgetRevisions.getVerticalBar().setSelection(scrollY);
    widgetRevisions.redraw();
  }

  /** run dialog
   */
  public void run()
  {
    widgetFind.setFocus();
    Dialogs.run(dialog);
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "CommandRevisions {}";
  }

  //-----------------------------------------------------------------------

  /** get max. width of revision id text
   * @param revisionDataTree revision data tree
   * @return with of revision id text [pixel]
   */
  private int getMaxRevisionIdTextWidth(RevisionDataTree revisionDataTree)
  {
    int maxWidth = 0;

    int width;
    for (RevisionData revisionData : revisionDataTree)
    {
//      if (   (!Settings.revisionGraphBranchesMergesOnly || revisionData.isBranchedOrMerged())
//          && (data.author.isEmpty() || revisionData.isFromAuthor(data.author))
//         )
      {
        width = Widgets.getTextWidth(widgetRevisions,revisionData.revisionId);
        maxWidth = Math.max(maxWidth,width);

        if (!revisionData.branchDataList.isEmpty())
        {
          for (BranchData branchData : revisionData.branchDataList)
          {
            width = getMaxAuthorTextWidth(branchData.revisionDataTree);
            maxWidth = Math.max(maxWidth,width);
          }
        }
      }
    }

    return maxWidth;
  }

  /** get max. width of author text
   * @param revisionDataTree revision data tree
   * @return with of author text [pixel]
   */
  private int getMaxAuthorTextWidth(RevisionDataTree revisionDataTree)
  {
    int maxWidth = 0;

    int width;
    for (RevisionData revisionData : revisionDataTree)
    {
//      if (   (!Settings.revisionGraphBranchesMergesOnly || revisionData.isBranchedOrMerged())
//          && (data.author.isEmpty() || revisionData.isFromAuthor(data.author))
//         )
      {
        width = Widgets.getTextWidth(widgetRevisions,revisionData.author);
        maxWidth = Math.max(maxWidth,width);

        if (!revisionData.branchDataList.isEmpty())
        {
          for (BranchData branchData : revisionData.branchDataList)
          {
            width = getMaxAuthorTextWidth(branchData.revisionDataTree);
            maxWidth = Math.max(maxWidth,width);
          }
        }
      }
    }

    return maxWidth;
  }

  /** get revision info text
   * @param revisionData revision data
   * @param maxTextLength max. text length [characters] or 0
   * @return revision info text
   */
  private String[] getRevisionInfoText(RevisionData revisionData, int maxTextLength)
  {
    String[] text = new String[2];

    StringBuilder buffer = new StringBuilder();
    if (Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.REVISION))
    {
      buffer.append(revisionData.revisionId);
    }
    if (Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.AUTHOR))
    {
      if (buffer.length() > 0) buffer.append(' ');
      buffer.append(String.format("(%s)",revisionData.author));
    }
    text[0] = buffer.toString();
    if ((maxTextLength > 0) && (text[0].length() > maxTextLength)) text[0] = text[0].substring(0,maxTextLength)+"\u2026";

    if (Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.MESSAGE))
    {
      text[1] = ((revisionData.commitMessageList != null) && (revisionData.commitMessageList.size() > 0)) ? revisionData.commitMessageList.getFirst() : "";
      if ((maxTextLength > 0) && (text[1].length() > maxTextLength)) text[1] = text[1].substring(0,maxTextLength)+"\u2026";
    }
    else
    {
      text[1] = "";
    }

    return text;
  }

  /** get revision info text
   * @param revisionData revision data
   * @return revision info text
   */
  private String[] getRevisionInfoText(RevisionData revisionData)
  {
    return getRevisionInfoText(revisionData,0);
  }

  /** get max. size of revision tree info text
   * @param revisionData revision data
   * @return size of revision info [pixel]
   */
  private Point getMaxRevisionInfoTextSize(RevisionData revisionData)
  {
    Point size;
    if (data.author.isEmpty() || revisionData.isFromAuthor(data.author))
    {
      size = Widgets.getTextSize(widgetRevisions,getRevisionInfoText(revisionData,MAX_TEXT_LENGTH));
    }
    else
    {
      size = new Point(0,0);
    }

    return size;
  }

  /** get max. size of revision tree info text
   * @param revisionDataTree revision data tree
   * @return size of revision info [pixel]
   */
  private Point getMaxRevisionInfoTextSize(RevisionDataTree revisionDataTree)
  {
    Point maxSize = new Point(0,0);

    Point size;
    for (RevisionData revisionData : revisionDataTree)
    {
      if (   (!Settings.revisionGraphBranchesMergesOnly || revisionData.isBranchedOrMerged())
          && (data.author.isEmpty() || revisionData.isFromAuthor(data.author))
         )
      {
        size = Widgets.getTextSize(widgetRevisions,getRevisionInfoText(revisionData,MAX_TEXT_LENGTH));
        maxSize.x = Math.max(maxSize.x,size.x);

        if (!revisionData.branchDataList.isEmpty())
        {
          for (BranchData branchData : revisionData.branchDataList)
          {
            if (branchData.name != null)
            {
              maxSize.x = Math.max(maxSize.x,
                                   Widgets.getTextWidth(widgetRevisions,branchData.name)
                                  );
            }
            size = getMaxRevisionInfoTextSize(branchData.revisionDataTree);
            maxSize.x = Math.max(maxSize.x,size.x);
            maxSize.y = Math.max(maxSize.y,size.y);
          }
        }
      }
    }

    return maxSize;
  }

  /** get size of revision tree
   * @param revisionDataTree revision data tree
   * @return size of revision tree [pixel]
   */
  private Point getTreeSize(RevisionDataTree revisionDataTree)
  {
    // get item width/height
    final int width,height;
    switch (Settings.revisionGraphType)
    {
      case BOXES:
        width  = Math.max(Settings.geometryRevisionBox.x,CONTAINER_MIN_WIDTH );
        height = Math.max(Settings.geometryRevisionBox.y,CONTAINER_MIN_HEIGHT);
        break;
      case DOTS:
        width  = CONTAINER_MARGIN+DOT_SIZE+maxRevisionInfoTextSize.x+CONTAINER_MARGIN;
        height = DOT_SIZE;
        break;
      case LIST:
        // Note: only tree (no text)
        width  = CONTAINER_MARGIN+DOT_SIZE+CONTAINER_MARGIN;
        height = Math.max(DOT_SIZE,fontHeight);
        break;
      default:
        width  = 0;
        height = 0;
        break;
    }

    Point treeSize = new Point(0,0);

    boolean firstFlag = true;
    for (RevisionData revisionData : revisionDataTree)
    {
      if (   (!Settings.revisionGraphBranchesMergesOnly || firstFlag || revisionData.isBranchedOrMerged())
          && (data.author.isEmpty() || revisionData.isFromAuthor(data.author))
         )
      {
        if (!Settings.revisionGraphBranchesMergesOnly || revisionData.isBranchedOrMerged()) treeSize.y += PADDING_Y;

        treeSize.x = Math.max(treeSize.x,width);
        if (!Settings.revisionGraphBranchesMergesOnly || revisionData.isBranchedOrMerged()) treeSize.y += height;

        if (!revisionData.branchDataList.isEmpty())
        {
          Point maxSubTreeSize = new Point(width,0);
          for (BranchData branchData : revisionData.branchDataList)
          {
            Point subTreeSize = getTreeSize(branchData.revisionDataTree);
            maxSubTreeSize.x += PADDING_X+subTreeSize.x;
            maxSubTreeSize.y = Math.max(maxSubTreeSize.y,subTreeSize.y);
          }

          // next column/row
          treeSize.x = Math.max(treeSize.x,maxSubTreeSize.x);
          if (!Settings.revisionGraphBranchesMergesOnly || revisionData.isBranchedOrMerged()) treeSize.y += maxSubTreeSize.y;
        }

        firstFlag = false;
      }
    }

    // add space for local revision
    treeSize.y += PADDING_Y+height;

    return treeSize;
  }

  /** get next column
   * @param x current x-position
   * @param subSize size of sub-tree
   * @return next x-position
   */
  private int getNextColumn(int x, Point subSize)
  {
    switch (Settings.revisionGraphType)
    {
      case BOXES:
        x = x+PADDING_X+subSize.x;
        break;
      case DOTS:
        x = x+PADDING_X+subSize.x;
        break;
      case LIST:
        x = x+PADDING_X+subSize.x;
        break;
      default:
        break;
    }

    return x;
  }

  /** get next row
   * @param y current y-position
   * @param subSize size of sub-tree
   * @return next y-position
   */
  private int getNextRow(int y, Point subSize)
  {
    switch (Settings.revisionGraphType)
    {
      case BOXES:
        y = Math.max(y,PADDING_Y+subSize.y);
        break;
      case DOTS:
        y = Math.max(y,PADDING_Y+subSize.y);
        break;
      case LIST:
        y = y+PADDING_Y+subSize.y;
        break;
      default:
        break;
    }

    return y;
  }

  /** get closest revision in revision tree
   * @param revisionDataTree revision data tree
   * @param revisionId revision id string
   * @return closest revision data or null
   */
  private RevisionData getClosestRevision(RevisionDataTree revisionDataTree, String revisionId)
  {
    RevisionData closestRevisionData = null;

    if (revisionId != null)
    {
      if (revisionId.equals(repositoryTab.repository.getLastRevision()))
      {
        closestRevisionData = (revisionDataTree.size() > 0) ? revisionDataTree.last() : null;
      }
      else
      {
        for (RevisionData revisionData : revisionDataTree)
        {
          if (   (revisionData.revisionId.compareTo(revisionId) <= 0)
              && ((closestRevisionData == null) || (revisionData.revisionId.compareTo(closestRevisionData.revisionId) > 0))
             )
          {
            closestRevisionData = revisionData;
          }
          else
          {
            if (!revisionData.branchDataList.isEmpty())
            {
              for (BranchData branchData : revisionData.branchDataList)
              {
                RevisionData subRevisionData = getClosestRevision(branchData.revisionDataTree,revisionId);
                if (subRevisionData != null)
                {
                  closestRevisionData = subRevisionData;
                }
              }
            }
          }
        }
      }
    }

    return closestRevisionData;
  }

  /** get next revision in revision tree
   * @param revisionDataTree revision data tree
   * @param revisionData revision data
   * @return next revision data or null
   */
  private RevisionData getNextRevision(RevisionDataTree revisionDataTree, RevisionData revisionData)
  {
    RevisionData nextRevisionData = null;

    Iterator<RevisionData> iterator = revisionDataTree.iterator();
    while (iterator.hasNext())
    {
      if (iterator.next() == revisionData)
      {
        nextRevisionData = iterator.next();
        break;
      }
      else
      {
        if (!revisionData.branchDataList.isEmpty())
        {
          for (BranchData branchData : revisionData.branchDataList)
          {
            RevisionData subRevisionData = getNextRevision(branchData.revisionDataTree,revisionData);
            if (subRevisionData != null)
            {
              nextRevisionData = subRevisionData;
              break;
            }
          }
        }
      }
    }

    return nextRevisionData;
  }

  /** get origin (x0,y0) of revision in revision tree
   * @param revisionDataTree revision data tree
   * @param revisionId revision id string
   * @return point [pixel] or null
   */
  private Point getRevisionX0Y0(RevisionDataTree revisionDataTree, String revisionId)
  {
    final int width  = Settings.geometryRevisionBox.x;
    final int height = Settings.geometryRevisionBox.y;

    Point point = new Point(0,0);

    boolean firstFlag = true;
    for (RevisionData revisionData : revisionDataTree)
    {
      if (!Settings.revisionGraphBranchesMergesOnly || revisionData.isBranchedOrMerged())
      {
        if (revisionData.revisionId.equals(revisionId))
        {
          return point;
        }
        else
        {
          point.y += PADDING_Y;

          point.y += height;
//Dprintf.dprintf("revisionData=%s: point=%s",revisionData,point);

          if (!revisionData.branchDataList.isEmpty())
          {
            Point maxSubSize = new Point(0,0);
            for (BranchData branchData : revisionData.branchDataList)
            {
              Point subPoint = getRevisionX0Y0(branchData.revisionDataTree,revisionId);
              if (subPoint != null)
              {
                point.x += width+PADDING_X+subPoint.x;
                point.y += subPoint.y;

                return point;
              }
              else
              {
                Point subTreeSize = getTreeSize(branchData.revisionDataTree);

                maxSubSize.x = Math.max(maxSubSize.x,PADDING_X+subTreeSize.x);
                maxSubSize.y = Math.max(maxSubSize.y,subTreeSize.y);
              }
            }
//Dprintf.dprintf("maxSubSize=%s",maxSubSize);

            // next column, get max. dy
//            point.x += Math.max(point.x,width+PADDING_X+maxSubSize.x);
            point.y += maxSubSize.y;
          }

          firstFlag = false;
        }
      }
    }

    return null;
  }

  /** get origin (x0,y0) of revision in revision tree
   * @param revisionId revision id string
   * @return point (in pixel) or null
   */
  private Point getRevisionX0Y0(String revisionId)
  {
    return getRevisionX0Y0(data.revisionDataTree,revisionId);
  }

  /** get origin (x0,y0) of revision in revision tree
   * @param revisionDataTree revision data tree
   * @param revisionData revision data
   * @return point (in pixel) or null
   */
  private Point getRevisionX0Y0(RevisionData revisionData)
  {
    return getRevisionX0Y0(revisionData.revisionId);
  }

  /** get revsion at position
   * @param x,y position
   * @return revision or null
   */
  private RevisionData getRevisionAt(RevisionDataTree revisionDataTree, int x, int y)
  {
    for (DrawInfo drawInfo : data.drawInfoList)
    {
      if (   (drawInfo.container.contains(x,y))
          || ((drawInfo.handle != null) && drawInfo.handle.contains(x,y))
         )
      {
        return drawInfo.revisionData;
      }
    }

    return null;
  }

  /** get revsion at position
   * @param x,y position
   * @return revision or null
   */
  private RevisionData getRevisionAt(int x, int y)
  {
    return getRevisionAt(data.revisionDataTree,x,y);
  }

  /** show UUID index tool tip
   * @param entityIndexData entity index data
   * @param x,y positions
   */
  private void showRevisionInfoToolTip(RevisionData revisionData, int x, int y)
  {
    int   row;
    Label label;

    if (widgetRevisionInfoToolTip != null)
    {
      widgetRevisionInfoToolTip.dispose();
      widgetRevisionInfoToolTip = null;
    }

    if (revisionData != null)
    {
      widgetRevisionInfoToolTip = new Shell(dialog,SWT.ON_TOP|SWT.NO_FOCUS|SWT.TOOL);
      widgetRevisionInfoToolTip.setBackground(COLOR_INFO_BACKGROUND);
      widgetRevisionInfoToolTip.setLayout(new TableLayout(new double[]{0.0,0.0,0.0,1.0},new double[]{0.0,1.0},2));
      Widgets.layout(widgetRevisionInfoToolTip,0,0,TableLayoutData.NSWE);

      row = 0;

      label = Widgets.newLabel(widgetRevisionInfoToolTip,"Revision"+":");
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,0,TableLayoutData.W);
      label = Widgets.newLabel(widgetRevisionInfoToolTip,revisionData.getRevisionText());
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,1,TableLayoutData.WE);
      row++;

      label = Widgets.newLabel(widgetRevisionInfoToolTip,"Date"+":");
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,0,TableLayoutData.W);
      label = Widgets.newLabel(widgetRevisionInfoToolTip,Onzen.DATETIME_FORMAT.format(revisionData.date));
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,1,TableLayoutData.WE);
      row++;

      label = Widgets.newLabel(widgetRevisionInfoToolTip,"Autor"+":");
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,0,TableLayoutData.W);
      label = Widgets.newLabel(widgetRevisionInfoToolTip,revisionData.author);
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,1,TableLayoutData.WE);
      row++;

      label = Widgets.newLabel(widgetRevisionInfoToolTip,"Message"+":");
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,0,TableLayoutData.NW);
      label = Widgets.newLabel(widgetRevisionInfoToolTip,StringUtils.join(revisionData.commitMessageList,"\n"));
      label.setForeground(COLOR_INFO_FOREGROUND);
      label.setBackground(COLOR_INFO_BACKGROUND);
      Widgets.layout(label,row,1,TableLayoutData.NSWE);
      row++;

      Point size = widgetRevisionInfoToolTip.computeSize(SWT.DEFAULT,SWT.DEFAULT);
      widgetRevisionInfoToolTip.setBounds(x,y,size.x,size.y);
      widgetRevisionInfoToolTip.setVisible(true);

      dialog.addMouseTrackListener(new MouseTrackListener()
      {
        @Override
        public void mouseEnter(MouseEvent mouseEvent)
        {
        }

        @Override
        public void mouseExit(MouseEvent mouseEvent)
        {
          if (widgetRevisionInfoToolTip != null)
          {
            // check if inside widget
            Point point = dialog.toDisplay(new Point(mouseEvent.x,mouseEvent.y));
            if (widgetRevisionInfoToolTip.getBounds().contains(point))
            {
              return;
            }

            // check if inside sub-widget
            for (Control control : widgetRevisionInfoToolTip.getChildren())
            {
              if (control.getBounds().contains(point))
              {
                return;
              }
            }

            // close tooltip
            widgetRevisionInfoToolTip.dispose();
            widgetRevisionInfoToolTip = null;
          }
        }

        @Override
        public void mouseHover(MouseEvent mouseEvent)
        {
        }
      });
    }
  }

  /** draw revision item as box
   * @param gc graphics context
   * @param p base position
   * @param width,height revision box width/height
   * @param revisionData revision data
   * @param drawInfoList draw info list or null
   */
  private void drawBoxItem(GC                  gc,
                           Point               p,
                           int                 width,
                           int                 height,
                           RevisionData        revisionData,
                           ArrayList<DrawInfo> drawInfoList
                          )
  {
    final int WIDTH_COLUMN_0 = Widgets.getTextWidth(widgetRevisions,new String[]{"Revision:","Date:","Autor:","Message:"});

    // init variables
    Image image   = new Image(display,width,height);
    GC    imageGC = new GC(image);

    // clear image
    imageGC.setBackground(COLOR_BACKGROUND);
    imageGC.fillRectangle(image.getBounds());

    // draw box (into image)
    imageGC.setBackground(COLOR_CONTAINER);
    imageGC.fillRoundRectangle(0,0,width,height,10,10);
    if      (revisionData == data.selectedRevisionData0)
    {
      imageGC.setLineWidth(3);
      imageGC.setForeground(COLOR_CONTAINER_SELECTED0);
    }
    else if (revisionData == data.selectedRevisionData1)
    {
      imageGC.setLineWidth(3);
      imageGC.setForeground(COLOR_CONTAINER_SELECTED1);
    }
    else
    {
      imageGC.setLineWidth(1);
      imageGC.setForeground(COLOR_CONTAINER_NORMAL);
    }
    imageGC.drawRoundRectangle(0,0,width-1,height-1,10,10);

    // draw text (into image)
    imageGC.setBackground(COLOR_BACKGROUND);
    imageGC.setForeground(COLOR_TEXT);
    imageGC.drawString("Revision:",CONTAINER_MARGIN+0,CONTAINER_MARGIN+0*fontHeight,true); imageGC.drawString(revisionData.getRevisionText(),                                                                CONTAINER_MARGIN+WIDTH_COLUMN_0+4,CONTAINER_MARGIN+0*fontHeight,true);
    imageGC.drawString("Date:",    CONTAINER_MARGIN+0,CONTAINER_MARGIN+1*fontHeight,true); imageGC.drawString((revisionData != LOCAL_REVISION) ? Onzen.DATETIME_FORMAT.format(revisionData.date)       : "-",CONTAINER_MARGIN+WIDTH_COLUMN_0+4,CONTAINER_MARGIN+1*fontHeight,true);
    imageGC.drawString("Autor:",   CONTAINER_MARGIN+0,CONTAINER_MARGIN+2*fontHeight,true); imageGC.drawString((revisionData != LOCAL_REVISION) ? revisionData.author                                   : "-",CONTAINER_MARGIN+WIDTH_COLUMN_0+4,CONTAINER_MARGIN+2*fontHeight,true);
    imageGC.drawString("Message:", CONTAINER_MARGIN+0,CONTAINER_MARGIN+3*fontHeight,true); imageGC.drawText  ((revisionData != LOCAL_REVISION) ? StringUtils.join(revisionData.commitMessageList,"\n") : "-",CONTAINER_MARGIN+WIDTH_COLUMN_0+4,CONTAINER_MARGIN+3*fontHeight,true);

    // draw image
    gc.drawImage(image,p.x-width/2,p.y-height/2);

    // free resources
    imageGC.dispose();

    // draw handle
    gc.setLineWidth(1);
    gc.setLineStyle(SWT.LINE_SOLID);
    gc.setBackground(COLOR_HANDLE);
    gc.fillRectangle(p.x+width/2-1,p.y+height/2-1,HANDLE_SIZE,HANDLE_SIZE);
    gc.setForeground(Onzen.COLOR_BLACK);
    gc.drawRectangle(p.x+width/2-1,p.y+height/2-1,HANDLE_SIZE,HANDLE_SIZE);

    // add container, handle coordinates to draw info
    if (drawInfoList != null)
    {
      drawInfoList.add(new DrawInfo(revisionData,
                                    new Rectangle(p.x-width/2,
                                                  p.y-height/2,
                                                  width,
                                                  height
                                                 ),
                                    new Rectangle(p.x+width/2-1,
                                                  p.y+height/2-1,
                                                  HANDLE_SIZE,
                                                  HANDLE_SIZE
                                                 )
                                   )
                      );
    }
  }

  /** draw revision item as dot
   * @param gc graphics context
   * @param p base position
   * @param dotSize size of dot
   * @param height revision height
   * @param revisionData revision data
   * @param drawInfoList draw info list or null
   */
  private void drawDotItem(GC                  gc,
                           Point               p,
                           int                 dotSize,
                           int                 height,
                           RevisionData        revisionData,
                           ArrayList<DrawInfo> drawInfoList
                          )
  {
    int dy;

    // get info text
    String text[] = getRevisionInfoText(revisionData,MAX_TEXT_LENGTH);

    gc.setLineStyle(SWT.LINE_SOLID);

    // draw dot
    gc.setBackground(COLOR_CONTAINER);
    gc.fillOval(p.x-dotSize/2,p.y-dotSize/2+(height-dotSize)/2,dotSize,dotSize);
    if      (revisionData == data.selectedRevisionData0)
    {
      gc.setLineWidth(3);
      gc.setForeground(COLOR_CONTAINER_SELECTED0);
    }
    else if (revisionData == data.selectedRevisionData1)
    {
      gc.setLineWidth(3);
      gc.setForeground(COLOR_CONTAINER_SELECTED1);
    }
    else
    {
      gc.setLineWidth(1);
      gc.setForeground(COLOR_CONTAINER_NORMAL);
    }
    gc.drawOval(p.x-dotSize/2,p.y-dotSize/2+(height-dotSize)/2,dotSize-1,dotSize-1);

    // draw text
    gc.setBackground(COLOR_BACKGROUND);
    gc.setForeground(COLOR_TEXT);
    dy = 0;
    for (String text_ : text)
    {
      gc.drawString(text_,p.x+CONTAINER_MARGIN+dotSize,p.y-fontHeight/2+dy,false);
      dy += fontHeight+4;
    }

    // add text
    if (drawInfoList != null)
    {
      drawInfoList.add(new DrawInfo(revisionData,
                                    new Rectangle(p.x-dotSize/2,
                                                  p.y-dotSize/2+(height-dotSize)/2,
                                                  dotSize/2+CONTAINER_MARGIN+dotSize+Widgets.getTextWidth(gc,text),
                                                  Math.max(dotSize,height)
                                                 )
                                   )
                      );
    }
  }

  /** draw revision item as dot
   * @param gc graphics context
   * @param p base position
   * @param dotSize size of dot
   * @param height revision height
   * @param revisionData revision data
   * @param drawInfoList draw info list or null
   */
  private void drawListItem(GC                  gc,
                            Point               p,
                            int                 dotSize,
                            int                 height,
                            RevisionData        revisionData,
                            ArrayList<DrawInfo> drawInfoList
                           )
  {
    int    dx;
    String text;

    gc.setLineStyle(SWT.LINE_SOLID);

    // draw dot
    gc.setBackground(COLOR_CONTAINER);
    gc.fillOval(p.x-dotSize/2,p.y-dotSize/2+(height-dotSize)/2,dotSize,dotSize);
    if      (revisionData == data.selectedRevisionData0)
    {
      gc.setLineWidth(3);
      gc.setForeground(COLOR_CONTAINER_SELECTED0);
    }
    else if (revisionData == data.selectedRevisionData1)
    {
      gc.setLineWidth(3);
      gc.setForeground(COLOR_CONTAINER_SELECTED1);
    }
    else
    {
      gc.setLineWidth(1);
      gc.setForeground(COLOR_CONTAINER_NORMAL);
    }
    gc.drawOval(p.x-dotSize/2,p.y-dotSize/2+(height-dotSize)/2,dotSize-1,dotSize-1);

    // draw text
    gc.setBackground(COLOR_BACKGROUND);
    gc.setForeground(COLOR_TEXT);
    dx = 0;
    if (Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.REVISION))
    {
      text = revisionData.revisionId;
      gc.drawString(text,maxTreeSize.x+CONTAINER_MARGIN+dotSize+dx,p.y-fontHeight/2,false);
      dx += maxRevisionIdTextWidth+CONTAINER_MARGIN;
    }
    if (Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.REVISION))
    {
      text = revisionData.author;
      gc.drawString(text,maxTreeSize.x+CONTAINER_MARGIN+dotSize+dx,p.y-fontHeight/2,false);
      dx += maxAuthorTextWidth+CONTAINER_MARGIN;
    }
    if (Settings.revisionGraphTitle.contains(Settings.RevisionGraphTitles.MESSAGE))
    {
      text = ((revisionData.commitMessageList != null) && (revisionData.commitMessageList.size() > 0)) ? ": "+revisionData.commitMessageList.getFirst() : "";
      gc.drawString(text,maxTreeSize.x+CONTAINER_MARGIN+dotSize+dx,p.y-fontHeight/2,false);
      dx += Widgets.getTextWidth(gc,text)+CONTAINER_MARGIN;
    }

    // add text
    if (drawInfoList != null)
    {
      drawInfoList.add(new DrawInfo(revisionData,
                                    new Rectangle(p.x-dotSize/2,
                                                  p.y-dotSize/2+(height-dotSize)/2,
                                                  dotSize/2+CONTAINER_MARGIN+dotSize+dx,
                                                  Math.max(dotSize,height)
                                                 )
                                   )
                      );
    }
  }

  /** draw revision item
   * @param drawState draw state; see DrawStates
   * @param gc graphics context
   * @param p base position
   * @param width,height revision item width/height
   * @param dotSize size of dot
   * @param revisionData revision data
   * @param drawInfoList draw info list or null
   */
  private void drawItem(DrawStates            drawState,
                        GC                    gc,
                        Point                 p,
                        int                   width,
                        int                   height,
                        RevisionData          revisionData,
                        ArrayList<DrawInfo>   drawInfoList
                       )
  {
    switch (Settings.revisionGraphType)
    {
      case BOXES:
        drawBoxItem(gc,
                    p,
                    width,
                    height,
                    revisionData,
                    drawInfoList
                   );
        break;
      case DOTS:
        drawDotItem(gc,
                    p,
                    DOT_SIZE,
                    DOT_SIZE,
                    revisionData,
                    drawInfoList
                   );
        break;
      case LIST:
        drawListItem(gc,
                     p,
                     DOT_SIZE,
                     DOT_SIZE,
                     revisionData,
                     drawInfoList
                    );
        break;
    }
  }

  /** draw vertical line
   * @param gc graphics context
   * @param p0,p1 position
   * @param lineStyle line style
   * @param color line color
   */
  private void drawVLine(GC    gc,
                         Point p0,
                         Point p1,
                         int   lineStyle,
                         Color color
                        )
  {
    assert p0.x == p1.x;

    gc.setLineWidth(LINES_WIDTH);
    gc.setLineStyle(lineStyle);
    gc.setForeground(color);
    gc.drawLine(p0.x,p0.y,
                p1.x,p1.y
               );
  }

  /** draw vertical line
   * @param gc graphics context
   * @param p0,p1 position
   */
  private void drawVLine(GC    gc,
                         Point p0,
                         Point p1
                        )
  {
    drawVLine(gc,p0,p1,SWT.LINE_SOLID,COLOR_LINES);
  }

  /** draw L-line
   * @param gc graphics context
   * @param p0,p1 position
   * @param lineStyle line style
   * @param color line color
   */
  private void drawLLine(GC    gc,
                         Point p0,
                         Point p1,
                         int   lineStyle,
                         Color color
                        )
  {
    final int PADDING = Math.min(PADDING_X,PADDING_Y);

    gc.setLineWidth(LINES_WIDTH);
    gc.setLineStyle(lineStyle);
    gc.setForeground(color);
    if      (p0.x > p1.x)
    {
      gc.drawLine(p0.x        ,p0.y,
                  p0.x        ,p1.y-PADDING
                 );
      gc.drawLine(p0.x        ,p1.y-PADDING,
                  p0.x-PADDING,p1.y
                 );
      gc.drawLine(p0.x-PADDING,p1.y,
                  p1.x        ,p1.y
                 );
    }
    else if (p0.x < p1.x)
    {
      gc.drawLine(p0.x        ,p0.y,
                  p0.x        ,p1.y-PADDING
                 );
      gc.drawLine(p0.x        ,p1.y-PADDING,
                  p0.x+PADDING,p1.y
                 );
      gc.drawLine(p0.x+PADDING,p1.y,
                  p1.x        ,p1.y
                 );
    }
    else
    {
      gc.drawLine(p0.x,p0.y,
                  p1.x,p1.y
                 );
    }
  }

  /** draw L-line
   * @param gc graphics context
   * @param p0,p1 position
   */
  private void drawLLine(GC    gc,
                         Point p0,
                         Point p1
                        )
  {
    drawLLine(gc,p0,p1,SWT.LINE_SOLID,COLOR_LINES);
  }

  /** draw branch L-line
   * @param gc graphics context
   * @param p0,p1 position
   */
  private void drawBranchLLine(GC    gc,
                               Point p0,
                               Point p1
                              )
  {
    final int PADDING = Math.min(PADDING_X,PADDING_Y);

    gc.setLineWidth(LINES_WIDTH);
    gc.setLineDash(LINES_DASH_PATTERN1);
    gc.setForeground(COLOR_BRANCH_LINES);
    gc.drawLine(p0.x        ,p0.y,
                p1.x-PADDING,p0.y
               );
    gc.drawLine(p1.x-PADDING,p0.y,
                p1.x        ,p0.y+PADDING
               );
    gc.drawLine(p1.x        ,p0.y+PADDING,
                p1.x        ,p1.y
               );
  }

  /** draw branch name
   * @param gc graphics context
   * @param p position
   * @param lx,ly line relative position
   * @param width,height revision box width/height
   * @param name branch name
   */
  private void drawBranchName(GC     gc,
                              Point  p,
                              int    lx,
                              int    ly,
                              int    width,
                              int    height,
                              String name
                             )
  {
    gc.setForeground(Onzen.COLOR_BLACK);
    gc.drawString(name,p.x+PADDING_X+width+lx,p.y+PADDING_Y+height/2,true);
  }

  /** draw branch L-line
   * @param gc graphics context
   * @param p0,p1 position
   */
  private void drawMergeLLine(GC    gc,
                              Point p0,
                              Point p1
                             )
  {
    final int PADDING = Math.min(PADDING_X,PADDING_Y);

    gc.setLineWidth(LINES_WIDTH);
    gc.setLineDash(LINES_DASH_PATTERN2);
    gc.setForeground(COLOR_MERGE_LINES);
    gc.setBackground(COLOR_MERGE_LINES);

    if      (p0.x > p1.x)
    {
      if (p1.y >= p0.y)
      {
        /*     p0
               |
           p1--/
        */
        gc.drawLine(p0.x        ,p0.y,
                    p0.x        ,p1.y-PADDING
                   );
        gc.drawLine(p0.x        ,p1.y-PADDING,
                    p0.x-PADDING,p1.y
                   );
        gc.drawLine(p0.x-PADDING,p1.y,
                    p1.x        ,p1.y
                   );
        gc.fillPolygon(new int[]{p1.x+DOT_SIZE/2+8,p1.y-5,
                                 p1.x+DOT_SIZE/2+8,p1.y+5,
                                 p1.x+DOT_SIZE/2  ,p1.y
                                }
                      );
      }
      else
      {
        /* p1--\
               |
               p0
        */
        gc.drawLine(p0.x        ,p0.y,
                    p0.x        ,p1.y+PADDING
                   );
        gc.drawLine(p0.x        ,p1.y+PADDING,
                    p0.x-PADDING,p1.y
                   );
        gc.drawLine(p0.x-PADDING,p1.y,
                    p1.x        ,p1.y
                   );
        gc.fillPolygon(new int[]{p1.x+DOT_SIZE/2+8,p1.y-5,
                                 p1.x+DOT_SIZE/2+8,p1.y+5,
                                 p1.x+DOT_SIZE/2  ,p1.y
                                }
                      );
      }
    }
    else if (p0.x < p1.x)
    {
      if (p1.y >= p0.y)
      {
        /* p0
           |
           \--p1
        */
        gc.drawLine(p0.x        ,p0.y,
                    p0.x        ,p1.y-PADDING
                   );
        gc.drawLine(p0.x        ,p1.y-PADDING,
                    p0.x+PADDING,p1.y
                   );
        gc.drawLine(p0.x+PADDING,p1.y,
                    p1.x        ,p1.y
                   );
        gc.fillPolygon(new int[]{p1.x+DOT_SIZE/2-8,p1.y-5,
                                 p1.x+DOT_SIZE/2-8,p1.y+5,
                                 p1.x+DOT_SIZE/2  ,p1.y
                                }
                      );
      }
      else
      {
        /* /--p1
           |
           p0
        */
        gc.drawLine(p0.x        ,p0.y,
                    p0.x        ,p1.y+PADDING
                   );
        gc.drawLine(p0.x        ,p1.y+PADDING,
                    p0.x+PADDING,p1.y
                   );
        gc.drawLine(p0.x+PADDING,p1.y,
                    p1.x        ,p1.y
                   );
        gc.fillPolygon(new int[]{p1.x+DOT_SIZE/2-8,p1.y-5,
                                 p1.x+DOT_SIZE/2-8,p1.y+5,
                                 p1.x+DOT_SIZE/2  ,p1.y
                                }
                      );
      }
    }
    else
    {
      /* p0
         |
         p1
      */
      gc.drawLine(p0.x        ,p0.y,
                  p1.x        ,p1.y
                 );
      gc.fillPolygon(new int[]{p1.x-5,p1.y-DOT_SIZE/2-8,
                               p1.x+5,p1.y+DOT_SIZE/2-8,
                               p1.x  ,p1.y+DOT_SIZE/2
                              }
                    );
    }
  }

  /** check if intersection with client draw area
   * @param clientArea client draw area
   * @param p position
   * @param width,height widht/height
   */
  private boolean intersects(Rectangle clientArea, Point p, int width, int height)
  {
    Rectangle rectangle = new Rectangle(p.x,
                                        p.y,
                                        width,
                                        height
                                       );
    return clientArea.intersects(rectangle);
  }

  /** check if intersection with client draw area
   * @param clientArea client draw area
   * @param p0,p1 positions
   */
  private boolean intersects(Rectangle clientArea, Point p0, Point p1)
  {
    int x      = Math.min(p0.x,p1.x);
    int y      = Math.min(p0.y,p1.y);
    int width  = Math.abs(p1.x-p0.x);
    int height = Math.abs(p1.y-p0.y);
    Rectangle rectangle = new Rectangle(x,
                                        y,
                                        width,
                                        height
                                       );
    return clientArea.intersects(rectangle);
  }

  /** redraw revision data tree
   * @param clientArea client draw area
   * @param drawState draw state; see DrawStates
   * @param gc graphics context
   * @param revisionDataTree revision data tree
   * @param x,y base position
   * @param containerDeltaWidth,containerDeltaHeight container delta for resize containers or 0,0
   * @param drawInfoList draw info list or null
   * @return next drawing position
   */
  private void redraw(Rectangle             clientArea,
                      DrawStates            drawState,
                      GC                    gc,
                      RevisionDataTree      revisionDataTree,
                      Point                 base,
                      int                   containerDeltaWidth,
                      int                   containerDeltaHeight,
                      boolean               drawLocalRevision,
                      HashMap<String,Point> positionMap,
                      ArrayList<DrawInfo>   drawInfoList
                     )
  {
    // get item width/height, line x/y-offset
    final int width,height,lx,ly;
    switch (Settings.revisionGraphType)
    {
      case BOXES:
        width  = Math.max(Settings.geometryRevisionBox.x+containerDeltaWidth, CONTAINER_MIN_WIDTH );
        height = Math.max(Settings.geometryRevisionBox.y+containerDeltaHeight,CONTAINER_MIN_HEIGHT);
        lx     = width /2;
        ly     = height/2;
        break;
      case DOTS:
        width  = CONTAINER_MARGIN+DOT_SIZE+maxRevisionInfoTextSize.x+CONTAINER_MARGIN;
        height = DOT_SIZE;
        lx     = DOT_SIZE/2;
        ly     = DOT_SIZE/2;
        break;
      case LIST:
        width  = CONTAINER_MARGIN+DOT_SIZE+CONTAINER_MARGIN;
        height = DOT_SIZE;
        lx     = DOT_SIZE/2;
        ly     = DOT_SIZE/2;
        break;
      default:
        width  = 0;
        height = 0;
        lx     = 0;
        ly     = 0;
        break;
    }

//Dprintf.dprintf("base=%s",base);
//Dprintf.dprintf("view=%s",view);

    // draw revision tree
    boolean firstFlag    = true;
    Point   prevPosition = new Point(base.x,base.y);
    Point   position     = new Point(base.x,base.y);
    int     dx,dy;
    for (RevisionData revisionData : revisionDataTree)
    {
      if (   (!Settings.revisionGraphBranchesMergesOnly || firstFlag || revisionData.isBranchedOrMerged())
          && (data.author.isEmpty() || revisionData.isFromAuthor(data.author))
         )
      {
        dy = 0;
//Dprintf.dprintf("revisionData=%s %d",revisionData.revisionId,revisionDataTree.size());

        // get text size
        Point size = getMaxRevisionInfoTextSize(revisionData);

        if (drawState == DrawStates.LINES)
        {
          // draw vertical connection line
          if (   !firstFlag
              && intersects(clientArea,prevPosition,position)
             )
          {
            drawVLine(gc,
                      prevPosition,
                      position
                     );
          }
        }
        dy += PADDING_Y;

        if (drawState == DrawStates.ENTRIES)
        {
          // draw entry
          if (intersects(clientArea,position,width,height))
          {
            drawItem(drawState,
                     gc,
                     position,
                     width,
                     height,
                     revisionData,
                     drawInfoList
                    );
          }
        }
        dy += Math.max(height,size.y);

//Dprintf.dprintf("revisionData.getRevisionText()=%s %s",revisionData.getRevisionText(),position);

        // draw branches/merges
        if (   !revisionData.branchDataList.isEmpty()
            && (data.author.isEmpty() || revisionData.branchDataList.isFromAuthor(data.author))
           )
        {
          dx = 0;
//Dprintf.dprintf("rev=%s: %d branches",revisionData.revisionId,revisionData.branchDataList.size());
          for (BranchData branchData : revisionData.branchDataList)
          {
//Dprintf.dprintf("branchData=%s",branchData);
            Point subPosition = new Point(position.x+dx+PADDING_X+width,position.y+dy+PADDING_Y+height);

            // get size of sub-tree
            Point subTreeSize = getTreeSize(branchData.revisionDataTree);

            if (drawState == DrawStates.LINES)
            {
              /* draw branch connection L-line:
                   _
                    \
                     |
              */
              if (intersects(clientArea,position,subPosition))
              {
                drawBranchLLine(gc,
                                position,
                                subPosition
                               );
              }
            }

            if (drawState == DrawStates.ENTRIES)
            {
              // draw branch name
              Point textPosition = new Point(position.x+dx,position.y-fontHeight);
              if (   (branchData.name != null)
                  && intersects(clientArea,textPosition,width,height)
                 )
              {
                drawBranchName(gc,
                               textPosition,
                               lx,
                               0,
                               width,
                               height,
                               branchData.name
                              );
//Dprintf.dprintf("%d %d %s",position.x+width,position.y+height/2,width,height,branchData.name);
              }
            }

            // draw branch sub-tree
            redraw(clientArea,
                   drawState,
                   gc,
                   branchData.revisionDataTree,
                   subPosition,
                   containerDeltaWidth,
                   containerDeltaHeight,
                   false,  // drawLocalRevision
                   positionMap,
                   drawInfoList
                  );

            // next column/row
            dx = getNextColumn(dx,subTreeSize);
            dy = getNextRow(dy,subTreeSize);
          }
        }
        if (revisionData.isMerge())
        {
          if (drawState == DrawStates.LINES)
          {
            // draw merge connection line
            RevisionData parentRevisionData0 = revisionData.parentRevisionData0;
            RevisionData parentRevisionData1 = revisionData.parentRevisionData1;
//Dprintf.dprintf("merge line: %s+%s -> %s",parentRevisionData0.revisionId,parentRevisionData1.revisionId,revisionData.revisionId);
            RevisionData fromParentRevisionData = revisionDataTree.contains(parentRevisionData0)
                                                    ? parentRevisionData1
                                                    : parentRevisionData0;
            if (fromParentRevisionData != null)
            {
//Dprintf.dprintf("merge %s -> %s",revisionData,fromParentRevisionData);
              Point mergePosition = positionMap.get(fromParentRevisionData.revisionId);
              if (   (mergePosition != null)
                  && intersects(clientArea,mergePosition,position)
                 )
              {
                drawMergeLLine(gc,
                               mergePosition,
                               position
                              );
              }
//else { Dprintf.halt("not found %s",revisionData.parentRevisionData0); }
            }
          }
        }

        // save position
        if (drawState == DrawStates.ENTRY_POSITIONS)
        {
          positionMap.put(revisionData.revisionId,new Point(position.x,position.y));
        }

        // next row
        prevPosition.x = position.x;
        prevPosition.y = position.y;
        position.y += dy;
        firstFlag = false;
      }
    }

    // local revision
    if (drawLocalRevision && (data.lastRevisionData != null))
    {
      prevPosition = positionMap.get(data.lastRevisionData.revisionId);
      if (prevPosition != null)
      {
        // draw below last revision
        position.x = prevPosition.x;

        if (!firstFlag)
        {
          if (drawState == DrawStates.LINES)
          {
            if (intersects(clientArea,prevPosition,position))
            {
              drawLLine(gc,
                        prevPosition,
                        position
                       );
            }
          }
        }
        if (drawState == DrawStates.ENTRIES)
        {
          if (intersects(clientArea,prevPosition,position))
          {
            drawItem(drawState,
                     gc,
                     position,
                     width,
                     height,
                     LOCAL_REVISION,
                     drawInfoList
                    );
          }
        }

        // save position
//        if (drawState == DrawStates.ENTRY_POSITIONS)
        {
          positionMap.put(LOCAL_REVISION.revisionId,new Point(position.x,position.y));
        }
      }
    }
  }

  /** redraw revision tree with container delta width/height (no update of draw infos)
   */
  private void redraw(int containerDeltaWidth, int containerDeltaHeight)
  {
    // redraw
    if (data.revisionDataTree != null)
    {
      // get item width/height
      final int width,height;
      switch (Settings.revisionGraphType)
      {
        case BOXES:
          width  = Math.max(Settings.geometryRevisionBox.x+containerDeltaWidth ,CONTAINER_MIN_WIDTH );
          height = Math.max(Settings.geometryRevisionBox.y+containerDeltaHeight,CONTAINER_MIN_HEIGHT);
          break;
        case DOTS:
          width  = CONTAINER_MARGIN+DOT_SIZE+maxRevisionInfoTextSize.x+CONTAINER_MARGIN;
          height = DOT_SIZE;
          break;
        case LIST:
          width  = CONTAINER_MARGIN+DOT_SIZE+maxRevisionInfoTextSize.x+CONTAINER_MARGIN;
          height = DOT_SIZE;
          break;
        default:
          width  = 0;
          height = 0;
          break;
      }

      // init variables
      Rectangle clientArea = widgetRevisions.getClientArea();
      GC        gc         = new GC(widgetRevisions);

      // clear
      gc.setBackground(Onzen.COLOR_WHITE);
      gc.fillRectangle(0,0,data.view.width,data.view.height);

      // redraw
      for (DrawStates drawState : new DrawStates[]{DrawStates.ENTRY_POSITIONS,DrawStates.LINES,DrawStates.ENTRIES})
      {
        redraw(clientArea,
               drawState,
               gc,
               data.revisionDataTree,
               new Point(data.view.x+MARGIN+width/2,data.view.y+MARGIN+height/2),
               containerDeltaWidth,
               containerDeltaHeight,
               false,  // drawLocalRevision
               (HashMap<String,Point>)null,
               (ArrayList<DrawInfo>)null
              );
      }

      // free resources
      gc.dispose();
    }
  }

  /** redraw revision tree and update draw info
   * @param clearFlag true to clear background
   */
  private void redraw()
  {
    if (data.revisionDataTree != null)
    {
      // get item width/height
      final int xOffset,yOffset;
      switch (Settings.revisionGraphType)
      {
        case BOXES:
          xOffset = 5+Math.max(Settings.geometryRevisionBox.x,CONTAINER_MIN_WIDTH )/2;
          yOffset = 5+Math.max(Settings.geometryRevisionBox.y,CONTAINER_MIN_HEIGHT)/2;
          break;
        case DOTS:
          xOffset = 5+DOT_SIZE/2;
          yOffset = 5+DOT_SIZE/2;
          break;
        case LIST:
          xOffset = 5+DOT_SIZE/2;
          yOffset = 5+DOT_SIZE/2;
          break;
        default:
          xOffset = 0;
          yOffset = 0;
          break;
      }

      // init variables
      Rectangle clientArea = widgetRevisions.getClientArea();
      GC        gc         = new GC(widgetRevisions);

      // clear
      data.drawInfoList.clear();
      gc.setBackground(Onzen.COLOR_WHITE);
      gc.fillRectangle(0,0,clientArea.width,clientArea.height);

      // draw
      HashMap<String,Point> positionMap = new HashMap<String,Point>();
      for (DrawStates drawState : new DrawStates[]{DrawStates.ENTRY_POSITIONS,DrawStates.LINES,DrawStates.ENTRIES})
//      for (DrawStates drawState : new DrawStates[]{DrawStates.LINES})
      {
//Dprintf.dprintf("******************************************** drawState=%s",drawState);
        redraw(clientArea,
               drawState,
               gc,
               data.revisionDataTree,
               new Point(data.view.x+xOffset,data.view.y+yOffset),
               0,  // containerDeltaWidth
               0,  // containerDeltaHeight
               true,  // drawLocalRevision
               positionMap,
               data.drawInfoList
              );
      }

      // free resources
      gc.dispose();
    }
  }

  /** set canvas size from revision tree size
   */
  private void setSize()
  {
    if (!widgetRevisions.isDisposed())
    {
      // get max. revisionid/author/revision text width
      maxRevisionIdTextWidth  = getMaxRevisionIdTextWidth(data.revisionDataTree);
      maxAuthorTextWidth      = getMaxAuthorTextWidth(data.revisionDataTree);
      maxRevisionInfoTextSize = getMaxRevisionInfoTextSize(data.revisionDataTree);

      // get tree size
      maxTreeSize = getTreeSize(data.revisionDataTree);

      // get size
      data.size = getTreeSize(data.revisionDataTree);
      data.size.x += 2*MARGIN;
      data.size.y += 2*MARGIN;

      // set scroll bars
      Rectangle clientArea = widgetRevisions.getClientArea();
      widgetHorizontalScrollBar.setMaximum(data.size.x);
      widgetVerticalScrollBar.setMaximum(data.size.y);
      widgetHorizontalScrollBar.setThumb(Math.min(clientArea.width,data.size.x));
      widgetVerticalScrollBar.setThumb(Math.min(clientArea.height,data.size.y));

      widgetHorizontalScrollBar.setIncrement(Settings.geometryRevisionBox.x+PADDING_X);
      widgetHorizontalScrollBar.setPageIncrement(clientArea.width);
      widgetHorizontalScrollBar.setSelection(clientArea.x);
      widgetVerticalScrollBar.setIncrement(Settings.geometryRevisionBox.y+PADDING_Y);
      widgetVerticalScrollBar.setPageIncrement(clientArea.height);
      widgetVerticalScrollBar.setSelection(clientArea.y);

      // redraw
      data.view.x      = clientArea.x;
      data.view.y      = clientArea.y;
      data.view.width  = clientArea.width;
      data.view.height = clientArea.height;
      widgetRevisions.redraw();
    }
  }

  /** append revision data to revision data array
   * @param revisionDataArray array
   * @param revisionData_ revision data to append
   */
  private void revisionDataTreeToArray(ArrayList<RevisionData> revisionDataArray, RevisionDataTree revisionDataTree)
  {
    for (RevisionData revisionData : revisionDataTree)
    {
      revisionDataArray.add(revisionData);

      for (BranchData branchData : revisionData.branchDataList)
      {
        revisionDataTreeToArray(revisionDataArray,branchData.revisionDataTree);
      }
    }
  }

  /** scroll to revision
   * @param revisionData revision to show
   */
  private void scrollTo(final RevisionData revisionData)
  {
    if (!dialog.isDisposed())
    {
      display.syncExec(new Runnable()
      {
        public void run()
        {
          final int width  = Settings.geometryRevisionBox.x;
          final int height = Settings.geometryRevisionBox.y;

          Point point = getRevisionX0Y0(revisionData);
          if (point != null)
          {
            // get x,y-center of revision
            point.x += width /2;
            point.y += height/2;
//Dprintf.dprintf("point=%s",point);

            // scroll to center of revision container view
            Rectangle clientArea = widgetRevisions.getClientArea();
            clientArea.x = widgetHorizontalScrollBar.getSelection();
            clientArea.y = widgetVerticalScrollBar.getSelection();
//Dprintf.dprintf("clientArea=%s point=%s data.view=%s",clientArea,point,data.view);
            if ((point.x < clientArea.x) || (point.x > clientArea.x+clientArea.width )) data.view.x = -(point.x-width /2);
            if ((point.y < clientArea.y) || (point.y > clientArea.y+clientArea.height)) data.view.y = -(point.y-height/2);

/*
???
//Dprintf.dprintf("revisionData=%s",revisionData);
            RevisionData nextRevisionData = getNextRevision(data.revisionDataTree,revisionData);;
            int          verticalSpace    = clientArea.height-(PADDING_Y+height);
            while ((verticalSpace > 0) && (nextRevisionData != null))
            {
              nextRevisionData = getNextRevision(data.revisionDataTree,nextRevisionData);
//Dprintf.dprintf("nextRevisionData=%s",nextRevisionData);
              verticalSpace -= (PADDING_Y+height);
            }
//data.view.y += verticalSpace;
Dprintf.dprintf("data.view=%s",data.view);
*/

            widgetRevisions.scroll(-data.view.x,-data.view.y,0,0,data.size.x,data.size.y,false);
            widgetRevisions.redraw();
            widgetHorizontalScrollBar.setSelection(-data.view.x);
            widgetVerticalScrollBar.setSelection(-data.view.y);
          }
        }
      });
    }
  }

  /** show revisions
   * @param revision revision to show
   */
  private void show(String revision)
  {
    // clear
    if (!dialog.isDisposed())
    {
      display.syncExec(new Runnable()
      {
        public void run()
        {
          data.revisionDataTree = null;
          Widgets.modified(data);
         }
      });
    }

    // start show revision tree
    final BusyDialog busyDialog = new BusyDialog(dialog,
                                                 "Create revision tree",
                                                 400,SWT.DEFAULT,
                                                 BusyDialog.TEXT0|BusyDialog.PROGRESS_BAR0|BusyDialog.LIST|BusyDialog.AUTO_ANIMATE|BusyDialog.ABORT_CLOSE
                                                );
    Background.run(new BackgroundRunnable(fileData,revision,busyDialog)
    {
      public void run(FileData fileData, final String revision, final BusyDialog busyDialog)
      {
        if (!dialog.isDisposed())
        {
          // get revision tree, local revision id
          Widgets.setCursor(dialog,Onzen.CURSOR_WAIT);
          repositoryTab.setStatusText("Create revision tree"+((fileData != null) ?  String.format(" for '%s'...",fileData.getName(repositoryTab.repository)) : ""));
          try
          {
            data.revisionDataTree = repositoryTab.repository.getRevisionDataTree(fileData,busyDialog);
            data.lastRevisionData = repositoryTab.repository.getRevisionData(fileData);
          }
          catch (Exception exception)
          {
exception.printStackTrace();
            final String exceptionMessage = exception.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(dialog,"Getting file revisions fail: %s",exceptionMessage);
                busyDialog.close();
                dialog.dispose();
              }
            });
            if (Settings.debugFlag)
            {
              Onzen.printStacktrace(exception);
            }
            return;
          }
          catch (Throwable throwable)
          {
            final String exceptionMessage = throwable.getMessage();
            display.syncExec(new Runnable()
            {
              public void run()
              {
                Dialogs.error(dialog,"Internal error: %s",exceptionMessage);
                busyDialog.close();
                dialog.dispose();
              }
            });
            Onzen.printInternalError(throwable);
            return;
          }
          finally
          {
            repositoryTab.clearStatusText();
            Widgets.resetCursor(dialog);
          }
//Dprintf.dprintf(""); printRevisionDataTree(data.revisionDataTree);

          // convert tree into an list for find function
          data.revisionDataList = new ArrayList<RevisionData>();
          revisionDataTreeToArray(data.revisionDataList,data.revisionDataTree);

          // get authors
          HashSet<String> authorSet = new HashSet<String>();
          for (RevisionData revisionData : data.revisionDataList)
          {
            authorSet.add(revisionData.author);
          }
          final ArrayList<String> authorList = new ArrayList<String>(authorSet);
          Collections.sort(authorList);
          display.syncExec(new Runnable()
          {
            public void run()
            {
              if (!widgetAuthor.isDisposed())
              {
                widgetAuthor.add("");
                for (String author : authorList)
                {
                  widgetAuthor.add(author);
                }
              }
            }
          });

          // show
          if (!dialog.isDisposed())
          {
            display.syncExec(new Runnable()
            {
              public void run()
              {
                if (!busyDialog.isAborted())
                {
                  // set canvas size
                  setSize();

                  // redraw/scroll to selected revision
                  data.selectedRevisionData1 = getClosestRevision(data.revisionDataTree,revision);
                  if (data.selectedRevisionData1 != null)
                  {
                    scrollTo(data.selectedRevisionData1);
                  }
                  else
                  {
                    widgetRevisions.redraw();
                  }

                  // notify modification
                  Widgets.modified(data);

                  // close busy dialog
                  if (busyDialog.isListEmpty())
                  {
                    busyDialog.close();
                  }
                  else
                  {
                    busyDialog.done();
                  }
                }
                else
                {
                  busyDialog.close();
                  dialog.dispose();
                }
              }
            });
          }
        }
      }
    });
  }

  /** show last revision
   */
  private void show()
  {
    show(repositoryTab.repository.getLastRevision());
  }

  /** search previous revision
   * @param widgetFind search text widget
   * @return revision data or null
   */
  private RevisionData findPrev(Text widgetFind)
  {
    RevisionData revisionData = null;

    if (!widgetFind.isDisposed() && (data.revisionDataList != null))
    {
      String findText = widgetFind.getText().toLowerCase();
      if (!findText.isEmpty())
      {
        // get selected/first revision
        RevisionData selectedRevisionData = (data.selectedRevisionData1 != null) ? data.selectedRevisionData1 : data.revisionDataList.get(0);

        // find current revision
        int z = 0;
        while ((z < data.revisionDataList.size()) && (data.revisionDataList.get(z) != selectedRevisionData))
        {
          z++;
        }

        // find previous matching revision
        do
        {
//Dprintf.dprintf("data.revisionDataList.get(z)=%s",data.revisionDataList.get(z));
          z--;
        }
        while (   (z >= 0)
               && !data.revisionDataList.get(z).match(findText)
              );

        if (z >= 0)
        {
//Dprintf.dprintf("data.revisionDataList.get(z)=%s",data.revisionDataList.get(z));
          // select previsous revision
          revisionData = data.revisionDataList.get(z);
        }
        else
        {
          // not found
          Widgets.flash(widgetFind);
        }
      }
    }

    return revisionData;
  }

  /** search next revision
   * @param widgetFind search text widget
   * @return revision data or null
   */
  private RevisionData findNext(Text widgetFind)
  {
    RevisionData revisionData = null;

    if (!widgetFind.isDisposed() && (data.revisionDataList != null))
    {
      String findText = widgetFind.getText().toLowerCase();
      if (!findText.isEmpty())
      {
        // find current revision index or -1 if non selected
        int z;
        if (data.selectedRevisionData1 != null)
        {
          z = 0;
          while ((z < data.revisionDataList.size()) && (data.revisionDataList.get(z) != data.selectedRevisionData1))
          {
            z++;
          }
        }
        else
        {
          z = -1;
        }

        // find next matching revision
        do
        {
          z++;
        }
        while (   (z < data.revisionDataList.size())
               && !data.revisionDataList.get(z).match(findText)
              );

        if (z < data.revisionDataList.size())
        {
          // select next revision
          revisionData = data.revisionDataList.get(z);
        }
        else
        {
          // not found
          Widgets.flash(widgetFind);
        }
      }
    }

    return revisionData;
  }

  /** print revision tree (for debugging)
   * @param revisionDataTree revision data tree to print
   * @param indent indentation
   */
  private void printRevisionDataTree(RevisionDataTree revisionDataTree, int indent)
  {
    for (RevisionData revisionData : revisionDataTree)
    {
      System.out.print(StringUtils.repeat(' ',indent));
      System.out.println(revisionData.revisionId+": "+revisionData.date);
      if (revisionData.parentRevisionData0 != null)
      {
        System.out.print(StringUtils.repeat(' ',indent+2)+"parent ");
        System.out.println(revisionData.parentRevisionData0.revisionId);
      }
      if (revisionData.parentRevisionData1 != null)
      {
        System.out.print(StringUtils.repeat(' ',indent+2)+"parent ");
        System.out.println(revisionData.parentRevisionData1.revisionId);
      }

      if (!revisionData.branchDataList.isEmpty())
      {
        System.out.print(StringUtils.repeat(' ',indent+2)+"branch:");
        for (BranchData branchData : revisionData.branchDataList)
        {
          printRevisionDataTree(branchData.revisionDataTree,indent+4);
        }
      }
    }
  }

  /** print revision tree (for debugging)
   * @param revisionDataTree revision data tree to print
   */
  private void printRevisionDataTree(RevisionDataTree revisionDataTree)
  {
    System.out.println("Revision tree:");
    printRevisionDataTree(revisionDataTree,2);
  }
}

/* end of file */
