/***********************************************************************\
*
* $Revision: 2169 $
* $Date: 2017-09-28 09:10:06 +0200 (Thu, 28 Sep 2017) $
* $Author: torsten $
* Contents: repository
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.File;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;

import java.lang.ref.WeakReference;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import java.text.SimpleDateFormat;
import java.text.ParseException;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.Enumeration;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.TreeSet;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;

import javax.activation.MimetypesFileTypeMap;

import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlList;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.bind.annotation.XmlValue;
import javax.xml.bind.annotation.XmlSeeAlso;
import javax.xml.bind.annotation.adapters.XmlAdapter;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

import java.lang.annotation.Annotation;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Field;
import java.lang.reflect.Constructor;
import static java.lang.annotation.ElementType.FIELD;
import static java.lang.annotation.ElementType.TYPE;

// graphics
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.TreeItem;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;

/****************************** Classes ********************************/

class RepositoryFile extends File
{
  public String rootPath;
  public String name;

  RepositoryFile(String rootPath, String name)
  {
    super(rootPath,(name != null) ? name : "");
    this.rootPath = rootPath;
    this.name     = (name != null) ? name : "";
  }

  RepositoryFile(RepositoryFile file, String name)
  {
    super(file.rootPath,(name != null) ? name : "");
    this.rootPath = file.rootPath;
    this.name     = !file.name.isEmpty()? file.name+'/'+name : name;
  }

  RepositoryFile(String name)
  {
    super(name);
    this.rootPath = "";
    this.name     = name;
  }

  @Override
  public String getName()
  {
    return new File(name).getName();
  }

  @Override
  public String getAbsolutePath()
  {
    return !rootPath.isEmpty() ? rootPath+'/'+name : name;
  }

  @Override
  public RepositoryFile getParentFile()
  {
    return new RepositoryFile(rootPath,new File(name).getParent());
  }

  @Override
  public boolean isDirectory()
  {
    return true;
  }

  @Override
  public String toString()
  {
    return "{"+rootPath+", "+name+"}";
  }
}

/** file data
 */
class FileData
{
  // --------------------------- constants --------------------------------
  /** file types
   */
  enum Types
  {
    NONE,

    FILE,
    DIRECTORY,
    LINK,
    DEVICE,
    SPECIAL,
    FIFO,
    SOCKET,

    UNKNOWN
  };

  /** file states
   */
  enum States
  {
    NONE,

    OK,                                    // local copy OK and up-to-date
    UNKNOWN,                               // unknown, not under revison control
    MODIFIED,                              // local copy modified
    CHECKOUT,                              // does not exists local, checkout required
    UPDATE,                                // local copy not up-to-date, updated required
    MERGE,                                 // merge required
    CONFLICT,                              // conflict, resolve required
    ADDED,                                 // added, but still not commited
    REMOVED,                               // removed, but still not commited
    RENAMED,                               // renamed, but still not commited
    NOT_EXISTS,                            // not found
    WAITING,                               // waiting for operation
    ERROR;

    /** parse type string
     * @param string state string
     * @return state
     */
    static States parse(String string)
    {
      States state;

      if      (string.equalsIgnoreCase("ok")        ) state = States.OK;
      else if (string.equalsIgnoreCase("unknown")   ) state = States.UNKNOWN;
      else if (string.equalsIgnoreCase("modified")  ) state = States.MODIFIED;
      else if (string.equalsIgnoreCase("checkout")  ) state = States.CHECKOUT;
      else if (string.equalsIgnoreCase("update")    ) state = States.UPDATE;
      else if (string.equalsIgnoreCase("merge")     ) state = States.MERGE;
      else if (string.equalsIgnoreCase("conflict")  ) state = States.CONFLICT;
      else if (string.equalsIgnoreCase("added")     ) state = States.ADDED;
      else if (string.equalsIgnoreCase("removed")   ) state = States.REMOVED;
      else if (string.equalsIgnoreCase("renamed")   ) state = States.RENAMED;
      else if (string.equalsIgnoreCase("not exists")) state = States.NOT_EXISTS;
      else if (string.equalsIgnoreCase("waiting")   ) state = States.WAITING;
      else if (string.equalsIgnoreCase("error")     ) state = States.ERROR;
      else                                            state = States.OK;

      return state;
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case OK:         return "ok";
        case UNKNOWN:    return "unknown";
        case MODIFIED:   return "modified";
        case CHECKOUT:   return "checkout";
        case UPDATE:     return "update";
        case MERGE:      return "merge";
        case CONFLICT:   return "conflict";
        case ADDED:      return "added";
        case REMOVED:    return "removed";
        case RENAMED:    return "renamed";
        case NOT_EXISTS: return "not exists";
        case WAITING:    return "waiting";
        case ERROR:      return "error";
        default:         return "unknown state";
      }
    }
  };

  /** state sets
   */
  final static EnumSet<States> STATES_ALL   = EnumSet.allOf(States.class);
  final static EnumSet<States> STATES_KNOWN = EnumSet.of(States.OK,
                                                         States.MODIFIED,
                                                         States.CHECKOUT,
                                                         States.UPDATE,
                                                         States.MERGE,
                                                         States.CONFLICT,
                                                         States.ADDED,
                                                         States.REMOVED,
                                                         States.RENAMED
                                                        );
  final static EnumSet<States> STATES_NONE  = EnumSet.allOf(States.class);

  /** file modes
   */
  enum Modes
  {
    NONE,

    TEXT,
    BINARY,

    UNKNOWN;

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case TEXT:   return "text";
        case BINARY: return "binary";
        default:     return "unknown";
      }
    }
  };

  // --------------------------- variables --------------------------------
  public final   String   rootPath;
//  public         String   name;                  // file name (from root path)
  public         Types    type;                  // file type
  public         long     size;                  // size [bytes]
  public         Date     date;
  public         States   state;                 // current state
  public         Modes    mode;
  public         boolean  locked;
  public         String   workingRevision;
  public         String   repositoryRevision;
  public         String   branch;

  static private HashMap<String,WeakReference<FileData>> fileDataMap = new HashMap<String,WeakReference<FileData>>();

  private        File     file;

//  public       TreeItem treeItem;              // tree item or null
  private        HashSet<TreeItem> treeItemSet;  // tree item set

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get file data instance
   * @param rootPath root path
   * @param name relative name
   * @param type file type (see FileData.Types)
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   * @param size file size [bytes]
   * @param date file date/time
   * @param workingRevision working revsion
   * @param repositoryRevision repository revision
   * @param branch branch name
   */
  public static FileData getInstance(String rootPath, String name, FileData.Types type, FileData.States state, FileData.Modes mode, boolean locked, long size, Date date, String workingRevision, String repositoryRevision, String branch)
  {
    assert !rootPath.isEmpty();

    FileData fileData;

    // get absolute name
    File file = new File(name);
    String absoluteName = file.isAbsolute() ? name : new File(rootPath,name).getAbsolutePath();

    // get/create file data instance
    WeakReference<FileData> fileDataReference = fileDataMap.get(absoluteName);
    if ((fileDataReference == null) || (fileDataReference.get() == null))
    {
      fileData = new FileData(rootPath,absoluteName);
      fileDataMap.put(absoluteName,new WeakReference<FileData>(fileData));
    }
    else
    {
      fileData = fileDataReference.get();
    }

    // save data
    fileData.type               = type;
    fileData.state              = state;
    fileData.mode               = mode;
    fileData.locked             = locked;
    fileData.size               = size;
    fileData.date               = date;
    fileData.workingRevision    = workingRevision;
    fileData.repositoryRevision = repositoryRevision;
    fileData.branch             = branch;

/*
{
System.gc();
int n = 0;
for (WeakReference<FileData> w : fileDataMap.values()) { if (w.get() != null) n++; }
Dprintf.dprintf("n=%d",n);
}
*/

    return fileData;
  }

  /** get file data instance
   * @param rootPath root path
   * @param name name
   * @param type file type (see FileData.Types)
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   * @param size file size [bytes]
   * @param date file date/time
   * @param workingRevision working revsion
   * @param repositoryRevision repository revision
   */
  public static FileData getInstance(String rootPath, String name, FileData.Types type, FileData.States state, FileData.Modes mode, boolean locked, long size, Date date, String workingRevision, String repositoryRevision)
  {
    return getInstance(rootPath,name,type,state,mode,locked,size,date,workingRevision,repositoryRevision,"");
  }

  /** get file data instance
   * @param rootPath root path
   * @param name name
   * @param type file type (see FileData.Types)
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   * @param size file size [bytes]
   * @param date file date/time
   */
  public static FileData getInstance(String rootPath, String name, FileData.Types type, FileData.States state, FileData.Modes mode, boolean locked, long size, Date date)
  {
    return getInstance(rootPath,name,type,state,mode,locked,size,date,"","");
  }

  /** get file data instance
   * @param rootPath root path
   * @param name name
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   */
  public static FileData getInstance(String rootPath, String name, FileData.States state, FileData.Modes mode, boolean locked)
  {
    return getInstance(rootPath,name,FileData.Types.FILE,state,mode,locked,0L,new Date());
  }

  /** get file data instance
   * @param rootPath root path
   * @param name name
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   */
  public static FileData getInstance(String rootPath, String name, FileData.States state, FileData.Modes mode)
  {
    return getInstance(rootPath,name,state,mode,false);
  }

  /** get file data instance
   * @param rootPath root path
   * @param name name
   * @param type file type (see FileData.Types)
   * @param size file size [bytes]
   * @param date file date/time
   */
  public static FileData getInstance(String rootPath, String name, FileData.Types type, long size, Date date)
  {
    return getInstance(rootPath,name,type,FileData.States.UNKNOWN,FileData.Modes.UNKNOWN,false,size,date);
  }

  /** get file data instance
   * @param rootPath root path
   * @param name file name
   * @param type file type (see FileData.Types)
   * @param date file date/time
   */
  public static FileData getInstance(String rootPath, String name, FileData.Types type, Date date)
  {
    return getInstance(rootPath,name,type,0L,date);
  }

  /** get file data instance
   * @param rootPath root path
   * @param name name
   * @param type file type (see FileData.Types)
   */
  public static FileData getInstance(String rootPath, String name, FileData.Types type)
  {
    return getInstance(rootPath,name,type,0L,new Date());
  }


  /** get file data instance
   * @param rootPath root path
   * @param name name
   * @param state file state
   */
  public static FileData getInstance(String rootPath, String name, FileData.States state)
  {
    return getInstance(rootPath,name,state,FileData.Modes.UNKNOWN,false);
  }

  /** get file data instance
   * @param rootPath root path
   * @param name name
   */
  public static FileData getInstance(String rootPath, String name)
  {
    return getInstance(rootPath,name,FileData.States.UNKNOWN);
  }

  /** get file data instance
   * @param rootPath root path
   * @param file file
   */
  public static FileData getInstance(String rootPath, File file)
  {
//TODO:which?
    return getInstance(rootPath,FileUtils.getCanonicalName(file));
//    return getInstance(rootPath,file.getPath());
  }

  /** get get parent file data
   */
  public FileData getParent()
  {
    String parentName = file.getParent();
    return (parentName != null) ? getInstance(rootPath,parentName,Types.DIRECTORY) : null;
  }

  /** get name relative to root path
   * @param rootPath root path
   * @return name
   */
  public String getName(String rootPath)
  {
    String absoluteName     = file.getAbsolutePath();
    String absoluteRootPath = new File(rootPath).getAbsolutePath();
    return (absoluteName.length() > absoluteRootPath.length()) ? absoluteName.substring(absoluteRootPath.length()+1) : "";
  }

  /** get name relative to root path
   * @param repository repository
   * @return name
   */
  public String getName(Repository repository)
  {
    return getName(repository.rootPath);
  }

  /** get absolute name
   * @return name
   */
  public String getName()
  {
    return file.getAbsolutePath();
  }

  /** get base name (name without path)
   * @return name
   */
  public String getBaseName()
  {
    return file.getName();
  }

  /** set repository relative file name
   * @param name file name
   */
  public void setFileName(String newName)
  {
    // remove file data with old name
    fileDataMap.remove(file.getAbsolutePath());

    // create file instance with new name
    file = new File(newName);
    if (!file.isAbsolute())
    {
      file = new File(rootPath,newName);
    }

    // add file data with new name
    fileDataMap.put(file.getAbsolutePath(),new WeakReference<FileData>(this));
  }

  /** get absolute directory name
   * @param rootPath root path
   * @return directory name
   */
  public String getDirectoryName(String rootPath)
  {
    File parentFile = file.getParentFile();
    if (parentFile != null)
    {
      String absoluteName     = parentFile.getAbsolutePath();
      String absoluteRootPath = new File(rootPath).getAbsolutePath();
      return (absoluteName.length() > absoluteRootPath.length()) ? absoluteName.substring(absoluteRootPath.length()+1) : "";
    }
    else
    {
      return "";
    }
  }

  /** get absolute directory name
   * @param repository repository
   * @return directory name
   */
  public String getDirectoryName(Repository repository)
  {
    return getDirectoryName(repository.rootPath);
  }

  /** get absolute directory name
   * @return directory name
   */
  public String getDirectoryName()
  {
    File parentFile = file.getParentFile();
    return (parentFile != null) ? parentFile.getAbsolutePath() : "";
  }

  /** get file instance
   * @return file
   */
  public File getFile()
  {
    return file;
  }

  /** check if file exists
   * @return true iff exists
   */
  public boolean exists()
  {
    return file.exists();
  }

  /** check if file data reference same file
   * @param fileData other file data
   * @return true if same file is referenced
   */
  public boolean equalFiles(FileData fileData)
  {
    if (fileData != null)
    {
      return file.equals(fileData.file);
    }
    else
    {
      return false;
    }
  }

  /** get file permissions string
   * @return file permissions string
   */
  public String getPermissions()
  {
    StringBuilder buffer = new StringBuilder();

    if (file.canRead()) buffer.append("read");
    if (file.canWrite())
    {
      if (buffer.length() > 0) buffer.append('/');
      buffer.append("write");
    }
    if (file.canExecute())
    {
      if (buffer.length() > 0) buffer.append('/');
      buffer.append("execute");
    }

    return buffer.toString();
  }

  /** get file mime type
   * @return mime type or null if no mime type found
   */
  public String getMimeType()
  {
    return Onzen.getMimeType(file);
  }

  /** convert this to set
   * @return file data set
   */
  public Set<FileData> toSet()
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(this);

    return fileDataSet;
  }

  /** convert to sorted set
   * @param file data set
   * @param comparator comparator
   * @return sorted file data set
   */
  public static TreeSet<FileData> toSortedSet(Set<FileData> fileDataSet, Comparator comparator)
  {
    TreeSet<FileData> fileDataTreeSet = new TreeSet<FileData>(comparator);
    fileDataTreeSet.addAll(fileDataSet);
    return fileDataTreeSet;
  }

  /** convert to sorted set
   * @param file data set
   * @return sorted file data set
   */
  public static TreeSet<FileData> toSortedSet(Set<FileData> fileDataSet)
  {
    return toSortedSet(fileDataSet,new FileDataComparator());
  }

  /** convert to set
   * @param rootPath root path
   * @param fileNames file name array
   * @return file data set
   */
  public static Set<FileData> toSet(String rootPath, String[] fileNames)
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    for (String fileName : fileNames)
    {
      fileDataSet.add(getInstance(rootPath,fileName));
    }

    return fileDataSet;
  }

  /** convert to set
   * @param rootPath root path
   * @param fileNameSet file name set
   * @return file data set
   */
  public static Set<FileData> toSet(String rootPath, Set<String> fileNameSet)
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    for (String fileName : fileNameSet)
    {
      fileDataSet.add(getInstance(rootPath,fileName));
    }

    return fileDataSet;
  }

  /** convert to set
   * @param rootPath root path
   * @param fileName file name
   * @param otherFileNames optional other file names
   * @return file data set
   */
  public static Set<FileData> toSet(String rootPath, String fileName, String... otherFileNames)
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(FileData.getInstance(rootPath,fileName));
    for (String otherFileName : otherFileNames)
    {
      fileDataSet.add(FileData.getInstance(rootPath,otherFileName));
    }

    return fileDataSet;
  }

  /** convert to name list
   * @param file data set
   * @return file name list
   */
//TODO: use toSortedNameList()?
  public static List<String> toNameList(Set<FileData> fileDataSet, Repository repository)
  {
    List<String> nameList = new LinkedList<String>();
    for (FileData fileData : fileDataSet)
    {
      nameList.add(fileData.getName(repository));
    }

    return nameList;
  }

  /** convert to sorted name list
   * @param file data set
   * @param repository repository
   * @return file name list
   */
  public static List<String> toSortedNameList(Set<FileData> fileDataSet, Repository repository)
  {
    ArrayList<String> nameList = new ArrayList<String>();
    for (FileData fileData : fileDataSet)
    {
      nameList.add(fileData.getName(repository));
    }

    Collections.sort(nameList);

    return nameList;
  }

  /** convert to name array
   * @param file data set
   * @return file name list
   */
  public static List<String> toSortedNameList(Set<FileData> fileDataSet)
  {
    ArrayList<String> nameList = new ArrayList<String>();
    for (FileData fileData : fileDataSet)
    {
      nameList.add(fileData.getName());
    }

    Collections.sort(nameList);

    return nameList;
  }

  /** add tree item using file data
   * @param treeItem tree item to add
   */
  public void addTreeItem(TreeItem treeItem)
  {
    treeItemSet.add(treeItem);
  }

  /** remove tree item using file data
   * @param treeItem tree item to remove
   */
  public void removeTreeItem(TreeItem treeItem)
  {
    treeItemSet.remove(treeItem);
  }

  /** get all tree items using file data
   * @return tree items set
   */
  public HashSet<TreeItem> getTreeItems()
  {
    return treeItemSet;
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "FileData {name: '"+file.getAbsolutePath()+"', type: "+type+", state: "+state+", mode: "+mode+", revision: '"+workingRevision+"', branch: '"+branch+"'}";
  }

  // ------------------------------------------------------------------------

  /** create file data
   * @param name file name
   * @param type file type (see FileData.Types)
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   * @param size file size [bytes]
   * @param date file date/time
   * @param workingRevision working revsion
   * @param repositoryRevision repository revision
   * @param branch branch
   */
  private FileData(String rootPath, String absoluteName)
  {
    this.rootPath    = rootPath;
    this.file        = new File(absoluteName);
    this.treeItemSet = new HashSet<TreeItem>();
  }
}

/** file data comparator
 */
class FileDataComparator implements Comparator<FileData>
{
  // --------------------------- constants --------------------------------
  enum SortModes
  {
    NAME,
    TYPE,
    SIZE,
    DATE,
    STATE,
    MODE,
    WORKING_REVISION,
    REPOSITORY_REVISION,
    BRANCH
  };

  // --------------------------- variables --------------------------------
  private SortModes sortMode;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create file data comparator
   * @param sortMode sort mode
   */
  FileDataComparator(SortModes sortMode)
  {
    this.sortMode = sortMode;
  }

  /** create file data comparator
   */
  FileDataComparator()
  {
    this(SortModes.NAME);
  }

  /** get sort mode
   * @return sort mode
   */
  public SortModes getSortMode()
  {
    return sortMode;
  }

  /** set sort mode
   * @param sortMode sort mode
   */
  public void setSortMode(SortModes sortMode)
  {
    this.sortMode = sortMode;
  }

  /** compare file data
   * @param fileData0, fileData1 file data to compare
   * @return -1 iff fileData0 < fileData1,
              0 iff fileData0 = fileData1,
              1 iff fileData0 > fileData1
   */
  public int compare(FileData fileData0, FileData fileData1)
  {
    switch (sortMode)
    {
      case NAME:
        return fileData0.getName().compareTo(fileData1.getName());
      case TYPE:
        return fileData0.type.compareTo(fileData1.type);
      case SIZE:
        if      (fileData0.size < fileData1.size) return -1;
        else if (fileData0.size > fileData1.size) return  1;
        else                                      return  0;
      case DATE:
        if      (fileData0.date.before(fileData1.date)) return -1;
        else if (fileData0.date.after(fileData1.date))  return  1;
        else                                            return  0;
      case STATE:
        return fileData0.state.compareTo(fileData1.state);
      case MODE:
        return fileData0.type.compareTo(fileData1.type);
      case WORKING_REVISION:
        return fileData0.workingRevision.compareTo(fileData1.workingRevision);
      case REPOSITORY_REVISION:
        return fileData0.repositoryRevision.compareTo(fileData1.repositoryRevision);
      case BRANCH:
        return fileData0.branch.compareTo(fileData1.branch);
      default:
        return 0;
    }
  }
}

// ------------------------------------------------------------------------

/** diff data
 */
class DiffData
{
  // --------------------------- constants --------------------------------

  /** diff block types
   */
  enum Types
  {
    NONE,

    KEEP,
    ADDED,
    DELETED,
    CHANGED,
    CHANGED_WHITESPACES,

    UNKNOWN;

    /** parse type string
     * @param string state string
     * @return block type
     */
    static Types parse(String string)
    {
      Types type;

      if      (string.equalsIgnoreCase("keep")              ) type = Types.KEEP;
      else if (string.equalsIgnoreCase("added")             ) type = Types.ADDED;
      else if (string.equalsIgnoreCase("deleted")           ) type = Types.DELETED;
      else if (string.equalsIgnoreCase("changed")           ) type = Types.CHANGED;
      else if (string.equalsIgnoreCase("changedWhitespaces")) type = Types.CHANGED_WHITESPACES;
      else                                                    type = Types.UNKNOWN;

      return type;
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case KEEP:                return "ok";
        case ADDED:               return "added";
        case DELETED:             return "deleted";
        case CHANGED:             return "changed";
        case CHANGED_WHITESPACES: return "changedWhitespaces";
        default:                  return null;
      }
    }
  };

  // --------------------------- variables --------------------------------
  public final Types    type;               // diff block type
  public final String[] keepLines;          // lines not changed
  public final String[] addedLines;         // add lines
  public final String[] deletedLines;       // deleted lines

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create diff data
   * @param type block type
   * @param addedLines lines to add
   * @param deletedLines lines to delete
   */
  public DiffData(Types type, String[] addedLines, String[] deletedLines)
  {
    this.type = type;
    switch (type)
    {
      case KEEP:
        throw new Error("INTERNAL ERROR: invalid block type!");
      case ADDED:
        throw new Error("INTERNAL ERROR: invalid block type!");
      case DELETED:
        throw new Error("INTERNAL ERROR: invalid block type!");
      case CHANGED:
      case CHANGED_WHITESPACES:
        this.keepLines    = null;
        this.addedLines   = addedLines;
        this.deletedLines = deletedLines;
        break;
      default:
        throw new Error("INTERNAL ERROR: invalid block type!");
    }
  }

  /** create diff data
   * @param type block type
   * @param addedLines lines to add
   * @param deletedLines lines to delete
   */
  public DiffData(Types type, List<String> addedLinesList, List<String> deletedLinesList)
  {
    this(type,
         addedLinesList.toArray(new String[addedLinesList.size()]),
         deletedLinesList.toArray(new String[deletedLinesList.size()])
        );
  }

  /** create diff data
   * @param type block type
   * @param addedLines lines to add
   * @param deletedLinesList lines to delete
   */
  public DiffData(Types type, String[] addedLines, List<String> deletedLinesList)
  {
    this(type,
         addedLines,
         deletedLinesList.toArray(new String[deletedLinesList.size()])
        );
  }
  /** create diff data
   * @param type block type
   * @param addedLines lines to add
   * @param deletedLines lines to delete
   */
  public DiffData(Types type, List<String> addedLinesList, String[] deletedLines)
  {
    this(type,
         addedLinesList.toArray(new String[addedLinesList.size()]),
         deletedLines
        );
  }


  /** create diff data
   * @param type block type
   * @param lines lines to keep/add/delete
   */
  public DiffData(Types type, String[] lines)
  {
    this.type = type;
    switch (type)
    {
      case KEEP:
        this.keepLines    = lines;
        this.addedLines   = null;
        this.deletedLines = null;
        break;
      case ADDED:
        this.keepLines    = null;
        this.addedLines   = lines;
        this.deletedLines = null;
        break;
      case DELETED:
        this.keepLines    = null;
        this.addedLines   = null;
        this.deletedLines = lines;
        break;
      case CHANGED:
      case CHANGED_WHITESPACES:
        throw new Error("INTERNAL ERROR: invalid block type!");
      default:
        throw new Error("INTERNAL ERROR: invalid block type!");
    }
  }

  /** create diff data
   * @param type block type
   * @param lines lines to keep/add/delete
   */
  public DiffData(Types type, List<String> linesList)
  {
    this(type,
         linesList.toArray(new String[linesList.size()])
        );
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "DiffData {"+type+", keep "+((keepLines != null) ? keepLines.length : 0)+": '"+StringUtils.join(keepLines,"\\n")+"', added "+((addedLines != null) ? addedLines.length : 0)+": '"+StringUtils.join(addedLines,"\\n")+"' deleted "+((deletedLines != null) ? deletedLines.length : 0)+": '"+StringUtils.join(deletedLines,"\\n")+"'}";
  }
}

// ------------------------------------------------------------------------

/** branch data
 */
class BranchData
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
  public final String             name;
  public final Date               date;
  public final String             author;
  public final LinkedList<String> commitMessageList;
  public final RevisionDataTree   revisionDataTree;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create branch data
   * @param name name
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   * @param revisionDataTree branch revision data tree
   */
  public BranchData(String name, Date date, String author, LinkedList<String> commitMessageList, RevisionDataTree revisionDataTree)
  {
    this.name              = name;
    this.date              = date;
    this.author            = author;
    this.commitMessageList = commitMessageList;
    this.revisionDataTree  = revisionDataTree;
  }

  /** create branch data
   * @param name name
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   */
  public BranchData(String name, Date date, String author, LinkedList<String> commitMessageList)
  {
    this(name,date,author,commitMessageList,(RevisionDataTree)null);
  }

  /** check if revision or sub-revision is from author
   * @return true iff revision or some sub-revision is from author
   */
  public boolean isFromAuthor(String fromAuthor)
  {
    // check revision author
    if (author.equalsIgnoreCase(fromAuthor)) return true;

    // check revision data
    for (RevisionData revisionData : revisionDataTree)
    {
      if (revisionData.isFromAuthor(fromAuthor)) return true;
    }

    return false;
  }

  /** output branch
   * @param prefix prefix text
   */
  public void print(String prefix)
  {
    System.out.println(prefix+name);
    if (!revisionDataTree.isEmpty()) revisionDataTree.print(prefix+"  ");
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "Branch data {"+name+", date: "+date+", author: "+author+", message: "+StringUtils.join(commitMessageList,", ")+", revisions: "+revisionDataTree+"}";
  }
}

/** branch data list
 */
class BranchDataList extends ArrayList<BranchData>
{
  /** create branch data list
   * @param branchData branch data
   */
  BranchDataList(BranchData... branchData)
  {
    super();

    for (BranchData branchData_ : branchData)
    {
      add(branchData_);
    }
  }

  /** check if revision or sub-revision is from author
   * @return true iff revision or some sub-revision is from author
   */
  public boolean isFromAuthor(String fromAuthor)
  {
    for (BranchData branchData : this)
    {
      if (branchData.isFromAuthor(fromAuthor)) return true;
    }

    return false;
  }

  /** output revision tree
   * @param prefix prefix text
   */
  public void print(String prefix)
  {
    for (BranchData branchData : this)
    {
      branchData.print("  "+prefix);
    }
  }

  /** output revision tree
   */
  public void print()
  {
    print("");
  }
}

/** merge data list
 */
class MergeDataList extends ArrayList<RevisionData>
{
  /** create merge data list
   * @param revisionData revision data
   */
  MergeDataList(RevisionData... revisionData)
  {
    super();

    for (RevisionData revisionData_ : revisionData)
    {
      add(revisionData_);
    }
  }
}

// ------------------------------------------------------------------------

/** tag list
 */
class TagList extends ArrayList<String>
{
  /** create tag list
   * @param tags tag names
   */
  public TagList(String... tags)
  {
    super();

    for (String tag : tags)
    {
      add(tag);
    }
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return StringUtils.join(this);
  }
}

// ------------------------------------------------------------------------

/** revision data
 */
class RevisionData implements Comparable
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
  public final String             revisionId;                               // revision id string
  public RevisionData             parentRevisionData0,parentRevisionData1;  // parents or null
  public final TagList            tagList;                                  // tag names of revison
  public final Date               date;                                     // commit/push date
  public final String             author;                                   // author name
  public final LinkedList<String> commitMessageList;                        // comment message text
  public final ArrayList<String>  fileNameList;                             // names of changed files
  public final BranchDataList     branchDataList;
  public final MergeDataList      mergeDataList;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create revision data
   * @param revisionId revision id string
   * @param parentRevisionData0,parentRevisionData1 parents of this revision or null
   * @param tags tag names of this revision or null
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   * @param fileNames names of changed files
   */
  public RevisionData(String revisionId, RevisionData parentRevisionData0, RevisionData parentRevisionData1, TagList tagList, Date date, String author, LinkedList<String> commitMessageList, ArrayList<String> fileNameList)
  {
    this.revisionId          = revisionId;
    this.parentRevisionData0 = parentRevisionData0;
    this.parentRevisionData1 = parentRevisionData1;
    this.tagList             = tagList;
    this.date                = date;
    this.author              = author;
    this.commitMessageList   = commitMessageList;
    this.fileNameList        = fileNameList;
    this.branchDataList      = new BranchDataList();
    this.mergeDataList       = new MergeDataList();

    if (parentRevisionData0 != null) parentRevisionData0.mergeDataList.add(this);
    if (parentRevisionData1 != null) parentRevisionData1.mergeDataList.add(this);
  }

  /** create revision data
   * @param revisionId revision id string
   * @param parentRevisionData0,parentRevisionData1 parents of this revision or null
   * @param tag tag name or null
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   * @param fileNameList names of changed files
   */
  public RevisionData(String revisionId, RevisionData parentRevisionData0, RevisionData parentRevisionData1, String tag, Date date, String author, LinkedList<String> commitMessageList, ArrayList<String> fileNameList)
  {
    this(revisionId,
         parentRevisionData0,
         parentRevisionData1,
         (tag != null) ? new TagList(tag) : (TagList)null,
         date,
         author,
         commitMessageList,
         fileNameList
        );
  }

  /** create revision data
   * @param revisionId revision id string
   * @param tagList tag names of this revision or null
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   * @param fileNameList names of changed files
   */
  public RevisionData(String revisionId, TagList tagList, Date date, String author, LinkedList<String> commitMessageList, ArrayList<String> fileNameList)
  {
    this(revisionId,
         (RevisionData)null,  // parentRevisionData0
         (RevisionData)null,  // parentRevisionData1
         tagList,
         date,
         author,
         commitMessageList,
         fileNameList
        );
  }

  /** create revision data
   * @param revisionId revision id string
   * @param tag tag name or null
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   * @param fileNameList names of changed files
   */
  public RevisionData(String revisionId, String tag, Date date, String author, LinkedList<String> commitMessageList, ArrayList<String> fileNameList)
  {
    this(revisionId,
         (RevisionData)null,  // parentRevisionData0
         (RevisionData)null,  // parentRevisionData1
         new TagList(),
         date,
         author,
         commitMessageList,
         fileNameList
        );
    if (tag != null) tagList.add(tag);
  }

  /** create revision data
   * @param revisionId revision id string
   * @param date date
   * @param author author name
   */
  public RevisionData(String revisionId, Date date, String author)
  {
    this(revisionId,
         (String)null,  // tag
         date,
         author,
         (LinkedList<String>)null,  // commitMessageList
         (ArrayList<String>)null  // fileNameList
        );
  }

  /** create revision data
   * @param revisionId revision
   */
  public RevisionData(String revisionId)
  {
    this(revisionId,
         new Date(),
         ""
        );
  }

  /** compare revisions
   * @param data other revision data
   * @return always 1 (>)
   */
  public int compareTo(Object data)
  {
    return 1;
  }

  /** check if revision is a merge
   * @return true iff revision is a merge
   */
  public boolean isMerge()
  {
    return (parentRevisionData0 != null) && (parentRevisionData1 != null);
  }

  /** check if revision is part of a branch
   * @return true iff revision is part of a branch
   */
  public boolean isBranched()
  {
    return !branchDataList.isEmpty();
  }

  /** check if revision is part of a merge (either a merge or merged into some other branch)
   * @return true iff revision is part of a merge
   */
  public boolean isMerged()
  {
    return    ((parentRevisionData0 != null) && (parentRevisionData1 != null))
           || !mergeDataList.isEmpty();
  }

  /** check if revision is part of a branch or a merge
   * @return true iff revision is part of a branch or merge
   */
  public boolean isBranchedOrMerged()
  {
    return isBranched() || isMerged();
  }

  /** check if revision or sub-revision is from author
   * @return true iff revision or some sub-revision is from author
   */
  public boolean isFromAuthor(String fromAuthor)
  {
    // check revision author
    if (author.equalsIgnoreCase(fromAuthor)) return true;

    // check branches
    if (branchDataList.isFromAuthor(fromAuthor)) return true;

    return false;
  }

  /** get revision text for display
   * @return revision text
   */
  public String getRevisionText()
  {
    StringBuilder buffer = new StringBuilder();

    buffer.append(revisionId);
    if (!tagList.isEmpty())
    {
      buffer.append(" (");
      buffer.append(tagList.toString());
      buffer.append(")");
    }

    return buffer.toString();
  }

  /** add branch
   * @param name branch name
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   * @param subRevisionDataTree branch sub-revision data tree
   * @return branch data
   */
  public BranchData addBranch(String name, Date date, String author, LinkedList<String> commitMessageList, RevisionDataTree subRevisionDataTree)
  {
    BranchData branchData = new BranchData(name,date,author,commitMessageList,subRevisionDataTree);
    branchDataList.add(branchData);

    subRevisionDataTree.parentBranch = branchData;

    return branchData;
  }

  /** add branch
   * @param name branch name
   * @param date date
   * @param author author name
   * @param commitMessageList commit message
   * @param subRevisionData branch sub-revision data
   * @return branch data
   */
  public BranchData addBranch(String name, Date date, String author, LinkedList<String> commitMessageList, RevisionData subRevisionData)
  {
    RevisionDataTree revisionDataTree = new RevisionDataTree();
    revisionDataTree.add(subRevisionData);

    return addBranch(name,date,author,commitMessageList,revisionDataTree);
  }

  /** match author/commit message/branch name with text
   * @param text text
   * @return true if text match to (is contained in) author or commit message
   */
  public boolean match(String text)
  {
//TODO
    text = text.toLowerCase();
Dprintf.dprintf("find text=%s",text);

    if (getRevisionText().toLowerCase().indexOf(text) >= 0) { Dprintf.dprintf("getRevisionText"); return true; }
return false;

/*
    if (author.toLowerCase().indexOf(text) >= 0) { Dprintf.dprintf("author"); return true; }

    if (branchDataList != null)
    {
      for (TeeData teeData : branchDataList)
      {
        if (teeData.name.toLowerCase().indexOf(text) >= 0) { Dprintf.dprintf("branchDataList"); return true; }
      }
    }

    if (mergeDataList != null)
    {
      for (TeeData teeData : mergeDataList)
      {
        if (teeData.name.toLowerCase().indexOf(text) >= 0) { Dprintf.dprintf("mergeDataList"); return true; }
      }
    }

    if (commitMessageList != null)
    {
      for (String string : commitMessageList)
      {
        if (string.toLowerCase().indexOf(text) >= 0) { Dprintf.dprintf("commitMessageList"); return true; }
      }
    }

    return false;
*/
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "RevisionData {"+revisionId+", date: "+date+", author: "+author+", message: "+StringUtils.join(commitMessageList)+", branches: "+branchDataList.size()+", parents: "+parentRevisionData0+" "+parentRevisionData1+", files: "+StringUtils.join(fileNameList)+"}";
  }
}

/** revision data list
 */
//class RevisionDataList extends ArrayList<RevisionData>
class RevisionDataList extends TreeSet<RevisionData>
{
  /** create revision data list
   * @param revisionData revision data
   */
  RevisionDataList(RevisionData... revisionData)
  {
    super();

    for (RevisionData revisionData_ : revisionData)
    {
      add(revisionData_);
    }
  }
}

/** revision data tree
 */
//class RevisionDataTree extends LinkedList<RevisionData>
class RevisionDataTree extends TreeSet<RevisionData>
{
//TODO: required?
  BranchData parentBranch;

  /** create revision data treee
   * @param revisionData revision data
   */
  RevisionDataTree(RevisionData revisionData)
  {
    super();

    this.parentBranch = null;
    add(revisionData);
  }

  /** create revision data treee
   */
  RevisionDataTree()
  {
    super();

    this.parentBranch = null;
  }

  public RevisionData peekFirst()
  {
    return first();
  }

  public RevisionData peekLast()
  {
    return last();
  }

  /** output revision tree
   * @param prefix prefix text
   */
  public void print(String prefix)
  {
    int i = 0;
    for (RevisionData revisionData : this)
    {
      System.out.println(String.format("%s%3d: %s",prefix,i,revisionData.revisionId));
      if (!revisionData.branchDataList.isEmpty()) revisionData.branchDataList.print(prefix+"     ");
      i++;
    }
  }

  /** output revision tree
   * @param n number of nodes
   */
  public void print(int n)
  {
    System.out.println("--- Revision tree ---");
    int                    i        = 0;
    Iterator<RevisionData> iterator = iterator();
    while ((n > 0) && iterator.hasNext())
    {
      RevisionData revisionData = iterator.next();

      System.out.println(String.format("%3d: %s",i,revisionData.revisionId));
      if (!revisionData.branchDataList.isEmpty()) revisionData.branchDataList.print("     ");

      i++;
      n--;
    }
    System.out.println("---");
  }

  /** output revision tree
   */
  public void print()
  {
    print(size());
  }
}

/** log data
 */
class LogData
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
  public final String             revisionId;
  public final Date               date;
  public final String             author;              // name of author
  public final LinkedList<String> commitMessageList;   // commit message lines
  public final ArrayList<String>  fileNameList;        // names of changed files

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create log data
   * @param revisionId revision id
   * @param author author name
   * @param date date/time
   * @param commitMessageList commit message
   * @param fileNames names of changed files
   */
  public LogData(String revisionId, Date date, String author, LinkedList<String> commitMessageList, ArrayList<String> fileNameList)
  {
    this.revisionId        = revisionId;
    this.author            = author;
    this.date              = date;
    this.commitMessageList = (LinkedList<String>)commitMessageList.clone();
    this.fileNameList      = (ArrayList<String>)fileNameList.clone();
  }

  /** match revision/author/commit message/file name with text
   * @param text text
   * @return true if text match to (is contained in) author or commit message
   */
  public boolean match(String text)
  {
    text = text.toLowerCase();

    if (revisionId.toLowerCase().indexOf(text) >= 0) return true;

    if (author.toLowerCase().indexOf(text) >= 0) return true;

    if (commitMessageList != null)
    {
      for (String line : commitMessageList)
      {
        if (line.toLowerCase().indexOf(text) >= 0) return true;
      }
    }

    if (fileNameList != null)
    {
      for (String fileName : fileNameList)
      {
        if (fileName.toLowerCase().indexOf(text) >= 0) return true;
      }
    }

    return false;
  }

  /** get first commit message line or ""
   * @return commit message line or ""
   */
  public String getFirstCommitMessage()
  {
    return !commitMessageList.isEmpty() ? commitMessageList.get(0) : "";
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "LogData {"+revisionId+", author: "+author+", date/time: "+date+", commit message: "+StringUtils.join(commitMessageList,", ")+", file: "+StringUtils.join(fileNameList,", ")+"}";
  }
}

/** log data comparator
 */
class LogDataComparator implements Comparator<LogData>
{
  // --------------------------- constants --------------------------------
  enum SortModes
  {
    REVISION,
    DATETIME,
    AUTHOR,
    COMMIT_MESSAGE
  };

  // --------------------------- variables --------------------------------
  private SortModes sortMode;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create log data comparator
   * @param sortMode sort mode
   */
  LogDataComparator(SortModes sortMode)
  {
    this.sortMode = sortMode;
  }

  /** create log data comparator
   */
  LogDataComparator()
  {
    this(SortModes.REVISION);
  }

  /** get sort mode
   * @return sort mode
   */
  public SortModes getSortMode()
  {
    return sortMode;
  }

  /** set sort mode
   * @param sortMode sort mode
   */
  public void setSortMode(SortModes sortMode)
  {
    this.sortMode = sortMode;
  }

  /** compare log data
   * @param logData0, logData1 log data to compare
   * @return -1 iff logData0 < logData1,
              0 iff logData0 = logData1,
              1 iff logData0 > logData1
   */
  public int compare(LogData logData0, LogData logData1)
  {
    switch (sortMode)
    {
      case REVISION:
        try
        {
          long n0 = Long.parseLong(logData0.revisionId);
          long n1 = Long.parseLong(logData1.revisionId);
          if      (n0 < n1) return -1;
          else if (n0 > n1) return  1;
          else              return  0;
        }
        catch (NumberFormatException exception)
        {
          return logData0.revisionId.compareTo(logData1.revisionId);
        }
      case DATETIME:
        if      (logData0.date.before(logData1.date)) return -1;
        else if (logData0.date.after(logData1.date))  return  1;
        else                                          return  0;
      case AUTHOR:
        return logData0.author.compareTo(logData1.author);
      case COMMIT_MESSAGE:
        if ((logData0.commitMessageList != null) && !logData0.commitMessageList.isEmpty() && (logData1.commitMessageList != null) && !logData1.commitMessageList.isEmpty())
        {
          return logData0.commitMessageList.getFirst().compareTo(logData1.commitMessageList.getFirst());
        }
        else if ((logData0.commitMessageList != null) && !logData0.commitMessageList.isEmpty())
        {
          return -1;
        }
        else if ((logData1.commitMessageList != null) && !logData1.commitMessageList.isEmpty())
        {
          return 1;
        }
        else
        {
          return 0;
        }
      default:
        return 0;
    }
  }
}

/** annotation data
 */
class AnnotationData
{
  // --------------------------- constants --------------------------------

  // --------------------------- variables --------------------------------
  public final String revisionId;
  public final String author;
  public final Date   date;
  public final String line;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create annotation data
   * @param revisionId revision id string
   * @param author author name
   * @param date date
   * @param line text line
   */
  public AnnotationData(String revisionId, String author, Date date, String line)
  {
    this.revisionId = revisionId;
    this.author     = author;
    this.date       = date;
    this.line       = line;
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "AnnotationData {"+revisionId+", author: "+author+", date: "+date+", line: "+line+"}";
  }
}

// ------------------------------------------------------------------------

/** repository exception
 */
class RepositoryException extends Exception
{
  private String[] extendedMessage = new String[0];

  /** create repository exception
   * @param message message
   * @param cause exception cause
   * @param extendedErrorMessage extended error message lines
   * @param arguments optional format arguments for message
   */
  RepositoryException(String message, Exception cause, String[] extendedMessage, Object... arguments)
  {
    super(String.format(message,arguments),cause);
    this.extendedMessage = extendedMessage;
  }

  /** create repository exception
   * @param message message
   * @param cause exception cause
   * @param arguments optional format arguments for message
   */
  RepositoryException(String message, String[] extendedMessage, Object... arguments)
  {
    this(message,null,extendedMessage,arguments);
  }

  /** create repository exception
   * @param cause exception cause
   */
  RepositoryException(Exception cause)
  {
    super(cause.getMessage(),cause);
    this.extendedMessage = null;
  }

  /** create repository exception
   * @param message message
   * @param arguments optional format arguments for message
   */
  RepositoryException(String message, Object... arguments)
  {
    this(message,null,arguments);
  }

  /** get extended message lines
   * @return message lines array
   */
  public String[] getExtendedMessage()
  {
    return extendedMessage;
  }
}

//-----------------------------------------------------------------------

/** store local files into database
 */
class StoredFiles
{
  // --------------------------- constants --------------------------------
  private static final String FILES_DATABASE_NAME    = "files";
  private static final int    FILES_DATABASE_VERSION = 1;

  // --------------------------- variables --------------------------------
  private File file;   // temporary zip storage file

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** create storage of files
   * @param rootPath root path
   * @param fileDataSet file data set to store
   */
  public StoredFiles(String rootPath, Set<FileData> fileDataSet)
    throws RepositoryException
  {
    this.file = null;

    ZipOutputStream output = null;
    try
    {
      if (fileDataSet.size() > 0)
      {
        // create temporary file
        file = File.createTempFile("onzen",".zip",new File(Settings.tmpDirectory));

        // create ZIP storage
        output = new ZipOutputStream(new FileOutputStream(file));
        output.setLevel(ZipOutputStream.STORED);

        // store files
        byte[] buffer = new byte[64*1024];
        for (FileData fileData : fileDataSet)
        {
          File file = new File(fileData.getName(rootPath));
          if (file.exists())
          {
            // add ZIP entry to output stream.
            output.putNextEntry(new ZipEntry(fileData.getName(rootPath)));

            // store file data
            FileInputStream input = new FileInputStream(file);
            int n;
            while ((n = input.read(buffer)) > 0)
            {
              output.write(buffer,0,n);
            }
            input.close();

            // close ZIP entry
            output.closeEntry();
          }
        }

        // close ZIP storage
        output.close(); output = null;
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(exception);
    }
    finally
    {
      try { if (output != null) output.close(); } catch (IOException exception) { /* ignored */ }
    }
  }

  /** restore files stored in database
   * @return true iff changes restored
   */
  protected boolean restore()
    throws RepositoryException
  {
    ZipFile zipFile = null;
    try
    {
      if (file != null)
      {
        // open ZIP storage
        zipFile = new ZipFile(file);

        // read entries
        byte[] buffer = new byte[64*1024];
        Enumeration zipEntries = zipFile.entries();
        while (zipEntries.hasMoreElements())
        {
          // get entry
          ZipEntry zipEntry = (ZipEntry)zipEntries.nextElement();

          // restore file data
          InputStream      input  = zipFile.getInputStream(zipEntry);
          FileOutputStream output = new FileOutputStream(zipEntry.getName());
          int n;
          while ((n = input.read(buffer)) > 0)
          {
            output.write(buffer,0,n);
          }
          input.close();
        }

        // close ZIP storage
        zipFile.close();
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(exception);
    }
    finally
    {
      try { if (zipFile != null) zipFile.close(); } catch (IOException exception) { /* ignored */ }
    }

    return true;
  }

  /** discard stored files in database
   */
  protected void discard()
    throws RepositoryException
  {
    if (file != null)
    {
      file.delete();
    }
  }
}

// ------------------------------------------------------------------------

/** store repository URLs into database
 */
class RepositoryURL implements Cloneable
{
  private static final String REPOSITORY_URLS_DATABASE_NAME    = "repositoryURLsoooo";
  private static final int    REPOSITORY_URLS_DATABASE_VERSION = 1;

  enum Types
  {
    NONE,
    AUTO,
    MANUAL,
    UNKNOWN;

    /** parse string
     * @param string string
     * @return repository URL add type
     */
    public static Types parse(String string)
    {
      Types type;

      if      (string.equalsIgnoreCase("none")  ) type = NONE;
      else if (string.equalsIgnoreCase("auto")  ) type = AUTO;
      else if (string.equalsIgnoreCase("manual")) type = MANUAL;
      else                                        type = UNKNOWN;

      return type;
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case NONE:   return "none";
        case AUTO:   return "auto";
        case MANUAL: return "manual";
        default:     return "unknown";
      }
    }
  };

  public Types            addType;
  public Repository.Types repositoryType;
  public String           path;
  public String           moduleName;
  public String           userName;

  /** create repository URL
   * @param addType add type
   * @param repositoryType repository type
   * @param path path
   * @param moduleName module name
   * @param userName user login name or ""
   */
  RepositoryURL(Types addType, Repository.Types repositoryType, String path, String moduleName, String userName)
  {
    this.addType        = addType;
    this.repositoryType = repositoryType;
    this.path           = path;
    this.moduleName     = moduleName;
    this.userName       = userName;
  }

  /** create repository URL
   * @param addType add type
   * @param repositoryType repository type
   * @param path path
   */
  RepositoryURL(Types addType, Repository.Types repositoryType, String path)
  {
    this(addType,repositoryType,path,"","");
  }

  /** create repository URL
   * @param repositoryType repository type
   * @param path path
   * @param moduleName module name
   * @param userName user login name or ""
   */
  RepositoryURL(Repository.Types repositoryType, String path, String moduleName, String userName)
  {
    this(Types.NONE,repositoryType,path,moduleName,userName);
  }

  /** create repository URL
   * @param repositoryType repository type
   * @param path path
   */
  RepositoryURL(Repository.Types repositoryType, String path)
  {
    this(repositoryType,path,"","");
  }

  /** create repository URL
   * @param path path
   * @param moduleName module name
   * @param userName user login name or ""
   */
  RepositoryURL(String path, String moduleName, String userName)
  {
    this(Repository.Types.NONE,path,moduleName,userName);
  }

  /** create repository URL
   * @param path path
   */
  RepositoryURL(String path)
  {
    this(path,"","");
  }

  /** create repository URL
   */
  RepositoryURL()
  {
    this(Types.UNKNOWN,Repository.Types.UNKNOWN,"");
  }

  /** clone object
   * @return cloned object
   */
  public RepositoryURL clone()
  {
    return new RepositoryURL(addType,repositoryType,path,moduleName,userName);
  }

  /** check if repository URL is equal
   * @return true iff equals
   */
  public boolean equals(Object object)
  {
    RepositoryURL repositoryURL = (RepositoryURL)object;

    return    (repositoryURL.repositoryType == repositoryType)
           && repositoryURL.path.equals(path)
           && repositoryURL.userName.equals(userName);
  }

  /** get repository URL history
   * @return repository URL history list
   */
  public static List<RepositoryURL> getHistoryURLs()
  {
    LinkedList<RepositoryURL> repositoryURLList = new LinkedList<RepositoryURL>();

    HistoryDatabase historyDatabase = null;
    try
    {
      // open database
      historyDatabase = openHistoryDatabase();

      // get history
      repositoryURLList = historyDatabase.getHistory();

      // close database
      historyDatabase.close(); historyDatabase = null;
    }
    catch (SQLException exception)
    {
      Onzen.printWarning("Cannot get repository URLs from database (error: %s)",exception.getMessage());
      return new LinkedList<RepositoryURL>();
    }
    finally
    {
      try { if (historyDatabase != null) closeHistoryDatabase(historyDatabase); } catch (SQLException exception) { /* ignored */ }
    }

    return repositoryURLList;
  }

  /** set repository URL history
   * @param repositoryURLs repository URL history list
   */
  public static void setHistoryURLs(RepositoryURL[] repositoryURLs)
  {
    HistoryDatabase historyDatabase = null;
    try
    {
      // open database
      historyDatabase = openHistoryDatabase();

      // set history
      historyDatabase.setHistory(repositoryURLs);

      // close database
      closeHistoryDatabase(historyDatabase); historyDatabase = null;
    }
    catch (SQLException exception)
    {
      Onzen.printWarning("Cannot set repository URLs from database (error: %s)",exception.getMessage());
      return;
    }
    finally
    {
      try { if (historyDatabase != null) closeHistoryDatabase(historyDatabase); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** add repository URL to history
   */
  public void addToHistory()
  {
    HistoryDatabase historyDatabase = null;
    try
    {
      // open database
      historyDatabase = openHistoryDatabase();

      // add repository URL
      historyDatabase.add(this);

      // close database
      closeHistoryDatabase(historyDatabase); historyDatabase = null;
    }
    catch (SQLException exception)
    {
      Onzen.printWarning("Cannot store repository URL into database (error: %s)",exception.getMessage());
      return;
    }
    finally
    {
      try { if (historyDatabase != null) closeHistoryDatabase(historyDatabase); } catch (SQLException exception) { /* ignored */ }
    }
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "RepositoryURL {"+addType+", "+repositoryType+", module: "+moduleName+", path: "+path+", user name: "+userName+"}";
  }

  /** open repository URL history database
   * @return history database
   */
  private static HistoryDatabase openHistoryDatabase()
    throws SQLException
  {
    return new HistoryDatabase<RepositoryURL>("repositoryURLs",
                                              20,
                                              HistoryDatabase.Directions.SORTED
                                             )
    {
      public PreparedStatement prepareInit(Database database)
        throws SQLException
      {
        PreparedStatement preparedStatement = database.prepareStatement("CREATE TABLE IF NOT EXISTS repositoryURLs ( "+
                                                                        "  id         INTEGER PRIMARY KEY, "+
                                                                        "  datetime   INTEGER DEFAULT (DATETIME('now')), "+
                                                                        "  type       INTEGER, "+
                                                                        "  path       TEXT, "+
                                                                        "  moduleName TEXT, "+
                                                                        "  userName   TEXT "+
                                                                        ");"
                                                                       );
        return preparedStatement;
      }
      public PreparedStatement prepareInsert(Database database, RepositoryURL repositoryURL)
        throws SQLException
      {
        PreparedStatement preparedStatement = database.prepareStatement("INSERT INTO repositoryURLs (type,path,moduleName,userName) VALUES (?,?,?,?);");
        preparedStatement.setInt(1,repositoryURL.repositoryType.toOrdinal());
        preparedStatement.setString(2,repositoryURL.path);
        preparedStatement.setString(3,repositoryURL.moduleName);
        preparedStatement.setString(4,repositoryURL.userName);
        return preparedStatement;
      }
      public PreparedStatement prepareDelete(Database database, int id)
        throws SQLException
      {
        PreparedStatement preparedStatement = database.prepareStatement("DELETE FROM repositoryURLs WHERE id=?;");
        preparedStatement.setInt(1,id);
        return preparedStatement;
      }
      public RepositoryURL getResult(ResultSet resultSet)
        throws SQLException
      {
        Repository.Types repositoryType = Repository.Types.toEnum(resultSet.getInt("type"));
        String           path           = resultSet.getString("path"); if (path == null) path = "";
        String           moduleName     = resultSet.getString("moduleName"); if (moduleName == null) moduleName = "";
        String           userName       = resultSet.getString("userName"); if (userName == null) userName = "";

        return new RepositoryURL(repositoryType,path,moduleName,userName);
      }
      public int dataCompareTo(RepositoryURL repositoryURL0, RepositoryURL repositoryURL1)
      {
        int result;

        if      ((repositoryURL0 != null) && ((repositoryURL1 == null) || (repositoryURL0.repositoryType.toOrdinal() < repositoryURL1.repositoryType.toOrdinal())))
        {
          result = -1;
        }
        else if ((repositoryURL0 != null) && ((repositoryURL1 == null) || (repositoryURL0.repositoryType.toOrdinal() > repositoryURL1.repositoryType.toOrdinal())))
        {
          result = 1;
        }
        else if ((repositoryURL0 != null) && (repositoryURL1 != null))
        {
          if      ((repositoryURL0.path != null) && (repositoryURL1.path == null))
          {
            result = -1;
          }
          else if ((repositoryURL1.path != null) && (repositoryURL0.path == null))
          {
            result = 1;
          }
          else
          {
            result = repositoryURL0.path.compareTo(repositoryURL1.path);
            if (result == 0)
            {
              if      ((repositoryURL0.moduleName != null) && (repositoryURL1.moduleName == null))
              {
                result = -1;
              }
              else if ((repositoryURL1.moduleName != null) && (repositoryURL0.moduleName == null))
              {
                result = 1;
              }
              else
              {
                result =repositoryURL0.moduleName.compareTo(repositoryURL1.moduleName);
                if (result == 0)
                {
                  if      ((repositoryURL0.userName != null) && (repositoryURL1.userName == null))
                  {
                    result = -1;
                  }
                  else if ((repositoryURL1.userName != null) && (repositoryURL0.userName == null))
                  {
                    result = 1;
                  }
                  else
                  {
                    result =repositoryURL0.userName.compareTo(repositoryURL1.userName);
                  }
                }
              }
            }
          }
        }
        else
        {
          result = 0;
        }
//Dprintf.dprintf("repositoryURL0=%s",repositoryURL0);
//Dprintf.dprintf("repositoryURL1=%s",repositoryURL1);
//Dprintf.dprintf("result=%s",result);

        return result;
      }
    };
  }

  /** close repository URL history database
   * @param historyDatabase history database
   */
  private static void closeHistoryDatabase(HistoryDatabase historyDatabase)
    throws SQLException
  {
    historyDatabase.close();
  }
}

// ------------------------------------------------------------------------

// ------------------------------------------------------------------------

/** config value annotation for repository
 */
@Target({TYPE,FIELD})
@Retention(RetentionPolicy.RUNTIME)
@interface RepositoryValue
{
  String  name()         default "";                 // name of value
  String  defaultValue() default "";                 // default value
  Class   type()         default DEFAULT.class;      // adapter class
  boolean pathSelector() default false;              // true for path selector widget
  int     min()          default Integer.MIN_VALUE;  // min. value
  int     max()          default Integer.MAX_VALUE;  // max. value
  String  title()        default "";                 // title of value
  String  text()         default "";                 // text of value (boolean)
  String  tooltip()      default "";                 // tooltip for value

  static final class DEFAULT
  {
  }
}

/** repository value adapter
 */
abstract class RepositoryValueAdapter
{
  /** create widget
   * @param composite composite
   * @param value value
   * @return widget
   */
  public abstract Control createWidget(Composite composite, String value);

  /** get value from widget
   * @param control control
   * @return value
   */
  public abstract String get(Control control);
}

/** password handler
 */
abstract class PasswordHandler
{
  /** get password
   * @param loginName login name
   * @param hostName host name
   * @return password or null
   */
  public abstract String getPassword(String loginName, String hostName);
}

/** repository
 */
@XmlType(propOrder={"title",
                    "selected",
                    "rootPath",
                    "comment",
                    "openDirectories",
                    "skipConvertWhitespacesFilePatterns",
                    "ignoreFilePatterns",

                    "patchTests",

                    "mailSMTPHost",
                    "mailSMTPPort",
                    "mailSMTPSSL",
                    "mailLogin",
                    "mailFrom",

                    "patchMailFlag",
                    "patchMailTo",
                    "patchMailCC",
                    "patchMailSubject",
                    "patchMailText",

                    "reviewServerFlag",
                    "reviewServerHost",
                    "reviewServerLogin",
                    "reviewServerSummary",
                    "reviewServerRepository",
                    "reviewServerGroups",
                    "reviewServerPersons",
                    "reviewServerDescription",

                    "autoSummaryFlag"
                   }
        )
@XmlSeeAlso({RepositoryNone.class,RepositoryDirectory.class,RepositoryCVS.class,RepositorySVN.class,RepositoryHG.class,RepositoryGIT.class})
@XmlAccessorType(XmlAccessType.NONE)
abstract class Repository implements Serializable
{
  /** special
   */
  class Special
  {
  }

  /** special commands
   */
  abstract class SpecialCommand extends Special
  {
    // shortcut
    int accelerator;

    /** create special command
     * @param title menu title
     * @param accelerator accelerator or SWT.NONE
     */
    SpecialCommand(int accelerator)
    {
      this.accelerator = accelerator;
    }

    /** create special command
     * @param title menu title
     */
    SpecialCommand()
    {
      this(SWT.NONE);
    }

    /** get title text
     * @return title text
     */
    public abstract String getTitle();

    /** execute special command
     * @param shell shell
     * @return true iff command executed
     */
    public abstract boolean run(Shell shell)
      throws RepositoryException;
  }

  /** special sub commands
   */
  class SpecialSubCommands extends Special
  {
    private String  title;
    private Special specials[];

    /** create special command
     * @param title menu title
     * @param accelerator accelerator or SWT.NONE
     */
    SpecialSubCommands(String title, Special... specials)
    {
      this.title    = title;
      this.specials = specials;
    }

    /** get title text
     * @return title text
     */
    public String getTitle()
    {
      return title;
    }

    /** get special commands for repository
     * @return array with special commands
     */
    public Special[] getSpecials()
    {
      return specials;
    }
  }

  /** special commands separator
   */
  class SpecialSeparator extends Special
  {
  }

  /** store local changes into into a temporary patch file
   */
  class StoredChanges
  {
    // --------------------------- constants --------------------------------

    // --------------------------- variables --------------------------------
    private File patchFile;

    // ------------------------ native functions ----------------------------

    // ---------------------------- methods ---------------------------------

    /** create stored chnages of local changes
     * @param fileDataSet file data set to store changes
     */
    public StoredChanges(Set<FileData> fileDataSet)
      throws RepositoryException
    {
      patchFile = getPatchFile(fileDataSet);
    }

    /** create stored chnages of all local changes
     */
    public StoredChanges()
      throws RepositoryException
    {
      this(getChangedFiles(FileData.STATES_KNOWN));
    }

    /** restore changes stored in database
     * @return true iff changes restored
     */
    public boolean restore()
      throws RepositoryException
    {
      // create patch from patch file
      Patch patch = new Patch(rootPath,patchFile);

      // apply patch to restore changes
      if (!patch.apply())
      {
        return false;
      }

      return true;
    }

    /** restore changes stored in database
     */
    public void discard()
      throws RepositoryException
    {
      if (patchFile != null)
      {
        patchFile.delete(); patchFile = null;
      }
    }
  }

  // --------------------------- constants --------------------------------

  /** repository types
   */
  enum Types
  {
    NONE,

    DIRECTORY,
    CVS,
    SVN,
    HG,
    GIT,

    UNKNOWN;

    /** parse string
     * @param string string
     * @return repository type
     */
    public static Types parse(String string)
    {
      Types type;

      if      (string.equalsIgnoreCase("none")     ) type = NONE;
      else if (string.equalsIgnoreCase("directory")) type = DIRECTORY;
      else if (string.equalsIgnoreCase("cvs")      ) type = CVS;
      else if (string.equalsIgnoreCase("svn")      ) type = SVN;
      else if (string.equalsIgnoreCase("hg")       ) type = HG;
      else if (string.equalsIgnoreCase("git")      ) type = GIT;
      else                                           type = UNKNOWN;

      return type;
    }

    /** convert ordinal value to enum
     * @param value ordinal value
     * @return enum
     */
    static Types toEnum(int value)
    {
      switch (value)
      {
        case 0:  return NONE;
        case 1:  return DIRECTORY;
        case 2:  return CVS;
        case 3:  return SVN;
        case 4:  return HG;
        case 5:  return GIT;
        default: return NONE;
      }
    }

    /** convert to ordinal value
     * @return ordinal value
     */
    public int toOrdinal()
    {
      switch (this)
      {
        case NONE:      return 0;
        case DIRECTORY: return 1;
        case CVS:       return 2;
        case SVN:       return 3;
        case HG:        return 4;
        case GIT:       return 5;
        default:        return 0;
      }
    }

    /** convert to string
     * @return string
     */
    public String toString()
    {
      switch (this)
      {
        case NONE:      return "none";
        case DIRECTORY: return "DIRECTORY";
        case CVS:       return "CVS";
        case SVN:       return "SVN";
        case HG:        return "HG";
        case GIT:       return "GIT";
        default:        return "unknown";
      }
    }
  };

  // --------------------------- variables --------------------------------
  @XmlAttribute(name = "title", required=true)
  public String title;
  @XmlAttribute(name = "selected")
  public boolean selected;

  @XmlElement(name = "rootPath")
  public String rootPath;

  @XmlElement(name = "comment")
  public String comment;

  @XmlElementWrapper(name = "openDirectories")
  @XmlElement(name = "path")
  private HashSet<String> openDirectories;

  @XmlElementWrapper(name = "skipConvertWhitespacesFilePatterns")
  @XmlElement(name = "pattern")
  private HashSet<String> skipConvertWhitespacesFilePatterns;

  @XmlElementWrapper(name = "ignoreFilePatterns")
  @XmlElement(name = "pattern")
  private HashSet<String> ignoreFilePatterns;

  @XmlElementWrapper(name = "patchTests")
  @XmlElement(name = "patchTests", defaultValue = "")
  public String[] patchTests = new String[0];

  @XmlElement(name = "mailSMTPHost")
  public String  mailSMTPHost;
  @XmlElement(name = "mailSMTPPort")
  public int     mailSMTPPort;
  @XmlElement(name = "mailSMTPSSL")
  public boolean mailSMTPSSL;
  @XmlElement(name = "mailLogin")
  public String  mailLogin;
  @XmlElement(name = "mailFrom")
  public String  mailFrom;

  @XmlElement(name = "patchMail")
  public boolean patchMailFlag;
  @XmlElement(name = "patchMailTo", defaultValue = "")
  public String patchMailTo = "";
  @XmlElement(name = "patchMailCC", defaultValue = "")
  public String patchMailCC = "";
  @XmlElement(name = "patchMailSubject", defaultValue = "Patch #${n %04d}: ${summary}")
  public String patchMailSubject = "Patch #${n %04d}: ${summary}";
  @XmlElement(name = "patchMailText", defaultValue = "${message}\n\n${tests - %s}\n\nComment: ${comment}")
  public String patchMailText = "${message}\n\n${tests - %s}";

  @XmlElement(name = "reviewServer")
  public boolean reviewServerFlag;
  @XmlElement(name = "reviewServerHost")
  public String  reviewServerHost;
  @XmlElement(name = "reviewServerLogin")
  public String  reviewServerLogin;
  @XmlElement(name = "reviewServerSummary", defaultValue = "Patch #${n %04d}: ${summary}")
  public String reviewServerSummary = "Patch #${n %04d}: ${summary}";
  @XmlElement(name = "reviewServerRepository")
  public String  reviewServerRepository;
  @XmlElement(name = "reviewServerGroups")
  public String  reviewServerGroups;
  @XmlElement(name = "reviewServerPersons")
  public String  reviewServerPersons;
  @XmlElement(name = "reviewServerDescription", defaultValue = "${message}\n\n${tests - %s}\n\nComment: ${comment}")
  public String reviewServerDescription = "${message}\n\n${tests - %s}";

  @XmlElement(name = "autoSummary")
  public boolean autoSummaryFlag;

  protected PasswordHandler passwordHandler;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get type of repository in path
   * @param rootPath root path
   * @return repository type or UNKNOWN
   */
  public static Types getType(String rootPath)
  {
    Types type = Types.UNKNOWN;

/*
    // Note: check if multiple types are detected
    File directory = new File(rootPath);
    while (directory != null)
    {
      File file;

      file = new File(directory,"CVS");
      if (file.exists() && file.isDirectory())
      {
        if (type != Types.UNKNOWN) return Types.UNKNOWN;
        type = Types.CVS;
      }
      file = new File(directory,".svn");
      if (file.exists() && file.isDirectory())
      {
        if (type != Types.UNKNOWN) return Types.UNKNOWN;
        type = Types.SVN;
      }
      file = new File(new File(directory,".hg"),"hgrc");
      if (file.exists() && file.isDirectory())
      {
        if (type != Types.UNKNOWN) return Types.UNKNOWN;
        type = Types.HG;
      }
      file = new File(new File(directory,".git"),"config");
      if (file.exists() && file.isFile())
      {
        if (type != Types.UNKNOWN) return Types.UNKNOWN;
        type = Types.GIT;
      }

      directory = directory.getParentFile();
    }
*/
    if (RepositoryCVS.isRepository(rootPath)) type = Types.CVS;
    if (RepositorySVN.isRepository(rootPath)) type = Types.SVN;
    if (RepositoryHG.isRepository(rootPath) ) type = Types.HG;
    if (RepositoryGIT.isRepository(rootPath)) type = Types.GIT;

    return type;
  }

  /** create new repository instance
   * @param type repository type; see Repository.Types
   * @param rootPath root path
   * @param userName user login name or ""
   * @param passwordHandler password handler to get password
   * @param comment comment text
   * @return repository or null if no repository found
   */
  public static Repository newInstance(Types type, String rootPath, String userName, PasswordHandler passwordHandler, String comment)
    throws RepositoryException
  {
    Repository repository;

    switch (type)
    {
      case DIRECTORY: repository = new RepositoryDirectory(rootPath,comment);                    break;
      case CVS:       repository = new RepositoryCVS(rootPath,userName,passwordHandler,comment); break;
      case SVN:       repository = new RepositorySVN(rootPath,userName,passwordHandler,comment); break;
      case HG:        repository = new RepositoryHG(rootPath,userName,passwordHandler,comment);  break;
      case GIT:       repository = new RepositoryGIT(rootPath,userName,passwordHandler,comment); break;
      default:        repository = new RepositoryNone(rootPath,comment);                         break;
    }

    return repository;
  }

  /** create new repository instance
   * @param type repository type; see Repository.Types
   * @param rootPath root path
   * @param userName user login name or ""
   * @param passwordHandler password handler to get password
   * @return repository or null if no repository found
   */
  public static Repository newInstance(Types type, String rootPath, String userName, PasswordHandler passwordHandler)
    throws RepositoryException
  {
    return newInstance(type,rootPath,userName,passwordHandler,"");
  }

  /** create new repository instance
   * @param type repository type; see Repository.Types
   * @param rootPath root path
   * @param passwordHandler password handler to get password
   * @return repository or null if no repository found
   */
  public static Repository newInstance(Types type, String rootPath, PasswordHandler passwordHandler)
    throws RepositoryException
  {
    return newInstance(type,rootPath,(String)null,passwordHandler);
  }

  /** create new repository instance
   * @param type repository type; see Repository.Types
   * @param rootPath root path
   * @param comment comment text
   * @return repository or null if no repository found
   */
  public static Repository newInstance(Types type, String rootPath, String comment)
    throws RepositoryException
  {
    return newInstance(type,rootPath,(String)null,(PasswordHandler)null,comment);
  }

  /** create new repository instance
   * @param type repository type; see Repository.Types
   * @param rootPath root path
   * @return repository or null if no repository found
   */
  public static Repository newInstance(Types type, String rootPath)
    throws RepositoryException
  {
    return newInstance(type,rootPath,"");
  }

  /** create new repository instance
   * @param type repository type; see Repository.Types
   * @return repository or null if no repository found
   */
  public static Repository newInstance(Types type)
    throws RepositoryException
  {
    return newInstance(type,"");
  }

  /** create new repository instance
   * @param rootPath root path
   * @param passwordHandler password handler to get password
   * @return repository or null if no repository found
   */
  public static Repository newInstance(String rootPath, PasswordHandler passwordHandler)
    throws RepositoryException
  {
    return newInstance(Repository.getType(rootPath),rootPath,passwordHandler);
  }

  /** create new repository instance
   * @param rootPath root path
   * @return repository or null if no repository found
   */
  public static Repository newInstance(String rootPath)
    throws RepositoryException
  {
    return newInstance(rootPath,(PasswordHandler)null);
  }

  /** create repository
   * @param rootPath root path
   * @param passwordHandler password handler
   * @param comment comment
   */
  Repository(String rootPath, PasswordHandler passwordHandler, String comment)
  {
    int z;

    this.title                              = (rootPath != null) ? new File(rootPath).getName() : null;
    this.rootPath                           = rootPath;
    this.passwordHandler                    = passwordHandler;
    this.comment                            = comment;
    this.openDirectories                    = new HashSet<String>();
    this.skipConvertWhitespacesFilePatterns = new HashSet<String>();
    this.ignoreFilePatterns                 = new HashSet<String>();
  }

  /** create repository
   * @param rootPath root path
   * @param comment comment
   */
  Repository(String rootPath, String comment)
  {
    this(rootPath,(PasswordHandler)null,comment);
  }

  /** create repository
   */
  Repository()
  {
    this((String)null,(String)null);
  }

  /** verify settings
   * @return null or error message
   */
  public String verifySettings()
  {
    return null;
  }

  /** set user name
   * @param userName user login name or ""
   */
  public void setUserName(String userName)
  {
  }

  /** set password handler
   * @param passwordHandle password handler
   */
  public void setPasswordHandler(PasswordHandler passwordHandler)
  {
    this.passwordHandler = passwordHandler;
  }

  /** check if repository support edit commit message
   * @return true iff edit commit message is supported
   */
  public boolean supportEditCommit()
  {
    return false;
  }

  /** check if repository support setting file mode
   * @return true iff file modes are supported
   */
  public boolean supportSetFileMode()
  {
    return false;
  }

  /** check if repository support change branch command
   * @return true iff change branch command is supported
   */
  public boolean supportChangeBranch()
  {
    return false;
  }

  /** check if repository support merge branch command
   * @return true iff merge branch command is supported
   */
  public boolean supportMergeBranch()
  {
    return false;
  }

  /** check if repository support merge commit command (cherry picking)
   * @return true iff merge commit command is supported
   */
  public boolean supportMergeCommit()
  {
    return false;
  }

  /** check if repository support incoming/outgoing commands
   * @return true iff incoming/outgoing commands are supported
   */
  public boolean supportIncomingOutgoing()
  {
    return false;
  }

  /** check if repository support pull/push commands
   * @return true iff pull/push commands are supported
   */
  public boolean supportPullPush()
  {
    return false;
  }

  /** check if repository support patch queues
   * @return true iff patch queues are supported
   */
  public boolean supportPatchQueues()
  {
    return false;
  }

  /** check if repository support lock/unlock
   * @return true iff lock/unlock is supported
   */
  public boolean supportLockUnlock()
  {
    return false;
  }

  /** check if repository support posting reviews
   * @return true iff posting reviews is supported
   */
  public boolean supportPostReview()
  {
    return false;
  }

  /** get special commands for repository
   * @return array with special commands
   */
  public Special[] getSpecials()
  {
    return new Special[0];
  }

  /** check if commit message is valid and acceptable
   * @return true iff commit message accepted
   */
  public boolean validCommitMessage(CommitMessage commitMessage)
  {
    return true;
  }

  /** get absolute+unique root path
   * @return absolute+unique root path
   */
  public final String getCanonicalRootPath()
  {
    return FileUtils.getCanonicalName(rootPath);
  }

  /** get open directories
   * @return open directories set
   */
  public Set<String> getOpenDirectories()
  {
    return (Set<String>)openDirectories.clone();
  }

  /** open sub-directory
   * @param directory sub-directory to open
   * @return file data set of sub-directory
   */
  public final Set<FileData> openDirectory(String directory)
  {
    // list files
    Set<FileData> fileDataSet = listFileData(directory);

    // add sub-directory to list with open directories
    synchronized(openDirectories)
    {
      openDirectories.add(directory);
    }

    return fileDataSet;
  }

  /** close sub-directory
   * @param directory sub-directory to close
   */
  public final void closeDirectory(String directory)
  {
    synchronized(openDirectories)
    {
      // remove directory and sub-directories from list with open directories
      String[] subDirectories = openDirectories.toArray(new String[openDirectories.size()]);
      for (String subDirectory : subDirectories)
      {
        if (subDirectory.startsWith(directory))
        {
          openDirectories.remove(subDirectory);
        }
      }
    }
  }

  /** get file data instance
   * @param name name
   * @param type file type (see FileData.Types)
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   * @param size file size [bytes]
   * @param date file date/time
   * @param workingRevision working revsion
   * @param repositoryRevision repository revision
   * @param branch branch
   */
  public final FileData getFileDataInstance(String name, FileData.Types type, FileData.States state, FileData.Modes mode, boolean locked, long size, Date date, String workingRevision, String repositoryRevision, String branch)
  {
    return FileData.getInstance(rootPath,name,type,state,mode,locked,size,date,workingRevision,repositoryRevision,branch);
  }

  /** get file data instance
   * @param name name
   * @param type file type (see FileData.Types)
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   * @param size file size [bytes]
   * @param date file date/time
   * @param workingRevision working revsion
   * @param repositoryRevision repository revision
   */
  public final FileData getFileDataInstance(String name, FileData.Types type, FileData.States state, FileData.Modes mode, boolean locked, long size, Date date, String workingRevision, String repositoryRevision)
  {
    return FileData.getInstance(rootPath,name,type,state,mode,locked,size,date,workingRevision,repositoryRevision);
  }

  /** get file data instance
   * @param name name
   * @param type file type (see FileData.Types)
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   * @param size file size [bytes]
   * @param date file date/time
   */
  public final FileData getFileDataInstance(String name, FileData.Types type, FileData.States state, FileData.Modes mode, boolean locked, long size, Date date)
  {
    return FileData.getInstance(rootPath,name,type,state,mode,locked,size,date);
  }

  /** get file data instance
   * @param name name
   * @param state state (see FileData.States)
   * @param mode mode (see FileData.Modes)
   * @param locked TRUE iff locked
   */
  public final FileData getFileDataInstance(String name, FileData.States state, FileData.Modes mode, boolean locked)
  {
    return FileData.getInstance(rootPath,name,state,mode,locked);
  }

  /** get file data instance
   * @param name name
   * @param type file type (see FileData.Types)
   * @param size file size [bytes]
   * @param date file date/time
   */
  public final FileData getFileDataInstance(String name, FileData.Types type, long size, Date date)
  {
    return FileData.getInstance(rootPath,name,type,size,date);
  }

  /** get file data instance
   * @param name file name
   * @param type file type (see FileData.Types)
   * @param date file date/time
   */
  public final FileData getFileDataInstance(String name, FileData.Types type, Date date)
  {
    return FileData.getInstance(rootPath,name,type,date);
  }

  /** get file data instance
   * @param name name
   * @param type file type (see FileData.Types)
   */
  public final FileData getFileDataInstance(String name, FileData.Types type)
  {
    return FileData.getInstance(rootPath,name,type);
  }


  /** get file data instance
   * @param name name
   * @param state file state
   */
  public final FileData getFileDataInstance(String name, FileData.States state)
  {
    return FileData.getInstance(rootPath,name,state);
  }

  /** get file data instance
   * @param name name
   */
  public final FileData getFileDataInstance(String name)
  {
    return FileData.getInstance(rootPath,name);
  }

  /** get file data instance
   * @param file file
   */
  public final FileData getFileDataInstance(File file)
  {
    return FileData.getInstance(rootPath,file);
  }

  /** convert to set
   * @param fileNames file name array
   * @return file data set
   */
  public Set<FileData> toSet(String[] fileNames)
  {
    return FileData.toSet(rootPath,fileNames);
  }

  /** convert to set
   * @param fileNameSet file name set
   * @return file data set
   */
  public Set<FileData> toSet(Set<String> fileNameSet)
  {
    return FileData.toSet(rootPath,fileNameSet);
  }

  /** convert to set
   * @param fileName file name
   * @return file data set
   */
  public final Set<FileData> toSet(String fileName, String... otherFileNames)
  {
    return FileData.toSet(rootPath,fileName,otherFileNames);
  }

  /** get skip convert whitespaces file patterns
   * @return skip convert whitespaces patterns set
   */
  public Set<String> getSkipConvertWhitespacesFilePatterns()
  {
    return skipConvertWhitespacesFilePatterns;
  }

  /** set skip convert whitespaces file patterns
   * @param filePatterns set convert whitespaces patterns array
   */
  public void setSkipConvertWhitespacesFilePatterns(String[] filePatterns)
  {
    synchronized(skipConvertWhitespacesFilePatterns)
    {
      skipConvertWhitespacesFilePatterns.clear();
      for (String filePattern : filePatterns)
      {
        skipConvertWhitespacesFilePatterns.add(filePattern);
      }
    }
  }

  /** remove skip convert whitespaces file pattern
   * @param filePattern file pattern to add
   */
  public void addSkipConvertWhitespacesFilePattern(String filePattern)
  {
    synchronized(skipConvertWhitespacesFilePatterns)
    {
      skipConvertWhitespacesFilePatterns.add(filePattern);
    }
  }

  /** remove skip convert whitespaces file pattern
   * @param filePattern file pattern to remove
   */
  public void removeSkipConvertWhitespacesFilePattern(String filePattern)
  {
    synchronized(skipConvertWhitespacesFilePatterns)
    {
      skipConvertWhitespacesFilePatterns.remove(filePattern);
    }
  }

  /** get ignore file patterns
   * @return ignore patterns set
   */
  public Set<String> getIgnoreFilePatterns()
  {
    return ignoreFilePatterns;
  }

  /** set ignore file patterns
   * @param filePatterns ignore patterns array
   */
  public void setIgnoreFilePatterns(String[] filePatterns)
  {
    synchronized(ignoreFilePatterns)
    {
      ignoreFilePatterns.clear();
      for (String filePattern : filePatterns)
      {
        ignoreFilePatterns.add(filePattern);
      }
    }
  }

  /** remove ignore file pattern
   * @param filePattern file pattern to add
   */
  public void addIgnoreFilePattern(String filePattern)
  {
    synchronized(ignoreFilePatterns)
    {
      ignoreFilePatterns.add(filePattern);
    }
  }

  /** remove ignore file pattern
   * @param filePattern file pattern to remove
   */
  public void removeIgnoreFilePattern(String filePattern)
  {
    synchronized(ignoreFilePatterns)
    {
      ignoreFilePatterns.remove(filePattern);
    }
  }

  /** get file name relative to root path
   * @param fileName file name
   * @return file name relative to root path or unchanged file name
   */
  public final String getRelativeFileName(String fileName)
  {
    if (fileName.startsWith(rootPath))
    {
      int index = rootPath.length();
      if ((fileName.length() > index) && (fileName.charAt(index) == File.separatorChar)) index++;
      fileName = fileName.substring(index);
    }

    return fileName;
  }

  /** get file name relative to root path
   * @param file file
   * @return file name relative to root path or unchanged file name
   */
  public final String getRelativeFileName(File file)
  {
    return getRelativeFileName(file.getPath());
  }

  /** get file type
   * @param file file
   * @return file type
   */
  public final FileData.Types getFileType(File file)
  {
    FileData.Types type = FileData.Types.UNKNOWN;
    try
    {
      // check if link, directory, file
      if      (!file.getCanonicalFile().equals(file.getAbsoluteFile())) type = FileData.Types.LINK;
      else if (file.isDirectory())                                      type = FileData.Types.DIRECTORY;
      else                                                              type = FileData.Types.FILE;
//Dprintf.dprintf("file=%s: '%s'=='%s' --> %s",file,checkFile.getCanonicalFile(),checkFile.getAbsoluteFile(),type);
    }
    catch (IOException exception)
    {
      // ignored
    }

    return type;
  }

  /** get file type
   * @param fileName file name
   * @return file type
   */
  public final FileData.Types getFileType(String fileName)
  {
    return getFileType(new File(fileName));
  }

  /** check if file is hidden
   * @param fileName file name
   * @param type file type; see FileData.Types
   * @return true iff hidden
   */
  public final boolean isHiddenFile(String fileName, FileData.Types type)
  {
    boolean hiddenFlag = false;
    switch (type)
    {
      case DIRECTORY:
        for (Settings.FilePattern filePattern : Settings.hiddenDirectoryPatterns)
        {
          if (filePattern.pattern.matcher(fileName).matches())
          {
            hiddenFlag = true;
            break;
          }
        }
        break;
      case FILE:
      case LINK:
        for (Settings.FilePattern filePattern : Settings.hiddenFilePatterns)
        {
          if (filePattern.pattern.matcher(fileName).matches())
          {
            hiddenFlag = true;
            break;
          }
        }
        break;
      default:
        break;
    }

    return hiddenFlag;
  }

  /** check if file is hidden
   * @param fileName file name
   * @return true iff hidden
   */
  public final boolean isHiddenFile(String fileName)
  {
    return isHiddenFile(fileName,getFileType(fileName));
  }

  /** check if file is hidden
   * @param file file
   * @param type file type; see FileData.Types
   * @return true iff hidden
   */
  public final boolean isHiddenFile(File file, FileData.Types type)
  {
    return isHiddenFile(file.getName(),type);
  }

  /** check if file is hidden
   * @param file file
   * @return true iff hidden
   */
  public final boolean isHiddenFile(File file)
  {
    return isHiddenFile(file.getName(),getFileType(file));
  }

  /** check if file is hidden
   * @param directory directory
   * @param baseName file base name
   * @return true iff hidden
   */
  public final boolean isHiddenFile(String directory, String baseName)
  {
    String fileName = !directory.isEmpty() ? directory+File.separator+baseName : baseName;
    return isHiddenFile(fileName,getFileType(fileName));
  }

  /** check if file should be skipped for convert whitespaces
   * @param fileName file name
   * @param true iff file should be skipped
   */
  public final boolean isSkipConvertWhitespacesFile(String fileName)
  {
    synchronized(skipConvertWhitespacesFilePatterns)
    {
      for (String skipConvertWhitespacesFilePattern : skipConvertWhitespacesFilePatterns)
      {
        if (   fileName.equals(skipConvertWhitespacesFilePattern)
            || fileName.matches(StringUtils.globToRegex(skipConvertWhitespacesFilePattern))
           )
        {
          return true;
        }
      }
    }

    return false;
  }

  /** check if file should be ignored
   * @param fileName file name
   * @param true iff file should be ignored
   */
  public final boolean isIgnoreFile(String fileName)
  {
    synchronized(ignoreFilePatterns)
    {
      for (String ignoreFilePattern : ignoreFilePatterns)
      {
        if (   fileName.equals(ignoreFilePattern)
            || fileName.matches(StringUtils.globToRegex(ignoreFilePattern))
           )
        {
          return true;
        }
      }
    }

    return false;
  }

  /** check if file should be ignored
   * @param file file
   * @param true iff file should be ignored
   */
  public final boolean isIgnoreFile(File file)
  {
    return isIgnoreFile(file.getPath());
  }

  /** check if file should be skipped for whitespace-ckeck
   * @param fileName file name
   * @return true iff file should be skipped
   */
  public boolean isSkipWhitespaceCheckFile(String fileName)
  {
    boolean skipFlag = false;
    File file = new File(fileName);
    for (Settings.FilePattern filePattern : Settings.skipWhitespaceCheckFilePatterns)
    {
      if (   filePattern.pattern.matcher(file.getName()).matches()
          || filePattern.pattern.matcher(file.getPath()).matches()
         )
      {
        skipFlag = true;
        break;
      }
    }

    return skipFlag;
  }

  /** get repository type
   * @return repository type
   */
  public Types getType()
  {
    return Types.UNKNOWN;
  }

  /** get list of files (exclude hidden files)
   * @param subDirectory sub-directory
   * @return hash with file data
   */
  public final List<File> listFiles(String subDirectory)
  {
    ArrayList<File> fileList = new ArrayList<File>();

    File[] files = (((subDirectory != null) && !subDirectory.isEmpty()) ? new File(rootPath,subDirectory) : new File(rootPath)).listFiles();
    if (files != null)
    {
      for (File file : files)
      {
        if (!isHiddenFile(file))
        {
          fileList.add(file);
        }
      }
    }

    return fileList;
  }

  /** get list of files (exclude hidden files)
   * @param subDirectory sub-directory
   * @return hash with file data
   */
  public final Set<FileData> listFileData(String subDirectory)
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    File[] files = (((subDirectory != null) && !subDirectory.isEmpty()) ? new File(rootPath,subDirectory) : new File(rootPath)).listFiles();
    if (files != null)
    {
      for (File file : files)
      {
        if (!isHiddenFile(file))
        {
          // get relative name
          String name = file.getAbsolutePath().substring(rootPath.length()+1);

          // add file data
          FileData fileData = getFileDataInstance(name,
                                                  getFileType(file),
                                                  file.length(),
                                                  new Date(file.lastModified())
                                                 );
          fileDataSet.add(fileData);
        }
      }
    }

    return fileDataSet;
  }

  /** list repository content
   * @param repositoryURL respository server URL
   * @return repository file list
   */
  public abstract List<RepositoryFile> list(String repositoryURL)
    throws RepositoryException;

  /** create new repository module
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param importPath import directory
   */
  public abstract void create(String repositoryURL, String moduleName, String importPath)
    throws RepositoryException;

  /** checkout repository
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param revisionId revision to checkout
   * @param userName user login name or ""
   * @param destinationPath destination path
   * @param busyDialog busy dialog or null
   */
  public abstract void checkout(String repositoryURL, String moduleName, String revisionId, String userName, String destinationPath, BusyDialog busyDialog)
    throws RepositoryException;

  /** checkout repository
   * @param repositoryURL repository server URLS
   * @param moduleName module name
   * @param revisionId revision to checkout
   * @param userName user login name or ""
   * @param destinationPath destination path
   */
  public final void checkout(String repositoryURL, String moduleName, String revisionId, String userName, String destinationPath)
    throws RepositoryException
  {
    checkout(repositoryURL,moduleName,revisionId,userName,destinationPath,(BusyDialog)null);
  }

  /** checkout repository
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param revisionId revision to checkout
   * @param userName user name or ""
   * @param password password or ""
   * @param destinationPath destination path
   */
  public final void checkout(String repositoryURL, String moduleName, String revisionId, String destinationPath)
    throws RepositoryException
  {
    checkout(repositoryURL,moduleName,revisionId,"",destinationPath);
  }

  /** update file states
   * @param fileDataSet file data set to update
   * @param fileDirectorySet relative directory name set to check for new/missing files
   * @param newFileDataSet new file data set or null
   */
  public abstract void updateStates(Set<FileData> fileDataSet, Set<String> fileDirectorySet, Set<FileData> newFileDataSet)
    throws RepositoryException;

  /** update file states
   * @param fileDataSet file data set to update
   * @param newFileDataSet new file data set or null
   */
  public final void updateStates(Set<FileData> fileDataSet, Set<FileData> newFileDataSet)
    throws RepositoryException
  {
/*
    // get directories
    HashSet<String> fileDirectorySet = new HashSet<String>();
    for (FileData fileData : fileDataSet)
    {
      switch (fileData.type)
      {
        case FILE:
          fileDirectorySet.add(fileData.getDirectoryName(rootPath));
          break;
        case DIRECTORY:
          fileDirectorySet.add(fileData.getName(rootPath));
          break;
      }
    }

    // udpate states
    updateStates(fileDataSet,fileDirectorySet,newFileDataSet);
/**/
///*
    // get directories
    HashSet<String> fileDirectorySet = new HashSet<String>();
    for (FileData fileData : fileDataSet)
    {
      fileDirectorySet.add(fileData.getDirectoryName(rootPath));
    }
    updateStates(fileDataSet,fileDirectorySet,newFileDataSet);
/**/
  }

  /** update files states
   * @param fileDataSet file data set to update
   */
  public void updateStates(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    updateStates(fileDataSet,(Set<FileData>)null);
  }

  /** update file state
   * @param fileDataSet file data to update
   */
  public final void updateStates(FileData fileData)
    throws RepositoryException
  {
    updateStates(fileData.toSet());
  }

  /** update file state
   * @param fileNames files to update
   */
  public final void updateStates(String[] fileNames)
    throws RepositoryException
  {
    updateStates(toSet(fileNames));
  }

  /** get repository root URL
   * @return repository root URL
   */
  public String getRepositoryRootURL()
  {
    return "";
  }

  /** get repository URL
   * @return repository URL
   */
  public String getRepositoryURL()
  {
    return "";
  }

  /** get master repository URL
   * @return masterRepository master repository URL or null
   */
  public RepositoryURL getMasterRepositoryURL()
  {
    return null;
  }

  /** get current branch name
   * @return branch name
   */
  public String getCurrentBranchName()
  {
    return "";
  }

  /** get first revision name
   * @return first revision name
   */
  public abstract String getFirstRevision();

  /** get last revision name
   * @return last revision name
   */
  public abstract String getLastRevision();

  /** get current revision
   * @paran repositoryURL repository server URL
   * @return current revision or null
   */
  public abstract String getCurrentRevision(String repositoryURL)
    throws RepositoryException;

  /** get current revision
   * @return current revision or null
   */
  public final String getCurrentRevision()
    throws RepositoryException
  {
    return getCurrentRevision(getRepositoryURL());
  }

  /** get revision names of file
   * @param name file name or URL
   * @return revision name list
   */
  public abstract List<String> getRevisionNames(String name)
    throws RepositoryException;

  /** get revision names of file
   * @param fileData file data
   * @return revision name list
   */
  public final List<String> getRevisionNames(FileData fileData)
    throws RepositoryException
  {
    return getRevisionNames(getFileDataName(fileData));
  }

    /** get revision data
   * @param fileData file data (can be null)
   * @param revisionId revision
   * @return revision data or null
   */
  public abstract RevisionData getRevisionData(FileData fileData, String revisionId)
    throws RepositoryException;

  /** get revision data of last commit
   * @param fileData file data
   * @return revision data or null
   */
  public final RevisionData getRevisionData(FileData fileData)
    throws RepositoryException
  {
    String currentRevision = getCurrentRevision();
    if (currentRevision != null)
    {
      return getRevisionData(fileData,currentRevision);
    }
    else
    {
      return null;
    }
  }

  /** get revision data
   * @param fileData file data
   * @param revisionId revision or null
   */
  public final RevisionData getRevisionData(String revisionId)
    throws RepositoryException
  {
    return getRevisionData((FileData)null,revisionId);
  }

  /** get revision data tree (grow down, root = initial entry, leaves = newest entries)
   * @param fileData file data
   * @param busyDialog busy dialog or null
   * @return revision data tree
   */
  public abstract RevisionDataTree getRevisionDataTree(FileData fileData, BusyDialog busyDialog)
    throws RepositoryException;

  /** get file data (text lines)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  public abstract String[] getFileLines(FileData fileData, String revisionId)
    throws RepositoryException;

  /** get file data (text lines)
   * @param fileData file data
   * @return file data (array of lines)
   */
  public final String[] getFileLines(FileData fileData)
    throws RepositoryException
  {
    return getFileLines(fileData,null);
  }

  /** get file data (byte array)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of bytes)
   */
  public abstract byte[] getFileBytes(FileData fileData, String revisionId)
    throws RepositoryException;

  /** get file data (text lines)
   * @param fileData file data
   * @return file data (array of lines)
   */
  public final byte[] getFileBytes(FileData fileData)
    throws RepositoryException
  {
    return getFileBytes(fileData,null);
  }

  /** get all changed/unknown files
   * @param stateSet state set
   * @return fileDataSet file data set with modified files
   */
  public abstract Set<FileData> getChangedFiles(EnumSet<FileData.States> stateSet)
    throws RepositoryException;

  /** get all changed/unknown files
   * @return fileDataSet file data set with modified files
   */
  public final Set<FileData> getChangedFiles()
    throws RepositoryException
  {
    return getChangedFiles(FileData.STATES_ALL);
  }

  /** get diff of file
   * @param fileData file data
   * @param revisionId1,revisionId2 revisions to get diff for
   * @return diff data
   */
  public abstract List<DiffData> getDiff(FileData fileData, String revisionId1, String revisionId2)
    throws RepositoryException;

  /** get diff of file
   * @param fileData file data
   * @param revisionId revision to get diff for
   * @return diff data
   */
  public final List<DiffData> getDiff(FileData fileData, String revisionId)
    throws RepositoryException
  {
    return getDiff(fileData,revisionId,(String)null);
  }

  /** get unified patch lines for files
   * Note: must return unified patch data with file names relative to rootPath
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @param output patch output or null
   * @param lineLine patch data lines or null
   */
  public abstract void getPatch(Set<FileData> fileDataSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces, PrintWriter output, List<String> lineList)
    throws RepositoryException;

  /** get unified patch lines for files
   * Note: must return unified patch data with file names relative to rootPath
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return file with patch
   */
  public final File getPatchFile(Set<FileData> fileDataSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    File file = null;
    try
    {
      file = File.createTempFile("patch",".patch",new File(Settings.tmpDirectory));
      PrintWriter output = new PrintWriter(new FileWriter(file));
      getPatch(fileDataSet,revisionId1,revisionId2,ignoreWhitespaces,output,null);
      output.close();
    }
    catch (IOException exception)
    {
      throw new RepositoryException(exception);
    }

    return file;
  }

  /** get unified patch lines for file
   * @param fileNameSet file name set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return file with patch
   */
  public final File getPatchFile(Collection<String> fileNameSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (String fileName : fileNameSet)
    {
      fileDataSet.add(getFileDataInstance(fileName));
    }
    updateStates(fileDataSet);

    return getPatchFile(fileDataSet,revisionId1,revisionId2,ignoreWhitespaces);
  }

  /** get unified patch lines for file
   * @param fileNames file names
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return file with patch
   */
  public final File getPatchFile(String[] fileNames, String revisionId1, String revisionId2, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (String fileName : fileNames)
    {
      fileDataSet.add(getFileDataInstance(fileName));
    }
    updateStates(fileDataSet);

    return getPatchFile(fileDataSet,revisionId1,revisionId2,ignoreWhitespaces);
  }

  /** get unified patch for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return file with patch
   */
  public final File getPatchFile(Set<FileData> fileDataSet, String revisionId1, String revisionId2)
    throws RepositoryException
  {
    return getPatchFile(fileDataSet,revisionId1,revisionId2,false);
  }

  /** get unified patch for file
   * @param fileDataSet file data set
   * @param ignoreWhitespaces true to ignore white spaces
   * @return file with patch
   */
  public final File getPatchFile(Set<FileData> fileDataSet, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    return getPatchFile(fileDataSet,getLastRevision(),null,ignoreWhitespaces);
  }

  /** get unified patch for file
   * @param fileDataSet file data set
   * @return file with patch
   */
  public final File getPatchFile(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    return getPatchFile(fileDataSet,false);
  }

  /** get patch for all changed files
   * @return file with patch
   */
  public final File getPatchFile()
    throws RepositoryException
  {
    return getPatchFile((Set<FileData>)null);
  }

  /** get unified patch lines for file
   * Note: must return unified patch data with file names relative to rootPath
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(Set<FileData> fileDataSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    List<String> lineList = new ArrayList<String>();
    getPatch(fileDataSet,revisionId1,revisionId2,ignoreWhitespaces,null,lineList);
    return lineList.toArray(new String[lineList.size()]);
  }

  /** get unified patch lines for file
   * @param fileNameSet file name set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(Collection<String> fileNameSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (String fileName : fileNameSet)
    {
      fileDataSet.add(getFileDataInstance(fileName));
    }
    updateStates(fileDataSet);

    return getPatchLines(fileDataSet,revisionId1,revisionId2,ignoreWhitespaces);
  }

  /** get unified patch lines for file
   * @param fileNames file names
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(String[] fileNames, String revisionId1, String revisionId2, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (String fileName : fileNames)
    {
      fileDataSet.add(getFileDataInstance(fileName));
    }
    updateStates(fileDataSet);

    return getPatchLines(fileDataSet,revisionId1,revisionId2,ignoreWhitespaces);
  }

  /** get unified patch for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(Set<FileData> fileDataSet, String revisionId1, String revisionId2)
    throws RepositoryException
  {
    return getPatchLines(fileDataSet,revisionId1,revisionId2,false);
  }

  /** get unified patch for file
   * @param fileDataSet file data set
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(Set<FileData> fileDataSet, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    return getPatchLines(fileDataSet,getLastRevision(),null,ignoreWhitespaces);
  }

  /** get unified patch for file
   * @param fileDataSet file data set
   * @return patch data lines
   */
  public final String[] getPatchLines(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    return getPatchLines(fileDataSet,false);
  }

  /** get patch for all changed files
   * @return patch data lines
   */
  public final String[] getPatchLines()
    throws RepositoryException
  {
    return getPatchLines((Set<FileData>)null);
  }

  /** get unified patch for file
   * @param fileData file data
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(FileData fileData, String revisionId1, String revisionId2, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);

    return getPatchLines(fileDataSet,revisionId1,revisionId2,false);
  }

  /** get unified patch for file
   * @param fileData file data
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(FileData fileData, String revisionId1, String revisionId2)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);

    return getPatchLines(fileDataSet,revisionId1,revisionId2,false);
  }

  /** get unified patch for file
   * @param fileData file data
   * @param ignoreWhitespaces true to ignore white spaces
   * @return patch data lines
   */
  public final String[] getPatchLines(FileData fileData, boolean ignoreWhitespaces)
    throws RepositoryException
  {
    return getPatchLines(fileData,getLastRevision(),null,ignoreWhitespaces);
  }

  /** get unified patch for file
   * @param fileData file data
   * @return patch data lines
   */
  public final String[] getPatchLines(FileData fileData)
    throws RepositoryException
  {
    return getPatchLines(fileData,false);
  }

// NYI: is getPatchBytes() obsolete?
  /** get patch data for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @return patch data bytes
   */
  public abstract byte[] getPatchBytes(Set<FileData> fileDataSet, String revisionId1, String revisionId2)
    throws RepositoryException;

  /** get patch data for file
   * @param fileDataSet file data set
   * @return patch data bytes
   */
  public final byte[] getPatchBytes(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    return getPatchBytes(fileDataSet,null,null);
  }

  /** get log to file
   * @param fileData file data
   * @return log data list
   */
  public abstract List<LogData> getLog(FileData fileData)
    throws RepositoryException;

  /** get complete log
   * @return log data list
   */
  public final List<LogData> getLog()
    throws RepositoryException
  {
    return getLog((FileData)null);
  }

  /** get annotations to file
   * @param fileData file data
   * @param revisionId revision
   * @return annotation list
   */
  public abstract List<AnnotationData> getAnnotations(FileData fileData, String revisionId)
    throws RepositoryException;

  /** get annotations to file of last revision
   * @param fileData file data
   * @return annotation list
   */
  public final List<AnnotationData> getAnnotations(FileData fileData)
    throws RepositoryException
  {
    return getAnnotations(fileData,getLastRevision());
  }

  /** update file from respository
   * @param fileDataSet file data set or null
   * @param busyDialog busy dialog or null
   */
  public abstract void update(Set<FileData> fileDataSet, BusyDialog busyDialog)
    throws RepositoryException;

  /** update file from respository
   * @param fileDataSet file data set
   */
  public final void update(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    update(fileDataSet,(BusyDialog)null);
  }

  /** update all files from respository
   * @param busyDialog busy dialog or null
   */
  public final void updateAll(BusyDialog busyDialog)
    throws RepositoryException
  {
    update((HashSet<FileData>)null,busyDialog);
  }

  /** update all files from respository
   */
  public final void updateAll()
    throws RepositoryException
  {
    updateAll(null);
  }

  /** commit files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  public abstract void commit(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException;

  /** commit file
   * @param fileData file data
   * @param commitMessage commit message
   */
  public final void commit(FileData fileData, CommitMessage commitMessage)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);

    commit(fileDataSet,commitMessage);
  }

  /** add files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary files, false otherwise
   */
  public abstract void add(Set<FileData> fileDataSet, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException;

  /** add files
   * @param fileNameSet file name set
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary files, false otherwise
   */
  public final void add(Collection<String> fileNameSet, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (String fileName : fileNameSet)
    {
      fileDataSet.add(getFileDataInstance(fileName));
    }

    add(fileDataSet,commitMessage,binaryFlag);
  }

  /** add files
   * @param fileNames file names
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary files, false otherwise
   */
  public final void add(String[] fileNames, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    for (String fileName : fileNames)
    {
      fileDataSet.add(getFileDataInstance(fileName));
    }

    add(fileDataSet,commitMessage,binaryFlag);
  }

  /** add file
   * @param fileData file data
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary file, false otherwise
   */
  public final void add(FileData fileData, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);

    add(fileDataSet,commitMessage,binaryFlag);
  }

  /** remove files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  public abstract void remove(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException;

  /** remove file
   * @param fileData file data
   * @param commitMessage commit message
   */
  public final void remove(FileData fileData, CommitMessage commitMessage)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);

    remove(fileDataSet,commitMessage);
  }

  /** copy files
   * @param fileDataSet files to copy
   * @param destination destination
   * @param commitMessage commit message
   */
  public abstract void copy(Set<FileData> fileDataSet, String destination, CommitMessage commitMessage)
    throws RepositoryException;

    /** rename file
   * @param fileData file data to rename
   * @param newName new name
   * @param commitMessage commit message
   */
  public abstract void rename(FileData fileData, String newName, CommitMessage commitMessage)
    throws RepositoryException;

  /** revert files
   * @param fileDataSet file data set or null for all files
   * @param revisionId revision to revert to
   * @param recursive true for recursive revert, false otherwise
   */
  public abstract void revert(Set<FileData> fileDataSet, String revisionId, boolean recursive)
    throws RepositoryException;

  /** revert files
   * @param fileData file data
   * @param revisionId revision to revert to
   * @param recursive true for recursive revert, false otherwise
   */
  public final void revert(FileData fileData, String revisionId, boolean recursive)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);

    revert(fileDataSet,revisionId,recursive);
  }

  /** revert files
   * @param fileData file data
   * @param revisionId revision to revert to
   */
  public final void revert(FileData fileData, String revisionId)
    throws RepositoryException
  {
    revert(fileData,revisionId,false);
  }

  /** revert files
   * @param fileDataSet file data set or null for all files
   */
  public final void revert(Set<FileData> fileDataSet)
    throws RepositoryException
  {
   revert(fileDataSet,getLastRevision(),false);
  }

  /** set conflicts resolved
   * @param fileDataSet file data set or null for all files
   */
  public abstract void resolve(Set<FileData> fileDataSet)
    throws RepositoryException;

  /** change current branch
   * @param newBranchName new branch name
   */
  public void changeBranch(String newBranchName)
    throws RepositoryException
  {
    throw new InternalError("change branch not supported");
  }

  /** merge other branch into current branch
   * @param otherBranchName other branch name
   */
  public void mergeBranch(String otherBranchName)
    throws RepositoryException
  {
    throw new InternalError("merge branch not supported");
  }

  /** merge commit into current branch
   * @param revision revision
   */
  public void mergeCommit(String revision)
    throws RepositoryException
  {
    throw new InternalError("merge commit not supported");
  }

  /** get incoming changes list
   * @param incoming changes list
   */
  public List<LogData> getIncomingChanges(String masterRepository)
    throws RepositoryException
  {
    throw new InternalError("get incoming changes not supported");
  }

  /** get outgoing changes list
   * @param outgoing changes list
   */
  public List<LogData> getOutgoingChanges(String masterRepository)
    throws RepositoryException
  {
    throw new InternalError("get outgoing changes not supported");
  }

  /** get incoming/outgoing changes lines
   * @param revisionId revision to get changes lines for
   */
  public String[] getChanges(String revisionId)
    throws RepositoryException
  {
    throw new InternalError("get incoming/outgoing changes not supported");
  }

  /** pull changes
   * @param masterRepositoryURL master repository URL
   * @param moduleName module name
   * @param userName user login name or null
   */
  public void pullChanges(String masterRepositoryURL, String moduleName, String userName)
    throws RepositoryException
  {
    throw new InternalError("pull changes not supported");
  }

  /** push changes
   * @param masterRepositoryURL master repository URL
   * @param moduleName module name
   * @param userName user login name or null
   */
  public void pushChanges(String masterRepositoryURL, String moduleName, String userName)
    throws RepositoryException
  {
    throw new InternalError("push changes not supported");
  }

  /** push changes
   * @param masterRepositoryURL master repository URL
   */
  public final void pushChanges(String masterRepositoryURL)
    throws RepositoryException
  {
    pushChanges(masterRepositoryURL,"","");
  }

  /** push changes
   */
  public final void pushChanges()
    throws RepositoryException
  {
    RepositoryURL repositoryURL = getMasterRepositoryURL();
    pushChanges(repositoryURL.path,repositoryURL.moduleName,repositoryURL.userName);
  }

  /** apply patch queue patches
   */
  public void applyPatches()
    throws RepositoryException
  {
    throw new InternalError("apply patches not supported");
  }

  /** unapply patch queue patches
   */
  public void unapplyPatches()
    throws RepositoryException
  {
    throw new InternalError("unapplay patches not supported");
  }

  /** lock files
   * @param fileDataSet file data set
   */
  public void lock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    throw new InternalError("lock not supported");
  }

  /** unlock files
   * @param fileDataSet file data set
   */
  public void unlock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    throw new InternalError("unlock not supported");
  }

  /** set commit message
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  public abstract void setCommitMessage(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException;

  /** set files mode
   * @param fileDataSet file data set
   * @param mode file mode
   * @param commitMessage commit message
   */
  public abstract void setFileMode(Set<FileData> fileDataSet, FileData.Modes mode, CommitMessage commitMessage)
    throws RepositoryException;

  /** set files mode
   * @param fileData file data
   * @param mode file mode
   * @param commitMessage commit message
   */
  public final void setFileMode(FileData fileData, FileData.Modes mode, CommitMessage commitMessage)
    throws RepositoryException
  {
    HashSet<FileData> fileDataSet = new HashSet<FileData>();
    fileDataSet.add(fileData);

    setFileMode(fileDataSet,mode,commitMessage);
  }

  /** get default name of root
   * @return default root name or null if there are not root names
   */
  public abstract String getDefaultRootName();

  /** get default branch/tag name
   * @return default branch/tag name
   */
  public abstract String getDefaultBranchTagName();

  /** get names of existing branches/tags
   * @param pathName path name
   * @return branch/tag name list
   */
  public abstract List<String> getBranchTagNames(String pathName)
    throws RepositoryException;

  /** get names of existing branches/tags
   * @return branch/tag name list
   */
  public final List<String> getBranchTagNames()
    throws RepositoryException
  {
//TODO? getRepositoryURL() correct?
//    return getBranchTagNames(getRepositoryRootURL());
    return getBranchTagNames(getRepositoryRootURL());
  }

    /** create new branch
   * @param rootName root name (source)
   * @param branchName branch name
   * @param fromRevision from revision name or null
   * @param commitMessage commit message
   * @param buysDialog busy dialog or null
   */
  public abstract void createBranch(String rootName, String branchName, String fromRevision, CommitMessage commitMessage, BusyDialog busyDialog)
    throws RepositoryException;

  /** post to review server
   * @param password review server password
   * @param reference reference string
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param testSet test set
   */
  protected String postReview(String password, String reference, Set<FileData> fileDataSet, CommitMessage commitMessage, LinkedHashSet<String> testSet)
    throws RepositoryException
  {
    if (reference.isEmpty())
    {
      if (!Settings.commandPostReviewServer.isEmpty())
      {
        File patchFile = null;
        try
        {
          // get patch file
          patchFile = getPatchFile(fileDataSet);

          // create command
          Macro macro = new Macro(StringUtils.split(new ArrayList<String>(),Settings.commandPostReviewServer,StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS,false));
          macro.expand("server",     reviewServerHost);
          macro.expand("login",      reviewServerLogin);
          macro.expand("password",   macro.hidden((password != null) ? password : ""));
          macro.expand("summary",    commitMessage.getSummary());
          macro.expand("groups",     reviewServerGroups);
          macro.expand("persons",    reviewServerPersons);
          macro.expand("description",commitMessage.getMessage());
          macro.expand("tests",      testSet,",");
          macro.expand("file",       patchFile.getAbsolutePath());
          Command command = new Command(macro);
//Dprintf.dprintf("command=%s",command);

          Exec exec = null;
          try
          {
            // execute
            exec = new Exec(command);
            String line;
            while ((line = exec.getStdout()) != null)
            {
Dprintf.dprintf("stdout %s",line);
            }
            while ((line = exec.getStderr()) != null)
            {
Dprintf.dprintf("stderr %s",line);
            }
            int exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              throw new RepositoryException("Cannot post patch to review server (exitcode: "+exitCode+")");
            }
            exec.done(); exec = null;
          }
          catch (IOException exception)
          {
            throw new RepositoryException(exception);
          }
          finally
          {
            if (exec != null) exec.done();
          }

          patchFile.delete(); patchFile = null;
        }
        finally
        {
          if (patchFile != null) patchFile.delete();
        }
      }
      else
      {
        throw new RepositoryException("No review server post command configured.");
      }
    }
    else
    {
      if (!Settings.commandUpdateReviewServer.isEmpty())
      {
        File patchFile = null;
        try
        {
          // get patch file
          patchFile = getPatchFile(fileDataSet);

          // create command
          Macro macro = new Macro(StringUtils.split(new ArrayList<String>(),Settings.commandUpdateReviewServer,StringUtils.WHITE_SPACES,StringUtils.QUOTE_CHARS,false));
          macro.expand("server",     reviewServerHost);
          macro.expand("login",      reviewServerLogin);
          macro.expand("password",   macro.hidden((password != null) ? password : ""));
          macro.expand("reference",  reference);
          macro.expand("summary",    commitMessage.getSummary());
          macro.expand("groups",     reviewServerGroups);
          macro.expand("persons",    reviewServerPersons);
          macro.expand("description",commitMessage.getMessage());
          macro.expand("tests",      testSet,",");
          macro.expand("file",       patchFile.getAbsolutePath());
          Command command = new Command(macro);
//Dprintf.dprintf("command=%s",command);

          Exec exec = null;
          try
          {
            // execute
            exec = new Exec(command);
            String line;
            while ((line = exec.getStdout()) != null)
            {
Dprintf.dprintf("stdout %s",line);
            }
            while ((line = exec.getStderr()) != null)
            {
Dprintf.dprintf("stderr %s",line);
            }
            int exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              throw new RepositoryException("Cannot post patch to review server (exitcode: "+exitCode+")");
            }
            exec.done(); exec = null;
          }
          catch (IOException exception)
          {
            throw new RepositoryException(exception);
          }
          finally
          {
            if (exec != null) exec.done();
          }

          patchFile.delete(); patchFile = null;
        }
        finally
        {
          if (patchFile != null) patchFile.delete();
        }
      }
      else
      {
        throw new RepositoryException("No review server update command configured.");
      }
    }

    return reference;
  }

  /** post to review server
   * @param password review server password
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param testSet test set
   */
  public final String postReview(String password, Set<FileData> fileDataSet, CommitMessage commitMessage, LinkedHashSet<String> testSet)
    throws RepositoryException
  {
    return postReview(password,"",fileDataSet,commitMessage,testSet);
  }

  /** update review server
   * @param password review server password
   * @param reference reference string
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param testSet test set
   */
  public final String updateReview(String password, String reference, Set<FileData> fileDataSet, CommitMessage commitMessage, LinkedHashSet<String> testSet)
    throws RepositoryException
  {
    return postReview(password,reference,fileDataSet,commitMessage,testSet);
  }

  /** convert data to string
   * @return string
   */
  public String toString()
  {
    return "Repository {type: "+getType()+", root: "+rootPath+", selected: "+selected+"}";
  }

  //-----------------------------------------------------------------------

  /** parse date
   * @param string date/time string
   * @return date/time or current date/time
   */
  protected Date parseDate(String string)
  {
    final String[] FORMATS = new String[]
    {
      "yyyy-MM-dd'T'HH:mm:ss.S'Z'",  // 2011-01-25T15:19:06.12345Z
      "yyyy-MM-dd HH:mm:ss Z",       // 2011-01-25 15:19:06 +0900
      "yyyy-MM-dd HH:mm Z",          // 2011-01-25 15:19 +0900
      "yyyy/MM/dd HH:mm:ss",         // 2010/05/13 07:09:36
      "EEE MMM dd HH:mm:ss yyyy Z",  // Wed Dec 17 15:41:19 2008 +0100
      "yyyy-MM-dd",                  // 2011-01-25
      "dd-MMM-yy",                   // 25-Jan-11
    };

    final Locale[] LOCALES = new Locale[]
    {
      Locale.US,
      Locale.GERMAN,
      Locale.GERMANY,

      Locale.CANADA,
      Locale.CANADA_FRENCH,
      Locale.CHINA,
      Locale.CHINESE,
      Locale.ENGLISH,
      Locale.FRANCE,
      Locale.FRENCH,
      Locale.ITALIAN,
      Locale.ITALY,
      Locale.JAPAN,
      Locale.JAPANESE,
      Locale.KOREA,
      Locale.KOREAN,
      Locale.PRC,
      Locale.ROOT,
      Locale.SIMPLIFIED_CHINESE,
      Locale.TAIWAN,
      Locale.TRADITIONAL_CHINESE,
      Locale.UK,
    };

    Date date = null;

    if (string != null)
    {
      for (String format : FORMATS)
      {
        for (Locale locale : LOCALES)
        {
          try
          {
            date = new SimpleDateFormat(format,locale).parse(string);
            return date;
          }
          catch (ParseException exception)
          {
            // ignored, try next format
          }
        }
      }
      if (Settings.debugFlag)
      {
        Onzen.printWarning("Cannot parse date '%s'",string);
      }
    }

    return date;
  }

  /** find file data
   * @param fileDataSet file data set
   * @param directory directory
   * @param baseName base name
   * @return file data or null
   */
  protected FileData findFileData(Set<FileData> fileDataSet, String directory, String baseName)
  {
    return findFileData(fileDataSet,!directory.isEmpty() ? directory+File.separator+baseName : baseName);
  }

  /** find file data
   * @param fileDataSet file data set
   * @param name name
   * @return file data or null
   */
  protected FileData findFileData(Set<FileData> fileDataSet, String name)
  {
    for (FileData fileData : fileDataSet)
    {
      if (name.equals(fileData.getName())) return fileData;
    }
    return null;
  }

  /** check if file data set contain file data
   * @param fileDataSet file data set
   * @param fileName file name
   * @return true iff found
   */
  protected boolean containFile(Set<FileData> fileDataSet, String fileName)
  {
    return findFileData(fileDataSet,fileName) != null;
  }

  /** check if file data set contain file data
   * @param fileDataSet file data set
   * @param fileData file data
   * @return true iff found
   */
  protected boolean containFileData(Set<FileData> fileDataSet, FileData fileData)
  {
    return containFile(fileDataSet,fileData.getName());
  }

  /** get relative file names from file data map
   * @param fileDataMap file data map
   * @return file names list
   */
  protected List<String> getFileDataNames(HashMap<String,FileData> fileDataMap)
  {
    List<String> fileNameList = new ArrayList<String>(fileDataMap.size());

    for (FileData fileData : fileDataMap.values())
    {
      fileNameList.add(fileData.getName(rootPath));
    }

    return fileNameList;
  }

  /** get relative file names from file data set
   * @param fileDataSet file data set
   * @return file names list
   */
  protected List<String> getFileDataNames(Set<FileData> fileDataSet)
  {
    ArrayList<String> fileNameList = new ArrayList<String>(fileDataSet.size());

    for (FileData fileData : fileDataSet)
    {
      fileNameList.add(fileData.getName(rootPath));
    }

    return fileNameList;
  }

  /** get relative file name from file data
   * @param fileData file data
   * @return file name
   */
  protected String getFileDataName(FileData fileData)
  {
    return fileData.getName(rootPath);
  }

  //-----------------------------------------------------------------------
}

/* end of file */
