/***********************************************************************\
*
* $Revision: 2167 $
* $Date: 2017-09-15 17:25:09 +0200 (Fri, 15 Sep 2017) $
* $Author: torsten $
* Contents: repository
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

import java.text.DateFormat;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Set;

import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Shell;

/****************************** Classes ********************************/

/** Git repository
 */
class RepositoryGIT extends Repository
{
  /** GIT revision data
   */
  static class RevisionDataGIT extends RevisionData
  {
    // revision types
    static enum Types
    {
      NONE,
      COMMIT,
      BRANCH,
      MERGE
    };

    Types   type;
    String  changeSetId;
    String  branchName;
    String  parentRevisionId1,parentRevisionId2;

    /** create HG revision data
     * @param revisionId revision id string
     * @param changeSetId change-set id string
     * @param branchName branch name
     * @param parentRevisionId1 parent revision id1 (!= null if branch or merge)
     * @param parentRevisionId2 parent revision id2 (!= null if merge)
     * @param tagList tag names or null
     * @param date date
     * @param author author name
     * @param commitMessageList commit message
     * @param fileNameList names of changed files
     */
    public RevisionDataGIT(String             revisionId,
                           String             branchName,
                           String             parentRevisionId1,
                           String             parentRevisionId2,
                           TagList            tagList,
                           Date               date,
                           String             author,
                           LinkedList<String> commitMessageList,
                           ArrayList<String>  fileNameList
                          )
    {
      super(revisionId,tagList,date,author,commitMessageList,fileNameList);
      this.type              = Types.NONE;
      this.changeSetId       = changeSetId;
      this.branchName        = branchName;
      this.parentRevisionId1 = parentRevisionId1;
      this.parentRevisionId2 = parentRevisionId2;
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return "GIT revision data {"+type+", "+revisionId+", date: "+date+", author: "+author+", message: "+StringUtils.join(commitMessageList,", ")+"}";
    }
  }

  /** parent data
   */
  class ParentData
  {
    private long revisionId1,revisionId2;

    /** create parent data
     */
    ParentData(long revisionId1, long revisionId2)
    {
      this.revisionId1 = revisionId1;
      this.revisionId2 = revisionId2;
    }

    /** create parent data
     */
    ParentData(long revisionId1)
    {
      this(revisionId1,0L);
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      if (revisionId2 != 0)
      {
        return Long.toString(revisionId1)+","+Long.toString(revisionId2);
      }
      else
      {
        return Long.toString(revisionId1);
      }
    }
  }

  // --------------------------- constants --------------------------------
  public final static String[] DEFAULT_REVISION_NAMES = new String[]{"HEAD"};
  public final static String[] DEFAULT_BRANCH_NAMES   = new String[]{"master","branches"};
  public final static String   DEFAULT_ROOT_NAME      = "master";
  public final static String   DEFAULT_BRANCHES_NAME  = "branches";
  public final static String   FIRST_REVISION_NAME    = "0";
  public final static String   LAST_REVISION_NAME     = "HEAD";

  /* log format template:
    <changeset id> <date> <time> <timezone> <author>
    <parents>
    <branch>
    <description>
    <description long>
    -----
    ...
  */
  private final String LOG_TEMPLATE_SHORT = "%h%n%ci%n%cn%n%p%n%d%n%w(,3,3)%B%n";
  private final String LOG_TEMPLATE_LONG  = "%H%n%ci%n%cn%n%P%n%d%n%w(,3,3)%B%n";

  // --------------------------- variables --------------------------------
  private final static RepositoryGIT staticInstance = new RepositoryGIT();

  @XmlElement(name = "userName")
  @RepositoryValue(title = "User name:", tooltip="GIT server user login name.")
  public String userName;

  private String repositoryURL     = null;
  private String repositoryRootURL = null;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get static instance
   * @return static instance
   */
  public final static RepositoryGIT getInstance()
  {
    return staticInstance;
  }

  /** create repository
   * @param rootPath root path
   */
  RepositoryGIT(String rootPath, String userName, PasswordHandler passwordHandler, String comment)
  {
    super(rootPath,passwordHandler,comment);

    assert !rootPath.isEmpty();

    this.userName = userName;
  }

  /** create repository
   */
  RepositoryGIT()
  {
    super();
  }

  /** get repository type
   * @return repository type
   */
  @Override
  public Types getType()
  {
    return Types.GIT;
  }

  /** check if repository support change branch command
   * @return true iff change branch command is supported
   */
  @Override
  public boolean supportChangeBranch()
  {
    return true;
  }

  /** check if repository support merge branch command
   * @return true iff merge branch command is supported
   */
  @Override
  public boolean supportMergeBranch()
  {
    return true;
  }

  /** check if repository support merge commit command (cherry picking)
   * @return true iff merge commit command is supported
   */
  public boolean supportMergeCommit()
  {
    return true;
  }

  /** check if repository support incoming/outgoing commands
   * @return true iff incoming/outgoing commands are supported
   */
  @Override
  public boolean supportIncomingOutgoing()
  {
    return true;
  }

  /** check if repository support pull/push commands
   * @return true iff pull/push commands are supported
   */
  @Override
  public boolean supportPullPush()
  {
    return true;
  }

  /** get special commands for repository
   * @return array with special commands
   */
  @Override
  public Special[] getSpecials()
  {
    return new Special[]
    {
      new SpecialCommand()
      {
        @Override
        public String getTitle()
        {
          return "Pull and update all sub-modules...";
        }

        @Override
        public boolean run(Shell shell)
          throws RepositoryException
        {
          pullUpdateAllSubModules(this,shell);

          return true;
        }
      },
      new SpecialSeparator(),
      new SpecialSubCommands("Stash",
        new SpecialCommand()
        {
          @Override
          public String getTitle()
          {
            return "Save";
          }

          @Override
          public boolean run(Shell shell)
            throws RepositoryException
          {
            stashSave(this,shell);

            return true;
          }
        },
        new SpecialCommand()
        {
          @Override
          public String getTitle()
          {
            return "Pop";
          }

          @Override
          public boolean run(Shell shell)
            throws RepositoryException
          {
            stashPop(this,shell);

            return true;
          }
        },
        new SpecialCommand()
        {
          @Override
          public String getTitle()
          {
            return "Reset";
          }

          @Override
          public boolean run(Shell shell)
            throws RepositoryException
          {
            reset(this,shell);

            return true;
          }
        }
      )
    };
  }

  /** check if repository
   * @param repositoryURL respository server URL
   * @return repository file list
   */
  public static boolean isRepository(String rootPath)
  {
    boolean result = false;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      command.clear();
      command.append(Settings.gitCommand,"rev-parse","--show-toplevel");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      result = (exec.waitFor() == 0);

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return result;
  }

  /** list repository content
   * @param name name
   * @return repository file list
   */
  @Override
  public List<RepositoryFile> list(String name)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,rootPath) : null;

    ArrayList<RepositoryFile> repositoryFileList = new ArrayList<RepositoryFile>();

    Exec exec = null;
    try
    {
      Command command = new Command();

      // checkout
      command.clear();
      command.append(Settings.gitCommand,"ls-remote");
      command.append(rootPath+(!name.isEmpty() ? '/'+name : ""));
      exec = new Exec(command);
      exec.closeStdin();

      // read output
      while (!exec.eofStdout())
      {
        String line;

        // read stdout
        line = exec.pollStdout();
        if (line != null)
        {
          if (line.endsWith("/"))
          {
            // directory
            repositoryFileList.add(new RepositoryFile(rootPath,(!name.isEmpty() ? name+'/' : "")+line.substring(0,line.length()-1)));
          }
          else
          {
            // file
            repositoryFileList.add(new RepositoryFile(rootPath,(!name.isEmpty() ? name+'/' : "")+line));
          }
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // sort by name
    Collections.sort(repositoryFileList);

    return repositoryFileList;
  }

  /** create new repository module
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param importPath import directory
   */
  @Override
  public void create(String repositoryURL, String moduleName, String importPath)
    throws RepositoryException
  {
Dprintf.dprintf("NYI");
throw new RepositoryException("NYI");
  }

  /** checkout repository
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param revisionId revision to checkout
   * @param userName user login name or ""
   * @param destinationPath destination path
   * @param busyDialog busy dialog or null
   */
  @Override
  public void checkout(String repositoryURL, String moduleName, String revisionId, String userName, String destinationPath, BusyDialog busyDialog)
    throws RepositoryException
  {
    final Pattern PATTERN_URI1 = Pattern.compile("^([^:/]+)://([^:]*):([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI2 = Pattern.compile("^([^:/]+)://([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI3 = Pattern.compile("^([^:/]+)://(.*)",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,repositoryURL) : null;

    File destinationDirectory = new File(destinationPath);
    File tmpDirectory         = null;
    Exec exec                 = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      // create destination directory
      destinationDirectory.mkdir();

      // create temporary directory
      tmpDirectory = FileUtils.createTempDirectory("git",".tmp",destinationDirectory.getParentFile());

      // checkout into temporary directory
      command.clear();
      command.append(Settings.gitCommand,"clone","--verbose","--progress");
      if ((revisionId != null) && !revisionId.isEmpty()) command.append("--branch",revisionId);
      Matcher matcher;
      if      ((matcher = PATTERN_URI1.matcher(repositoryURL)).matches())
      {
        command.appendConcat(matcher.group(1),
                             "://",
                             matcher.group(2),
                             ":",
                             command.hidden(matcher.group(3)),
                             "@",
                             matcher.group(4)
                            );
      }
      else if ((matcher = PATTERN_URI2.matcher(repositoryURL)).matches())
      {
        if ((password == null) || password.isEmpty())
        {
          throw new RepositoryException("no password given");
        }
        command.appendConcat(matcher.group(1),
                             "://",
                             matcher.group(2),
                             ":",
                             command.hidden(password),
                             "@",
                             matcher.group(3)
                            );
      }
      else if ((matcher = PATTERN_URI3.matcher(repositoryURL)).matches())
      {
        if ((userName == null) || userName.isEmpty())
        {
          userName = System.getProperty("user.name");
        }
        if ((password == null) || password.isEmpty())
        {
          throw new RepositoryException("no password given");
        }
        command.appendConcat(matcher.group(1),
                             "://",
                             userName,
                             ":",
                             command.hidden(password),
                             "@",
                             matcher.group(2)
                            );
      }
      else
      {
        command.appendConcat(repositoryURL);
      }
      command.append(tmpDirectory);
      exec = new Exec(tmpDirectory.getAbsolutePath(),command,10*1000);
      exec.closeStdin();

      // read output
      while (   ((busyDialog == null) || !busyDialog.isAborted())
             && !exec.isTerminated()
            )
      {
        // read stdout
        line = exec.pollStdout();
        if (line != null)
        {
//Dprintf.dprintf("out: %s",line);
          if (busyDialog != null) busyDialog.updateList(line);
        }
      }
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
      }

      // done
      exec.done(); exec = null;

      // move to destination directory
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        try
        {
          if (busyDialog != null) busyDialog.updateList("Move files into '"+destinationPath+"'...");
          FileUtils.moveDirectoryTree(tmpDirectory,destinationDirectory);
        }
        catch (IOException exception)
        {
          throw new RepositoryException(Onzen.reniceIOException(exception));
        }
      }

      // switch to branch
      if (!moduleName.isEmpty())
      {
        command.clear();
        command.append(Settings.gitCommand,"checkout",moduleName);
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          // wait for termination
          exitCode = exec.waitFor();
          if (exitCode != 0)
          {
            throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
          }
        }
        else
        {
          // abort
          exec.destroy();
        }

        // done
        exec.done(); exec = null;
      }

      // get sub-modules
      command.clear();
      command.append(Settings.gitCommand,"submodule","init");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
      }

      // done
      exec.done(); exec = null;

      command.clear();
      command.append(Settings.gitCommand,"submodule","update");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
      if (tmpDirectory != null) try { FileUtils.deleteDirectoryTree(tmpDirectory); } catch (IOException ignoredException) { /* ignored */ }
    }
  }

  /** update file states
   * @param fileDataSet file data set to update
   * @param directorySet relative directory name set to check for new/missing files
   * @param newFileDataSet new file data set or null
   */
  @Override
  public void updateStates(Set<FileData> fileDataSet, Set<String> directorySet, Set<FileData> newFileDataSet)
    throws RepositoryException
  {
    final Pattern PATTERN_STATUS         = Pattern.compile("^\\s*(.)(.)\\s+(.*?)/*\\s*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_CURRENT_BRANCH = Pattern.compile("^\\*\\s+(.*?)/*\\s*",Pattern.CASE_INSENSITIVE);

    // all not reported files have state OK
    for (FileData fileData : fileDataSet)
    {
      fileData.state = FileData.States.OK;
    }
//for (FileData d : fileDataSet) Dprintf.dprintf("d=%s",d);

    // get repository absolute path
    String absolutePath = getRootPath();

    // get status of files
    for (String fileDirectory : directorySet)
    {
      Exec exec = null;
      try
      {
        FileData.Types  type;
        String          name               = null;
        FileData.States state              = FileData.States.UNKNOWN;
        FileData.Modes  mode               = FileData.Modes.UNKNOWN;
        long            size;
        Date            date;
        String          workingRevision    = "";
        String          repositoryRevision = "";

        Command         command            = new Command();
        String          line;
        Matcher         matcher;

        // get status
        command.clear();
        command.append(Settings.gitCommand,"status","--porcelain","--untracked-files=all","--ignored");
        command.append("--");
        command.append(!fileDirectory.isEmpty() ? fileDirectory : ".");
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // parse status data
        String xdirectory = new File(rootPath,fileDirectory).getPath();
        while ((line = exec.getStdout()) != null)
        {
//Dprintf.dprintf("line=%s",line);
          // match state, file name
          if      ((matcher = PATTERN_STATUS.matcher(line)).matches())
          {
            // get state, name
            state = parseState(matcher.group(1),matcher.group(2));
            name  = StringUtils.replace(StringUtils.unescape(matcher.group(3)),'/',File.separatorChar);

            // filter files by name prefix (Note: git unfortunatelly report always all files :-(...)
            String absoluteName = new File(absolutePath,name).getPath();
            if (absoluteName.startsWith(rootPath))
            {
              // get file type, size, date/time
              File file = new File(absolutePath,name);
              type = getFileType(file);
              size = file.length();
              date = new Date(file.lastModified());

              FileData fileData = findFileData(fileDataSet,file.getAbsolutePath());
              if      (fileData != null)
              {
                fileData.type               = type;
                fileData.state              = state;
                fileData.mode               = mode;
                fileData.size               = size;
                fileData.date               = date;
                fileData.workingRevision    = workingRevision;
                fileData.repositoryRevision = repositoryRevision;
              }
              else if (   (newFileDataSet != null)
                       && !isHiddenFile(name)
                       && !isIgnoreFile(name)
                       && !containFile(newFileDataSet,file.getAbsolutePath())
                      )
              {
                // create file data
//Dprintf.dprintf("add new %s, %s",name,state);
                newFileDataSet.add(FileData.getInstance(rootPath,
                                                        name,
                                                        type,
                                                        state,
                                                        mode,
                                                        false,  // locked
                                                        size,
                                                        date,
                                                        workingRevision,
                                                        repositoryRevision
                                                       )
                                  );
              }
            }
          }
          else
          {
            // unknown line
            Onzen.printWarning("No match for update states line '%s'",line);
          }
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(exception);
      }
      finally
      {
        if (exec != null) exec.done();
      }
//Dprintf.dprintf("fileDataSet=%s",fileDataSet);
//Dprintf.dprintf("newFileDataSet=%s",newFileDataSet);

      // set branch name
      String branchName = getCurrentBranchName();
      for (FileData fileData : fileDataSet)
      {
//Dprintf.dprintf("fileData=%s",fileData);
        fileData.branch = branchName;
      }
    }
  }

  /** get repository root URL
   * @return repository root URL
   */
  @Override
  public String getRepositoryRootURL()
  {
    if (repositoryRootURL == null)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;
        int     exitCode;

        command.clear();
        command.append(Settings.gitCommand,"rev-parse","--show-toplevel");
        command.append("--");
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        line = exec.getStdout();
        if (line != null) repositoryRootURL = line;

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        // ignored
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    return repositoryRootURL;
  }

  /** get repository URL
   * @return repository URL
   */
  @Override
  public String getRepositoryURL()
  {
    if (repositoryURL == null)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;

        command.clear();
        command.append(Settings.gitCommand,"config","--local","remote.origin.url");
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // parse output
        line = exec.getStdout();
        if (line != null) repositoryURL = line;

        // wait for termination
        exec.waitFor();

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        // ignored
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    return repositoryURL;
  }

  /** get master repository URL
   * @return masterRepository master repository URL or null
   */
  @Override
  public RepositoryURL getMasterRepositoryURL()
  {
    return new RepositoryURL(Types.GIT,getRepositoryURL());
  }

  /** get current branch name
   * @return branch name
   */
  @Override
  public String getCurrentBranchName()
  {
    final Pattern PATTERN_CURRENT_BRANCH = Pattern.compile("^\\*\\s+(.*?)\\s*",Pattern.CASE_INSENSITIVE);

    String branchName = "";

    // get root
    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;
      int     exitCode;

      command.clear();
      command.append(Settings.gitCommand,"branch");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // parse branch
      while ((line = exec.getStdout()) != null)
      {
        if ((matcher = PATTERN_CURRENT_BRANCH.matcher(line)).matches())
        {
          branchName = matcher.group(1);
          break;
        }
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return branchName;
  }

  /** get first revision name
   * @return first revision name
   */
  @Override
  public String getFirstRevision()
  {
    return FIRST_REVISION_NAME;
  }

  /** get last revision name
   * @return last revision name
   */
  @Override
  public String getLastRevision()
  {
    return LAST_REVISION_NAME;
  }

  /** get current revision
   * @paran repositoryURL repository server URL
   * @return current revision or null
   */
  public String getCurrentRevision(String repositoryURL)
    throws RepositoryException
  {
//TODO
return null;
  }

  /** get revision names of file
   * @param name file name or URL
   * @return revision name list
   */
  @Override
  public List<String> getRevisionNames(String name)
    throws RepositoryException
  {
    ArrayList<String> revisionList = new ArrayList<String>();

    // get revision info list
    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;


      // get log (Note: execute in correct directory to support GIT sub-modules, too)
      command.clear();
      command.append(Settings.gitCommand,"log","--format="+(Settings.gitShortCommitNumbers ? "%h" : "%H"));
      command.append("--");
      if ((name != null) && name.isEmpty())
      {
        File file = new File(rootPath,name);
        exec = new Exec(!file.isDirectory() ? file.getParent() : file.getAbsolutePath(),command);
      }
      else
      {
        exec = new Exec(rootPath,command);
      }
      exec.closeStdin();

      // parse revisions in log output
      while ((line = exec.getStdout()) != null)
      {
        revisionList.add(line);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // sort
    Collections.sort(revisionList);

    return revisionList;
  }

  /** get revision data
   * @param fileData file data (not used)
   * @param revisionId revision
   * @return revision data or null
   */
  @Override
  public RevisionData getRevisionData(FileData fileData, String revisionId)
    throws RepositoryException
  {
    RevisionDataGIT revisionData = null;

    assert revisionId != null;

    // get revision data
    Exec exec = null;
    try
    {
      Command            command           = new Command();
      String             line;
      Date               date              = new Date();
      String             author            = "";
      String             copyFromPath      = "";
      String             copyFromRevision  = "";
      ArrayList<String>  fileNameList      = new ArrayList<String>();
      LinkedList<String> commitMessageList = new LinkedList<String>();

      // get single log entry (Note: execute in correct directory to support GIT sub-modules, too)
      command.clear();
      command.append(Settings.gitCommand,"log","-1","--format="+(Settings.gitShortCommitNumbers ? LOG_TEMPLATE_SHORT : LOG_TEMPLATE_LONG));
      if (revisionId != null) command.append(revisionId);
      command.append("--");
      if (fileData != null)
      {
Dprintf.dprintf("--------------- %s",fileData.getFile().getParent());
        exec = new Exec((fileData.type != FileData.Types.DIRECTORY) ? fileData.getFile().getParent() : fileData.getName(),command);
      }
      else
      {
        exec = new Exec(rootPath,command);
      }
      exec.closeStdin();
//Dprintf.dprintf("command=%s",command);

      // parse
      revisionData = parseNextRevisionData(exec);

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return revisionData;
  }

  /** get revision data tree
   * @param fileData file data
   * @param busyDialog busy dialog or null
   * @return revision data tree
   */
  @Override
  public RevisionDataTree getRevisionDataTree(FileData fileData, BusyDialog busyDialog)
    throws RepositoryException
  {
    /* Note: for an unknown reason e. g. "| | * \ \   " do not match with "^[|/\\ ]*\\*[|/\\ ]*\\s*$" - why? Use temporary a more relaxed pattern
    final Pattern PATTERN_COMMIT = Pattern.compile("^[|/\\ ]*\\*[|/\\ ]*\\s*$",Pattern.CASE_INSENSITIVE);
    */
    final Pattern PATTERN_COMMIT = Pattern.compile("^[|/\\ ]*\\*[|/\\ ]*(.*?)$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_BRANCH = Pattern.compile("^[|/\\ ]*\\|/[|/\\ ]*\\s*$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_MERGE  = Pattern.compile("^[|/\\ ]*\\|\\\\[|/\\ ]*\\s*$",Pattern.CASE_INSENSITIVE);

    LinkedList<RevisionDataGIT> revisionDataList = new LinkedList<RevisionDataGIT>();

    // init busy dialog
    if (busyDialog != null) busyDialog.setMaximum(getTotalCommitCount());

    // get revision info list
    if (busyDialog != null) busyDialog.updateText("Get revisions...");
    Exec exec = null;
    try
    {
      Command command  = new Command();
      String  line;
      Matcher matcher;

      /* get log
         Note 1: execute in correct directory to support GIT sub-modules, too
         Note 2: because GIT does not offer any useful command to list a complete log including
                 branch and merge information, use the graph log view and parse that output :-(
      */
      command.clear();
      command.append(Settings.gitCommand,"log","--all","--graph","--no-color","--format=%n%n%n"+(Settings.gitShortCommitNumbers ? LOG_TEMPLATE_SHORT : LOG_TEMPLATE_LONG));
      command.append("--");
      if (fileData != null)
      {
        command.append(fileData.getName());
        exec = new Exec((fileData.type != FileData.Types.DIRECTORY) ? fileData.getFile().getParent() : fileData.getName(),command);
      }
      else
      {
        exec = new Exec(rootPath,command);
      }
      exec.closeStdin();
//Dprintf.dprintf("command=%s",command);

      // parse header
      if (parseLogHeader(exec))
      {
        // parse data
        int                   i            = 0;
        String                skipPrefix;
        RevisionDataGIT.Types type;
        RevisionDataGIT       revisionData = null;
        while (   ((line = exec.getStdout()) != null)
               && ((busyDialog == null) || !busyDialog.isAborted())
              )
        {
//Dprintf.dprintf("line=%s",line);
          if ((matcher = PATTERN_COMMIT.matcher(line)).matches())
          {
            // get last type
            line = exec.getStdout();
            if      ((matcher = PATTERN_BRANCH.matcher(line)).matches()) type = RevisionDataGIT.Types.BRANCH;
            else if ((matcher = PATTERN_MERGE.matcher(line)).matches() ) type = RevisionDataGIT.Types.MERGE;
            else                                                         type = RevisionDataGIT.Types.COMMIT;

            // get line prefix
            skipPrefix = exec.getStdout();

            // parse and add log entry
            revisionData = parseNextRevisionData(exec,skipPrefix);
            if (revisionData != null)
            {
              revisionData.type = type;
              revisionDataList.addFirst(revisionData);
            }

            // next entry, update busy dialog
            i++;
            if (busyDialog != null) busyDialog.updateProgressBar(i);
          }
          else
          {
            if (Settings.debugFlag)
            {
              Dprintf.halt("Unknown line: '%s'",line);
            }
          }
        }
      }
//for (RevisionDataGIT revisionData : revisionDataList) Dprintf.dprintf("%s: %s",revisionData.revisionId,revisionData.type);

      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        exec.destroy();
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
//Dprintf.dprintf("***************************");
//for (RevisionDataGIT revisionData : revisionDataList) Dprintf.dprintf("revisionData=%s\n",revisionData);

    // create revision data tree from list
    if (busyDialog != null) busyDialog.updateText("Build revision tree...");
    return createRevisionDataTree(revisionDataList,busyDialog);
  }

  /** get file data (text lines)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public String[] getFileLines(FileData fileData, String revisionId)
    throws RepositoryException
  {
    List<String> lineList = new ArrayList<String>();

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;

      // get file (Note: execute in sub-directory!)
      command.clear();
      command.append(Settings.gitCommand,"show");
      command.append(((revisionId != null) ? revisionId : "HEAD")+":"+fileData.getName(rootPath));
      exec = new Exec(fileData.getFile().getParent(),command);
      exec.closeStdin();

      // read file data
      while ((line = exec.getStdout()) != null)
      {
        lineList.add(line);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return lineList.toArray(new String[lineList.size()]);
  }

  /** get file data (byte array)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public byte[] getFileBytes(FileData fileData, String revisionId)
    throws RepositoryException
  {
    ByteArrayOutputStream output = new ByteArrayOutputStream(64*1024);

    Exec exec = null;
    try
    {
      Command command = new Command();
      int     n;
      byte[]  buffer  = new byte[64*1024];

      // get file data (Note: execute in sub-directory!)
      command.clear();
      command.append(Settings.gitCommand,"show");
      command.append(((revisionId != null) ? revisionId : "HEAD")+":"+fileData.getName(rootPath));
      exec = new Exec(fileData.getFile().getParent(),command,true);
      exec.closeStdin();

      // read file bytes into byte array stream
      while ((n = exec.readStdout(buffer)) > 0)
      {
        output.write(buffer,0,n);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // convert byte array stream into array
    return output.toByteArray();
  }

  /** get all changed/unknown files
   * @param stateSet state set
   * @return fileDataSet file data set with modified files
   */
  @Override
  public Set<FileData> getChangedFiles(EnumSet<FileData.States> stateSet)
    throws RepositoryException
  {
    final Pattern PATTERN_STATUS         = Pattern.compile("^\\s*(.)(.)\\s+(.*?)\\s*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_CURRENT_BRANCH = Pattern.compile("^\\*\\s+(.*?)\\s*",Pattern.CASE_INSENSITIVE);

    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    // get repository absolute path
    String absolutePath = getRootPath();

    Exec exec = null;
    try
    {
      Command         command            = new Command();
      String          line;
      Matcher         matcher;
      String          name               = null;
      FileData.States state              = FileData.States.UNKNOWN;
      String          workingRevision    = "";
      String          repositoryRevision = "";
      String          author             = "";

      // get list of files which may be updated or which are locally changed
      command.clear();
      command.append(Settings.gitCommand,"status","--porcelain","--untracked-files=all");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read list
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        // match name, state
        if      ((matcher = PATTERN_STATUS.matcher(line)).matches())
        {
          state = parseState(matcher.group(1),matcher.group(2));
          name  = StringUtils.replace(StringUtils.unescape(matcher.group(3)),'/',File.separatorChar);

          File file = new File(absolutePath,name);
          if (   !isHiddenFile(file)
              && !isIgnoreFile(file)
              && file.isFile()
              && stateSet.contains(state)
              && file.getAbsolutePath().startsWith(rootPath)
             )
          {
            // get relative name
            name = file.getAbsolutePath().substring(rootPath.length()+1);

            FileData fileData = FileData.getInstance(rootPath,
                                                     name,
                                                     state,
                                                     FileData.Modes.BINARY
                                                    );
            fileDataSet.add(fileData);
          }
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for changed files line '%s'",line);
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
//Dprintf.dprintf("fileDataSet=%s",fileDataSet);

    return fileDataSet;
  }

  /** get diff of file
   * @param fileData file data
   * @param revision revision to get diff for
   * @return diff data
   */
  @Override
  public List<DiffData> getDiff(FileData fileData, String oldRevision, String newRevision)
    throws RepositoryException
  {
    final Pattern PATTERN_DIFF_START1   = Pattern.compile("^\\-\\-\\-.*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF_START2   = Pattern.compile("^\\+\\+\\+.*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF_OLD_MODE = Pattern.compile("^old mode\\s+.*$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF_NEW_MODE = Pattern.compile("^new mode\\s+.*$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF_INDEX    = Pattern.compile("^index\\s+.*$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF          = Pattern.compile("^@@\\s+\\-([\\d,]*)\\s+\\+([\\d,]*)\\s+@@(.*)$",Pattern.CASE_INSENSITIVE);

    ArrayList<DiffData> diffDataList = new ArrayList<DiffData>();

    String[] newFileLines = null;
    if (newRevision != null)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;
        Matcher matcher;

        File file = new File(rootPath,getFileDataName(fileData));

        // check out new revision (Note: execute in root path)
        command.clear();
        command.append(Settings.gitCommand,"show");
        command.append(newRevision+":"+getFileDataName(fileData));
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // read content
        List<String> newFileLineList = new ArrayList<String>();
        while ((line = exec.getStdout()) != null)
        {
          newFileLineList.add(line);
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;

        // convert to lines array
        newFileLines = newFileLineList.toArray(new String[newFileLineList.size()]);
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }
    else
    {
      // use local revision
      try
      {
        // open file
        BufferedReader bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));

        // read content
        List<String> newFileLineList = new ArrayList<String>();
        String line;
        while ((line = bufferedReader.readLine()) != null)
        {
          newFileLineList.add(line);
        }

        // close file
        bufferedReader.close();

        // convert to lines array
        newFileLines = newFileLineList.toArray(new String[newFileLineList.size()]);
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
    }

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      // diff file (Note: execute in correct directory to support GIT sub-modules, too)
      command.clear();
      command.append(Settings.gitCommand,"diff");
      if (oldRevision != null) command.append("-r",oldRevision);
      if (newRevision != null) command.append("-r",newRevision);
      command.append("--");
      command.append(fileData.getName());
      exec = new Exec((fileData.type != FileData.Types.DIRECTORY) ? fileData.getFile().getParent() : fileData.getName(),command);
      exec.closeStdin();

      /* skip diff header (1 line)
           diff ...
      */
      exec.getStdout();

      /* parse diff output
           Format:
             @@ -<i> +<j> @@
      */
      int          lineNb = 1;
      DiffData     diffData;
      List<String> keepLinesList    = new ArrayList<String>();
      List<String> addedLinesList   = new ArrayList<String>();
      List<String> deletedLinesList = new ArrayList<String>();
      String[]     lines;
      while ((line = exec.getStdout()) != null)
      {
        if      ((matcher = PATTERN_DIFF_START1.matcher(line)).matches())
        {
          // skipped
        }
        else if ((matcher = PATTERN_DIFF_START2.matcher(line)).matches())
        {
          // skipped
        }
        else if ((matcher = PATTERN_DIFF_OLD_MODE.matcher(line)).matches())
        {
          // skipped
        }
        else if ((matcher = PATTERN_DIFF_NEW_MODE.matcher(line)).matches())
        {
          // skipped
        }
        else if ((matcher = PATTERN_DIFF_INDEX.matcher(line)).matches())
        {
          // skipped
        }
        else if ((matcher = PATTERN_DIFF.matcher(line)).matches())
        {
          int[] oldIndex = parseDiffIndex(matcher.group(1));
          int[] newIndex = parseDiffIndex(matcher.group(2));
//Dprintf.dprintf("oldIndex=lineNb %d, lines %d",oldIndex[0],oldIndex[1]);
//Dprintf.dprintf("newIndex=lineNb %d, lines %d",newIndex[0],newIndex[1]);
//Dprintf.dprintf("matcher.group(3)=%s",matcher.group(3));

          // git is odd: there is a line of the diff appended to the diff-header line, but it is not needed?...
          // exec.ungetStdout(matcher.group(3));

          // read until @@ is found
          while (   ((line = exec.peekStdout()) != null)
                 && !line.startsWith("@@")
                )
          {
//Dprintf.dprintf("line=#%s#",line);
            // skip unknown lines
            while (   ((line = exec.getStdout()) != null)
                   && !line.isEmpty()
                   && !line.startsWith(" ")
                   && !line.startsWith("-")
                   && !line.startsWith("+")
                  )
            {
//Dprintf.dprintf("skip=#%s#",line);
            }
            exec.ungetStdout(line);

            // get keep lines
            keepLinesList.clear();
            while ((lineNb < newIndex[0]) && (lineNb <= newFileLines.length))
            {
//Dprintf.dprintf("keep=#%s#",newFileLines[lineNb-1]);
              keepLinesList.add(newFileLines[lineNb-1]);
              lineNb++;
            }
            while (   ((line = exec.getStdout()) != null)
                   && (line.isEmpty() || line.startsWith(" "))
                  )
            {
              if (lineNb <= newFileLines.length)
              {
//Dprintf.dprintf("keep=#%s#",newFileLines[lineNb-1]);
                keepLinesList.add(newFileLines[lineNb-1]);
              }
              else
              {
Dprintf.dprintf("ups keep lineNb=%d newFileLines.length=%d",lineNb,newFileLines.length);
                keepLinesList.add("???");
              }
              lineNb++;
            }
            exec.ungetStdout(line);
            diffData = new DiffData(DiffData.Types.KEEP,keepLinesList);
            diffDataList.add(diffData);

            // get deleted lines
            deletedLinesList.clear();
            while (   ((line = exec.getStdout()) != null)
                   && line.startsWith("-")
                  )
            {
//Dprintf.dprintf("del=#%s#",line);
              deletedLinesList.add(line.substring(1));
            }
            exec.ungetStdout(line);

            // get added lines
            addedLinesList.clear();
            while (   ((line = exec.getStdout()) != null)
                   && line.startsWith("+")
                  )
            {
//Dprintf.dprintf("add=#%s#",line);
              addedLinesList.add(line.substring(1));
              lineNb++;
            }
            exec.ungetStdout(line);

            int deletedLinesCount = deletedLinesList.size();
            int addedLinesCount   = addedLinesList.size();
            if      (deletedLinesCount < addedLinesCount)
            {
              // changed
              if (deletedLinesCount > 0)
              {
                lines = new String[deletedLinesCount];
                for (int z = 0; z < deletedLinesCount; z++)
                {
                  lines[z] = addedLinesList.get(z);
                }
                diffData = new DiffData(DiffData.Types.CHANGED,lines,deletedLinesList);
                diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",lines.length,deletedLinesCount);
              }

              // added
              lines = new String[addedLinesCount-deletedLinesCount];
              for (int z = 0; z < addedLinesCount-deletedLinesCount; z++)
              {
                 lines[z] = addedLinesList.get(deletedLinesCount+z);
              }
              diffData = new DiffData(DiffData.Types.ADDED,lines);
              diffDataList.add(diffData);
//Dprintf.dprintf("a %d",lines.length);
            }
            else if (deletedLinesCount > addedLinesCount)
            {
              // changed
              if (addedLinesCount > 0)
              {
                lines = new String[addedLinesCount];
                for (int z = 0; z < addedLinesCount; z++)
                {
                  lines[z] = deletedLinesList.get(z);
                }
                diffData = new DiffData(DiffData.Types.CHANGED,addedLinesList,lines);
                diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",addedLinesCount,lines.length);
              }

              // deleted
              lines = new String[deletedLinesCount-addedLinesCount];
              for (int z = 0; z < deletedLinesCount-addedLinesCount; z++)
              {
                 lines[z] = deletedLinesList.get(addedLinesCount+z);
              }
              diffData = new DiffData(DiffData.Types.DELETED,lines);
              diffDataList.add(diffData);
//Dprintf.dprintf("d %d",lines.length);
            }
            else if ((deletedLinesCount > 0) && (addedLinesCount > 0))
            {
              // changed
              diffData = new DiffData(DiffData.Types.CHANGED,addedLinesList,deletedLinesList);
              diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",addedLinesCount,deletedLinesCount);
            }
          }
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for diff line '%s'",line);
        }
      }

      // get rest of keep lines
      if (lineNb <= newFileLines.length)
      {
        keepLinesList.clear();
        while (lineNb <= newFileLines.length)
        {
          keepLinesList.add(newFileLines[lineNb-1]);
          lineNb++;
        }
        diffData = new DiffData(DiffData.Types.KEEP,keepLinesList);
        diffDataList.add(diffData);
      }
//Dprintf.dprintf("diffData=%s",diffData);

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
/*
int lineNb=1;
for (DiffData d : diffDataList)
{
Dprintf.dprintf("%s %d: %d %d %d",d.blockType,
lineNb,
(d.keepLines!=null)?d.keepLines.length:0,
(d.addedLines!=null)?d.addedLines.length:0,
(d.deletedLines!=null)?d.deletedLines.length:0
);
if (d.blockType==DiffData.Types.CHANGE) for (int z = 0; z < d.deletedLines.length; z++) Dprintf.dprintf("%s -----> %s",d.deletedLines[z],d.addedLines[z]);
if (d.blockType==DiffData.Types.KEEP) lineNb += d.keepLines.length;
if (d.blockType==DiffData.Types.ADDED) lineNb += d.addedLines.length;
}
*/

    return diffDataList;
  }

  /** get patch for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @param output patch output or null
   * @param lineList patch data lines or null
   */
  @Override
  public void getPatch(Set<FileData> fileDataSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces, PrintWriter output, List<String> lineList)
    throws RepositoryException
  {
    final Pattern PATTERN_OLD_FILE = Pattern.compile("^\\-\\-\\-\\s+(a[/\\\\])(.*?)(\\t.*){0,1}",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_NEW_FILE = Pattern.compile("^\\+\\+\\+\\s+(b[/\\\\])(.*?)(\\t.*){0,1}",Pattern.CASE_INSENSITIVE);

    // get existing/new files
    HashMap<String,FileData> existingFileDataMap = new HashMap<String,FileData>();
    HashSet<FileData>        newFileDataSet      = new HashSet<FileData>();
    if (fileDataSet != null)
    {
      for (FileData fileData : fileDataSet)
      {
        if (fileData.state != FileData.States.UNKNOWN)
        {
          existingFileDataMap.put(getFileDataName(fileData),fileData);
        }
        else
        {
          newFileDataSet.add(fileData);
        }
      }
    }

    // get patch for existing files
    if ((fileDataSet == null) || (existingFileDataMap.size() > 0))
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;

        // get patch
        command.clear();
        command.append(Settings.gitCommand,"diff","--patch");
        if (ignoreWhitespaces)
        {
          command.append("-w","-b","--ignore-space-at-eol");
        }
        else
        {
        }
        if (revisionId1 != null) command.append("-r",revisionId1);
        if (revisionId2 != null) command.append("-r",revisionId2);
        command.append("--");
        if (fileDataSet != null) command.append(getFileDataNames(existingFileDataMap));
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // read patch data
        HashSet<FileData> patchFileDataSet = new HashSet<FileData>();
        Matcher matcher;
        while ((line = exec.getStdout()) != null)
        {
//Dprintf.dprintf("line=%s",line);
          // fix +++/--- lines: check if absolute path and convert or remove prefixes
          if      ((matcher = PATTERN_OLD_FILE.matcher(line)).matches())
          {
            String prefix     = matcher.group(1);
            String name       = matcher.group(2);
            String dateString = matcher.group(3);

            // store
            patchFileDataSet.add(existingFileDataMap.get(name));

            // get file name
            String fileName;
            if      (name.startsWith(rootPath+File.separator))
            {
              // absolute path -> convert to relative path to root path
              fileName = name.substring(rootPath.length()+1);
            }
            else
            {
              // use original name
              fileName = prefix+name;
            }

            line = "--- "+fileName+((dateString != null)?"\t"+dateString:"");
          }
          else if ((matcher = PATTERN_NEW_FILE.matcher(line)).matches())
          {
            String prefix     = matcher.group(1);
            String name       = matcher.group(2);
            String dateString = matcher.group(3);

            // store
            patchFileDataSet.add(existingFileDataMap.get(name));

            // get file name
            String fileName;
            if      (name.startsWith(rootPath+File.separator))
            {
              // absolute path -> convert to relative path to root path
              fileName = name.substring(rootPath.length()+1);
            }
            else
            {
              // use original name
              fileName = prefix+name;
            }

            line = "+++ "+fileName+((dateString != null)?"\t"+dateString:"");
          }

          if      (output   != null) output.println(line);
          else if (lineList != null) lineList.add(line);
        }

        // get file names for not modified files
        for (FileData fileData : existingFileDataMap.values())
        {
          if (!patchFileDataSet.contains(fileData))
          {
            if      (output   != null) output.println("File: "+getFileDataName(fileData));
            else if (lineList != null) lineList.add("File: "+getFileDataName(fileData));
          }
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    // get complete patches for new files
    for (FileData fileData : newFileDataSet)
    {
      try
      {
        BufferedReader bufferedReader;
        String         line;

        // count number of lines in file
        int lineCount = 0;
        bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));
        while ((line = bufferedReader.readLine()) != null)
        {
          lineCount++;
        }
        bufferedReader.close();

        // add as patch
        bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));
        String dateString = DateFormat.getDateInstance().format(new Date());
        if      (output   != null)
        {
          output.println(String.format("diff -u %s",fileData.getName()));
          output.println(String.format("--- /dev/null\t%s",dateString));
          output.println(String.format("+++ %s\t%s",fileData.getName(),dateString));
          output.println(String.format("@@ -1,%d +1,%d @@",lineCount,lineCount));
          while ((line = bufferedReader.readLine()) != null)
          {
            output.println("+"+line);
          }
        }
        else if (lineList != null)
        {
          lineList.add(String.format("diff -u %s",fileData.getName()));
          lineList.add(String.format("--- /dev/null\t%s",dateString));
          lineList.add(String.format("+++ %s\t%s",fileData.getName(),dateString));
          lineList.add(String.format("@@ -1,%d +1,%d @@",lineCount,lineCount));
          while ((line = bufferedReader.readLine()) != null)
          {
            lineList.add("+"+line);
          }
        }
        bufferedReader.close();
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
    }
  }

  /** get patch data for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @return patch data bytes
   */
  @Override
  public byte[] getPatchBytes(Set<FileData> fileDataSet, String revisionId1, String revisionId2)
    throws RepositoryException
  {
Dprintf.dprintf("");
    return null;
  }

  /** get log to file
   * @param fileData file data
   * @return log array
   */
  @Override
  public List<LogData> getLog(FileData fileData)
    throws RepositoryException
  {
    final Pattern PATTERN_FILENAME = Pattern.compile("^[ACDMRTUXB]\\s+(.*)",Pattern.CASE_INSENSITIVE);

    List<LogData> logDataList = new ArrayList<LogData>();

    Exec exec = null;
    try
    {
      Command            command = new Command();
      String             line;
      String             revision          = null;
      Date               date              = new Date();
      String             author            = "";
      String             copyFromPath      = "";
      String             copyFromRevision  = "";
      ArrayList<String>  fileNameList      = new ArrayList<String>();
      LinkedList<String> commitMessageList = new LinkedList<String>();

      // get log (Note: execute in correct directory to support GIT sub-modules, too)
      command.clear();
      command.append(Settings.gitCommand,"log","--name-status","--format=---%n"+(Settings.gitShortCommitNumbers ? "%h" : "%H")+"%n%cn%n%ci%n%w(,3,3)%B");
      command.append("--");
      if (fileData != null)
      {
        command.append(fileData.getName());
        exec = new Exec((fileData.type != FileData.Types.DIRECTORY) ? fileData.getFile().getParent() : fileData.getName(),command);
      }
      else
      {
        exec = new Exec(rootPath,command);
      }
      exec.closeStdin();

      Matcher matcher;
      while (   (line = exec.getStdout()) != null
             && (line.equals("---"))
            )
      {
        // get revision, date, author
        revision = exec.getStdout();
        author   = exec.getStdout();
        date     = parseDate(exec.getStdout());

        // get commit message
        commitMessageList.clear();
        while (   ((line = exec.getStdout()) != null)
               && line.startsWith("  ")
              )
        {
          commitMessageList.add(line.substring(3));
        }
        while (!commitMessageList.isEmpty() && commitMessageList.peekFirst().trim().isEmpty())
        {
          commitMessageList.removeFirst();
        }
        while (!commitMessageList.isEmpty() && commitMessageList.peekLast().trim().isEmpty())
        {
          commitMessageList.removeLast();
        }
        exec.ungetStdout(line);

        // skip empty lines
        while (   ((line = exec.getStdout()) != null)
               && line.trim().isEmpty()
              )
        {
          // nothing to do
        }
        exec.ungetStdout(line);

        // get file names
        fileNameList.clear();
        while (   ((line = exec.getStdout()) != null)
               && (matcher = PATTERN_FILENAME.matcher(line)).matches()
              )
        {
          fileNameList.add(matcher.group(1));
        }
        exec.ungetStdout(line);

        // add log info entry
        logDataList.add(new LogData(revision,
                                    date,
                                    author,
                                    commitMessageList,
                                    fileNameList
                                   )
                       );
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
//for (LogData logData : logDataList) Dprintf.dprintf("logData=%s",logData);

    return logDataList;
  }

  /** get annotations to file
   * @param fileData file data
   * @param revisionId revision
   * @return annotation list
   */
  @Override
  public List<AnnotationData> getAnnotations(FileData fileData, String revisionId)
    throws RepositoryException
  {
    final Pattern PATTERN_ANNOTATION = Pattern.compile("^(\\S+)\\s+\\((.*?)\\s+(\\S+\\s+\\S+\\s+\\S+)\\s+\\d+\\)\\s(.*)",Pattern.CASE_INSENSITIVE);
//    final Pattern PATTERN_COMMIT = Pattern.compile("^(\\S+)\\s+(\\d+)\\s+(\\d+)\\s(\\d+)",Pattern.CASE_INSENSITIVE);
//    final Pattern PATTERN_AUTHOR = Pattern.compile("^author\\s+(.+)",Pattern.CASE_INSENSITIVE);
//    final Pattern PATTERN_DATE   = Pattern.compile("^commit-date\\s+(.+)",Pattern.CASE_INSENSITIVE);
//    final Pattern PATTERN_LINE   = Pattern.compile("^\\t(.*)",Pattern.CASE_INSENSITIVE);

    ArrayList<AnnotationData> annotationDataList = new ArrayList<AnnotationData>();

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      File file = new File(rootPath,getFileDataName(fileData));

      // get annotations (Note: execute in correct directory to support GIT sub-modules, too)
      command.clear();
      command.append(Settings.gitCommand,"blame","--date","iso");
      if (revisionId != null) command.append(revisionId);
      command.append("--");
      command.append(file.getName());
      exec = new Exec((fileData.type != FileData.Types.DIRECTORY) ? fileData.getFile().getParent() : fileData.getName(),command);
      exec.closeStdin();

      /* parse annotation output
           Format:
             <revision> (<author> <date> <time> <time zone> <line nb>) <line>
      */
      while ((line = exec.getStdout()) != null)
      {
        if      ((matcher = PATTERN_ANNOTATION.matcher(line)).matches())
        {
          annotationDataList.add(new AnnotationData(matcher.group(1),
                                                    matcher.group(2),
                                                    parseDate(matcher.group(3)),
                                                    matcher.group(4))
                                                   );
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for annotations line '%s'",line);
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return annotationDataList;
  }

  /** update file from respository
   * @param fileDataSet file data set or null
   * @param busyDialog busy dialog or null
   */
  @Override
  public void update(Set<FileData> fileDataSet, BusyDialog busyDialog)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      if (Settings.gitStashOnUpdate)
      {
        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          // stash local changes
          command.clear();
          command.append(Settings.gitCommand,"stash","save");
          exec = new Exec(rootPath,command);
          exec.closeStdin();

          if ((busyDialog == null) || !busyDialog.isAborted())
          {
            exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
            }
          }
          else
          {
            // abort
            exec.destroy();
          }

          // done
          exec.done(); exec = null;
        }
      }

      try
      {
        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          // update files
          command.clear();
          command.append(Settings.gitCommand,"pull","--rebase");
          exec = new Exec(rootPath,command);
          exec.closeStdin();

          // read output
          while (   ((busyDialog == null) || !busyDialog.isAborted())
                 && !exec.isTerminated()
                )
          {
            // read stdout
            line = exec.pollStdout();
            if ((line != null) && line.startsWith("getting "))
            {
              if (busyDialog != null) busyDialog.updateList(line);
            }
          }
          if ((busyDialog == null) || !busyDialog.isAborted())
          {
            exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
            }
          }
          else
          {
            // abort
            exec.destroy();
          }

          // done
          exec.done(); exec = null;
        }

        if (Settings.gitAutoPullSubModules)
        {
          if ((busyDialog == null) || !busyDialog.isAborted())
          {
            // update files
            command.clear();
            command.append(Settings.gitCommand,"submodule","foreach",Settings.gitCommand,"pull","--rebase","origin","master");
            exec = new Exec(rootPath,command);
            exec.closeStdin();

            // read output
            while (   ((busyDialog == null) || !busyDialog.isAborted())
                   && !exec.isTerminated()
                  )
            {
              // read stdout
              line = exec.pollStdout();
              if ((line != null) && line.startsWith("getting "))
              {
      //Dprintf.dprintf("out: %s",line);
                if (busyDialog != null) busyDialog.updateList(line);
              }
            }
            if ((busyDialog == null) || !busyDialog.isAborted())
            {
              exitCode = exec.waitFor();
              if (exitCode != 0)
              {
                throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
              }
            }
            else
            {
              // abort
              exec.destroy();
            }

            // done
            exec.done(); exec = null;
          }
        }
      }
      finally
      {
        if (Settings.gitStashOnUpdate)
        {
          if ((busyDialog == null) || !busyDialog.isAborted())
          {
            // restore local changes
            command.clear();
            command.append(Settings.gitCommand,"stash","pop");
            exec = new Exec(rootPath,command);
            exec.closeStdin();

            if ((busyDialog == null) || !busyDialog.isAborted())
            {
              exitCode = exec.waitFor();
              if (exitCode != 0)
              {
                throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
              }
            }
            else
            {
              // abort
              exec.destroy();
            }

            // done
            exec.done(); exec = null;
          }
        }
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** commit files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void commit(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // commit files (Note: execute in correct directory to support GIT sub-modules, too)
      command.clear();
      command.append(Settings.gitCommand,"commit","-F",commitMessage.getFileName());
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** add files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary files, false otherwise
   */
  @Override
  public void add(Set<FileData> fileDataSet, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException
  {
    for (FileData fileData : fileDataSet)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        int     exitCode;

        // add files (Note: execute in correct directory to support GIT sub-modules, too)
        command.clear();
        command.append(Settings.gitCommand,"add");
        command.append("--");
        command.append(fileData.getName());
        exec = new Exec((fileData.type != FileData.Types.DIRECTORY) ? fileData.getFile().getParent() : fileData.getName(),command);
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;

        // immediate commit when message is given
        if ((commitMessage != null) && !commitMessage.isEmpty())
        {
          commit(fileData,commitMessage);
        }
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }
  }

  /** remove files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void remove(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      // delete local files
      for (FileData fileData : fileDataSet)
      {
        fileData.getFile().delete();
      }

      // remove from repository
      Command command = new Command();
      int     exitCode;

//      File file = new File(rootPath,getFileDataName(fileData));
//TODO
      // remove files (Note: execute in correct directory to support GIT sub-modules, too)
      command.clear();
      command.append(Settings.gitCommand,"rm","-r");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        // commit removed files
        commit(fileDataSet,commitMessage);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** copy files
   * @param fileDataSet files to copy
   * @param destination destination
   * @param commitMessage commit message
   */
  @Override
  public void copy(Set<FileData> fileDataSet, String destination, CommitMessage commitMessage)
    throws RepositoryException
  {
Dprintf.dprintf("TODO");
  }

  /** rename file
   * @param fileData file data to rename
   * @param newName new name
   * @param commitMessage commit message
   */
  @Override
  public void rename(FileData fileData, String newName, CommitMessage commitMessage)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // rename file
      command.clear();
      command.append(Settings.gitCommand,"mv");
      command.append("--");
      command.append(fileData.getName());
      command.append(newName);
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        Set<FileData> fileDataSet = FileData.toSet(rootPath,
                                                   getFileDataName(fileData),
                                                   newName
                                                  );
        commit(fileDataSet,commitMessage);
      }

      // set new name
      fileData.setFileName(newName);
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** revert files
   * @param fileDataSet file data set or null for all files
   * @param revisionId revision to revert to
   * @param recursive true for recursive revert, false otherwise
 */
  @Override
  public void revert(Set<FileData> fileDataSet, String revisionId, boolean recursive)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // delete local files
      for (FileData fileData : fileDataSet)
      {
        fileData.getFile().delete();
      }

      if (fileDataSet.size() > 0)
      {
        // revert files
        command.clear();
        command.append(Settings.gitCommand,"reset");
        command.append("--");
        command.append(getFileDataNames(fileDataSet));
        exec = new Exec(rootPath,command);

        // wait for termination
        exec.waitFor();

        // done
        exec.done(); exec = null;

        command.clear();
        command.append(Settings.gitCommand,"checkout");
        command.append("--");
        command.append(getFileDataNames(fileDataSet));
        exec = new Exec(rootPath,command);

        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** set conflicts resolved
   * @param fileDataSet file data set or null for all files
   */
  @Override
  public void resolve(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    for (FileData fileData : fileDataSet)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        int     exitCode;

        File file = new File(rootPath,getFileDataName(fileData));

        // "add" files to resolve
        command.clear();
        command.append(Settings.gitCommand,"add");
        command.append("--");
        command.append(file.getName());
        exec = new Exec(file.getParent(),command);
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;

/*
        // immediate commit when message is given
        if ((commitMessage != null) && !commitMessage.isEmpty())
        {
          commit(fileData,commitMessage);
        }*/
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }
  }

  /** change current branch
   * @param newBranchName new branch name
   */
  @Override
  public void changeBranch(String newBranchName)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // checkout branch
      command.clear();
      command.append(Settings.gitCommand,"checkout",newBranchName);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** merge other branch into current branch
   * @param otherBranchName other branch name
   */
  @Override
  public void mergeBranch(String otherBranchName)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // merge branch
      command.clear();
      command.append(Settings.gitCommand,"merge",otherBranchName);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** merge commit into current branch
   * @param revision revision
   */
  @Override
  public void mergeCommit(String revision)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // merge branch
      command.clear();
      command.append(Settings.gitCommand,"cherry-pick",revision);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** get incoming changes list
   * @param incoming changes list
   */
  @Override
  public List<LogData> getIncomingChanges(String masterRepository)
    throws RepositoryException
  {
    final Pattern PATTERN_CHANGE = Pattern.compile("^(\\d+)\\s+(\\S+)\\s+(\\S+\\s+\\S+\\s+\\S+)\\s+(.*)\\s*",Pattern.CASE_INSENSITIVE);

    ArrayList<LogData> logDataList = new ArrayList<LogData>();

    Exec exec = null;
    try
    {
      Command            command           = new Command();
      String             line;
      Matcher            matcher;
      String             revision          = null;
      String             changeSetId       = null;
      Date               date              = null;
      String             author            = null;
      LinkedList<String> commitMessageList = new LinkedList<String>();
      ArrayList<String>  fileNameList      = new ArrayList<String>();

      // get log
      command.clear();
      command.append(Settings.gitCommand,"log","..@{u}");
      exec = new Exec(rootPath,command);

      // parse revisions in log output
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        // match revision, node, date author
        if      ((matcher = PATTERN_CHANGE.matcher(line)).matches())
        {
          revision    = matcher.group(1);
          changeSetId = matcher.group(2);
          date        = parseDate(matcher.group(3));
          author      = matcher.group(4);

          // get files
          fileNameList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            fileNameList.add(line);
          }

          // get commit message
          commitMessageList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            commitMessageList.add(line);
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekFirst().trim().isEmpty())
          {
            commitMessageList.removeFirst();
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekLast().trim().isEmpty())
          {
            commitMessageList.removeLast();
          }

          logDataList.add(new LogData(revision,
                                      date,
                                      author,
                                      commitMessageList,
                                      fileNameList
                                     )
                         );
        }
      }

      // check exitcode (Note: exitcode 1 is OK and mean "no incoming changes")
      int exitCode = exec.waitFor();
      if (exitCode > 1)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return logDataList;
  }

  /** get outgoing changes list
   * @param outgoing changes list
   */
  @Override
  public List<LogData> getOutgoingChanges(String masterRepository)
    throws RepositoryException
  {
    final Pattern PATTERN_CHANGE = Pattern.compile("^(\\d+)\\s+(\\S+)\\s+(\\S+\\s+\\S+\\s+\\S+)\\s+(.*)\\s*",Pattern.CASE_INSENSITIVE);

    ArrayList<LogData> logDataList = new ArrayList<LogData>();

    Exec exec = null;
    try
    {
      Command            command           = new Command();
      String             line;
      Matcher            matcher;
      String             revision          = null;
      String             changeSetId       = null;
      Date               date              = null;
      String             author            = null;
      LinkedList<String> commitMessageList = new LinkedList<String>();
      ArrayList<String>  fileNameList      = new ArrayList<String>();

      // get log
      command.clear();
      command.append(Settings.gitCommand,"log","@{u}..");
      exec = new Exec(rootPath,command);

      // parse revisions in log output
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        // match revision, node, date author
        if      ((matcher = PATTERN_CHANGE.matcher(line)).matches())
        {
          revision    = matcher.group(1);
          changeSetId = matcher.group(2);
          date        = parseDate(matcher.group(3));
          author      = matcher.group(4);

          // get files
          fileNameList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            fileNameList.add(line);
          }

          // get commit message
          commitMessageList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            commitMessageList.add(line);
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekFirst().trim().isEmpty())
          {
            commitMessageList.removeFirst();
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekLast().trim().isEmpty())
          {
            commitMessageList.removeLast();
          }

          logDataList.add(new LogData(revision,
                                      date,
                                      author,
                                      commitMessageList,
                                      fileNameList
                                     )
                         );
        }
      }

      // check exitcode (Note: exitcode 1 is OK and mean "no outing changes")
      int exitCode = exec.waitFor();
      if (exitCode > 1)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return logDataList;
  }

  /** get incoming/outgoing changes lines
   * @param revisionId revision to get changes lines for
   */
  @Override
  public String[] getChanges(String revisionId)
    throws RepositoryException
  {
    return null;
  }

  /** pull changes
   * @param masterRepositoryURL master repository server URL
   * @param moduleName module name
   * @param userName user login name or ""
   * @param password password or ""
   */
  @Override
  public void pullChanges(String masterRepositoryURL, String moduleName, String userName)
    throws RepositoryException
  {
    final Pattern PATTERN_URI1 = Pattern.compile("^([^:/]+)://([^:]*):([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI2 = Pattern.compile("^([^:/]+)://([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI3 = Pattern.compile("^([^:/]+)://(.*)",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      Matcher matcher;
      int     exitCode;

      command.clear();
      command.append(Settings.gitCommand,"pull");
/*
      if ((masterRepositoryURL != null) && !masterRepositoryURL.isEmpty())
      {
        if      ((matcher = PATTERN_URI1.matcher(masterRepositoryURL)).matches())
        {
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(matcher.group(3)),
                               "@",
                               matcher.group(4)
                              );
        }
        else if ((matcher = PATTERN_URI2.matcher(masterRepositoryURL)).matches())
        {
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(3)
                              );
        }
        else if ((matcher = PATTERN_URI3.matcher(masterRepositoryURL)).matches())
        {
          if ((userName != null) && !userName.isEmpty())
          {
            if ((password != null) && !password.isEmpty())
            {
              command.appendConcat(matcher.group(1),
                                   "://",
                                   userName,
                                   ":",
                                   command.hidden(password),
                                   "@",
                                   matcher.group(2)
                                  );
            }
            else
            {
              command.appendConcat(matcher.group(1),
                                   "://",
                                   userName,
                                   "@",
                                   matcher.group(2)
                                  );
            }
          }
          else
          {
            command.appendConcat(matcher.group(1),
                                 "@",
                                 matcher.group(2)
                                );
          }
        }
        else
        {
          command.appendConcat(masterRepositoryURL);
        }
      }
      if (!moduleName.isEmpty()) command.append(moduleName);
*/
      command.append(!moduleName.isEmpty() ? moduleName : "origin");
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      if (Settings.gitAutoPullSubModules)
      {
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** push changes
   * @param masterRepositoryURL master repository URL
   * @param moduleName module name
   * @param userName user login name or ""
   * @param password password or ""
   */
  @Override
  public void pushChanges(String masterRepositoryURL, String moduleName, String userName)
    throws RepositoryException
  {
    final Pattern PATTERN_URI1 = Pattern.compile("^([^:/]+)://([^:]*):([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI2 = Pattern.compile("^([^:/]+)://([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI3 = Pattern.compile("^([^:/]+)://(.*)",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;
      int     exitCode;

      // push changes
      command.clear();
      command.append(Settings.gitCommand,"push");
/*
      if ((masterRepositoryURL != null) && !masterRepositoryURL.isEmpty())
      {
        if      ((matcher = PATTERN_URI1.matcher(masterRepositoryURL)).matches())
        {
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(matcher.group(3)),
                               "@",
                               matcher.group(4)
                              );
        }
        else if ((matcher = PATTERN_URI2.matcher(masterRepositoryURL)).matches())
        {
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(3)
                              );
        }
        else if ((matcher = PATTERN_URI3.matcher(masterRepositoryURL)).matches())
        {
          if ((userName == null) || userName.isEmpty())
          {
            userName = System.getProperty("user.name");
          }
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               userName,
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(2)
                              );
        }
        else
        {
          command.append(masterRepositoryURL);
        }
      }
      if (!moduleName.isEmpty()) command.append(moduleName);
*/
      command.append(!moduleName.isEmpty() ? moduleName : "origin");
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // push tags
      command.clear();
      command.append(Settings.gitCommand,"push","--tags");
/*
      if ((masterRepositoryURL != null) && !masterRepositoryURL.isEmpty())
      {
        if      ((matcher = PATTERN_URI1.matcher(masterRepositoryURL)).matches())
        {
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(matcher.group(3)),
                               "@",
                               matcher.group(4)
                              );
        }
        else if ((matcher = PATTERN_URI2.matcher(masterRepositoryURL)).matches())
        {
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(3)
                              );
        }
        else if ((matcher = PATTERN_URI3.matcher(masterRepositoryURL)).matches())
        {
          if ((userName == null) || userName.isEmpty())
          {
            userName = System.getProperty("user.name");
          }
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               userName,
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(2)
                              );
        }
        else
        {
          command.append(masterRepositoryURL);
        }
      }
      if (!moduleName.isEmpty()) command.append(moduleName);
*/
      command.append(!moduleName.isEmpty() ? moduleName : "origin");
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** apply patches
   */
  @Override
  public void applyPatches()
    throws RepositoryException
  {
Dprintf.dprintf("NYI");
throw new RepositoryException("NYI");
  }

  /** unapply patches
   */
  @Override
  public void unapplyPatches()
    throws RepositoryException
  {
Dprintf.dprintf("NYI");
throw new RepositoryException("NYI");
  }

  /** lock files
   * @param fileDataSet file data set
   */
  @Override
  public void lock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
Dprintf.dprintf("NYI");
throw new RepositoryException("NYI");
  }

  /** unlock files
   * @param fileDataSet file data set
   */
  @Override
  public void unlock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
Dprintf.dprintf("NYI");
throw new RepositoryException("NYI");
  }

  /** set commit message
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void setCommitMessage(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** set files mode
   * @param fileDataSet file data set
   * @param mode file mode
   * @param commitMessage commit message
   */
  @Override
  public void setFileMode(Set<FileData> fileDataSet, FileData.Modes mode, CommitMessage commitMessage)
    throws RepositoryException
  {
Dprintf.dprintf("NYI");
throw new RepositoryException("NYI");
  }

  /** get default name of root
   * @return default root name
   */
  @Override
  public String getDefaultRootName()
  {
    return DEFAULT_ROOT_NAME;
  }

  /** get default branch/tag name
   * @return default branch/tag name
   */
  @Override
  public String getDefaultBranchTagName()
  {
    return DEFAULT_BRANCHES_NAME;
  }

  /** get names of existing branches/tags
   * @param pathName path name
   * @return branch/tag name list
   */
  @Override
  public List<String> getBranchTagNames(String pathName)
    throws RepositoryException
  {
    final Pattern PATTERN_BRANCH = Pattern.compile("^(\\S+)\\s+refs/heads/(.*)",Pattern.CASE_INSENSITIVE);

    HashSet<String> branchTagNameSet = new HashSet<String>();

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;
      int     exitCode;

      // list tags
      command.clear();
//      command.append(Settings.gitCommand,"tag","-l");
      command.append(Settings.gitCommand,"ls-remote","--heads","origin");
      exec = new Exec(rootPath,command);

      // read output
      while ((line = exec.getStdout()) != null)
      {
        if      ((matcher = PATTERN_BRANCH.matcher(line)).matches())
        {
          branchTagNameSet.add(matcher.group(2));
          branchTagNameSet.add("origin/"+matcher.group(2));
        }
        else
        {
          branchTagNameSet.add(line);
        }
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // convert to list and sort
    ArrayList<String> branchTagNameList = new ArrayList<String>(branchTagNameSet);
    Collections.sort(branchTagNameList);

    return branchTagNameList;
  }

  /** create new branch
   * @param rootName root name (source)
   * @param branchTagName branch/tag name
   * @param fromRevision from revision name or null
   * @param commitMessage commit message
   * @param buysDialog busy dialog or null
   */
  @Override
  public void createBranch(String rootName, String branchTagName, String fromRevision, CommitMessage commitMessage, BusyDialog busyDialog)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      // create branch
      command.clear();
      command.append(Settings.gitCommand,"checkout","-b",branchTagName);
      if (!fromRevision.isEmpty()) command.append(fromRevision);
      exec = new Exec(rootPath,command);

      // read output
      while (   ((busyDialog == null) || !busyDialog.isAborted())
             && !exec.isTerminated()
            )
      {
        // read stdout
        line = exec.pollStdout();
        if (line != null)
        {
//Dprintf.dprintf("out: %s",line);
          if (busyDialog != null) busyDialog.updateList(line);
        }
      }
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
      }

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        // push branch
        command.clear();
        command.append(Settings.gitCommand,"push","-u","origin",branchTagName);
        exec = new Exec(rootPath,command);

        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  //-----------------------------------------------------------------------

  /** get absolute local repository path
   * @return absolute path
   */
  private String getRootPath()
  {
    String absolutePath = "";

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      command.clear();
      command.append(Settings.gitCommand,"rev-parse","--show-toplevel");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      absolutePath = exec.getStdout();

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return absolutePath;
  }

  /** get total commit count
   * @return total commit count
   */
  private int getTotalCommitCount()
    throws RepositoryException
  {
    int commitCount = 0;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      command.clear();
      command.append(Settings.gitCommand,"rev-list","--all","--count");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // parse
      if ((line = exec.getStdout()) != null)
      {
        commitCount = Integer.parseInt(line);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return commitCount;
  }

  /** get branch map
   * @return branch map
   */
  private HashMap<String,String> getBranchMap()
    throws RepositoryException
  {
    final Pattern PATTERN_BRANCH = Pattern.compile("^.\\s+(\\S+)\\s+(\\S+)\\s+.*",Pattern.CASE_INSENSITIVE);

    HashMap<String,String> branchMap = new HashMap<String,String>();

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      command.clear();
      command.append(Settings.gitCommand,"branch","--list","-vv");
      if (!Settings.gitShortCommitNumbers) command.append("--no-abrev");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // parse
      while ((line = exec.getStdout()) != null)
      {
        if ((matcher = PATTERN_BRANCH.matcher(line)).matches())
        {
          branchMap.put(matcher.group(1),matcher.group(2));
        }
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
//for (String k : branchMap.keySet()) Dprintf.dprintf("b %s: %s",k,branchMap.get(k));

    return branchMap;
  }

  /** parse git state string
   * @param remoteStateString remote state state string
   * @param localStateString local state state string
   * @return state
   */
  private FileData.States parseState(String remoteStateString, String localStateString)
  {
    if      (remoteStateString.equalsIgnoreCase(" ") && localStateString.equalsIgnoreCase(" ")) return FileData.States.OK;
    else if (remoteStateString.equalsIgnoreCase("M") || localStateString.equalsIgnoreCase("M")) return FileData.States.MODIFIED;
    else if (remoteStateString.equalsIgnoreCase("A") || localStateString.equalsIgnoreCase("A")) return FileData.States.ADDED;
    else if (remoteStateString.equalsIgnoreCase("D") || localStateString.equalsIgnoreCase("D")) return FileData.States.REMOVED;
    else if (remoteStateString.equalsIgnoreCase("R") || localStateString.equalsIgnoreCase("R")) return FileData.States.RENAMED;
    else if (remoteStateString.equalsIgnoreCase("C") || localStateString.equalsIgnoreCase("C")) return FileData.States.MODIFIED;
    else if (remoteStateString.equalsIgnoreCase("U") || localStateString.equalsIgnoreCase("U")) return FileData.States.MERGE;
    else if (remoteStateString.equalsIgnoreCase("?") || localStateString.equalsIgnoreCase("?")) return FileData.States.UNKNOWN;
    else if (remoteStateString.equalsIgnoreCase("!") || localStateString.equalsIgnoreCase("!")) return FileData.States.UNKNOWN;
    else                                                                                        return FileData.States.UNKNOWN;
  }

  /** parse HG diff index
   * @param string diff index string
   * @return index array [lineNb,length] or null
   */
  private int[] parseDiffIndex(String string)
  {
    Object[] data = new Object[2];
    if      (StringParser.parse(string,"%d,%d",data))
    {
      return new int[]{(Integer)data[0],(Integer)data[1]-(Integer)data[0]-1};
    }
    else if (StringParser.parse(string,"%d",data))
    {
      return new int[]{(Integer)data[0],1};
    }
    else
    {
      return null;
    }
  }

  /** parse log header (currently nothing to do)
   * @param exec exec command
   * @return true iff header parsed
   */
  private boolean parseLogHeader(Exec exec)
    throws IOException
  {
    return true;
  }

  /** parse next log revision data
   * @param exec exec command
   * @param skipPrefix skip prefix or null
   * @return revision data info or null
   */
  private RevisionDataGIT parseNextRevisionData(Exec exec, String skipPrefix)
    throws IOException
  {
    final Pattern PATTERN_BRANCH_NAMES  = Pattern.compile("^\\s*\\((.*)\\)$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_MERGE_PARENTS = Pattern.compile("^(\\S+)\\s+(\\S+)$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_BRANCH_PARENT = Pattern.compile("^(\\S+)$",Pattern.CASE_INSENSITIVE);

    RevisionDataGIT    revisionData      = null;

    String             line;
    Matcher            matcher;
    String             revisionId        = null;
    Date               date              = null;
    String             author            = null;
    String             changeSetId       = null;
    String             branchName        = null;
    String             parentRevisionId1 = null;
    String             parentRevisionId2 = null;
    TagList            tagList           = new TagList();
    LinkedList<String> commitMessageList = new LinkedList<String>();
    ArrayList<String>  fileNameList      = new ArrayList<String>();
    if ((line = exec.getStdout(skipPrefix)) != null)
    {
//Dprintf.dprintf("line=%s",line);
      // get revision, change-set id, date, author
      revisionId = line;
      date       = parseDate(exec.getStdout(skipPrefix));
      author     = exec.getStdout(skipPrefix);

      // get parents
      if ((line = exec.getStdout(skipPrefix)) != null)
      {
try{
}catch (java.lang.StringIndexOutOfBoundsException e)
{
Dprintf.dprintf("line='%s'",line);
Dprintf.dprintf("skipPrefix='%s'",skipPrefix);
}

        if      ((matcher = PATTERN_MERGE_PARENTS.matcher(line)).matches())
        {
          // node with two parents: merge
          parentRevisionId1 = matcher.group(1);
          parentRevisionId2 = matcher.group(2);
        }
        else if ((matcher = PATTERN_BRANCH_PARENT.matcher(line)).matches())
        {
          // node with single parent: branch
          parentRevisionId1 = matcher.group(1);
        }
      }

      // get branch name
      // Note: cannot detect branch name of a branch commit in GIT? :-(
      if ((line = exec.getStdout(skipPrefix)) != null)
      {
/*
//get branch name from references?

//Dprintf.dprintf("line=%s",line);
        if ((matcher = PATTERN_BRANCH_NAMES.matcher(line)).matches())
        {
//Dprintf.dprintf("matcher.group(1)=%s",matcher.group(1));
          String branchNames[] = matcher.group(1).split(", ");
          branchName = (branchNames.length > 0) ? branchNames[branchNames.length-1] : "";
//Dprintf.dprintf("branchName=%s",branchName);
        }
        else
        {
          branchName = line.trim();
//Dprintf.dprintf("2222branchName=%s",branchName);
        }
//System.exit(1);
*/
        branchName = "";
      }

      // get tags
      tagList.clear();

      // get commit message lines
      commitMessageList.clear();
      while ((line = exec.getStdout(skipPrefix)) != null)
      {
        if (line.startsWith("   "))
        {
          commitMessageList.add(line.substring(3));
        }
        else
        {
          break;
        }
      }
      while (!commitMessageList.isEmpty() && commitMessageList.peekFirst().trim().isEmpty())
      {
        commitMessageList.removeFirst();
      }
      while (!commitMessageList.isEmpty() && commitMessageList.peekLast().trim().isEmpty())
      {
        commitMessageList.removeLast();
      }

      // create log info entry
      revisionData = new RevisionDataGIT(revisionId,
                                         branchName,
                                         parentRevisionId1,
                                         parentRevisionId2,
                                         tagList,
                                         date,
                                         author,
                                         commitMessageList,
                                         fileNameList
                                        );
    }
//Dprintf.dprintf("revisionData=%s",revisionData);

    return revisionData;
  }

  /** parse next log revision data
   * @param exec exec command
   * @return revision data info or null
   */
  private RevisionDataGIT parseNextRevisionData(Exec exec)
    throws IOException
  {
    return parseNextRevisionData(exec,(String)null);
  }

  /** create revision tree
   * @param revisionDataList revision data list
   * @param busyDialog busy dialog or null
   * @return revision tree
   */
  private RevisionDataTree createRevisionDataTree(LinkedList<RevisionDataGIT> revisionDataList,
                                                  BusyDialog                  busyDialog
                                                 )
  {
    RevisionDataTree revisionDataTree = new RevisionDataTree();

    HashMap<String,RevisionDataTree>  revisionDataTreeMap       = new HashMap<String,RevisionDataTree>();
//TODO: branchRevisionDataTreeMap not needed?
    HashMap<String,RevisionDataTree>  branchRevisionDataTreeMap = new HashMap<String,RevisionDataTree>();
    branchRevisionDataTreeMap.put("master",revisionDataTree);
    HashMap<String,RevisionData>      revisionDataMap           = new HashMap<String,RevisionData>();

    for (RevisionDataGIT revisionData : revisionDataList)
    {
      RevisionDataTree branchRevisionDataTree;

      switch (revisionData.type)
      {
        case COMMIT:
          // commit -> find branch and add revision
//Dprintf.dprintf("---- commit %s: %s",revisionData.revisionId,revisionData.parentRevisionId1);
          branchRevisionDataTree = (revisionData.parentRevisionId1 != null) ? revisionDataTreeMap.get(revisionData.parentRevisionId1) : revisionDataTree;
          if (branchRevisionDataTree != null)
          {
            branchRevisionDataTree.add(revisionData);
            revisionDataTreeMap.put(revisionData.revisionId,branchRevisionDataTree);
          }
          else
          {
            // no branch found? -> add revision at end, skip
            revisionDataTree.add(revisionData);
            revisionDataTreeMap.put(revisionData.revisionId,revisionDataTree);
            if (busyDialog != null) busyDialog.updateList("Branch not found for %s",revisionData.parentRevisionId1);
          }
          break;
        case BRANCH:
          // branch -> find branch and add new branch
//Dprintf.dprintf("---- branch %s: %s = %s",revisionData.revisionId,revisionData.parentRevisionId1,revisionData.branchName);
          assert revisionData.parentRevisionId1 != null;

          RevisionData branchRootRevisionData = revisionDataMap.get(revisionData.parentRevisionId1);
          if (branchRootRevisionData == null) branchRootRevisionData = revisionDataTree.peekLast();
          if (branchRootRevisionData != null)
          {
            BranchData branchData = branchRootRevisionData.addBranch(revisionData.branchName,
                                                                     revisionData.date,
                                                                     revisionData.author,
                                                                     revisionData.commitMessageList,
                                                                     revisionData
                                                                    );
            revisionDataTreeMap.put(revisionData.revisionId,branchData.revisionDataTree);
            if (!branchRevisionDataTreeMap.containsKey(revisionData.branchName))
            {
              branchRevisionDataTreeMap.put(revisionData.branchName,branchData.revisionDataTree);
            }
          }
          else
          {
            // no branch root found? -> add revision at end, skip
            revisionDataTree.add(revisionData);
            revisionDataTreeMap.put(revisionData.revisionId,revisionDataTree);
            if (busyDialog != null) busyDialog.updateList("Branch not found for %s",revisionData.parentRevisionId1);
          }
          break;
        case MERGE:
          // merge -> find merge from/to and add merge
//Dprintf.dprintf("--- merge %s: %s+%s -> %s",revisionData.revisionId,revisionData.parentRevisionId1,revisionData.parentRevisionId2,revisionData.branchName);
          assert revisionData.parentRevisionId1 != null;
          assert revisionData.parentRevisionId2 != null;

          RevisionData mergeRevisionDataFrom = revisionDataMap.get(revisionData.parentRevisionId1);
          RevisionData mergeRevisionDataTo   = revisionDataMap.get(revisionData.parentRevisionId2);
          if ((mergeRevisionDataFrom != null) && (mergeRevisionDataTo != null))
          {
            revisionData.parentRevisionData0 = mergeRevisionDataFrom;
            revisionData.parentRevisionData1 = mergeRevisionDataTo;
            mergeRevisionDataFrom.mergeDataList.add(revisionData);
            mergeRevisionDataTo.mergeDataList.add(revisionData);
          }

          branchRevisionDataTree = revisionDataTreeMap.get(revisionData.parentRevisionId2);
          if (branchRevisionDataTree != null)
          {
//Dprintf.dprintf("first/last on merge destination %s/%s",branchRevisionDataTree.getFirst().revisionId,branchRevisionDataTree.peekLast().revisionId);
            branchRevisionDataTree.add(revisionData);
            revisionDataTreeMap.put(revisionData.revisionId,branchRevisionDataTree);
          }
          else
          {
            // no branch found? -> add revision at end, skip
            revisionDataTree.add(revisionData);
            revisionDataTreeMap.put(revisionData.revisionId,revisionDataTree);
            if (busyDialog != null) busyDialog.updateList("Branch not found for %s",revisionData.parentRevisionId1);
          }
          break;
      }

      // store revision into map
      revisionDataMap.put(revisionData.revisionId,revisionData);
    }
//Dprintf.dprintf("revisionDataTree size %d",revisionDataTree.size());

    return revisionDataTree;
  }

  /** stash save command
   * @param specialCommand special command
   * @param shell shell
   */
  private void stashSave(SpecialCommand specialCommand, Shell shell)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();

      // stash save
      command.append(Settings.gitCommand,"stash","save");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** stash pop command
   * @param specialCommand special command
   * @param shell shell
   */
  private void stashPop(SpecialCommand specialCommand, Shell shell)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();

      // stash pop
      command.append(Settings.gitCommand,"stash","pop");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** reset command
   * @param specialCommand special command
   * @param shell shell
   */
  private void reset(SpecialCommand specialCommand, Shell shell)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();

      // reset
      command.append(Settings.gitCommand,"reset");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** pull and update all sub-modules command
   * @param specialCommand special command
   * @param shell shell
   */
  private void pullUpdateAllSubModules(SpecialCommand specialCommand, Shell shell)
    throws RepositoryException
  {
    BusyDialog busyDialog = null;
    Exec       exec       = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      // open busy dialog
      busyDialog = busyDialog = new BusyDialog(shell,
                                               "Execute pull and update command",
                                               Settings.geometrySpecialCommand,
                                               "Execute pull and update command"+"\u2026",
                                               BusyDialog.LIST|BusyDialog.AUTO_ANIMATE
                                              );

      // pull sub-modules
      command.append(Settings.gitCommand,"submodule","foreach",Settings.gitCommand,"pull","--rebase","origin","master");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read output
      while (   !busyDialog.isAborted()
             && !exec.isTerminated()
            )
      {
        // read stdout
        line = exec.pollStdout();
        if (line != null)
        {
//Dprintf.dprintf("out: %s",line);
          busyDialog.updateList(line);
        }
      }
      if (!busyDialog.isAborted())
      {
        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
      }

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // update sub-modules
      command.append(Settings.gitCommand,"submodule","foreach",Settings.gitCommand,"update");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
      busyDialog.close(); busyDialog = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
      if (busyDialog != null) busyDialog.close();
    }
  }

/*
  private void getBranchRootRevision()
    throws RepositoryException
  {

diff -u <(git rev-list --first-parent topic)              <(git rev-list --first-parent master) |      sed -ne 's/^ //p' | head -1
  }

*/
}

/* end of file */
