/***********************************************************************\
*
* $Revision: 2155 $
* $Date: 2017-08-09 10:16:04 +0200 (Wed, 09 Aug 2017) $
* $Author: torsten $
* Contents: Mecurial repository functions
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

import java.text.DateFormat;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Set;
import java.util.Stack;

import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlType;

import org.eclipse.swt.widgets.Shell;

/****************************** Classes ********************************/

/** Mercurial (hg) repository
 */
@XmlType(propOrder={"userName","masterRepository","outgoingRepository"})
@XmlAccessorType(XmlAccessType.NONE)
class RepositoryHG extends Repository
{
  /** HG revision data
   */
  class RevisionDataHG extends RevisionData
  {
    int     revision;
    String  changeSetId;
    String  branchName;
    int     parentRevision1,parentRevision2;

    /** create HG revision data
     * @param revisionId revision id string
     * @param changeSetId change-set id string
     * @param branchName branch name
     * @param parentRevision1 parent revision id1 (!= 0 if branch or merge)
     * @param parentRevision2 parent revision id2 (!= 0 if merge)
     * @param tagList tag names or null
     * @param date date
     * @param author author name
     * @param commitMessageList commit message
     * @param fileNameList names of changed files
     */
    public RevisionDataHG(String             revisionId,
                          int                revision,
                          String             changeSetId,
                          String             branchName,
                          int                parentRevision1,
                          int                parentRevision2,
                          TagList            tagList,
                          Date               date,
                          String             author,
                          LinkedList<String> commitMessageList,
                          ArrayList<String>  fileNameList
                         )
    {
      super(revisionId,tagList,date,author,commitMessageList,fileNameList);
      this.revision        = revision;
      this.changeSetId     = changeSetId;
      this.branchName      = branchName;
      this.parentRevision1 = parentRevision1;
      this.parentRevision2 = parentRevision2;
    }

    /** get revision text for display
     * @return revision text
     */
    public String getRevisionText()
    {
      StringBuilder buffer = new StringBuilder();

      buffer.append(revisionId);
      if (changeSetId != null)
      {
        buffer.append(" (");
        buffer.append(changeSetId);
        buffer.append(")");
      }

      return buffer.toString();
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return "HG revision data {"+revisionId+", date: "+date+", author: "+author+", message: "+StringUtils.join(commitMessageList,", ")+"}";
    }
  }

  /** parent data
   */
  class ParentData
  {
    private int revisionId1,revisionId2;

    /** create parent data
     */
    ParentData(int revisionId1, int revisionId2)
    {
      this.revisionId1 = revisionId1;
      this.revisionId2 = revisionId2;
    }

    /** create parent data
     */
    ParentData(int revisionId1)
    {
      this(revisionId1,0);
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      if (revisionId2 != 0)
      {
        return Integer.toString(revisionId1)+","+Integer.toString(revisionId2);
      }
      else
      {
        return Integer.toString(revisionId1);
      }
    }
  }

  // --------------------------- constants --------------------------------
  public final static String[] DEFAULT_REVISION_NAMES = new String[]{"tip"};
  public final static String[] DEFAULT_BRANCH_NAMES   = new String[]{"trunk","branches"};
  public final static String   DEFAULT_ROOT_NAME      = "trunk";
  public final static String   DEFAULT_BRANCHES_NAME  = "branches";
  public final static String   FIRST_REVISION_NAME    = "-1";
  public final static String   LAST_REVISION_NAME     = "tip";

  /* log format template:
    <rev> <changeset id> <date> <time> <timezone> <author>
    <parents>
    <branch>
    <tags>
    <description>
    -----
    ...
  */
  private final String LOG_TEMPLATE_SHORT = "{rev} {node|short} {date|isodate} {author}\\n{parents}\\n{branch}\\n{tags}\\n{desc}\\n-----\\n";
  private final String LOG_TEMPLATE_LONG  = "{rev} {node|long} {date|isodate} {author}\\n{parents}\\n{branch}\\n{tags}\\n{desc}\\n-----\\n";

  // --------------------------- variables --------------------------------
  private final static RepositoryHG staticInstance = new RepositoryHG();

  @XmlElement(name = "userName")
  @RepositoryValue(title = "User name:", tooltip="HG server user login name.")
  public String userName = "";

  @XmlElement(name = "masterRepository")
  @RepositoryValue(title = "Master repository:", pathSelector=true, tooltip="Path to master repository.")
  public String masterRepository;

  @XmlElement(name = "outgoingRepository")
  @RepositoryValue(title = "Outgoing repository:", pathSelector=true, tooltip="Path to outgoing repository.")
  public String outgoingRepository;

  private String                      repositoryURL     = null;
  private String                      repositoryRootURL = null;
  private HashMap<Integer,ParentData> parentMap         = null;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get static instance
   * @return static instance
   */
  public final static RepositoryHG getInstance()
  {
    return staticInstance;
  }

  /** create repository
   * @param rootPath root path
   */
  RepositoryHG(String rootPath, String userName, PasswordHandler passwordHandler, String comment)
  {
    super(rootPath,passwordHandler,comment);

    assert !rootPath.isEmpty();

    this.userName = userName;

    // start getting parent map
    Background.run(new BackgroundRunnable()
    {
      public void run()
      {
// NYI: is the global parent map needed?
        parentMap = getParentMap();
//for (int i : parentMap.keySet()) Dprintf.dprintf("%d -> %s",i,parentMap.get(i));
      }
    });
  }

  /** create repository
   */
  RepositoryHG()
  {
    super();
  }

  /** get repository type
   * @return repository type
   */
  @Override
  public Types getType()
  {
    return Types.HG;
  }

  /** check if repository support change branch command
   * @return true iff change branch command is supported
   */
  @Override
  public boolean supportChangeBranch()
  {
    return true;
  }

  /** check if repository support merge branch command
   * @return true iff merge branch command is supported
   */
  @Override
  public boolean supportMergeBranch()
  {
    return true;
  }

  /** check if repository support merge commit command (cherry picking)
   * @return true iff merge commit command is supported
   */
  public boolean supportMergeCommit()
  {
    return true;
  }

  /** check if repository support incoming/outgoing commands
   * @return true iff incoming/outgoing commands are supported
   */
  @Override
  public boolean supportIncomingOutgoing()
  {
    return true;
  }

  /** check if repository support pull/push commands
   * @return true iff pull/push commands are supported
   */
  @Override
  public boolean supportPullPush()
  {
    return true;
  }

  /** check if repository support patch queues
   * @return true
   */
  @Override
  public boolean supportPatchQueues()
  {
    return true;
  }

  /** check if repository support lock/unlock
   * @return true iff lock/unlock is supported
   */
  @Override
  public boolean supportLockUnlock()
  {
Dprintf.dprintf("TODO: check if lock extension is available?");
    return true;
  }

  /** check if commit message is valid and acceptable
   * @return true iff commit message accepted
   */
  @Override
  public boolean validCommitMessage(CommitMessage commitMessage)
  {
    return    !Settings.hgSingleLineCommitMessages
           || (Settings.hgSingleLineMaxCommitMessageLength == 0)
           || (commitMessage.getMessage(", ").length() <= Settings.hgSingleLineMaxCommitMessageLength);
  }

  /** get master repository URL
   * @return masterRepository master repository URL or null
   */
  @Override
  public RepositoryURL getMasterRepositoryURL()
  {
    return new RepositoryURL(Types.GIT,masterRepository);
  }

  /** get special commands for repository
   * @return array with special commands
   */
  @Override
  public Special[] getSpecials()
  {
    return new SpecialCommand[]
    {
      new SpecialCommand()
      {
        @Override
        public String getTitle()
        {
          return "Rollback";
        }

        @Override
        public boolean run(Shell shell)
          throws RepositoryException
        {
          rollback(this,shell);

          return true;
        }
      }
    };
  }

  /** check if repository
   * @param repositoryURL respository server URL
   * @return repository file list
   */
  public static boolean isRepository(String rootPath)
  {
    boolean result = false;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      command.clear();
      command.append(Settings.hgCommand,"root");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      result = (exec.waitFor() == 0);

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return result;
  }

  /** list repository content
   * @param name name
   * @return repository file list
   */
  @Override
  public List<RepositoryFile> list(String name)
    throws RepositoryException
  {
return null;
  }

  /** create new repository module
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param importPath import directory
   */
  @Override
  public void create(String repositoryURL, String moduleName, String importPath)
    throws RepositoryException
  {
Dprintf.dprintf("NYI");
throw new RepositoryException("NYI");
  }

  /** checkout repository
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param revisionId revision to checkout
   * @param userName user name or ""
   * @param destinationPath destination path
   * @param busyDialog busy dialog or null
   */
  @Override
  public void checkout(String repositoryURL, String moduleName, String revisionId, String userName, String destinationPath, BusyDialog busyDialog)
    throws RepositoryException
  {
    final Pattern PATTERN_URI1 = Pattern.compile("^([^:/]+)://([^:]*):([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI2 = Pattern.compile("^([^:/]+)://([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI3 = Pattern.compile("^([^:/]+)://(.*)",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,repositoryURL) : null;

    Exec exec = null;
    try
    {
      Command  command              = new Command();
      int      exitCode             = -1;
      String[] extendedErrorMessage = null;
      Matcher  matcher;

      if (exitCode != 0)
      {
        // checkout
        command.clear();
        command.append(Settings.hgCommand,"-y","-v",(Settings.hgUseForestExtension) ? "fclone" : "clone");
        if (!revisionId.isEmpty()) command.append("-r",revisionId);
        if (Settings.hgAlwaysTrustServerCertificate) command.append("--config web.cacerts=");
        if      ((matcher = PATTERN_URI1.matcher(repositoryURL)).matches())
        {
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(matcher.group(3)),
                               "@",
                               matcher.group(4),
                               "/",
                               moduleName
                              );
        }
        else if ((matcher = PATTERN_URI2.matcher(repositoryURL)).matches())
        {
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(3),
                               "/",
                               moduleName
                              );
        }
        else if ((matcher = PATTERN_URI3.matcher(repositoryURL)).matches())
        {
          if ((userName == null) || userName.isEmpty())
          {
            userName = System.getProperty("user.name");
          }
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               userName,
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(2),
                               "/",
                               moduleName
                              );
        }
        else
        {
          command.append("file://"+repositoryURL+"/"+moduleName);
        }
        command.append(destinationPath);
        exec = new Exec(rootPath,command);

        // read output
        while (   ((busyDialog == null) || !busyDialog.isAborted())
               && !exec.isTerminated()
              )
        {
          String line;

          // read stdout
          line = exec.pollStdout();
          if ((line != null) && line.startsWith("getting "))
          {
//Dprintf.dprintf("out: %s",line);
            if (busyDialog != null) busyDialog.updateList(line);
          }
        }
        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          // wait for termination
          exitCode             = exec.waitFor();
          extendedErrorMessage = exec.getExtendedErrorMessage();
        }
        else
        {
          // abort
          exec.destroy();
          try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
        }

        // done
        exec.done(); exec = null;
      }

      if (exitCode != 0)
      {
        // checkout insecure
        command.clear();
        command.append(Settings.hgCommand,"-y","-v",(Settings.hgUseForestExtension) ? "fclone" : "clone","--insecure");
        if (!revisionId.isEmpty()) command.append("-r",revisionId);
        if (Settings.hgAlwaysTrustServerCertificate) command.append("--config web.cacerts=");
        if      ((matcher = PATTERN_URI1.matcher(repositoryURL)).matches())
        {
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(matcher.group(3)),
                               "@",
                               matcher.group(4),
                               "/",
                               moduleName
                              );
        }
        else if ((matcher = PATTERN_URI2.matcher(repositoryURL)).matches())
        {
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               matcher.group(2),
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(3),
                               "/",
                               moduleName
                              );
        }
        else if ((matcher = PATTERN_URI3.matcher(repositoryURL)).matches())
        {
          if ((userName == null) || userName.isEmpty())
          {
            userName = System.getProperty("user.name");
          }
          if ((password == null) || password.isEmpty())
          {
            throw new RepositoryException("no password given");
          }
          command.appendConcat(matcher.group(1),
                               "://",
                               userName,
                               ":",
                               command.hidden(password),
                               "@",
                               matcher.group(2),
                               "/",
                               moduleName
                              );
        }
        else
        {
          command.append("file://"+repositoryURL+"/"+moduleName);
        }
        command.append(destinationPath);
        exec = new Exec(rootPath,command);

        // read output
        while (   ((busyDialog == null) || !busyDialog.isAborted())
               && !exec.isTerminated()
              )
        {
          String line;

          // read stdout
          line = exec.pollStdout();
          if ((line != null) && line.startsWith("getting "))
          {
//Dprintf.dprintf("out: %s",line);
            if (busyDialog != null) busyDialog.updateList(line);
          }
        }
        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          // wait for termination
          exitCode             = exec.waitFor();
          extendedErrorMessage = exec.getExtendedErrorMessage();
        }
        else
        {
          // abort
          exec.destroy();
          try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
        }

        // done
        exec.done(); exec = null;
      }

      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        if (exitCode != 0)
        {
          try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
          throw new RepositoryException("'%s', exit code: %d",extendedErrorMessage,command.toString(),exitCode);
        }
      }
    }
    catch (IOException exception)
    {
      try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** update file states
   * @param fileDataSet file data set to update
   * @param fileDirectorySet relative directory name set to check for new/missing files
   * @param newFileDataSet new file data set or null
   */
  @Override
  public void updateStates(Set<FileData> fileDataSet, Set<String> fileDirectorySet, Set<FileData> newFileDataSet)
    throws RepositoryException
  {
    final Pattern PATTERN_STATUS = Pattern.compile("^\\s*(.)\\s+(.*?)\\s*",Pattern.CASE_INSENSITIVE);

    // all not reported files have state OK
    for (FileData fileData : fileDataSet)
    {
      fileData.state = FileData.States.OK;
    }
//Dprintf.dprintf("fileDataSet=%s",fileDataSet);

    // get repository absolute path
    String absolutePath = getRootPath();

    for (String directory : fileDirectorySet)
    {
      Exec exec = null;
      try
      {
        FileData.Types  type;
        String          name               = null;
        FileData.States state              = FileData.States.UNKNOWN;
        FileData.Modes  mode               = FileData.Modes.UNKNOWN;
        long            size;
        Date            date;
        String          workingRevision    = "";
        String          repositoryRevision = "";

        Command         command            = new Command();
        String          line;
        Matcher         matcher;

        command.clear();
        command.append(Settings.hgCommand,"-y","status","-mardcu");
        command.append("--");
        if (!directory.isEmpty()) command.append(directory);
        exec = new Exec(rootPath,command);

        // parse status data
        while ((line = exec.getStdout()) != null)
        {
//Dprintf.dprintf("line=%s",line);
          // check if one entry is complete
          // match file name, state
          if      ((matcher = PATTERN_STATUS.matcher(line)).matches())
          {
            // get state, name
            state = parseState(matcher.group(1));
            name  = StringUtils.replace(StringUtils.unescape(matcher.group(2)),'/',File.separatorChar);

            // filter files by name prefix (Note: hg unfortunatelly report always all files :-(...)
            String absoluteName = new File(absolutePath,name).getPath();
            if (absoluteName.startsWith(rootPath))
            {
              // get file type, size, date/time
              File file = new File(absolutePath,name);
              type = getFileType(file);
              size = file.length();
              date = new Date(file.lastModified());

              FileData fileData = findFileData(fileDataSet,file.getAbsolutePath());
//Dprintf.dprintf("directory=%s file=%s fileData=%s type=%s",directory,file,fileData,type);
              if      (fileData != null)
              {
                fileData.type               = type;
                fileData.state              = state;
                fileData.mode               = mode;
                fileData.size               = size;
                fileData.date               = date;
                fileData.workingRevision    = workingRevision;
                fileData.repositoryRevision = repositoryRevision;
              }
              else if (   (newFileDataSet != null)
                      && !isHiddenFile(name)
                      && !isIgnoreFile(name)
                      && !containFile(newFileDataSet,file.getAbsolutePath())
                      )
              {
                // check if file not in sub-directory (hg list all files :-()
                String subDirectory = new File(absolutePath,name).getParent();
                if (   ((subDirectory == null) && directory.isEmpty())
                    || ((subDirectory != null) && subDirectory.equals(directory))
                  )
                {
                  // create file data
                  FileData newFileData = FileData.getInstance(rootPath,
                                                              name,
                                                              type,
                                                              state,
                                                              mode,
                                                              false,  // locked
                                                              size,
                                                              date,
                                                              workingRevision,
                                                              repositoryRevision
                                                             );
                  newFileDataSet.add(newFileData);
//Dprintf.dprintf("new file=%s name=%s fileData=%s",file,name,newFileData);
                }
              }
            }
          }
          else
          {
            // unknown line
            Onzen.printWarning("No match for update states line '%s'",line);
          }
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(exception);
      }
      finally
      {
        if (exec != null) exec.done();
      }
//Dprintf.dprintf("fileDataSet=%s",fileDataSet);

      // get revision (tag or identity)
      exec = null;
      try
      {
        Command command = new Command();
        String  line;

        command.clear();
        command.append(Settings.hgCommand,"identify","-t");
        command.append("--");
        exec = new Exec(rootPath,command);
        line = exec.getStdout();
        if ((line == null) || line.trim().isEmpty())
        {
          exec.waitFor();
          exec.done(); exec = null;
          command.clear();
          command.append(Settings.hgCommand,"identify","-i");
          command.append("--");
          exec = new Exec(rootPath,command);
          line = exec.getStdout();
        }
        if (line != null)
        {
          for (FileData fileData : fileDataSet)
          {
            fileData.workingRevision = line;
          }
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(exception);
      }
      finally
      {
        if (exec != null) exec.done();
      }

      // get branch
      exec = null;
      try
      {
        Command command = new Command();
        String  line;

        command.clear();
        command.append(Settings.hgCommand,"branch");
        command.append("--");
        exec = new Exec(rootPath,command);
        if ((line = exec.getStdout()) != null)
        {
          for (FileData fileData : fileDataSet)
          {
            fileData.branch = line;
          }
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(exception);
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }
  }

  /** get repository root URL
   * @return repository root URL
   */
  @Override
  public String getRepositoryRootURL()
  {
    final Pattern PATTERN_DEFAULT = Pattern.compile("^default\\s+=\\s*(.+)",Pattern.CASE_INSENSITIVE);

    if (repositoryRootURL == null)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;
        Matcher matcher;

        command.clear();
        command.append(Settings.hgCommand,"paths");
        command.append("--");
        exec = new Exec(rootPath,command);

        // parse info output
        while ((line = exec.getStdout()) != null)
        {
//Dprintf.dprintf("line=%s",line);
          // match name, state
          if      ((matcher = PATTERN_DEFAULT.matcher(line)).matches())
          {
            // first parent directory which does not have a .hg directory is repository directory
            File file = new File(matcher.group(1));
            while (   !file.getPath().isEmpty()
                   && new File(file,".hg").exists()
                  )
            {
              file = file.getParentFile();
            }
            repositoryRootURL = file.getPath();
          }
        }

        // wait for termination
        exec.waitFor();

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        // ignored
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    return repositoryURL;
  }

  /** get repository URL
   * @return repository URL
   */
  @Override
  public String getRepositoryURL()
  {
    final Pattern PATTERN_DEFAULT = Pattern.compile("^default\\s+=\\s*(.+)",Pattern.CASE_INSENSITIVE);

    if (repositoryURL == null)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;
        Matcher matcher;

        command.clear();
        command.append(Settings.hgCommand,"paths","default");
        command.append("--");
        exec = new Exec(rootPath,command);

//???
if (false) {
        // parse info output
        while ((line = exec.getStdout()) != null)
        {
  //Dprintf.dprintf("line=%s",line);
          // match name, state
          if      ((matcher = PATTERN_DEFAULT.matcher(line)).matches())
          {
            // first parent directory which does not have a .hg directory is repository directory
            File file = new File(matcher.group(1));
            while (   !file.getPath().isEmpty()
                   && new File(file,".hg").exists()
                  )
            {
              file = file.getParentFile();
            }
            repositoryURL = file.getPath();
          }
        }
} else {
        repositoryURL = exec.getStdout();
}

        // wait for termination
        exec.waitFor();

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        // ignored
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    return repositoryURL;
  }

  /** get first revision name
   * @return first revision name
   */
  @Override
  public String getFirstRevision()
  {
    return FIRST_REVISION_NAME;
  }

  /** get last revision name
   * @return last revision name
   */
  @Override
  public String getLastRevision()
  {
    return LAST_REVISION_NAME;
  }

  /** get current revision
   * @paran repositoryURL repository server URL
   * @return current revision or null
   */
  public String getCurrentRevision(String repositoryURL)
    throws RepositoryException
  {
//TODO
return null;
  }

  /** get revision names of file
   * @param name file name or URL
   * @return revision name list
   */
  @Override
  public List<String> getRevisionNames(String name)
    throws RepositoryException
  {
    final Pattern PATTERN_REVSION = Pattern.compile("^(\\d+).*",Pattern.CASE_INSENSITIVE);

    ArrayList<String> revisionList = new ArrayList<String>();

    // get revision info list
    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      // get log
      command.clear();
      command.append(Settings.hgCommand,"-y","-v","log","--follow","--template","{rev} {node|short}\\n");
      command.append("--");
      if (name != null) command.append(name);
      exec = new Exec(rootPath,command);

      // parse revisions in log output
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        // match name, state
        if      ((matcher = PATTERN_REVSION.matcher(line)).matches())
        {
          // add log info entry
          revisionList.add(matcher.group(1));
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // sort
    Collections.sort(revisionList,new Comparator<String>()
    {
      public int compare(String revisionId1, String revisionId2)
      {
        int number1 = Integer.parseInt(revisionId1);
        int number2 = Integer.parseInt(revisionId2);
        if      (number1 < number2)
        {
          return -1;
        }
        else if (number1 > number2)
        {
          return 1;
        }
        else
        {
          return 0;
        }
      }
    });

    return revisionList;
  }

  /** get revision data
   * @param fileData file data
   * @param revisionId revision
   * @return revision data or null
   */
  @Override
  public RevisionData getRevisionData(FileData fileData, String revisionId)
    throws RepositoryException
  {
    RevisionData revisionData = null;

    assert revisionId != null;

    // get revision data
    Exec exec = null;
    try
    {
      Command command = new Command();

      // get single log entry
      command.clear();
      command.append(Settings.hgCommand,"-y","-v","log","--follow");
      if (revisionId != null) command.append("-r",revisionId);
      command.append("-l","1","--template",Settings.gitShortCommitNumbers ? LOG_TEMPLATE_SHORT : LOG_TEMPLATE_LONG);
      command.append("--");
      if (fileData != null) command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);
      exec.closeStdin();
Dprintf.dprintf("command=%s",command);

      // parse header
      if (parseLogHeader(exec))
      {
        // parse data
        revisionData = parseNextRevisionData(exec);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return revisionData;
  }

  /** get revision data tree
   * @param fileData file data
   * @param busyDialog busy dialog or null
   * @return revision data tree
   */
  @Override
  public RevisionDataTree getRevisionDataTree(FileData fileData, BusyDialog busyDialog)
    throws RepositoryException
  {
    LinkedList<RevisionDataHG> revisionDataList = new LinkedList<RevisionDataHG>();

    // init busy dialog
    if (busyDialog != null) busyDialog.setMaximum(getTotalCommitCount());

    // get revision data info list
    if (busyDialog != null) busyDialog.updateText("Get revisions...");
    Exec exec = null;
    try
    {
      Command command = new Command();

      // get log
      command.clear();
      command.append(Settings.hgCommand,"-y","-v","log","--follow","--template",Settings.gitShortCommitNumbers ? LOG_TEMPLATE_SHORT : LOG_TEMPLATE_LONG);
      command.append("--");
      if (fileData != null) command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);

      // parse header
      if (parseLogHeader(exec))
      {
        // parse data
        int            i = 0;
        RevisionDataHG revisionData;
        while (   ((revisionData = parseNextRevisionData(exec)) != null)
               && ((busyDialog == null) || !busyDialog.isAborted())
              )
        {
          // add log entry
          revisionDataList.addFirst(revisionData);

          // next entry, update busy dialog
          i++;
          if (busyDialog != null) busyDialog.updateProgressBar(i);
        }
      }
      else
      {
        Onzen.printWarning("Cannot parse log output");
      }

      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        exec.destroy();
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
//for (RevisionDataHG revisionData : revisionDataList) Dprintf.dprintf("revisionData=%s",revisionData);

    // create revision data tree from list
    if (busyDialog != null) busyDialog.updateText("Build revision tree...");
    return createRevisionDataTree(revisionDataList,busyDialog);
  }

  /** get file data (text lines)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public String[] getFileLines(FileData fileData, String revisionId)
    throws RepositoryException
  {
    List<String> lineList = new ArrayList<String>();

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;

      // get file
      command.clear();
      command.append(Settings.hgCommand,"-y","cat");
      if (revisionId != null) command.append("-r",revisionId);
      command.append("--");
      command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);

      // read file data
      while ((line = exec.getStdout()) != null)
      {
        lineList.add(line);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return lineList.toArray(new String[lineList.size()]);
  }

  /** get file data (byte array)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public byte[] getFileBytes(FileData fileData, String revisionId)
    throws RepositoryException
  {
    ByteArrayOutputStream output = new ByteArrayOutputStream(64*1024);

    Exec exec = null;
    try
    {
      Command command = new Command();
      int     n;
      byte[]  buffer  = new byte[64*1024];

      // get file data
      command.clear();
      command.append(Settings.hgCommand,"-y","cat");
      if (revisionId != null) command.append("-r",revisionId);
      command.append("--");
      command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command,true);

      // read file bytes into byte array stream
      while ((n = exec.readStdout(buffer)) > 0)
      {
        output.write(buffer,0,n);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // convert byte array stream into array
    return output.toByteArray();
  }

  /** get all changed/unknown files
   * @param stateSet state set
   * @return fileDataSet file data set with modified files
   */
  @Override
  public Set<FileData> getChangedFiles(EnumSet<FileData.States> stateSet)
    throws RepositoryException
  {
    final Pattern PATTERN_STATUS = Pattern.compile("^\\s*(.)\\s+(.*?)\\s*",Pattern.CASE_INSENSITIVE);

    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    Exec exec = null;
    try
    {
      Command         command            = new Command();
      String          line;
      Matcher         matcher;
      String          name               = null;
      FileData.States state              = FileData.States.UNKNOWN;
      String          workingRevision    = "";
      String          repositoryRevision = "";
      String          author             = "";

      // get list of files which may be updated or which are locally changed
      command.clear();
      command.append(Settings.hgCommand,"-y","status","-mardu");
      command.append("--");
      command.append("glob:**");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read list
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        // match name, state
        if      ((matcher = PATTERN_STATUS.matcher(line)).matches())
        {
          state = parseState(matcher.group(1));
          name  = StringUtils.replace(StringUtils.unescape(matcher.group(2)),'/',File.separatorChar);

          File file = new File(rootPath,name);
          if (   !isHiddenFile(file)
              && !isIgnoreFile(file)
              && file.isFile()
              && stateSet.contains(state)
             )
          {
            // Note: 'glob:**' produce relative name

            fileDataSet.add(FileData.getInstance(rootPath,
                                                 name,
                                                 state,
                                                 FileData.Modes.BINARY
                                                )
                           );
          }
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for changed files line '%s'",line);
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return fileDataSet;
  }

  /** get diff of file
   * @param fileData file data
   * @param revision revision to get diff for
   * @return diff data
   */
  @Override
  public List<DiffData> getDiff(FileData fileData, String oldRevision, String newRevision)
    throws RepositoryException
  {
    final Pattern PATTERN_DIFF_START1 = Pattern.compile("^\\-\\-\\-.*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF_START2 = Pattern.compile("^\\+\\+\\+.*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF        = Pattern.compile("^@@\\s+\\-([\\d,]*)\\s+\\+([\\d,]*)\\s+@@$",Pattern.CASE_INSENSITIVE);

    ArrayList<DiffData> diffDataList = new ArrayList<DiffData>();

    String[] newFileLines = null;
    if (newRevision != null)
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        Matcher matcher;
        String  line;

        // check out new revision
        command.clear();
        command.append(Settings.hgCommand,"-y","cat","-r",newRevision);
        command.append("--");
        command.append(getFileDataName(fileData));
        exec = new Exec(rootPath,command);

        // read content
        List<String> newFileLineList = new ArrayList<String>();
        while ((line = exec.getStdout()) != null)
        {
          newFileLineList.add(line);
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;

        // convert to lines array
        newFileLines = newFileLineList.toArray(new String[newFileLineList.size()]);
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }
    else
    {
      // use local revision
      try
      {
        // open file
        BufferedReader bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));

        // read content
        List<String> newFileLineList = new ArrayList<String>();
        String line;
        while ((line = bufferedReader.readLine()) != null)
        {
          newFileLineList.add(line);
        }

        // close file
        bufferedReader.close();

        newFileLines = newFileLineList.toArray(new String[newFileLineList.size()]);
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
    }

    Exec exec = null;
    try
    {
      Command command = new Command();
      Matcher matcher;
      String  line;

      // diff file
      command.clear();
      command.append(Settings.hgCommand,"-y","diff");
      if (oldRevision != null) command.append("-r",oldRevision);
      if (newRevision != null) command.append("-r",newRevision);
      command.append("--");
      command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);

      /* skip diff header (3 lines)
           diff -r ...
           --- ...
           +++ ...
      */
      exec.getStdout();
      exec.getStdout();
      exec.getStdout();

      /* parse diff output
           Format:
             @@ -<i> +<j> @@
      */
      int          lineNb = 1;
      DiffData     diffData;
      List<String> keepLinesList    = new ArrayList<String>();
      List<String> addedLinesList   = new ArrayList<String>();
      List<String> deletedLinesList = new ArrayList<String>();
      String[]     lines;
      while ((line = exec.getStdout()) != null)
      {
        if      ((matcher = PATTERN_DIFF.matcher(line)).matches())
        {
          int[] oldIndex = parseDiffIndex(matcher.group(1));
          int[] newIndex = parseDiffIndex(matcher.group(2));
//Dprintf.dprintf("oldIndex=%d,%d",oldIndex[0],oldIndex[1]);
//Dprintf.dprintf("newIndex=%d,%d",newIndex[0],newIndex[1]);

          // read until @@ is found
          while (   ((line = exec.peekStdout()) != null)
                 && !line.startsWith("@@")
                )
          {
//Dprintf.dprintf("line=#%s#",line);
            // skip unknown lines
            while (   ((line = exec.getStdout()) != null)
                   && !line.isEmpty()
                   && !line.startsWith(" ")
                   && !line.startsWith("-")
                   && !line.startsWith("+")
                  )
            {
//Dprintf.dprintf("skip=#%s#",line);
            }
            exec.ungetStdout(line);

            // get keep lines
            keepLinesList.clear();
            while ((lineNb < newIndex[0]) && (lineNb <= newFileLines.length))
            {
              keepLinesList.add(newFileLines[lineNb-1]);
              lineNb++;
            }
            while (   ((line = exec.getStdout()) != null)
                   && (line.isEmpty() || line.startsWith(" "))
                  )
            {
              if (lineNb <= newFileLines.length)
              {
                keepLinesList.add(newFileLines[lineNb-1]);
              }
              else
              {
                keepLinesList.add("???");
              }
              lineNb++;
            }
            exec.ungetStdout(line);
            diffData = new DiffData(DiffData.Types.KEEP,keepLinesList);
            diffDataList.add(diffData);

            // get deleted lines
            deletedLinesList.clear();
            while (   ((line = exec.getStdout()) != null)
                   && line.startsWith("-")
                  )
            {
              deletedLinesList.add(line.substring(1));
            }
            exec.ungetStdout(line);

            // get added lines
            addedLinesList.clear();
            while (   ((line = exec.getStdout()) != null)
                   && line.startsWith("+")
                  )
            {
              addedLinesList.add(line.substring(1));
              lineNb++;
            }
            exec.ungetStdout(line);

            int deletedLinesCount = deletedLinesList.size();
            int addedLinesCount   = addedLinesList.size();
            if      (deletedLinesCount < addedLinesCount)
            {
              // changed
              if (deletedLinesCount > 0)
              {
                lines = new String[deletedLinesCount];
                for (int z = 0; z < deletedLinesCount; z++)
                {
                  lines[z] = addedLinesList.get(z);
                }
                diffData = new DiffData(DiffData.Types.CHANGED,lines,deletedLinesList);
                diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",lines.length,deletedLinesCount);
              }

              // added
              lines = new String[addedLinesCount-deletedLinesCount];
              for (int z = 0; z < addedLinesCount-deletedLinesCount; z++)
              {
                 lines[z] = addedLinesList.get(deletedLinesCount+z);
              }
              diffData = new DiffData(DiffData.Types.ADDED,lines);
              diffDataList.add(diffData);
//Dprintf.dprintf("a %d",lines.length);
            }
            else if (deletedLinesCount > addedLinesCount)
            {
              // changed
              if (addedLinesCount > 0)
              {
                lines = new String[addedLinesCount];
                for (int z = 0; z < addedLinesCount; z++)
                {
                  lines[z] = deletedLinesList.get(z);
                }
                diffData = new DiffData(DiffData.Types.CHANGED,addedLinesList,lines);
                diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",addedLinesCount,lines.length);
              }

              // deleted
              lines = new String[deletedLinesCount-addedLinesCount];
              for (int z = 0; z < deletedLinesCount-addedLinesCount; z++)
              {
                 lines[z] = deletedLinesList.get(addedLinesCount+z);
              }
              diffData = new DiffData(DiffData.Types.DELETED,lines);
              diffDataList.add(diffData);
//Dprintf.dprintf("d %d",lines.length);
            }
            else if ((deletedLinesCount > 0) && (addedLinesCount > 0))
            {
              // changed
              diffData = new DiffData(DiffData.Types.CHANGED,addedLinesList,deletedLinesList);
              diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",addedLinesCount,deletedLinesCount);
            }
          }
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for diff line '%s'",line);
        }
      }

      // get rest of keep lines
      if (lineNb <= newFileLines.length)
      {
        keepLinesList.clear();
        while (lineNb <= newFileLines.length)
        {
          keepLinesList.add(newFileLines[lineNb-1]);
          lineNb++;
        }
        diffData = new DiffData(DiffData.Types.KEEP,keepLinesList);
        diffDataList.add(diffData);
      }
//Dprintf.dprintf("diffData=%s",diffData);

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
/*
int lineNb=1;
for (DiffData d : diffDataList)
{
Dprintf.dprintf("%s %d: %d %d %d",d.blockType,
lineNb,
(d.keepLines!=null)?d.keepLines.length:0,
(d.addedLines!=null)?d.addedLines.length:0,
(d.deletedLines!=null)?d.deletedLines.length:0
);
if (d.blockType==DiffData.Types.CHANGE) for (int z = 0; z < d.deletedLines.length; z++) Dprintf.dprintf("%s -----> %s",d.deletedLines[z],d.addedLines[z]);
if (d.blockType==DiffData.Types.KEEP) lineNb += d.keepLines.length;
if (d.blockType==DiffData.Types.ADDED) lineNb += d.addedLines.length;
}
*/

    return diffDataList;
  }

  /** get patch lines for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @param output patch output or null
   * @param lineList patch data lines or null
   */
  @Override
  public void getPatch(Set<FileData> fileDataSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces, PrintWriter output, List<String> lineList)
    throws RepositoryException
  {
    final Pattern PATTERN_OLD_FILE = Pattern.compile("^\\-\\-\\-\\s+(a[/\\\\])(.*?)(\\t.*){0,1}",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_NEW_FILE = Pattern.compile("^\\+\\+\\+\\s+(b[/\\\\])(.*?)(\\t.*){0,1}",Pattern.CASE_INSENSITIVE);

    // get existing/new files
    HashMap<String,FileData> existingFileDataMap = new HashMap<String,FileData>();
    HashSet<FileData>        newFileDataSet      = new HashSet<FileData>();
    if (fileDataSet != null)
    {
      for (FileData fileData : fileDataSet)
      {
        if (fileData.state != FileData.States.UNKNOWN)
        {
          existingFileDataMap.put(getFileDataName(fileData),fileData);
        }
        else
        {
          newFileDataSet.add(fileData);
        }
      }
    }

    // get patch for existing files
    if ((fileDataSet == null) || (existingFileDataMap.size() > 0))
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;

        // get sub-directory relative to hg root path
        String subDirectory = getSubDirectory();

        // get patch
        command.clear();
        command.append(Settings.hgCommand);
        if (!Settings.hgDiffCommand.isEmpty())
        {
          // use external diff command
          command.append("extdiff","-p",Settings.hgDiffCommand);
          if (ignoreWhitespaces)
          {
            if (!Settings.hgDiffCommandOptionsIgnoreWhitespaces.isEmpty())
            {
              command.append("-o",Settings.hgDiffCommandOptionsIgnoreWhitespaces);
            }
          }
          else
          {
            if (!Settings.hgDiffCommandOptions.isEmpty())
            {
              command.append("-o",Settings.hgDiffCommandOptions);
            }
          }
        }
        else
        {
          // use internal diff
          command.append("diff");
          if (ignoreWhitespaces)
          {
            command.append("-w","-b","-B","--git");
          }
          else
          {
            command.append("--git");
          }
        }
        if (revisionId1 != null) command.append("-r",revisionId1);
        if (revisionId2 != null) command.append("-r",revisionId2);
        command.append("--");
        if (fileDataSet != null) command.append(getFileDataNames(existingFileDataMap));
        exec = new Exec(rootPath,command);

        // read patch data
        HashSet<FileData> patchFileDataSet = new HashSet<FileData>();
        Matcher matcher;
        while ((line = exec.getStdout()) != null)
        {
//Dprintf.dprintf("line=%s",line);
          // fix +++/--- lines: check if absolute path and convert or remove prefixes
          if      ((matcher = PATTERN_OLD_FILE.matcher(line)).matches())
          {
            String prefix     = matcher.group(1);
            String name       = matcher.group(2);
            String dateString = matcher.group(3);

            // store
            patchFileDataSet.add(existingFileDataMap.get(name));

            // get file name
            String fileName;
            if      (name.startsWith(rootPath+File.separator))
            {
              // absolute path -> convert to relative path to root path
              fileName = name.substring(rootPath.length()+1);
            }
            else if (Settings.hgRelativePatchPaths)
            {
              // relative hg path -> convert to relative path to root path
              fileName = name.startsWith(subDirectory+File.separator)?name.substring(subDirectory.length()+1):name;
            }
            else
            {
              // use original name
              fileName = prefix+name;
            }

            // remove sub-directory if needed
            if ((subDirectory.length() > 0) && (fileName.length() > subDirectory.length()) && fileName.startsWith(subDirectory)) fileName = fileName.substring(subDirectory.length()+1);

            line = "--- "+fileName+((dateString != null)?"\t"+dateString:"");
          }
          else if ((matcher = PATTERN_NEW_FILE.matcher(line)).matches())
          {
            String prefix     = matcher.group(1);
            String name       = matcher.group(2);
            String dateString = matcher.group(3);

            // store
            patchFileDataSet.add(existingFileDataMap.get(name));

            // get file name
            String fileName;
            if      (name.startsWith(rootPath+File.separator))
            {
              // absolute path -> convert to relative path to root path
              fileName = name.substring(rootPath.length()+1);
            }
            else if (Settings.hgRelativePatchPaths)
            {
              // relative hg path -> convert to relative path to root path
              fileName = name.startsWith(subDirectory+File.separator)?name.substring(subDirectory.length()+1):name;
            }
            else
            {
              // use original name
              fileName = prefix+name;
            }

            // remove sub-directory if needed
            if ((subDirectory.length() > 0) && (fileName.length() > subDirectory.length()) && fileName.startsWith(subDirectory)) fileName = fileName.substring(subDirectory.length()+1);

            line = "+++ "+fileName+((dateString != null)?"\t"+dateString:"");
          }

          if      (output   != null) output.println(line);
          else if (lineList != null) lineList.add(line);
        }

        // get file names for not modified files
        for (FileData fileData : existingFileDataMap.values())
        {
          if (!patchFileDataSet.contains(fileData))
          {
            if      (output   != null) output.println("File: "+getFileDataName(fileData));
            else if (lineList != null) lineList.add("File: "+getFileDataName(fileData));
          }
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    // get complete patches for new files
    for (FileData fileData : newFileDataSet)
    {
      try
      {
        BufferedReader bufferedReader;
        String         line;

        // count number of lines in file
        int lineCount = 0;
        bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));
        while ((line = bufferedReader.readLine()) != null)
        {
          lineCount++;
        }
        bufferedReader.close();

        // add as patch
        bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));
        String dateString = DateFormat.getDateInstance().format(new Date());
        if      (output   != null)
        {
          output.println(String.format("diff -u %s",fileData.getName()));
          output.println(String.format("--- /dev/null\t%s",dateString));
          output.println(String.format("+++ %s\t%s",fileData.getName(),dateString));
          output.println(String.format("@@ -1,%d +1,%d @@",lineCount,lineCount));
          while ((line = bufferedReader.readLine()) != null)
          {
            output.println("+"+line);
          }
        }
        else if (lineList != null)
        {
          lineList.add(String.format("diff -u %s",fileData.getName()));
          lineList.add(String.format("--- /dev/null\t%s",dateString));
          lineList.add(String.format("+++ %s\t%s",fileData.getName(),dateString));
          lineList.add(String.format("@@ -1,%d +1,%d @@",lineCount,lineCount));
          while ((line = bufferedReader.readLine()) != null)
          {
            lineList.add("+"+line);
          }
        }
        bufferedReader.close();
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
    }
  }

  /** get patch data for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @return patch data bytes
   */
  @Override
  public byte[] getPatchBytes(Set<FileData> fileDataSet, String revisionId1, String revisionId2)
    throws RepositoryException
  {
    ByteArrayOutputStream output = new ByteArrayOutputStream(64*1024);

    Exec exec = null;
    try
    {
      Command command = new Command();
      int     n;
      byte[]  buffer  = new byte[64*1024];

      // get patch
      command.clear();
      command.append(Settings.hgCommand,"diff");
      if (revisionId1 != null) command.append("-r",revisionId1);
      if (revisionId2 != null) command.append("-r",revisionId2);
      command.append("--");
      if (fileDataSet != null) command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);

      // read patch bytes into byte array stream
      while ((n = exec.readStdout(buffer)) > 0)
      {
        output.write(buffer,0,n);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // convert byte array stream into array
    return output.toByteArray();
  }

  /** get log to file
   * @param fileData file data
   * @return log array
   */
  @Override
  public List<LogData> getLog(FileData fileData)
    throws RepositoryException
  {
    List<LogData> logDataList = new ArrayList<LogData>();

    Exec exec = null;
    try
    {
      // get log
      Command command = new Command();
      command.clear();
      command.append(Settings.hgCommand,"-y","-v","log","--follow","--template",Settings.gitShortCommitNumbers ? LOG_TEMPLATE_SHORT : LOG_TEMPLATE_LONG);
      command.append("--");
      if (fileData != null) command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);

      // parse header
      if (parseLogHeader(exec))
      {
        // parse data
        RevisionDataHG revisionData;
        while ((revisionData = parseNextRevisionData(exec)) != null)
        {
          // add log info entry
          logDataList.add(new LogData(revisionData.revisionId,
                                      revisionData.date,
                                      revisionData.author,
                                      revisionData.commitMessageList,
                                      revisionData.fileNameList
                                     )
                         );
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
//for (LogData logData : logDataList) Dprintf.dprintf("logData=%s",logData);

    return logDataList;
  }

  /** get annotations to file
   * @param fileData file data
   * @param revisionId revision
   * @return annotation list
   */
  @Override
  public List<AnnotationData> getAnnotations(FileData fileData, String revisionId)
    throws RepositoryException
  {
    final Pattern PATTERN_ANNOTATION = Pattern.compile("^\\s*(\\S+)\\s+(\\S+)\\s+(.*?):\\s(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_EMPTY      = Pattern.compile("^\\s*",Pattern.CASE_INSENSITIVE);

    ArrayList<AnnotationData> annotationDataList = new ArrayList<AnnotationData>();

    Exec exec = null;
    try
    {
      Command command = new Command();
      Matcher matcher;
      String  line;

      // get annotations
      command.clear();
      command.append(Settings.hgCommand,"blame","-n","-u","-d");
      if (revisionId != null) command.append("-r",revisionId);
      command.append("--");
      command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);

      /* parse annotation output
           Format:
             <revision> <author> <date> <line>
      */
      while ((line = exec.getStdout()) != null)
      {
        if      ((matcher = PATTERN_ANNOTATION.matcher(line)).matches())
        {
          annotationDataList.add(new AnnotationData(matcher.group(2),
                                                    matcher.group(1),
                                                    parseDate(matcher.group(3)),
                                                    matcher.group(4)
                                                   )
                                );
        }
        else if (PATTERN_EMPTY.matcher(line).matches())
        {
          // skip empty lines
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for annotations line '%s'",line);
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return annotationDataList;
  }

  /** update file from respository
   * @param fileDataSet file data set or null
   * @param busyDialog busy dialog or null
   */
  @Override
  public void update(Set<FileData> fileDataSet, BusyDialog busyDialog)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      String[] trees;
      if (Settings.hgUseForestExtension)
      {
        // get trees of forest (if forest extension is enabled)
        trees = getTrees();
      }
      else
      {
        // only single tree
        trees = new String[]{rootPath};
      }

      if (Settings.hgUseQueueExtension)
      {
        // unapply patches for all trees in forest/single repository
        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          for (String tree : trees)
          {
            command.clear();
            command.append(Settings.hgCommand,"qpop","-a");
            command.append("--");
            exec = new Exec(tree,command);

            exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
            }

            exec.done(); exec = null;
          }
        }
      }

      // update files (Note: hg does not allow to update single files, thus update all files)
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        if      (Settings.hgUpdateWithFetch)
        {
          StoredFiles   storedFiles   = null;
          StoredChanges storedChanges = null;
          try
          {
            // create files backup of all local changes (in case there occur some error)
            storedFiles = new StoredFiles(rootPath,getChangedFiles());

            if (Settings.hgSafeUpdate)
            {
              // do 'safe' update when fetch is used: store local changes into patch
              storedChanges = new StoredChanges();
            }

            // revert all local changes
            command.clear();
            command.append(Settings.hgCommand,"revert","--no-backup","-a");
            command.append("--");
            exec = new Exec(rootPath,command);

            exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              if (Settings.hgUpdateWithFetch && Settings.hgSafeUpdate)
              {
                storedFiles.restore(); storedFiles = null;
              }
              throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
            }

            exec.done(); exec = null;

            // fetch+fpush
            command.clear();
            command.append(Settings.hgCommand,"-y","fetch");
            command.append("--");
            if ((masterRepository != null) && !masterRepository.isEmpty()) command.append(masterRepository);
            exec = new Exec(rootPath,command);

            exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              if (Settings.hgUpdateWithFetch && Settings.hgSafeUpdate)
              {
                storedFiles.restore(); storedFiles = null;
              }
              throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
            }

            exec.done(); exec = null;

            command.clear();
            command.append(Settings.hgCommand,"fpush");
            command.append("--");
            if ((masterRepository != null) && !masterRepository.isEmpty()) command.append(masterRepository);
            exec = new Exec(rootPath,command);

            exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              if (Settings.hgUpdateWithFetch && Settings.hgSafeUpdate)
              {
                storedFiles.restore(); storedFiles = null;
              }
              throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
            }

            exec.done(); exec = null;

            if (Settings.hgSafeUpdate)
            {
              // do 'safe' update when fetch is used: restore local changes and merge if needed
              if (!storedChanges.restore())
              {
                throw new RepositoryException("restore local changes fail");
              }

              // discard changes patch
              storedChanges.discard(); storedChanges = null;
            }

            // discard files backup
            storedFiles.discard(); storedFiles = null;
          }
          finally
          {
            if (storedChanges != null) storedChanges.discard();
            if (storedFiles != null) storedFiles.restore();
          }
        }
        else if (Settings.hgUseForestExtension)
        {
          command.clear();
          command.append(Settings.hgCommand,"fupdate");
          command.append("--");
          exec = new Exec(rootPath,command);

          exitCode = exec.waitFor();
          if (exitCode != 0)
          {
            throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
          }

          exec.done(); exec = null;
        }
        else
        {
          command.clear();
          command.append(Settings.hgCommand,"update");
          command.append("--");
          exec = new Exec(rootPath,command);

          exitCode = exec.waitFor();
          if (exitCode != 0)
          {
            throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
          }

          exec.done(); exec = null;
        }
      }

      if (Settings.hgUseQueueExtension)
      {
        // apply patches for all trees in forest/single repository
        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          for (String tree : trees)
          {
            command.clear();
            command.append(Settings.hgCommand,"qpush","-a");
            command.append("--");
            exec = new Exec(tree,command);

            exitCode = exec.waitFor();
            if (exitCode != 0)
            {
              throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
            }

            exec.done(); exec = null;
          }
        }
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** commit files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void commit(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // commit files
      command.clear();
      command.append(Settings.hgCommand,"commit");
      if (!commitMessage.isEmpty())
      {
        if (Settings.hgSingleLineCommitMessages)
        {
          command.append("-m",commitMessage.getMessage(", "));
        }
        else
        {
          command.append("-l",commitMessage.getFileName());
        }
      }
      else
      {
        command.append("-m","empty");
      }
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate push changes when configured
      if (Settings.hgImmediatePush)
      {
        pushChanges(masterRepository);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** add files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary files, false otherwise
   */
  @Override
  public void add(Set<FileData> fileDataSet, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // add files
      command.clear();
      command.append(Settings.hgCommand,"add");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        commit(fileDataSet,commitMessage);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** remove files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void remove(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      // delete local files
      for (FileData fileData : fileDataSet)
      {
        fileData.getFile().delete();
      }

      // remove from repository
      Command command = new Command();
      int     exitCode;

      // remove files
      command.clear();
      command.append(Settings.hgCommand,"remove");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        commit(fileDataSet,commitMessage);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** copy files
   * @param fileDataSet files to copy
   * @param destination destination
   * @param commitMessage commit message
   */
  @Override
  public void copy(Set<FileData> fileDataSet, String destination, CommitMessage commitMessage)
    throws RepositoryException
  {
Dprintf.dprintf("TODO");
  }

  /** rename file
   * @param fileData file data to rename
   * @param newName new name
   * @param commitMessage commit message
   */
  @Override
  public void rename(FileData fileData, String newName, CommitMessage commitMessage)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // rename file
      command.clear();
      command.append(Settings.hgCommand,"rename");
      command.append("--");
      command.append(fileData.getName());
      command.append(newName);
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        Set<FileData> fileDataSet = FileData.toSet(getFileDataName(fileData),
                                                   (!rootPath.isEmpty()) ? rootPath+File.separator+newName : newName
                                                  );
//        HashSet<FileData> fileDataSet = new HashSet<FileData>();
//        fileDataSet.add(fileData);
//        fileDataSet.add(new FileData(newName));
        commit(fileDataSet,commitMessage);
      }

      // set new name
      fileData.setFileName(newName);
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** revert files
   * @param fileDataSet file data set or null for all files
   * @param revisionId revision to revert to
   * @param recursive true for recursive revert, false otherwise
   */
  @Override
  public void revert(Set<FileData> fileDataSet, String revisionId, boolean recursive)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // revert files
      command.clear();
      command.append(Settings.hgCommand,"revert","--no-backup");
      // Note: hg revert to last pulled repository version when "tip" is given, not the local version
      if (!revisionId.equals(LAST_REVISION_NAME)) command.append(Settings.hgCommand,"-r",revisionId);
      if (fileDataSet != null)
      {
        command.append("--");
        command.append(getFileDataNames(fileDataSet));
      }
      else
      {
        command.append("-a");
        command.append("--");
      }
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** set conflicts resolved
   * @param fileDataSet file data set or null for all files
   */
  @Override
  public void resolve(Set<FileData> fileDataSet)
    throws RepositoryException
  {
Dprintf.dprintf("");
  }

  /** merge other branch into current branch
   * @param otherBranchName other branch name
   */
  @Override
  public void mergeBranch(String otherBranchName)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // merge branch
      command.clear();
      command.append(Settings.hgCommand,"merge",otherBranchName);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** merge commit into current branch
   * @param revision revision
   */
  @Override
  public void mergeCommit(String revision)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // merge branch
      command.clear();
      command.append(Settings.hgCommand,"graft","--rev",revision);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** get incoming changes list
   * @param incoming changes list
   */
  @Override
  public List<LogData> getIncomingChanges(String masterRepository)
    throws RepositoryException
  {
    final Pattern PATTERN_CHANGE = Pattern.compile("^(\\d+)\\s+(\\S+)\\s+(\\S+\\s+\\S+\\s+\\S+)\\s+(.*)\\s*",Pattern.CASE_INSENSITIVE);

    ArrayList<LogData> logDataList = new ArrayList<LogData>();

    Exec exec = null;
    try
    {
      Command            command           = new Command();
      String             line;
      Matcher            matcher;
      String             revision          = null;
      String             changeSetId       = null;
      Date               date              = null;
      String             author            = null;
      LinkedList<String> commitMessageList = new LinkedList<String>();
      ArrayList<String>  fileNameList      = new ArrayList<String>();
      int                exitCode;

      // get log
      command.clear();
      command.append(Settings.hgCommand,"-y","-v","incoming","--template","{rev} {node|short} {date|isodate} {author|person}\\n{desc}\\n-----\\n{files}\\n-----\\n");
      appendURI(command);
      command.append("--");
      if ((masterRepository != null) && !masterRepository.isEmpty()) command.append(masterRepository);
      exec = new Exec(rootPath,command);

      // parse revisions in log output
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        if      ((matcher = PATTERN_CHANGE.matcher(line)).matches())
        {
          // get revision, change-set id, date, author
          revision    = matcher.group(1);
          changeSetId = matcher.group(2);
          date        = parseDate(matcher.group(3));
          author      = matcher.group(4);

          // get commit message
          commitMessageList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            commitMessageList.add(line);
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekFirst().trim().isEmpty())
          {
            commitMessageList.removeFirst();
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekLast().trim().isEmpty())
          {
            commitMessageList.removeLast();
          }

          // get file names
          fileNameList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            fileNameList.add(line);
          }

          // add revision data entry
          logDataList.add(new LogData(revision,
                                      date,
                                      author,
                                      commitMessageList,
                                      fileNameList
                                     )
                         );
        }
      }

      // check exitcode (Note: exitcode 1 is OK and mean "no incoming changes")
      exitCode = exec.waitFor();
      if (exitCode > 1)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return logDataList;
  }

  /** get outgoing changes list
   * @param outgoing changes list
   */
  @Override
  public List<LogData> getOutgoingChanges(String masterRepository)
    throws RepositoryException
  {
    final Pattern PATTERN_CHANGE = Pattern.compile("^(\\d+)\\s+(\\S+)\\s+(\\S+\\s+\\S+\\s+\\S+)\\s+(.*)\\s*",Pattern.CASE_INSENSITIVE);

    ArrayList<LogData> logDataList = new ArrayList<LogData>();

    Exec exec = null;
    try
    {
      Command            command           = new Command();
      String             line;
      Matcher            matcher;
      String             revision          = null;
      String             changeSetId       = null;
      Date               date              = null;
      String             author            = null;
      LinkedList<String> commitMessageList = new LinkedList<String>();
      ArrayList<String>  fileNameList      = new ArrayList<String>();
      int                exitCode;

      // get log
      command.clear();
      command.append(Settings.hgCommand,"-y","-v","outgoing","--template","{rev} {node|short} {date|isodate} {author|person}\\\n{desc}\\n-----\n{files}\\n-----\\n");
      appendURI(command);
      command.append("--");
      if ((masterRepository != null) && !masterRepository.isEmpty()) command.append(masterRepository);
      exec = new Exec(rootPath,command);

      // parse revisions in log output
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        if      ((matcher = PATTERN_CHANGE.matcher(line)).matches())
        {
          // get revision, change-set id, date, author
          revision    = matcher.group(1);
          changeSetId = matcher.group(2);
          date        = parseDate(matcher.group(3));
          author      = matcher.group(4);

          // get commit message
          commitMessageList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            commitMessageList.add(line);
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekFirst().trim().isEmpty())
          {
            commitMessageList.removeFirst();
          }
          while (!commitMessageList.isEmpty() && commitMessageList.peekLast().trim().isEmpty())
          {
            commitMessageList.removeLast();
          }

          // get file names
          fileNameList.clear();
          while (  ((line = exec.getStdout()) != null)
                 && !line.startsWith("-----")
                )
          {
            fileNameList.add(line);
          }

          // add revision data entry
          logDataList.add(new LogData(revision,
                                      date,
                                      author,
                                      commitMessageList,
                                      fileNameList
                                     )
                         );
        }
      }

      // check exitcode (Note: exitcode 1 is OK and mean "no outing changes")
      exitCode = exec.waitFor();
      if (exitCode > 1)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return logDataList;
  }

  /** get incoming/outgoing changes lines
   * @param revisionId revision to get changes lines for
   */
  @Override
  public String[] getChanges(String revisionId)
    throws RepositoryException
  {
    List<String> lineList = new ArrayList<String>();

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;

      // get file
      command.clear();
      command.append(Settings.hgCommand,"-y","export","-r",revisionId);
      command.append("--");
      exec = new Exec(rootPath,command);

      // read file data
      while ((line = exec.getStdout()) != null)
      {
        lineList.add(line);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return lineList.toArray(new String[lineList.size()]);
  }

  /** pull changes
   * @param masterRepositoryURL master repository URL
   * @param moduleName module name
   * @param userName user login name or ""
   */
  @Override
  public void pullChanges(String masterRepositoryURL, String moduleName, String userName)
    throws RepositoryException
  {
Dprintf.dprintf("*********************** userName=%s",userName);
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      command.clear();
      command.append(Settings.hgCommand,"-y",Settings.hgUseForestExtension?"fpull":"pull");
      appendURI(command,userName,"/"+moduleName);
      command.append("--");
      if ((masterRepository != null) && !masterRepository.isEmpty()) command.append(masterRepository);
      exec = new Exec(rootPath,command);

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** push changes
   * @param masterRepositoryURL master repository URL
   * @param moduleName module name
   * @param userName user login name or ""
   */
  @Override
  public void pushChanges(String masterRepositoryURL, String moduleName, String userName)
    throws RepositoryException
  {
    final Pattern PATTERN_ABORT = Pattern.compile("^\\s*abort:.*",Pattern.CASE_INSENSITIVE);

    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;
      String  line;
      Matcher matcher;

      command.clear();
      command.append(Settings.hgCommand,"-y",Settings.hgUseForestExtension?"fpush":"push");
      appendURI(command,userName,"/"+moduleName);
      command.append("--");
      if ((masterRepository != null) && !masterRepository.isEmpty()) command.append(masterRepository);
      exec = new Exec(rootPath,command);

      // bug in hg: hg does not report an exitcode != 0 if something go wrong => check error output
      exitCode = 0;
      while ((line = exec.getStderr()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        if ((matcher = PATTERN_ABORT.matcher(line)).matches())
        {
          exitCode = 1;
        }
      }
      if (exitCode == 0)
      {
        exitCode = exec.waitFor();
      }
      else
      {
        exec.waitFor();
      }
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** apply patches
   */
  @Override
  public void applyPatches()
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      String[] trees;
      if (Settings.hgUseForestExtension)
      {
        // get trees of forest (if forest extension is enabled)
        trees = getTrees();
      }
      else
      {
        // only single tree
        trees = new String[]{rootPath};
      }

      // unapply patches for all trees in forest/single repository
      for (String tree : trees)
      {
        command.clear();
        command.append(Settings.hgCommand,"qpush","-a");
        command.append("--");
        exec = new Exec(tree,command);

        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }

        exec.done(); exec = null;
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** unapply patches
   */
  @Override
  public void unapplyPatches()
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      String[] trees;
      if (Settings.hgUseForestExtension)
      {
        // get trees of forest (if forest extension is enabled)
        trees = getTrees();
      }
      else
      {
        // only single tree
        trees = new String[]{rootPath};
      }

      // unapply patches for all trees in forest/single repository
      for (String tree : trees)
      {
        command.clear();
        command.append(Settings.hgCommand,"qpop","-a");
        command.append("--");
        exec = new Exec(tree,command);

        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }

        exec.done(); exec = null;
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** lock files
   * @param fileDataSet file data set
   */
  @Override
  public void lock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    try
    {
      Command command = new Command();
      int     exitCode;

      // copy file
      command.clear();
      command.append(Settings.hgCommand,"lock");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exitCode = new Exec(rootPath,command).waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
  }

  /** unlock files
   * @param fileDataSet file data set
   */
  @Override
  public void unlock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    try
    {
      Command command = new Command();
      int     exitCode;

      // copy file
      command.clear();
      command.append(Settings.hgCommand,"unlock");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exitCode = new Exec(rootPath,command).waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
  }

  /** set commit message
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void setCommitMessage(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** set files mode
   * @param fileDataSet file data set
   * @param mode file mode
   * @param commitMessage commit message
   */
  @Override
  public void setFileMode(Set<FileData> fileDataSet, FileData.Modes mode, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** get default name of root
   * @return default root name
   */
  @Override
  public String getDefaultRootName()
  {
    return DEFAULT_ROOT_NAME;
  }

  /** get default branch/tag name
   * @return default branch/tag name
   */
  @Override
  public String getDefaultBranchTagName()
  {
    return DEFAULT_BRANCHES_NAME;
  }

  /** get names of existing branches/tags
   * @param pathName path name
   * @return array with branch/tag names (always empty)
   */
  @Override
  public List<String> getBranchTagNames(String pathName)
    throws RepositoryException
  {
    return new ArrayList<String>();
  }

  /** create new branch
   * @param rootName root name (source)
   * @param branchTagName branch name
   * @param fromRevision from revision name or null
   * @param commitMessage commit message
   * @param buysDialog busy dialog or null
   */
  @Override
  public void createBranch(String rootName, String branchTagName, String fromRevision, CommitMessage commitMessage, BusyDialog busyDialog)
    throws RepositoryException
  {
    String repositoryURL = getRepositoryURL();

    Exec exec = null;
    try
    {
      Command command = new Command();

      // create branch
      command.clear();
      command.append(Settings.hgCommand,"-y","-v",(Settings.hgUseForestExtension) ? "fclone" : "clone");
      if (!fromRevision.isEmpty()) command.append("-r",fromRevision);
      command.append((!rootName.isEmpty()) ? repositoryURL+"/"+rootName : repositoryURL,repositoryURL+"/"+branchTagName);
      exec = new Exec(rootPath,command);

      // read output
      String line;
      while (   ((busyDialog == null) || !busyDialog.isAborted())
             && ((line = exec.getStdout()) != null)
            )
      {
//Dprintf.dprintf("line=%s",line);
        if (busyDialog != null) busyDialog.updateText(line);
      }
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",command.toString(),exitCode);
        }
      }
      else
      {
        exec.destroy();
      }

      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** post to review server/update review server
   */
  protected String postReview(String password, String reference, Set<FileData> fileDataSet, CommitMessage commitMessage, LinkedHashSet<String> testSet)
    throws RepositoryException
  {
    final Pattern PATTERN_REFERENCE = Pattern.compile("^review\\s+.*:\\s+.*/r/(\\d+.)/",Pattern.CASE_INSENSITIVE);

//    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;
      int     exitCode;

      // post review for files
      command.clear();
      command.append(Settings.hgCommand,"lpostreview");
      if (reviewServerHost != null) command.append("--server",reviewServerHost);
      if (reviewServerLogin != null) command.append("--username",reviewServerLogin);
      if (password != null) command.append("--password",command.hidden(password));
      if (reviewServerRepository != null) command.append("--repoid",reviewServerRepository);
      if (outgoingRepository != null) command.append("--outgoingrepo",outgoingRepository);
      if (!reference.isEmpty()) command.append("--existing",reference);
      command.append("--publish");
      command.append("--summary",commitMessage.getSummary());
      command.append("--description",commitMessage.getMessage());
      command.append("--tests",StringUtils.join(testSet,", "));
      if (reviewServerGroups != null) command.append("--target_groups",reviewServerGroups);
      if (reviewServerPersons != null) command.append("--target_people",reviewServerPersons);
//command.append("--debug");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));

      // execute
      exec = new Exec(rootPath,command);
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("stdout %s",line);
        if ((matcher = PATTERN_REFERENCE.matcher(line)).matches())
        {
          reference = matcher.group(1);
        }
      }
      while ((line = exec.getStderr()) != null)
      {
//Dprintf.dprintf("stderr %s",line);
      }
      exitCode = exec.done(); exec = null;
      if (exitCode != 0)
      {
        throw new RepositoryException("Cannot post patch to review server (exitcode: "+exitCode+")");
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return reference;
  }

  //-----------------------------------------------------------------------

  /** append URI with username and password to command
   * @param command command
   * @param userName user name or null
   * @param suffix suffix string
   */
  private void appendURI(Command command, String userName, String suffix)
    throws RepositoryException
  {
    final Pattern PATTERN_URI1 = Pattern.compile("^([^:/]+)://([^:]*):([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI2 = Pattern.compile("^([^:/]+)://([^@]*)@(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_URI3 = Pattern.compile("^([^:/]+)://(.*)",Pattern.CASE_INSENSITIVE);

Dprintf.dprintf("userName=%s, %s",userName,this.userName);
Dprintf.dprintf("passwordHandler=%s",passwordHandler);
    if ((userName == null) || userName.isEmpty()) userName = this.userName;
    if ((userName == null) || userName.isEmpty()) userName = System.getProperty("user.name");
    String password = ((userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,repositoryURL) : null;

    String repositoryURL = getRepositoryURL();

    Matcher matcher;
    if      ((matcher = PATTERN_URI1.matcher(repositoryURL)).matches())
    {
      command.appendConcat(matcher.group(1),
                           "://",
                           matcher.group(2),
                           ":",
                           command.hidden(matcher.group(3)),
                           "@",
                           matcher.group(4),
                           suffix
                          );
    }
    else if ((matcher = PATTERN_URI2.matcher(repositoryURL)).matches())
    {
      if ((password == null) || password.isEmpty())
      {
        if ((userName == null) || userName.isEmpty())
        {
          throw new RepositoryException("no user name given");
        }
        else
        {
          throw new RepositoryException("no password given");
        }
      }
      command.appendConcat(matcher.group(1),
                           "://",
                           matcher.group(2),
                           ":",
                           command.hidden(password),
                           "@",
                           matcher.group(3),
                           suffix
                          );
    }
    else if ((matcher = PATTERN_URI3.matcher(repositoryURL)).matches())
    {
      if ((password == null) || password.isEmpty())
      {
        if ((userName == null) || userName.isEmpty())
        {
          throw new RepositoryException("no user name given");
        }
        else
        {
          throw new RepositoryException("no password given");
        }
      }
      command.appendConcat(matcher.group(1),
                           "://",
                           userName,
                           ":",
                           command.hidden(password),
                           "@",
                           matcher.group(2),
                           suffix
                          );
    }
    else
    {
      command.appendConcat(repositoryURL,
                           suffix
                          );
    }
  }

  /** append URI with username and password to command
   * @param command command
   */
  private void appendURI(Command command)
    throws RepositoryException
  {
    appendURI(command,null,"");
  }

  /** parse hg state string
   * @param string state string
   * @return state
   */
  private FileData.States parseState(String string)
  {
    if      (string.equalsIgnoreCase(" ")) return FileData.States.OK;
    else if (string.equalsIgnoreCase("C")) return FileData.States.OK;
    else if (string.equalsIgnoreCase("A")) return FileData.States.ADDED;
    else if (string.equalsIgnoreCase("R")) return FileData.States.REMOVED;
    else if (string.equalsIgnoreCase("M")) return FileData.States.MODIFIED;
    else if (string.equalsIgnoreCase("!")) return FileData.States.CHECKOUT;
    else if (string.equalsIgnoreCase("I")) return FileData.States.UNKNOWN;
    else if (string.equalsIgnoreCase("?")) return FileData.States.UNKNOWN;
    else                                   return FileData.States.UNKNOWN;
  }

  /** get HG root path
   * @return root path
   */
  private String getRootPath()
  {
    String hgRootPath = "";

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;

      // get forest trees
      command.clear();
      command.append(Settings.hgCommand,"root");
      command.append("--");
      exec = new Exec(rootPath,command);

      // read
      hgRootPath = exec.getStdout();

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return hgRootPath;
  }

  /** get total commit count
   * @return total commit count
   */
  private int getTotalCommitCount()
    throws RepositoryException
  {
    final Pattern PATTERN_COUNT = Pattern.compile("^(\\d+).*",Pattern.CASE_INSENSITIVE);

    int commitCount = 0;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      command.clear();
      command.append(Settings.hgCommand,"id","--num");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();
//Dprintf.dprintf("command=%s",command);

      // parse
      if (   ((line = exec.getStdout()) != null)
          && (matcher = PATTERN_COUNT.matcher(line)).matches()
         )
      {
        commitCount = Integer.parseInt(matcher.group(1));
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return commitCount;
  }

  /** get sub-directory based on hg root path
   * @return sub-directory or ""
   */
  private String getSubDirectory()
    throws IOException,RepositoryException
  {
    // get hg root path
    String hgRootPath = getRootPath();

    // get sub-directory
    return ((rootPath.length() > hgRootPath.length()) && rootPath.startsWith(hgRootPath))
             ?rootPath.substring(hgRootPath.length()+1)
             :"";
  }

  /** get trees of hg forest
   * @return tree names
   */
  private String[] getTrees()
    throws IOException,RepositoryException
  {
    List<String> treeList = new ArrayList<String>();

    Command command = new Command();
    Exec    exec;
    String  line;

    // get forest trees
    command.clear();
    command.append(Settings.hgCommand,"ftree");
    command.append("--");
    exec = new Exec(rootPath,command);

    // parse
    while ((line = exec.getStdout()) != null)
    {
      treeList.add(line);
    }

    // done
    exec.done(); exec = null;

    return treeList.toArray(new String[treeList.size()]);
  }

// NYI: is the global parent map needed?
  private HashMap<Integer,ParentData> getParentMap()
  {
    HashMap<Integer,ParentData> parentMap = new HashMap<Integer,ParentData>();

    try
    {
      Command command = new Command();
      Exec    exec;
      String  line;

      // get parents
      command.clear();
      command.append(Settings.hgCommand,"-y","log","--follow","--template","{rev} {parents}\\n");
      command.append("--");
      exec = new Exec(rootPath,command);

      // parse
      Object[] data = new Object[5];
      while ((line = exec.getStdout()) != null)
      {
        if      (StringParser.parse(line,"%d %d:%s %d:%s",data))
        {
          // node with two parents: merge
          int revision        = (Integer)data[0];
          int parentRevision1 = (Integer)data[1];
          int parentRevision2 = (Integer)data[3];

          parentMap.put(revision,new ParentData(parentRevision1,parentRevision2));
        }
        else if (StringParser.parse(line,"%d %d:%s",data))
        {
// NYI: what is a single parent? a commit?
          // node with single parent: commit?
          int revision        = (Integer)data[0];
          int parentRevision1 = (Integer)data[1];

          parentMap.put(revision,new ParentData(parentRevision1));
        }
        else
        {
          // nodes with no parents are ignored
        }
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      // ignored
    }

    return parentMap;
  }

// get parent revision data
  private RevisionData getParentRevisionData(int revision, HashMap<Integer,RevisionData> revisionDataMap, HashMap<RevisionData,ParentData> fileParentMap)
  {
    RevisionData revisionData = null;

    if ((revision != 0) && (parentMap != null))
    {
      synchronized(parentMap)
      {
      Stack<Integer> revisionStack = new Stack<Integer>();
      do
      {
        // check if revision exists
        revisionData = revisionDataMap.get(revision);

        if (revisionData == null)
        {
          // push parents for check
          ParentData parentData = fileParentMap.get(revision);
          if (parentData != null)
          {
            if (parentData.revisionId2 != 0) revisionStack.push(parentData.revisionId2);

            // get next revision to check
            revision = parentData.revisionId1;
          }
          else
          {
            revision = revision-1;
          }
        }
      }
      while ((revisionData == null) && (revision != 0));
      }
    }

    return revisionData;
  }

  /** parse HG diff index
   * @param string diff index string
   * @return index array [lineNb,length] or null
   */
  private int[] parseDiffIndex(String string)
  {
    Object[] data = new Object[2];
    if      (StringParser.parse(string,"%d,%d",data))
    {
      return new int[]{(Integer)data[0],(Integer)data[1]-(Integer)data[0]-1};
    }
    else if (StringParser.parse(string,"%d",data))
    {
      return new int[]{(Integer)data[0],1};
    }
    else
    {
      return null;
    }
  }

  /** parse log header (currently nothing to do)
   * @param exec exec command
   * @return true iff header parsed
   */
  private boolean parseLogHeader(Exec exec)
    throws IOException
  {
    return true;
  }

  /** parse next log revision data
   * @param exec exec command
   * @return revision data info or null
   */
  private RevisionDataHG parseNextRevisionData(Exec exec)
    throws IOException
  {
    final Pattern PATTERN_REVISION      = Pattern.compile("^(\\d+)\\s+(\\S+)\\s+(\\S+\\s+\\S+\\s+\\S+)\\s+(.*)\\s*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_MERGE_PARENTS = Pattern.compile("^(\\d+):\\S+\\s+(\\d+):\\S+\\s*$",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_BRANCH_PARENT = Pattern.compile("^(\\d+):\\S+\\s*$",Pattern.CASE_INSENSITIVE);

    RevisionDataHG     revisionData      = null;

    boolean            dataDone          = false;
    Matcher            matcher;
    String             line;
    String             revisionId        = null;
    Date               date              = null;
    String             author            = null;
    int                revision          = 0;
    String             changeSetId       = null;
    String             branchName        = null;
    int                parentRevision1   = 0;
    int                parentRevision2   = 0;
    TagList            tagList           = new TagList();
    LinkedList<String> commitMessageList = new LinkedList<String>();
    ArrayList<String>  fileNameList      = new ArrayList<String>();
    while (   !dataDone
           && ((line = exec.getStdout()) != null)
          )
    {
//Dprintf.dprintf("line=%s",line);
      if      (line.startsWith("-----"))
      {
        dataDone = true;
      }
      else if ((matcher = PATTERN_REVISION.matcher(line)).matches())
      {
        // get revision, change-set id, date, author
        revisionId  = matcher.group(1);
        revision    = Integer.parseInt(revisionId);
        changeSetId = matcher.group(2);
        date        = parseDate(matcher.group(3));
        author      = matcher.group(4);

        // get parents
        if ((line = exec.getStdout()) != null)
        {
          if      ((matcher = PATTERN_MERGE_PARENTS.matcher(line)).matches())
          {
            // node with two parents: merge
            parentRevision1 = Integer.parseInt(matcher.group(1));
            parentRevision2 = Integer.parseInt(matcher.group(2));
          }
          else if ((matcher = PATTERN_BRANCH_PARENT.matcher(line)).matches())
          {
            // node with single parent: branch
            parentRevision1 = Integer.parseInt(matcher.group(1));
          }
        }

        // get branch name
        if ((line = exec.getStdout()) != null)
        {
          branchName = line;
        }

        // get tags
        tagList.clear();
        if ((line = exec.getStdout()) != null)
        {
          for (String tag : StringUtils.splitArray(line,StringUtils.WHITE_SPACES,false))
          {
            tagList.add(tag);
          }
        }

        // get commit message lines
        commitMessageList.clear();
        while (  ((line = exec.getStdout()) != null)
               && !line.startsWith("-----")
              )
        {
          commitMessageList.add(line);
        }
        while (!commitMessageList.isEmpty() && commitMessageList.peekFirst().trim().isEmpty())
        {
          commitMessageList.removeFirst();
        }
        while (!commitMessageList.isEmpty() && commitMessageList.peekLast().trim().isEmpty())
        {
          commitMessageList.removeLast();
        }

        // get file names
        fileNameList.clear();
//TODO

        // create log info entry
        revisionData = new RevisionDataHG(revisionId,
                                          revision,
                                          changeSetId,
                                          branchName,
                                          parentRevision1,
                                          parentRevision2,
                                          tagList,
                                          date,
                                          author,
                                          commitMessageList,
                                          fileNameList
                                         );

        dataDone = true;
      }
    }
//Dprintf.dprintf("revisionData=%s",revisionData);

    return revisionData;
  }

  /** create revision tree
   * @param revisionDataList revision data list
   * @param branchesMap branches
   * @param branchRevisionNumbers branch revision numbers
   * @param busyDialog busy dialog or null
   * @return revision tree
   */
  private RevisionDataTree createRevisionDataTree(LinkedList<RevisionDataHG> revisionDataList,
                                                  BusyDialog                 busyDialog
                                                 )
  {
    RevisionDataTree revisionDataTree = new RevisionDataTree();

    HashMap<Integer,RevisionDataTree> revisionDataTreeMap       = new HashMap<Integer,RevisionDataTree>();
    HashMap<String,RevisionDataTree>  branchRevisionDataTreeMap = new HashMap<String,RevisionDataTree>();
    branchRevisionDataTreeMap.put("default",revisionDataTree);
    HashMap<Integer,RevisionData>     revisionDataMap           = new HashMap<Integer,RevisionData>();

    for (RevisionDataHG revisionData : revisionDataList)
    {
      if      ((revisionData.parentRevision1 != 0) && (revisionData.parentRevision2 != 0))
      {
        // merge
//Dprintf.dprintf("--- merge %d: %d+%d -> %s",revisionData.revision,revisionData.parentRevision1,revisionData.parentRevision2,revisionData.branchName);
        RevisionData mergeRevisionDataFrom = revisionDataMap.get(revisionData.parentRevision1);
        RevisionData mergeRevisionDataTo   = revisionDataMap.get(revisionData.parentRevision2);
        if ((mergeRevisionDataFrom != null) && (mergeRevisionDataTo != null))
        {
          revisionData.parentRevisionData0 = mergeRevisionDataFrom;
          revisionData.parentRevisionData1 = mergeRevisionDataTo;
          mergeRevisionDataFrom.mergeDataList.add(revisionData);
          mergeRevisionDataTo.mergeDataList.add(revisionData);
        }

        RevisionDataTree branchRevisionDataTree = branchRevisionDataTreeMap.get(revisionData.branchName);
if (revisionData.revision==150)
//if (revisionData.revision==49)
{
//Dprintf.dprintf("%d -> prev %s/%s",revisionData.revision,branchRevisionDataTree.peekLast());
//branchRevisionDataTree.print(500);
//System.exit(1);
}
        if (branchRevisionDataTree != null)
        {
//Dprintf.dprintf("first/last on merge destination %s/%s",branchRevisionDataTree.getFirst().revisionId,branchRevisionDataTree.peekLast().revisionId);
          branchRevisionDataTree.add(revisionData);
          revisionDataTreeMap.put(revisionData.revision,branchRevisionDataTree);
        }
        else
        {
          // no branch found? -> add revision at end, skip
          revisionDataTree.add(revisionData);
          revisionDataTreeMap.put(revisionData.revision,revisionDataTree);
          if (busyDialog != null) busyDialog.updateList("Branch not found for %s",revisionData.branchName);
        }
      }
      else if (revisionData.parentRevision1 != 0)
      {
        assert revisionData.parentRevision2 == 0;

        // branch
//Dprintf.dprintf("---- branch %d: %d = %s",revisionData.revision,revisionData.parentRevision1,revisionData.branchName);
        RevisionData branchRootRevisionData = revisionDataMap.get(revisionData.parentRevision1);
        if (branchRootRevisionData == null) branchRootRevisionData = revisionDataTree.peekLast();
        if (branchRootRevisionData != null)
        {
          BranchData branchData = branchRootRevisionData.addBranch(revisionData.branchName,
                                                                   revisionData.date,
                                                                   revisionData.author,
                                                                   revisionData.commitMessageList,
                                                                   revisionData
                                                                  );
          revisionDataTreeMap.put(revisionData.revision,branchData.revisionDataTree);
          if (!branchRevisionDataTreeMap.containsKey(revisionData.branchName))
          {
            branchRevisionDataTreeMap.put(revisionData.branchName,branchData.revisionDataTree);
          }
        }
        else
        {
          // no branch root found? -> add revision at end, skip
          revisionDataTree.add(revisionData);
          revisionDataTreeMap.put(revisionData.revision,revisionDataTree);
          if (busyDialog != null) busyDialog.updateList("Branch not found for %s",revisionData.parentRevision1);
        }
      }
      else
      {
        assert revisionData.parentRevision1 == 0;
        assert revisionData.parentRevision2 == 0;

        // commit -> find branch and add revision
        if (revisionData.revision > 0)
        {
          RevisionDataTree branchRevisionDataTree = revisionDataTreeMap.get(revisionData.revision-1);
//if (revisionData.revision==47) Dprintf.dprintf("%d -> prev %s",revisionData.revision,branchRevisionDataTree);
          if (branchRevisionDataTree == null) branchRevisionDataTree = revisionDataTree;
          if (branchRevisionDataTree != null)
          {
            branchRevisionDataTree.add(revisionData);
//Dprintf.dprintf("1 revisionData.revision=%s parent=%s: %s",revisionData.revision,revisionData.revision-1,branchRevisionDataTree.get(0));
            revisionDataTreeMap.put(revisionData.revision,branchRevisionDataTree);
          }
          else
          {
            // no branch found? -> add revision at end, skip
            revisionDataTree.add(revisionData);
            revisionDataTreeMap.put(revisionData.revision,revisionDataTree);
            if (busyDialog != null) busyDialog.updateList("Branch not found for %d",revisionData.revision-1);
          }
        }
        else
        {
          revisionDataTree.add(revisionData);
          revisionDataTreeMap.put(revisionData.revision,revisionDataTree);
        }
      }

      // store revision into map
      revisionDataMap.put(revisionData.revision,revisionData);
    }

    return revisionDataTree;
  }

  /** rollback command
   * @param specialCommand special command
   * @param shell shell
   */
  private void rollback(SpecialCommand specialCommand, Shell shell)
    throws RepositoryException
  {
    final Pattern PATTERN_URI = Pattern.compile("^[^:/]+://.*",Pattern.CASE_INSENSITIVE);

    Exec exec = null;
    try
    {
      Command command = new Command();

      // rollback
      command.append(Settings.hgCommand,"rollback");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }
}

/* end of file */
