/***********************************************************************\
*
* $Revision: 2169 $
* $Date: 2017-09-28 09:10:06 +0200 (Thu, 28 Sep 2017) $
* $Author: torsten $
* Contents: Apache Subversion (SVN) repository
* Systems: all
*
\***********************************************************************/

/****************************** Imports ********************************/
// base
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

import java.text.DateFormat;

import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.util.Set;
import java.util.TreeMap;

import javax.xml.bind.annotation.XmlAccessorType;
import javax.xml.bind.annotation.XmlAccessType;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlType;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

// graphics
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Shell;

/****************************** Classes ********************************/

/** repository value adapter for version
 */
class RepositoryValueAdapterVersion extends RepositoryValueAdapter
{
  /** create widget
   * @param composite composite
   * @param value value
   * @return widget
   */
  public Control createWidget(Composite composite, String value)
  {
    Combo combo;

    combo = Widgets.newOptionMenu(composite);
    combo.setItems(Settings.svnCommands.names());
    if (value != null) combo.setText(value);

    return combo;
  }

  /** get value from widget
   * @param control control
   * @return value
   */
  public String get(Control control)
  {
    Combo combo = (Combo)control;

    return combo.getText();
  }
}

/** Apache Subversion (SVN) repository
 */
@XmlType(propOrder={"userName","logPrefix","version"})
@XmlAccessorType(XmlAccessType.NONE)
class RepositorySVN extends Repository
{
  /** SVN revision data
   */
  class RevisionDataSVN extends RevisionData
  {
    String branchName;
    String copyFromPath;
    String copyFromRevision;

    /** create CVS revision data
     * @param revisionId revision id string
     * @param branchName branch name
     * @param date date
     * @param author author name
     * @param copyFromPath
     * @param copyFromRevision
     * @param commitMessageList commit message lines
     * @param fileNameList names of changed files
     */
    RevisionDataSVN(String             revisionId,
                    String             branchName,
                    Date               date,
                    String             author,
                    String             copyFromPath,
                    String             copyFromRevision,
                    LinkedList<String> commitMessageList,
                    ArrayList<String>  fileNameList
                   )
    {
      super(revisionId,(String)null,date,author,commitMessageList,fileNameList);
      this.branchName       = branchName;
      this.copyFromPath     = copyFromPath;
      this.copyFromRevision = copyFromRevision;
    }

    /** convert data to string
     * @return string
     */
    public String toString()
    {
      return "SVN revision data {"+revisionId+", date: "+date+", author: "+author+", message: "+StringUtils.join(commitMessageList,", ")+"}";
    }
  }

  // --------------------------- constants --------------------------------
  public final static String[] DEFAULT_REVISION_NAMES = new String[]{"HEAD"};
  public final static String[] DEFAULT_BRANCH_NAMES   = new String[]{"trunk","tags","branches"};
  public final static String   DEFAULT_ROOT_NAME      = "trunk";
  public final static String   DEFAULT_BRANCH_NAME    = "branches";
  public final static String   FIRST_REVISION_NAME    = "1";
  public final static String   LAST_REVISION_NAME     = "HEAD";

  // --------------------------- variables --------------------------------
  private final static RepositorySVN staticInstance = new RepositorySVN();

  @XmlElement(name = "userName")
  @RepositoryValue(title = "User name:", tooltip="SVN server user login name.")
  public String userName = "";

  @XmlElement(name = "logPrefix")
  @RepositoryValue(title = "Log prefix:", defaultValue="/trunk", tooltip="Prefix of repository log path.")
  public String logPrefix;

  @XmlElement(name = "version")
  @RepositoryValue(title = "Version:", type=RepositoryValueAdapterVersion.class, tooltip="Repsitory version.")
  public String version = "";

  private String repositoryURL = null;
  private String repositoryRootURL = null;

  // ------------------------ native functions ----------------------------

  // ---------------------------- methods ---------------------------------

  /** get static instance
   * @return static instance
   */
  public final static RepositorySVN getInstance()
  {
Dprintf.dprintf("xxx=%s",Settings.svnCommands.get(staticInstance.version));
    return staticInstance;
  }

  /** create repository
   * @param rootPath root path
   */
  RepositorySVN(String rootPath, String userName, PasswordHandler passwordHandler, String comment)
  {
    super(rootPath,passwordHandler,comment);

    assert !rootPath.isEmpty();

    this.userName = userName;
  }

  /** create repository
   */
  RepositorySVN()
  {
    super();
  }

  /** verify settings
   * @return null or error message
   */
  public String verifySettings()
  {
    if (!Settings.svnCommands.contains(version))
    {
      return String.format("Command for SVN version '%s' does not exists",version);
    }

    return null;
  }

  /** get repository type
   * @return repository type
   */
  @Override
  public Types getType()
  {
    return Types.SVN;
  }

  /** check if repository support change branch command
   * @return true iff change branch command is supported
   */
  @Override
  public boolean supportChangeBranch()
  {
    return true;
  }

  /** check if repository support merge branch command
   * @return true iff merge branch command is supported
   */
  @Override
  public boolean supportMergeBranch()
  {
    return true;
  }

  /** check if repository support edit commit message
   * @return true iff edit commit message is supported
   */
  public boolean supportEditCommit()
  {
    return true;
  }

  /** check if repository support lock/unlock
   * @return true iff lock/unlock is supported
   */
  @Override
  public boolean supportLockUnlock()
  {
    return true;
  }

  /** get special commands for repository
   * @return array with special commands
   */
  @Override
  public Special[] getSpecials()
  {
    return new SpecialCommand[]
    {
      new SpecialCommand()
      {
        @Override
        public String getTitle()
        {
          return "Cleanup";
        }

        @Override
        public boolean run(Shell shell)
          throws RepositoryException
        {
          cleanup(this,shell);

          return true;
        }
      }
    };
  }

  /** check if repository
   * @param repositoryURL respository server URL
   * @return repository file list
   */
  public static boolean isRepository(String rootPath)
  {
    boolean result = false;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      int     exitCode;

      command.clear();
      command.append(Settings.svnCommands.get(""),"info");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      result = (exec.waitFor() == 0);

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      // ignored
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return result;
  }

  /** list repository content
   * @param name name
   * @return repository file list
   */
  @Override
  public List<RepositoryFile> list(String name)
    throws RepositoryException
  {
    final Pattern PATTERN_URI = Pattern.compile("^[^:/]+://.*",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,rootPath) : null;

    ArrayList<RepositoryFile> repositoryFileList = new ArrayList<RepositoryFile>();

    Exec exec = null;
    try
    {
      Command command = new Command();

      // checkout
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("list");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append(rootPath+(!name.isEmpty() ? '/'+name : ""));
      exec = new Exec(command);
      exec.closeStdin();

      // read output
      while (!exec.eofStdout())
      {
        String line;

        // read stdout
        line = exec.pollStdout();
        if (line != null)
        {
          if (line.endsWith("/"))
          {
            // directory
            repositoryFileList.add(new RepositoryFile(rootPath,(!name.isEmpty() ? name+'/' : "")+line.substring(0,line.length()-1)));
          }
          else
          {
            // file
            repositoryFileList.add(new RepositoryFile(rootPath,(!name.isEmpty() ? name+'/' : "")+line));
          }
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // sort by name
    Collections.sort(repositoryFileList);

    return repositoryFileList;
  }

  /** create new repository module
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param importPath import directory
   */
  @Override
  public void create(String repositoryURL, String moduleName, String importPath)
    throws RepositoryException
  {
    final Pattern PATTERN_URI = Pattern.compile("^[^:/]+://.*",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // get full repository path
      String path = (PATTERN_URI.matcher(repositoryURL).matches()
                       ? repositoryURL
                       : "file://"+repositoryURL
                    )+"/"+moduleName;

      // create repository
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("import",importPath,path,"-m","initial");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** checkout repository
   * @param repositoryURL repository server URL
   * @param moduleName module name
   * @param revisionId revision to checkout
   * @param userName user login name or ""
   * @param destinationPath destination path
   * @param busyDialog busy dialog or null
   */
  @Override
  public void checkout(String repositoryURL, String moduleName, String revisionId, String userName, String destinationPath, BusyDialog busyDialog)
    throws RepositoryException
  {
    final Pattern PATTERN_URI = Pattern.compile("^[^:/]+://.*",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,repositoryURL) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // get full repository path
      String path = (PATTERN_URI.matcher(repositoryURL).matches()
                       ? repositoryURL
                       : "file://"+repositoryURL
                    )+"/"+moduleName;

      // checkout
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("checkout");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      if ((revisionId != null) && !revisionId.isEmpty()) command.append("--revision",revisionId);
      command.append(path,destinationPath);
      exec = new Exec(destinationPath,command);
      exec.closeStdin();

      // read output
      int n = destinationPath.length();
      while (   ((busyDialog == null) || !busyDialog.isAborted())
             && !exec.eofStdout()
            )
      {
        String line;

        // read stdout
        line = exec.pollStdout();
        if (line != null)
        {
//Dprintf.dprintf("out: %s",line);
          if (busyDialog != null) busyDialog.updateList(line);
        }
      }
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
        try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      try { FileUtils.deleteDirectoryTree(destinationPath); } catch (IOException ignoredException) { /* ignored */ }
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** update file states
   * @param fileDataSet file data set to update
   * @param directorySet relative directory name set to check for new/missing files
   * @param newFileDataSet new file data set or null
   */
  @Override
  public void updateStates(Set<FileData> fileDataSet, Set<String> directorySet, Set<FileData> newFileDataSet)
    throws RepositoryException
  {
    String branch;

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

//Dprintf.dprintf("fileDataSet=%s",fileDataSet);
//Dprintf.dprintf("directorySet=%s",directorySet);
    for (String directory : directorySet)
    {
      Exec exec = null;
      try
      {
        FileData.Types  type;
        String          name;
        FileData.States state              = FileData.States.UNKNOWN;
        boolean         locked             = false;
        long            size;
        Date            date;
        String          workingRevision    = "";
        String          repositoryRevision = "";

        Command         command            = new Command();
        String          line;
        Matcher         matcher;
        String          path;
        String          author;

        // get branch
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        command.append("info","--xml");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        branch = "";
        try
        {
          DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
          Document document = documentBuilder.parse(exec.getStdoutStream());
          document.getDocumentElement().normalize();

          NodeList relativeURLList    = document.getElementsByTagName("relative-url");
          Element  relativeURLElement = (relativeURLList.getLength() > 0) ? (Element)relativeURLList.item(0) : null;
          if (relativeURLElement != null)
          {
            branch = relativeURLElement.getFirstChild().getNodeValue();
            if (branch.startsWith("^/")) branch = branch.substring(2);
          }
        }
        catch (org.xml.sax.SAXException exception)
        {
          // ignored
        }
        catch (javax.xml.parsers.ParserConfigurationException exception)
        {
          // ignored
        }

        // get status
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
        command.append("status","-uvN","--xml");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        exec = new Exec(rootPath,directory,command);
        exec.closeStdin();
//Dprintf.dprintf("command=%s directory=%s",command,directory);

        try
        {
          DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
          Document document = documentBuilder.parse(exec.getStdoutStream());
          document.getDocumentElement().normalize();

          NodeList entryList = document.getElementsByTagName("entry");
          for (int z = 0; z < entryList.getLength(); z++)
          {
            Element  entryElement = (Element)entryList.item(z);
            NodeList nodeList;

            path = entryElement.getAttribute("path");
            if (!path.equals("."))
            {
              // get relative name
              name = (!directory.isEmpty() ? directory+File.separator : "")+StringUtils.replace(StringUtils.unescape(path),'/',File.separatorChar);

              // get state, working revision
              nodeList = entryElement.getElementsByTagName("wc-status");
              Element workingCopyStatusElement = (nodeList.getLength() > 0) ? (Element)nodeList.item(0) : null;
              nodeList = entryElement.getElementsByTagName("repos-status");
              Element repositoryStatusElement = (nodeList.getLength() > 0) ? (Element)nodeList.item(0) : null;
              if (workingCopyStatusElement != null)
              {
                state = parseState(workingCopyStatusElement,repositoryStatusElement);
                workingRevision = workingCopyStatusElement.getAttribute("revision");
                if (workingRevision.equals("-1")) workingRevision = "";
              }
              else
              {
                state           = FileData.States.UNKNOWN;
                workingRevision = "";
              }
//Dprintf.dprintf("state=%s workingRevision=%s",state,workingRevision);

              // get repository revision, author
              nodeList = entryElement.getElementsByTagName("commit");
              if (nodeList.getLength() > 0)
              {
                Element commitElement = (Element)nodeList.item(0);

                repositoryRevision = commitElement.getAttribute("revision");
                if (repositoryRevision.equals("-1")) repositoryRevision = "";

                nodeList = commitElement.getElementsByTagName("author");
                if (nodeList.getLength() > 0)
                {
                  Element authorElement = (Element)nodeList.item(0);
                  author = authorElement.getFirstChild().getNodeValue();
                }
                else
                {
                  author = "";
                }
              }
              else
              {
                repositoryRevision = "";
                author             = "";
              }
//Dprintf.dprintf("repositoryRevision=%s author=%s",repositoryRevision,author);

              // get file type, size, date/time
              File file = new File(rootPath,name);
              type = getFileType(file);
              size = file.length();
              date = new Date(file.lastModified());

              FileData fileData = findFileData(fileDataSet,file.getAbsolutePath());
//Dprintf.dprintf("directory=%s file=%s fileData=%s type=%s",directory,file,fileData,type);
              if      (fileData != null)
              {
                // update file data
                fileData.type               = type;
                fileData.state              = state;
                fileData.mode               = FileData.Modes.BINARY;
                fileData.locked             = locked;
                fileData.size               = size;
                fileData.date               = date;
                fileData.workingRevision    = workingRevision;
                fileData.repositoryRevision = repositoryRevision;
                fileData.branch             = branch;
              }
              else if (   (newFileDataSet != null)
                       && !name.equals(".")
                       && !isHiddenFile(name)
                       && !isIgnoreFile(name)
                       && !containFile(newFileDataSet,file.getAbsolutePath())
                      )
              {
                // create file data
                FileData newFileData = FileData.getInstance(rootPath,
                                                            name,
                                                            type,
                                                            state,
                                                            FileData.Modes.BINARY,
                                                            locked,
                                                            size,
                                                            date,
                                                            workingRevision,
                                                            repositoryRevision,
                                                            branch
                                                           );
//Dprintf.dprintf("new1 file=%s name=%s fileData=%s",file,name,newFileData);
                newFileDataSet.add(newFileData);
              }
/*
Dprintf.dprintf("fileName=%s",fileName);
Dprintf.dprintf("state=%s",state);
Dprintf.dprintf("workingRevision=%s",workingRevision);
Dprintf.dprintf("repositoryRevision=%s",repositoryRevision);
Dprintf.dprintf("author=%s",author);
Dprintf.dprintf("branch=%s",branch);
Dprintf.dprintf("");
/**/
            }
          }
        }
        catch (org.xml.sax.SAXException exception)
        {
          // ignored
        }
        catch (javax.xml.parsers.ParserConfigurationException exception)
        {
          // ignored
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(exception);
      }
      finally
      {
        if (exec != null) exec.done();
      }
//Dprintf.dprintf("fileDataSet=%s",fileDataSet);
//Dprintf.dprintf("newFileDataSet=%s",newFileDataSet);

      // add unknown files
      if (newFileDataSet != null)
      {
        for (File file : listFiles(directory))
        {
          String name = file.getPath();
          if (   !name.equals(".")
              && !isHiddenFile(name)
              && !isIgnoreFile(name)
              && !containFile(fileDataSet,file.getAbsolutePath())
              && !containFile(newFileDataSet,file.getAbsolutePath())
             )
          {
            FileData newFileData = FileData.getInstance(rootPath,
                                                        name,
                                                        getFileType(file),
                                                        file.length(),
                                                        new Date(file.lastModified())
                                                       );
//Dprintf.dprintf("new2 file=%s fileData=%s",file,newFileData);
            newFileDataSet.add(newFileData);
          }
        }
      }
    }
  }

  /** get repository root URL
   * @return repository root URL
   */
  @Override
  public String getRepositoryRootURL()
  {
    if (repositoryRootURL == null)
    {
      Exec exec = null;
      try
      {
        // get info
        Command command = new Command();
        String  line;
        Matcher matcher;

        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        command.append("info","--xml");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        try
        {
          DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
          Document document = documentBuilder.parse(exec.getStdoutStream());
          document.getDocumentElement().normalize();

          NodeList rootList = document.getElementsByTagName("root");
          if (rootList.getLength() > 0)
          {
            repositoryRootURL = rootList.item(0).getFirstChild().getNodeValue();
          }
        }
        catch (org.xml.sax.SAXException exception)
        {
          // ignored
        }
        catch (javax.xml.parsers.ParserConfigurationException exception)
        {
          // ignored
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        // ignored
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    return repositoryRootURL;
  }

  /** get repository URL
   * @return repository URL
   */
  @Override
  public String getRepositoryURL()
  {
    if (repositoryURL == null)
    {
      Exec exec = null;
      try
      {
        // get info
        Command command = new Command();
        String  line;
        Matcher matcher;

        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        command.append("info","--xml");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        try
        {
          DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
          Document document = documentBuilder.parse(exec.getStdoutStream());
          document.getDocumentElement().normalize();

          NodeList urlList = document.getElementsByTagName("url");
          if (urlList.getLength() > 0)
          {
            repositoryURL = urlList.item(0).getFirstChild().getNodeValue();
          }
        }
        catch (org.xml.sax.SAXException exception)
        {
          // ignored
        }
        catch (javax.xml.parsers.ParserConfigurationException exception)
        {
          // ignored
        }

/*
        // parse info output
        while ((line = exec.getStdout()) != null)
        {
  //Dprintf.dprintf("line=%s",line);
          // match name, state
          if      ((matcher = PATTERN_REPOSITORTY_URL.matcher(line)).matches())
          {
            repositoryURL = matcher.group(1);
          }
        }
*/

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        // ignored
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    return repositoryURL;
  }

  /** get first revision name
   * @return first revision name
   */
  @Override
  public String getFirstRevision()
  {
    return FIRST_REVISION_NAME;
  }

  /** get last revision name
   * @return last revision name
   */
  @Override
  public String getLastRevision()
  {
    return LAST_REVISION_NAME;
  }

  /** get current revision
   * @paran repositoryURL repository server URL
   * @return current revision or null
   */
  public String getCurrentRevision(String repositoryURL)
    throws RepositoryException
  {
    String currentRevision = null;

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      command.append("info","--xml");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();
//Dprintf.dprintf("command=%s",command);

      try
      {
        DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document document = documentBuilder.parse(exec.getStdoutStream());
        document.getDocumentElement().normalize();

        NodeList commitList    = document.getElementsByTagName("commit");
        Element  commitElement = (commitList.getLength() > 0) ? (Element)commitList.item(0) : null;
        if (commitElement != null)
        {
          currentRevision = commitElement.getAttribute("revision");
        }
      }
      catch (org.xml.sax.SAXException exception)
      {
        // ignored
      }
      catch (javax.xml.parsers.ParserConfigurationException exception)
      {
        // ignored
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return currentRevision;
  }

  /** get revision names of file
   * @param name file name or URL
   * @return revision name list
   */
  @Override
  public List<String> getRevisionNames(String name)
    throws RepositoryException
  {
    ArrayList<String> revisionList = new ArrayList<String>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      // get revision data list
      Command command = new Command();
      String  line;
      Matcher matcher;

      // get log
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("log","--xml","-r","HEAD:0","--verbose");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      if ((name != null) && !name.isEmpty()) command.append(name);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      try
      {
        DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document document = documentBuilder.parse(exec.getStdoutStream());
        document.getDocumentElement().normalize();
        NodeList logList = document.getElementsByTagName("log");

        Element  logElement   = (Element)logList.item(0);
        NodeList logEntryList = logElement.getElementsByTagName("logentry");
        int      i            = 0;
        while (   (i < logEntryList.getLength())
//               && ((busyDialog == null) || !busyDialog.isAborted())
              )
        {
          // get log entry
          Element logEntryElement = (Element)logEntryList.item(i);

          // parse and add revision number
          revisionList.add(logEntryElement.getAttribute("revision"));

          // next entry, update busy dialog
          i++;
//          if (busyDialog != null) busyDialog.updateProgressBar(i);
        }
      }
      catch (org.xml.sax.SAXException exception)
      {
        // ignored
      }
      catch (javax.xml.parsers.ParserConfigurationException exception)
      {
        // ignored
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // sort
    Collections.sort(revisionList,new Comparator<String>()
    {
      public int compare(String revisionId1, String revisionId2)
      {
        int number1 = Integer.parseInt(revisionId1);
        int number2 = Integer.parseInt(revisionId2);
        if      (number1 < number2)
        {
          return -1;
        }
        else if (number1 > number2)
        {
          return 1;
        }
        else
        {
          return 0;
        }
      }
    });

    return revisionList;
  }

  /** get revision data
   * @param fileData file data
   * @param revisionId revision
   * @return revision data or null
   */
  @Override
  public RevisionData getRevisionData(FileData fileData, String revisionId)
    throws RepositoryException
  {
    RevisionData revisionData = null;

    assert revisionId != null;

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      // get revision data
      Command command = new Command();

      // get single log entry (Note: fileData unused)
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("log","--xml","-r",revisionId/*+":PREV"*/,"-l","1","--verbose");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getRepositoryRootURL());
      exec = new Exec(rootPath,command);
      exec.closeStdin();
//Dprintf.dprintf("command=%s",command);

      try
      {
        DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document document = documentBuilder.parse(exec.getStdoutStream());
        document.getDocumentElement().normalize();
        NodeList logList = document.getElementsByTagName("log");

        Element  logElement      = (Element)logList.item(0);

        NodeList nodeList        = logElement.getElementsByTagName("logentry");
        Element  logEntryElement = (nodeList.getLength() > 0) ? (Element)nodeList.item(0) : null;
        if (logEntryElement != null)
        {
          // parse log entry
          revisionData = parseRevisionData(logEntryElement);
        }
      }
      catch (org.xml.sax.SAXException exception)
      {
        // ignored
      }
      catch (javax.xml.parsers.ParserConfigurationException exception)
      {
        // ignored
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return revisionData;
  }

  /** get revision data tree
   * @param fileData file data
   * @param busyDialog busy dialog or null
   * @return revision data tree
   */
  @Override
  public RevisionDataTree getRevisionDataTree(FileData fileData, BusyDialog busyDialog)
    throws RepositoryException
  {
    LinkedList<RevisionDataSVN> revisionDataList = new LinkedList<RevisionDataSVN>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    // init busy dialog
    if (busyDialog != null) busyDialog.setMaximum(getTotalCommitCount());

    // get revision data info list
    if (busyDialog != null) busyDialog.updateText("Get revisions...");
    if (getRepositoryRootURL() != null)
    {
      Exec exec = null;
      try
      {
        // get revision data list
        Command command = new Command();

        // get log
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
        command.append("log","--xml","-r","HEAD:0","--verbose");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
Dprintf.dprintf("");
  //      command.append(getFileDataName(fileData));
        command.append(getRepositoryRootURL());
        exec = new Exec(rootPath,command);
        exec.closeStdin();
//Dprintf.dprintf("command=%s",command);

        try
        {
          DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
          Document document = documentBuilder.parse(exec.getStdoutStream());
          document.getDocumentElement().normalize();
          NodeList logList = document.getElementsByTagName("log");

          Element  logElement   = (Element)logList.item(0);
          NodeList logEntryList = logElement.getElementsByTagName("logentry");
          int      i            = 0;
          while (   (i < logEntryList.getLength())
                 && ((busyDialog == null) || !busyDialog.isAborted())
                )
          {
            // get log entry
            Element logEntryElement = (Element)logEntryList.item(i);

            // parse and add log entry
            revisionDataList.addFirst(parseRevisionData(logEntryElement));

            // next entry, update busy dialog
            i++;
            if (busyDialog != null) busyDialog.updateProgressBar(i);
          }
        }
        catch (org.xml.sax.SAXException exception)
        {
          // ignored
        }
        catch (javax.xml.parsers.ParserConfigurationException exception)
        {
          // ignored
        }

        if ((busyDialog == null) || !busyDialog.isAborted())
        {
          // wait for termination
          int exitCode = exec.waitFor();
          if (exitCode != 0)
          {
            throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
          }
        }
        else
        {
          exec.destroy();
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }
//for (RevisionDataSVN revisionData : revisionDataList) Dprintf.dprintf("revisionData=%s",revisionData);

    // create revision data tree from list
    if (busyDialog != null) busyDialog.updateText("Build revision tree...");
    return createRevisionDataTree(revisionDataList,busyDialog);
  }

  /** get file data (text lines)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public String[] getFileLines(FileData fileData, String revisionId)
    throws RepositoryException
  {
    List<String> lineList = new ArrayList<String>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;

      // get file
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("cat");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      if (revisionId != null) command.append("--revision",revisionId);
      command.append("--");
      command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read file data
      while ((line = exec.getStdout()) != null)
      {
        lineList.add(line);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return lineList.toArray(new String[lineList.size()]);
  }

  /** get file data (byte array)
   * @param fileData file data
   * @param revisionId revision to get
   * @return file data (array of lines)
   */
  @Override
  public byte[] getFileBytes(FileData fileData, String revisionId)
    throws RepositoryException
  {
    ByteArrayOutputStream output = new ByteArrayOutputStream(64*1024);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      int     n;
      byte[]  buffer  = new byte[64*1024];

      // get file data
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("cat");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      if (revisionId != null) command.append("--revision",revisionId);
      command.append("--");
      command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command,true);
      exec.closeStdin();

      // read file bytes into byte array stream
      while ((n = exec.readStdout(buffer)) > 0)
      {
        output.write(buffer,0,n);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // convert byte array stream into array
    return output.toByteArray();
  }

  /** get all changed/unknown files
   * @param stateSet state set
   * @return fileDataSet file data set with modified files
   */
  @Override
  public Set<FileData> getChangedFiles(EnumSet<FileData.States> stateSet)
    throws RepositoryException
  {
    final Pattern PATTERN_STATUS         = Pattern.compile("^(.).(.)..(.).\\s(.)\\s+(\\S+?)\\s+(\\S+?)\\s+(\\S+?)\\s+(.*?)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_UNKNOWN        = Pattern.compile("^\\?.......\\s+(.*?)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_STATUS_AGAINST = Pattern.compile("^Status against revision:.*",Pattern.CASE_INSENSITIVE);

    HashSet<FileData> fileDataSet = new HashSet<FileData>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command         command            = new Command();
      String          line;
      Matcher         matcher;
      String          name               = null;
      FileData.States state              = FileData.States.UNKNOWN;
      boolean         locked             = false;
      String          workingRevision    = "";
      String          repositoryRevision = "";
      String          author             = "";

      // get list of files which may be updated or which are locally changed
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("status","-uv");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read list
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        // match name, state
        if      ((matcher = PATTERN_STATUS.matcher(line)).matches())
        {
          state              = parseState(matcher.group(1),matcher.group(4));
          locked             = matcher.group(2).equals("K") || matcher.group(3).equals("L");
          workingRevision    = matcher.group(5);
          repositoryRevision = matcher.group(6);
          author             = matcher.group(7);
          name               = matcher.group(8);

          File file = new File(rootPath,name);
          if (   !isIgnoreFile(file)
              && file.isFile()
              && stateSet.contains(state)
             )
          {
            fileDataSet.add(FileData.getInstance(rootPath,
                                                 name,
                                                 state,
                                                 FileData.Modes.BINARY,
                                                 locked
                                                )
                           );
          }
        }
        else if ((matcher = PATTERN_UNKNOWN.matcher(line)).matches())
        {
          name = matcher.group(1);

          File file = new File(rootPath,name);
          if (   !isHiddenFile(file)
              && !isIgnoreFile(file)
              && file.isFile()
              && stateSet.contains(FileData.States.UNKNOWN)
             )
          {
            fileDataSet.add(FileData.getInstance(rootPath,name,FileData.States.UNKNOWN));
          }
        }
        else if (PATTERN_STATUS_AGAINST.matcher(line).matches())
        {
          // ignore "Pattern status against:..."
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for changed files line '%s'",line);
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return fileDataSet;
  }

  /** get diff of file
   * @param fileData file data
   * @param revision revision to get diff for
   * @return diff data
   */
  @Override
  public List<DiffData> getDiff(FileData fileData, String oldRevision, String newRevision)
    throws RepositoryException
  {
    final Pattern PATTERN_DIFF_START  = Pattern.compile("^\\+\\+\\+.*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_DIFF        = Pattern.compile("^@@\\s+\\-([\\d,]*)\\s+\\+([\\d,]*)\\s+@@$",Pattern.CASE_INSENSITIVE);

    ArrayList<DiffData> diffDataList = new ArrayList<DiffData>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      Matcher matcher;
      String  line;

      String[] newFileLines = null;
      if (newRevision != null)
      {
        // check out new revision
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
        command.append("cat","--revision",newRevision);
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        if (fileData != null) command.append(getFileDataName(fileData));
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // read content
        List<String> newFileLineList = new ArrayList<String>();
        while ((line = exec.getStdout()) != null)
        {
          newFileLineList.add(line);
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;

        // convert to lines array
        newFileLines = newFileLineList.toArray(new String[newFileLineList.size()]);
      }
      else
      {
        // use local revision
        try
        {
          // open file
          BufferedReader bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));

          // read content
          List<String> newFileLineList = new ArrayList<String>();
          while ((line = bufferedReader.readLine()) != null)
          {
            newFileLineList.add(line);
          }

          // close file
          bufferedReader.close();

          newFileLines = newFileLineList.toArray(new String[newFileLineList.size()]);
        }
        catch (IOException exception)
        {
          throw new RepositoryException(Onzen.reniceIOException(exception));
        }
      }

      // diff file
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("diff","--revision",((oldRevision != null) ? oldRevision : getLastRevision())+((newRevision != null) ? ":"+newRevision : ""));
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      if (fileData != null) command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // skip diff header
      while ((line = exec.getStdout()) != null)
      {
        if (PATTERN_DIFF_START.matcher(line).matches()) break;
      }

      /* parse diff output
           Format:
             @@ -<i> +<j> @@
      */
      int          lineNb = 1;
      DiffData     diffData;
      List<String> keepLinesList    = new ArrayList<String>();
      List<String> addedLinesList   = new ArrayList<String>();
      List<String> deletedLinesList = new ArrayList<String>();
      String[]          lines;
      while ((line = exec.getStdout()) != null)
      {
//Dprintf.dprintf("line=%s",line);
        if      ((matcher = PATTERN_DIFF.matcher(line)).matches())
        {
          int[] oldIndex = parseDiffIndex(matcher.group(1));
          int[] newIndex = parseDiffIndex(matcher.group(2));
//Dprintf.dprintf("oldIndex=%d,%d",oldIndex[0],oldIndex[1]);
//Dprintf.dprintf("newIndex=%d,%d",newIndex[0],newIndex[1]);

          while (   ((line = exec.getStdout()) != null)
                 && !line.startsWith("@@")
                )
          {
            exec.ungetStdout(line);
//Dprintf.dprintf("line=%s",line);

            // get keep lines
            keepLinesList.clear();
            while ((lineNb < newIndex[0]) && (lineNb <= newFileLines.length))
            {
              keepLinesList.add(newFileLines[lineNb-1]);
              lineNb++;
            }
            while (   ((line = exec.getStdout()) != null)
                   && (line.isEmpty() || line.startsWith(" "))
                  )
            {
              if (lineNb <= newFileLines.length)
              {
                keepLinesList.add(newFileLines[lineNb-1]);
              }
              else
              {
                keepLinesList.add("???");
              }
              lineNb++;
            }
            exec.ungetStdout(line);
            diffData = new DiffData(DiffData.Types.KEEP,keepLinesList);
            diffDataList.add(diffData);

            // get deleted lines
            deletedLinesList.clear();
            while (   ((line = exec.getStdout()) != null)
                   && line.startsWith("-")
                  )
            {
              deletedLinesList.add(line.substring(1));
            }
            exec.ungetStdout(line);

            // get added lines
            addedLinesList.clear();
            while (   ((line = exec.getStdout()) != null)
                   && line.startsWith("+")
                  )
            {
              addedLinesList.add(line.substring(1));
              lineNb++;
            }
            exec.ungetStdout(line);

            int deletedLinesCount = deletedLinesList.size();
            int addedLinesCount   = addedLinesList.size();
            if      (deletedLinesCount < addedLinesCount)
            {
              // changed
              if (deletedLinesCount > 0)
              {
                lines = new String[deletedLinesCount];
                for (int z = 0; z < deletedLinesCount; z++)
                {
                  lines[z] = addedLinesList.get(z);
                }
                diffData = new DiffData(DiffData.Types.CHANGED,lines,deletedLinesList);
                diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",lines.length,deletedLinesCount);
              }

              // added
              lines = new String[addedLinesCount-deletedLinesCount];
              for (int z = 0; z < addedLinesCount-deletedLinesCount; z++)
              {
                 lines[z] = addedLinesList.get(deletedLinesCount+z);
              }
              diffData = new DiffData(DiffData.Types.ADDED,lines);
              diffDataList.add(diffData);
//Dprintf.dprintf("a %d",lines.length);
            }
            else if (deletedLinesCount > addedLinesCount)
            {
              // changed
              if (addedLinesCount > 0)
              {
                lines = new String[addedLinesCount];
                for (int z = 0; z < addedLinesCount; z++)
                {
                  lines[z] = deletedLinesList.get(z);
                }
                diffData = new DiffData(DiffData.Types.CHANGED,addedLinesList,lines);
                diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",addedLinesCount,lines.length);
              }

              // deleted
              lines = new String[deletedLinesCount-addedLinesCount];
              for (int z = 0; z < deletedLinesCount-addedLinesCount; z++)
              {
                 lines[z] = deletedLinesList.get(addedLinesCount+z);
              }
              diffData = new DiffData(DiffData.Types.DELETED,lines);
              diffDataList.add(diffData);
//Dprintf.dprintf("d %d",lines.length);
            }
            else if ((deletedLinesCount > 0) && (addedLinesCount > 0))
            {
              // changed
              diffData = new DiffData(DiffData.Types.CHANGED,addedLinesList,deletedLinesList);
              diffDataList.add(diffData);
//Dprintf.dprintf("c %d %d",addedLinesCount,deletedLinesCount);
            }
            else
            {
              // unknown -> stop
              break;
            }
          }
          exec.ungetStdout(line);
        }
        else if (line.startsWith("Property "))
        {
          // skip "Property ..."
          do
          {
            lineNb++;
          }
          while (   ((line = exec.getStdout()) != null)
                 && !line.startsWith("@@")
                );
          exec.ungetStdout(line);
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for diff line '%s'",line);
        }
      }

      // get rest of keep lines
      if (lineNb <= newFileLines.length)
      {
        keepLinesList.clear();
        while (lineNb <= newFileLines.length)
        {
          keepLinesList.add(newFileLines[lineNb-1]);
          lineNb++;
        }
        diffData = new DiffData(DiffData.Types.KEEP,keepLinesList);
        diffDataList.add(diffData);
      }
//Dprintf.dprintf("diffData=%s",diffData);

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
/*
int lineNb=1;
for (DiffData d : diffDataList)
{
Dprintf.dprintf("%s %d: %d %d %d",d.blockType,
lineNb,
(d.keepLines!=null)?d.keepLines.length:0,
(d.addedLines!=null)?d.addedLines.length:0,
(d.deletedLines!=null)?d.deletedLines.length:0
);
if (d.blockType==DiffData.Types.CHANGED) for (int z = 0; z < d.deletedLines.length; z++) Dprintf.dprintf("%s -----> %s",d.deletedLines[z],d.addedLines[z]);
if (d.blockType==DiffData.Types.KEEP) lineNb += d.keepLines.length;
if (d.blockType==DiffData.Types.ADDED) lineNb += d.addedLines.length;
}
*/

    return diffDataList;
  }

  /** get patch lines for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @param ignoreWhitespaces true to ignore white spaces
   * @param output patch output or null
   * @param lineList patch data lines or null
   */
  @Override
  public void getPatch(Set<FileData> fileDataSet, String revisionId1, String revisionId2, boolean ignoreWhitespaces, PrintWriter output, List<String> lineList)
    throws RepositoryException
  {
    final Pattern PATTERN_OLD_FILE = Pattern.compile("^\\-\\-\\-\\s+(.*)",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_NEW_FILE = Pattern.compile("^\\+\\+\\+\\s+(.*)",Pattern.CASE_INSENSITIVE);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    // get existing/new files
    HashSet<FileData> existFileDataSet = new HashSet<FileData>();
    HashSet<FileData> newFileDataSet   = new HashSet<FileData>();
    if (fileDataSet != null)
    {
      for (FileData fileData : fileDataSet)
      {
        if (fileData.state != FileData.States.UNKNOWN)
        {
          existFileDataSet.add(fileData);
        }
        else
        {
          newFileDataSet.add(fileData);
        }
      }
    }

    // get patch for existing files
    if ((fileDataSet == null) || (existFileDataSet.size() > 0))
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;

        // get patch
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
        command.append("diff");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        if (!Settings.svnDiffCommand.isEmpty())
        {
          // use external diff command
          command.append("--diff-cmd",Settings.svnDiffCommand);
          if (ignoreWhitespaces)
          {
            if (!Settings.svnDiffCommandOptionsIgnoreWhitespaces.isEmpty())
            {
              command.append("-x",Settings.svnDiffCommandOptionsIgnoreWhitespaces);
            }
          }
          else
          {
            if (!Settings.svnDiffCommandOptions.isEmpty())
            {
              command.append("-x",Settings.svnDiffCommandOptions);
            }
          }
        }
        else
        {
          // use internal diff
          if (ignoreWhitespaces)
          {
            command.append("-x","-ubw");
          }
          else
          {
            command.append("-x","-u");
          }
        }
        command.append("--revision",((revisionId1 != null) ? revisionId1 : getLastRevision())+((revisionId2 != null) ? ":"+revisionId2 : ""));
        command.append("--");
        if (fileDataSet != null) command.append(getFileDataNames(existFileDataSet));
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // read patch data
        Matcher matcher;
        while ((line = exec.getStdout()) != null)
        {
          // fix +++/--- lines: strip out first part in name, e. g. "a/"/"b/", check if absolute path and convert
          if      ((matcher = PATTERN_OLD_FILE.matcher(line)).matches())
          {
            String fileName = matcher.group(1);
            if (fileName.startsWith(rootPath))
            {
              // absolute path -> convert to relative path
              fileName = matcher.group(1).substring(rootPath.length()+1);
            }
            else
            {
              // relative path -> strip out first path in name
              int index = fileName.indexOf(File.separator);
              fileName = (index >= 0) ? fileName.substring(index+1) : fileName;
            }
            line = "--- "+fileName;
          }
          else if ((matcher = PATTERN_NEW_FILE.matcher(line)).matches())
          {
            String fileName = matcher.group(1);
            if (fileName.startsWith(rootPath))
            {
              // absolute path -> convert to relative path
              fileName = matcher.group(1).substring(rootPath.length()+1);
            }
            else
            {
              // relative path -> strip out first path in name
              int index = fileName.indexOf(File.separator);
              fileName = (index >= 0) ? fileName.substring(index+1) : fileName;
            }
            line = "+++ "+fileName;
          }

          if      (output   != null) output.println(line);
          else if (lineList != null) lineList.add(line);
        }

        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    // get complete patches for new files
    for (FileData fileData : newFileDataSet)
    {
      try
      {
        BufferedReader bufferedReader;
        String         line;

        // count number of lines in file
        int lineCount = 0;
        bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));
        while ((line = bufferedReader.readLine()) != null)
        {
          lineCount++;
        }
        bufferedReader.close();

        // add as patch
        bufferedReader = new BufferedReader(new FileReader(fileData.getFile()));
        String dateString = DateFormat.getDateInstance().format(new Date());
        if      (output   != null)
        {
          output.println(String.format("diff -u %s",fileData.getName()));
          output.println(String.format("--- /dev/null\t%s",dateString));
          output.println(String.format("+++ %s\t%s",fileData.getName(),dateString));
          output.println(String.format("@@ -1,%d +1,%d @@",lineCount,lineCount));
          while ((line = bufferedReader.readLine()) != null)
          {
            output.println("+"+line);
          }
        }
        else if (lineList != null)
        {
          lineList.add(String.format("diff -u %s",fileData.getName()));
          lineList.add(String.format("--- /dev/null\t%s",dateString));
          lineList.add(String.format("+++ %s\t%s",fileData.getName(),dateString));
          lineList.add(String.format("@@ -1,%d +1,%d @@",lineCount,lineCount));
          while ((line = bufferedReader.readLine()) != null)
          {
            lineList.add("+"+line);
          }
        }
        bufferedReader.close();
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
    }
  }

  /** get patch data for file
   * @param fileDataSet file data set
   * @param revisionId1,revisionId2 revisions to get patch for
   * @return patch data bytes
   */
  @Override
  public byte[] getPatchBytes(Set<FileData> fileDataSet, String revisionId1, String revisionId2)
    throws RepositoryException
  {
    ByteArrayOutputStream output = new ByteArrayOutputStream(64*1024);

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      int     n;
      byte[]  buffer  = new byte[64*1024];

      // get patch
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("diff","--revision",((revisionId1 != null) ? revisionId1 : getLastRevision())+((revisionId2 != null) ? ":"+revisionId2 : ""));
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      if (fileDataSet != null) command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read patch bytes into byte array stream
      while ((n = exec.readStdout(buffer)) > 0)
      {
        output.write(buffer,0,n);
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    // convert byte array stream into array
    return output.toByteArray();
  }

  /** get log to file
   * @param fileData file data
   * @return log array
   */
  @Override
  public List<LogData> getLog(FileData fileData)
    throws RepositoryException
  {
    List<LogData> logDataList = new ArrayList<LogData>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      // get revision data list
      HashMap<String,String> symbolicNamesMap = new HashMap<String,String>();
      Command                command = new Command();

      // get log
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("log","--xml","-r","HEAD:0","--verbose");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      if (fileData != null) command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      try
      {
        DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document document = documentBuilder.parse(exec.getStdoutStream());
        document.getDocumentElement().normalize();
        NodeList logList = document.getElementsByTagName("log");

        Element  logElement   = (Element)logList.item(0);
        NodeList logEntryList = logElement.getElementsByTagName("logentry");
        int      i            = 0;
        while (   (i < logEntryList.getLength())
//               && ((busyDialog == null) || !busyDialog.isAborted())
              )
        {
          // get log entry
          Element logEntryElement = (Element)logEntryList.item(i);

          // parse and add log entry
          logDataList.add(parseLogData(logEntryElement));

          // next entry, update busy dialog
          i++;
//          if (busyDialog != null) busyDialog.updateProgressBar(i);
        }
      }
      catch (org.xml.sax.SAXException exception)
      {
        // ignored
      }
      catch (javax.xml.parsers.ParserConfigurationException exception)
      {
        // ignored
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (Exception exception)
    {
      throw new RepositoryException(exception);
    }
    finally
    {
      if (exec != null) exec.done();
    }
  //for (LogData logData : logDataList) Dprintf.dprintf("logData=%s",logData);

    return logDataList;
  }

  /** get annotations to file
   * @param fileData file data
   * @param revisionId revision
   * @return annotation list
   */
  @Override
  public List<AnnotationData> getAnnotations(FileData fileData, String revisionId)
    throws RepositoryException
  {
    final Pattern PATTERN_ANNOTATION = Pattern.compile("^\\s*(\\d+)\\s+(\\S+)\\s+(\\S+\\s+\\S+\\s+\\S+)\\s+\\([^\\)]*\\)\\s(.*?)\\s*",Pattern.CASE_INSENSITIVE);
    final Pattern PATTERN_EMPTY      = Pattern.compile("^\\s*",Pattern.CASE_INSENSITIVE);

    ArrayList<AnnotationData> annotationDataList = new ArrayList<AnnotationData>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      Matcher matcher;
      String  line;

      // get annotations
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("blame","-v");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      if (revisionId != null) command.append("-r",revisionId);
      command.append("--");
      command.append(getFileDataName(fileData));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      /* parse annotation output
           Format:
             <revision> <author> <date> <line>
      */
      while ((line = exec.getStdout()) != null)
      {
        if      ((matcher = PATTERN_ANNOTATION.matcher(line)).matches())
        {
          annotationDataList.add(new AnnotationData(matcher.group(1),
                                                    matcher.group(2),
                                                    parseDate(matcher.group(3)),
                                                    matcher.group(4))
                                                   );
        }
        else if (PATTERN_EMPTY.matcher(line).matches())
        {
          // skip empty lines
        }
        else
        {
          // unknown line
          Onzen.printWarning("No match for annotations line '%s'",line);
        }
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return annotationDataList;
  }

  /** update file from respository
   * @param fileDataSet file data set or null
   * @param busyDialog busy dialog or null
   */
  @Override
  public void update(Set<FileData> fileDataSet, BusyDialog busyDialog)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // update files
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("update","--accept","postpone");
      if (fileDataSet != null) command.append("--depth","immediates");
//      command.append(Settings.svnCommands.get(version),"--non-interactive","merge","--dry-run","-r","BASE:HEAD");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      if (fileDataSet != null) command.append(getFileDataNames(fileDataSet));
//      command.append(".");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read output
      while (   ((busyDialog == null) || !busyDialog.isAborted())
             && !exec.eofStdout()
            )
      {
        String line;

        line = exec.pollStdout();
        if (line != null)
        {
          if (busyDialog != null) busyDialog.updateList(line);
        }
      }
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** commit files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void commit(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // commit files
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("commit","-F",commitMessage.getFileName());
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** add files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   * @param binaryFlag true to add file as binary files, false otherwise
   */
  @Override
  public void add(Set<FileData> fileDataSet, CommitMessage commitMessage, boolean binaryFlag)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // add files
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("add");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        commit(fileDataSet,commitMessage);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** remove files
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void remove(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      // delete local files
      for (FileData fileData : fileDataSet)
      {
        fileData.getFile().delete();
      }

      // remove from repository
      Command command = new Command();

      // remove files
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("remove");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        commit(fileDataSet,commitMessage);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** copy files
   * @param fileDataSet files to copy
   * @param destination destination
   * @param commitMessage commit message
   */
  @Override
  public void copy(Set<FileData> fileDataSet, String destination, CommitMessage commitMessage)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // copy file
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("copy");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      command.append(destination);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        fileDataSet.add(FileData.getInstance(rootPath,destination));
        commit(fileDataSet,commitMessage);
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** rename file
   * @param fileData file data to rename
   * @param newName new name
   * @param commitMessage commit message
   */
  @Override
  public void rename(FileData fileData, String newName, CommitMessage commitMessage)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // copy file
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("rename");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(fileData.getName());
      command.append(newName);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      // immediate commit when message is given
      if ((commitMessage != null) && !commitMessage.isEmpty())
      {
        Set<FileData> fileDataSet = FileData.toSet(getFileDataName(fileData),
                                                   newName
                                                  );
        commit(fileDataSet,commitMessage);
      }

      // set new name
      fileData.setFileName(newName);
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** revert files
   * @param fileDataSet file data set
   * @param revisionId revision to revert to
   * @param recursive true for recursive revert, false otherwise
   */
  @Override
  public void revert(Set<FileData> fileDataSet, String revisionId, boolean recursive)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // delete local files
      for (FileData fileData : fileDataSet)
      {
        fileData.getFile().delete();
      }

      // revert files
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("revert");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      if (recursive) command.append("-R");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;

      if (revisionId != null)
      {
        // update to specifc revision
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
        command.append("update","-r",revisionId);
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        command.append(getFileDataNames(fileDataSet));
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** set conflicts resolved
   * @param fileDataSet file data set or null for all files
   */
  @Override
  public void resolve(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // copy file
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("resolve","--accept","working");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** change current branch
   * @param newBranchName new branch name
   */
  @Override
  public void changeBranch(String newBranchName)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // checkout branch
      command.clear();
      command.append(Settings.svnCommands.get(version),"switch","^/"+newBranchName);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** merge other branch into current branch
   * @param otherBranchName other branch name
   */
  @Override
  public void mergeBranch(String otherBranchName)
    throws RepositoryException
  {
    Exec exec = null;
    try
    {
      Command command = new Command();
      int     exitCode;

      // merge branch
      command.clear();
      command.append(Settings.svnCommands.get(version),"merge","--reintegrate","/"+otherBranchName);
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // wait for termination
      exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** lock files
   * @param fileDataSet file data set
   */
  @Override
  public void lock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // copy file
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("lock");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** unlock files
   * @param fileDataSet file data set
   */
  @Override
  public void unlock(Set<FileData> fileDataSet)
    throws RepositoryException
  {
    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // copy file
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("unlock");
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("--");
      command.append(getFileDataNames(fileDataSet));
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  /** set commit message
   * @param fileDataSet file data set
   * @param commitMessage commit message
   */
  @Override
  public void setCommitMessage(Set<FileData> fileDataSet, CommitMessage commitMessage)
    throws RepositoryException
  {
//svn propset -r 7173 --revprop svn:log "added master I/O, removed priority" http://localhost:4430/svn/bar/trunk/bar/bar
  }

  /** set files mode
   * @param fileDataSet file data set
   * @param mode file mode
   * @param commitMessage commit message
   */
  @Override
  public void setFileMode(Set<FileData> fileDataSet, FileData.Modes mode, CommitMessage commitMessage)
    throws RepositoryException
  {
  }

  /** get default name of root
   * @return default root name
   */
  @Override
  public String getDefaultRootName()
  {
    return DEFAULT_ROOT_NAME;
  }

  /** get default branch/tag name
   * @return default branch/tag name
   */
  @Override
  public String getDefaultBranchTagName()
  {
    return DEFAULT_BRANCH_NAME+"/";
  }

  /** get names of existing branches/tags
   * @param pathName path name
   * @return branch/tag name list
   */
  @Override
  public List<String> getBranchTagNames(String pathName)
    throws RepositoryException
  {
    HashSet<String> branchTagNameSet = new HashSet<String>();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    for (String branchName : DEFAULT_BRANCH_NAMES)
    {
      branchTagNameSet.add(branchName);
    }
    if (!pathName.isEmpty())
    {
      Exec exec = null;
      try
      {
        Command command = new Command();
        String  line;
        int     exitCode;

        // list branches
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
        command.append("list",pathName+"/branches");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        exec = new Exec(rootPath,command);
        exec.closeStdin();
//Dprintf.dprintf("command=%s",command);

        // read output
        while ((line = exec.getStdout()) != null)
        {
          branchTagNameSet.add("branches/"+StringUtils.trimEnd(line,"/\\"));
        }

        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;

        // list tags
        command.clear();
        command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
        if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
        if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
        command.append("list",pathName+"/tags");
        if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
        command.append("--");
        exec = new Exec(rootPath,command);
        exec.closeStdin();

        // read output
        while ((line = exec.getStdout()) != null)
        {
          branchTagNameSet.add("tags/"+StringUtils.trimEnd(line,"/\\"));
        }

        // wait for termination
        exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }

        // done
        exec.done(); exec = null;
      }
      catch (IOException exception)
      {
        throw new RepositoryException(Onzen.reniceIOException(exception));
      }
      finally
      {
        if (exec != null) exec.done();
      }
    }

    // convert to list and sort
    ArrayList<String> branchTagNameList = new ArrayList<String>(branchTagNameSet);
    Collections.sort(branchTagNameList);

    return branchTagNameList;
  }

  /** create new branch
   * @param rootName root name (source)
   * @param branchName branch name
   * @param fromRevision from revision name or null
   * @param commitMessage commit message
   * @param buysDialog busy dialog or null
   */
  @Override
  public void createBranch(String rootName, String branchName, String fromRevision, CommitMessage commitMessage, BusyDialog busyDialog)
    throws RepositoryException
  {
    String repositoryURL     = getRepositoryURL();
    String repositoryRootURL = getRepositoryRootURL();

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();

      // create branch
      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      if ((password != null) && !password.isEmpty()) command.append("--password",command.hidden(password));
      command.append("copy",repositoryURL+"/"+rootName,repositoryURL+"/"+branchName);
      if (!fromRevision.isEmpty()) command.append("-r",fromRevision);
      if (Settings.svnAlwaysTrustServerCertificate) command.append("--trust-server-cert");
      command.append("-F",commitMessage.getFileName());
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read output
      String line;
      while (   ((busyDialog == null) || !busyDialog.isAborted())
             && ((line = exec.getStdout()) != null)
            )
      {
//Dprintf.dprintf("line=%s",line);
        if (busyDialog != null) busyDialog.updateText(line);
      }
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        exec.destroy();
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }
  }

  //-----------------------------------------------------------------------

  /** get total commit count
   * @return total commit count
   */
  private int getTotalCommitCount()
    throws RepositoryException
  {
    int commitCount = 0;

    String password = ((passwordHandler != null) && (userName != null) && !userName.isEmpty()) ? passwordHandler.getPassword(userName,getRepositoryRootURL()) : null;

    Exec exec = null;
    try
    {
      Command command = new Command();
      String  line;
      Matcher matcher;

      command.clear();
      command.append(Settings.svnCommands.get(version),"--non-interactive","--config-option","config:auth:password-stores=");
      if ((userName != null) && !userName.isEmpty()) command.append("--username",userName);
      command.append("info","--xml");
      command.append("--");
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      try
      {
        DocumentBuilder documentBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
        Document document = documentBuilder.parse(exec.getStdoutStream());
        document.getDocumentElement().normalize();

        NodeList commitList    = document.getElementsByTagName("commit");
        Element  commitElement = (commitList.getLength() > 0) ? (Element)commitList.item(0) : null;
        if (commitElement != null)
        {
          commitCount = Integer.parseInt(commitElement.getAttribute("revision"));
        }
      }
      catch (org.xml.sax.SAXException exception)
      {
        // ignored
      }
      catch (javax.xml.parsers.ParserConfigurationException exception)
      {
        // ignored
      }

      // done
      exec.done(); exec = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
    }

    return commitCount;
  }

  /** parse SVN state string
   * @param stateFlag state flag string
   * @param updateFlag update flag string
   * @return state
   */
  private FileData.States parseState(String stateFlag, String updateFlag)
  {
    if      (stateFlag.equalsIgnoreCase(" "))
    {
      if (updateFlag.equalsIgnoreCase("*"))
      {
        return FileData.States.UPDATE;
      }
      else
      {
        return FileData.States.OK;
      }
    }
    else if (stateFlag.equalsIgnoreCase("A")) return FileData.States.ADDED;
    else if (stateFlag.equalsIgnoreCase("C")) return FileData.States.CONFLICT;
    else if (stateFlag.equalsIgnoreCase("D")) return FileData.States.REMOVED;
    else if (stateFlag.equalsIgnoreCase("M"))
    {
      if (updateFlag.equalsIgnoreCase("*"))
      {
        return FileData.States.MERGE;
      }
      else
      {
        return FileData.States.MODIFIED;
      }
    }
    else if (stateFlag.equalsIgnoreCase("!")) return FileData.States.CHECKOUT;
    else                                      return FileData.States.UNKNOWN;
  }

  /** parse SVN XML state string
   * @param element status element
   * @return state
   */
  private FileData.States parseState(Element workingCopyStatusElement, Element repositoryStatusElement)
  {
    FileData.States state  = FileData.States.UNKNOWN;

    String workingCopyItem = workingCopyStatusElement.getAttribute("item");
    String treeConflicted  = workingCopyStatusElement.getAttribute("tree-conflicted");
    String repositoryItem  = (repositoryStatusElement != null) ? repositoryStatusElement.getAttribute("item") : null;

    if      (workingCopyItem.equalsIgnoreCase("normal")           ) state = FileData.States.OK;
    else if (workingCopyItem.equalsIgnoreCase("update")           ) state = FileData.States.UPDATE;
    else if (workingCopyItem.equalsIgnoreCase("modified")         ) state = FileData.States.MODIFIED;
    else if (workingCopyItem.equalsIgnoreCase("added")            ) state = FileData.States.ADDED;
    else if (workingCopyItem.equalsIgnoreCase("conflict")         ) state = FileData.States.CONFLICT;
    else if (   workingCopyItem.equalsIgnoreCase("conflicted")
             || treeConflicted.equalsIgnoreCase("true")           ) state = FileData.States.CONFLICT;
    else if (workingCopyItem.equalsIgnoreCase("removed")          ) state = FileData.States.REMOVED;
    else if (workingCopyItem.equalsIgnoreCase("deleted")          ) state = FileData.States.REMOVED;
    else if (workingCopyItem.equalsIgnoreCase("merge")            ) state = FileData.States.MERGE;
    else if (workingCopyItem.equalsIgnoreCase("checkout")         ) state = FileData.States.CHECKOUT;
    else if (workingCopyItem.equalsIgnoreCase("missing")          ) state = FileData.States.CHECKOUT;
    else if (workingCopyItem.equalsIgnoreCase("unversioned")      ) state = FileData.States.UNKNOWN;

    if (repositoryItem != null)
    {
      if (repositoryItem.equalsIgnoreCase("added")) state = FileData.States.CHECKOUT;
      else                                          state = FileData.States.UPDATE;
    }

    return state;
  }

  /** parse SVN diff index
   * @param string diff index string
   * @return index array [lineNb,length] or null
   */
  private int[] parseDiffIndex(String string)
  {
    Object[] data = new Object[2];
    if      (StringParser.parse(string,"%d,%d",data))
    {
      return new int[]{(Integer)data[0],(Integer)data[1]-(Integer)data[0]-1};
    }
    else if (StringParser.parse(string,"%d",data))
    {
      return new int[]{(Integer)data[0],1};
    }
    else
    {
      return null;
    }
  }

  /** parse log XML data
   * @param logEntryElement log entry element
   * @return revision data or null
   */
  private LogData parseLogData(Element logEntryElement)
  {
    NodeList           nodeList;
    String             revision          = "";
    Date               date              = new Date();
    String             author            = "";
    LinkedList<String> commitMessageList = new LinkedList<String>();
    ArrayList<String>  fileNameList      = new ArrayList<String>();

    // get revision, date, author
    revision = logEntryElement.getAttribute("revision");
    nodeList = logEntryElement.getElementsByTagName("date");
    if (nodeList.getLength() > 0)
    {
      Element dateElement = (Element)nodeList.item(0);
      Node dateNode = dateElement.getFirstChild();
      if (dateNode != null)
      {
        date = parseDate(dateNode.getNodeValue());
      }
    }
    nodeList = logEntryElement.getElementsByTagName("author");
    if (nodeList.getLength() > 0)
    {
      Element authorElement = (Element)nodeList.item(0);
      Node authorNode = authorElement.getFirstChild();
      if (authorNode != null)
      {
        author = authorNode.getNodeValue();
      }
    }

    // get commit message
    nodeList = logEntryElement.getElementsByTagName("msg");
    if (nodeList.getLength() > 0)
    {
      Element msgElement = (Element)nodeList.item(0);
      Node msgNode = msgElement.getFirstChild();
      if (msgNode != null)
      {
        StringUtils.split(commitMessageList,msgNode.getNodeValue(),'\n');
      }
    }

    // get file names
    nodeList = logEntryElement.getElementsByTagName("paths");
    if (nodeList.getLength() > 0)
    {
      Element pathsElement = (Element)nodeList.item(0);
      nodeList = pathsElement.getElementsByTagName("path");
      for (int j = 0; j < nodeList.getLength(); j++)
      {
        Element pathElement = (Element)nodeList.item(j);

        String fileName = pathElement.getFirstChild().getNodeValue();
        fileNameList.add(fileName);
      }
    }

    // create log data entry
    return new LogData(revision,
                       date,
                       author,
                       commitMessageList,
                       fileNameList
                      );
  }

  /** parse log revision XML data
   * @param logEntryElement log entry element
   * @return revision data or null
   */
  private RevisionDataSVN parseRevisionData(Element logEntryElement)
  {
    NodeList           nodeList;
    String             revision          = "";
    Date               date              = new Date();
    String             author            = "";
    String             copyFromPath      = "";
    String             copyFromRevision  = "";
    String             branchName        = null;
    LinkedList<String> commitMessageList = new LinkedList<String>();
    ArrayList<String>  fileNameList      = new ArrayList<String>();

    // get revision, date, author
    revision = logEntryElement.getAttribute("revision");
    nodeList = logEntryElement.getElementsByTagName("date");
    if (nodeList.getLength() > 0)
    {
      Element dateElement = (Element)nodeList.item(0);
      Node dateNode = dateElement.getFirstChild();
      if (dateNode != null)
      {
        date = parseDate(dateNode.getNodeValue());
      }
    }
    nodeList = logEntryElement.getElementsByTagName("author");
    if (nodeList.getLength() > 0)
    {
      Element authorElement = (Element)nodeList.item(0);
      Node authorNode = authorElement.getFirstChild();
      if (authorNode != null)
      {
        author = authorNode.getNodeValue();
      }
    }

    // get commit message
    nodeList = logEntryElement.getElementsByTagName("msg");
    if (nodeList.getLength() > 0)
    {
      Element msgElement = (Element)nodeList.item(0);
      Node msgNode = msgElement.getFirstChild();
      if (msgNode != null)
      {
        StringUtils.split(commitMessageList,msgNode.getNodeValue(),'\n');
      }
    }

    // get file names
    nodeList = logEntryElement.getElementsByTagName("paths");
    if (nodeList.getLength() > 0)
    {
      Element pathsElement = (Element)nodeList.item(0);
      nodeList = pathsElement.getElementsByTagName("path");
      for (int j = 0; j < nodeList.getLength(); j++)
      {
        Element pathElement = (Element)nodeList.item(j);

        copyFromPath     = pathElement.getAttribute("copyfrom-path");
        copyFromRevision = pathElement.getAttribute("copyfrom-rev");

        String fileName = pathElement.getFirstChild().getNodeValue();
        fileNameList.add(fileName);
      }
    }

    // get branch name (first file name if a branch)
    if (!copyFromPath.isEmpty())
    {
      branchName = fileNameList.get(0);
      if (branchName.startsWith("/")) branchName = branchName.substring(1);
    }

    // create revision data entry
    return new RevisionDataSVN(revision,
                               branchName,
                               date,
                               author,
                               copyFromPath,
                               copyFromRevision,
                               commitMessageList,
                               fileNameList
                              );
  }

  /** create revision tree
   * @param revisionDataList revision data list
   * @param busyDialog busy dialog or null
   * @return revision tree
   */
  private RevisionDataTree createRevisionDataTree(LinkedList<RevisionDataSVN> revisionDataList,
                                                  BusyDialog                  busyDialog
                                                 )
  {
    RevisionDataTree revisionDataTree = new RevisionDataTree();

    HashMap<String,RevisionData>      branchDataMap             = new HashMap<String,RevisionData>();
    TreeMap<String,RevisionDataTree>  branchRevisionDataTreeMap = new TreeMap<String,RevisionDataTree>();
    branchRevisionDataTreeMap.put("/trunk",   revisionDataTree);
    branchRevisionDataTreeMap.put("/branches",revisionDataTree);
    branchRevisionDataTreeMap.put("/tags",    revisionDataTree);

    for (RevisionDataSVN revisionData : revisionDataList)
    {
//Dprintf.dprintf("revisionData=%s copyFromRevision=%s",revisionData,revisionData.copyFromRevision);

      RevisionData branchRevisionData = null;
      if (!revisionData.copyFromPath.isEmpty())
      {
        // copy from -> create tag/branch
//Dprintf.dprintf("copyFromPath=%s copyFromRevision=%s revisionData.fileNameList.get(0)=%s",revisionData.copyFromPath,revisionData.copyFromRevision,revisionData.fileNameList.get(0));
        branchRevisionData = branchDataMap.get(revisionData.copyFromPath);
        if (branchRevisionData == null) branchRevisionData = revisionDataTree.peekLast();
        if (branchRevisionData != null)
        {
          // add branch
          BranchData branchData = branchRevisionData.addBranch(revisionData.branchName,
                                                               revisionData.date,
                                                               revisionData.author,
                                                               revisionData.commitMessageList,
                                                               revisionData
                                                              );
          if (!branchRevisionDataTreeMap.containsKey(revisionData.fileNameList.get(0)+'/'))
          {
            branchRevisionDataTreeMap.put(revisionData.fileNameList.get(0)+'/',branchData.revisionDataTree);
          }
        }
        else
        {
          // no branch found? -> skip
          if (busyDialog != null) busyDialog.updateList("Branch not found for %s",revisionData.copyFromPath);
        }

        // store copyfrom revision data = internal branch name
//Dprintf.dprintf("new branch %s: parent %s",revisionData.fileNames[0],branchRevisionData);
        branchDataMap.put(revisionData.fileNameList.get(0),revisionData);
      }
      else
      {
        // commit -> find branch and add revision
//Dprintf.dprintf("commit revisionData=%s: %s",revisionData.revisionId,revisionData.fileNameList.get(0));
        Map.Entry<String,RevisionDataTree> branchRevisionDataTreeEntry = branchRevisionDataTreeMap.floorEntry(revisionData.fileNameList.get(0)+'/');
        if (branchRevisionDataTreeEntry != null)
        {
          branchRevisionDataTreeEntry.getValue().add(revisionData);
        }
        else
        {
          // no branch found? -> add revision at end, skip
          revisionDataTree.add(revisionData);
          if (busyDialog != null) busyDialog.updateList("Branch not found for %s",revisionData.fileNameList.get(0));
        }
      }
    }
//Dprintf.dprintf("------------------------------"); revisionDataTree.print();

    return revisionDataTree;
  }

  /** cleanup command
   * @param specialCommand special command
   * @param shell shell
   */
  private void cleanup(SpecialCommand specialCommand, Shell shell)
    throws RepositoryException
  {
    final Pattern PATTERN_URI = Pattern.compile("^[^:/]+://.*",Pattern.CASE_INSENSITIVE);

    BusyDialog busyDialog = null;
    Exec       exec       = null;
    try
    {
      Command command = new Command();

      // open busy dialog
      busyDialog = busyDialog = new BusyDialog(shell,
                                               "Execute clean-up command",
                                               Settings.geometrySpecialCommand,
                                               "Execute clean-up command"+"\u2026",
                                               BusyDialog.LIST|BusyDialog.AUTO_ANIMATE
                                              );
/*
              busyDialog.addListener(SWT.Dispose,new Listener()
              {
                public void handleEvent(Event event)
                {
                  Shell dialog = (Shell)event.widget;

                  Settings.geometrySpecialCommand = dialog.getSize();
                }
              });*/

      // cleanup
      command.append(Settings.svnCommands.get(version),"cleanup");
      busyDialog.updateList("%s: %s",rootPath,command.toString());
      exec = new Exec(rootPath,command);
      exec.closeStdin();

      // read output
      while (   !busyDialog.isAborted()
             && !exec.eofStdout()
            )
      {
        String line;

        // read stdout
        line = exec.pollStdout();
        if (line != null)
        {
//Dprintf.dprintf("out: %s",line);
          busyDialog.updateList(line);
        }
      }
      if ((busyDialog == null) || !busyDialog.isAborted())
      {
        // wait for termination
        int exitCode = exec.waitFor();
        if (exitCode != 0)
        {
          throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
        }
      }
      else
      {
        // abort
        exec.destroy();
      }

      // wait for termination
      int exitCode = exec.waitFor();
      if (exitCode != 0)
      {
        throw new RepositoryException("'%s', exit code: %d",exec.getExtendedErrorMessage(),command.toString(),exitCode);
      }

      // done
      exec.done(); exec = null;
      busyDialog.close(); busyDialog = null;
    }
    catch (IOException exception)
    {
      throw new RepositoryException(Onzen.reniceIOException(exception));
    }
    finally
    {
      if (exec != null) exec.done();
      if (busyDialog != null) busyDialog.close();
    }
  }
}

/* end of file */
